﻿* Encoding: UTF-8.
*MATRIX - END MATRIX functions.
*SPSS macros by Kirill Orlov
*kior@comtv.ru, ttnphns@gmail.com
*https://www.spsstools.net/en/KO-spssmacros
*All rights reserved.

********************* ФУНКЦИИ ДЛЯ MATRIX - END MATRIX *********************.

*1) Аргументы функции разделяются символом %, а не запятой.

*2) Функции пускаются как команды, первым словом в строке – а не как «справа в выражении», и имя
*возвращаемого результата надо писать среди аргументов функции, в качестве последнего из них.
*Т.е. правильно: !функция(аргумент % результат); неправильно: результат = !функция(аргумент).
*Некоторые функции выдают более одного результата.

*3) Любой аргумент, кроме результата, может быть выражением, не только именем или значением.
*Например, !функция(1.5*имя(2,4)+3 % результат) - допустимо. Выражение однако не может
*содержать функции, описанные в этом документе.

*4) Имя результата может совпадать с именем любого из аргументов.
*Т.е. можно, например: !функция(имя1 % имя2 % имя1). В этом случае результат будет имя1, а то, что
*прежде носило это имя, пропадет.

*5) Не используйте в вашем синтаксисе в MATRIX - END MATRIX имена, начинающиеся с символа @,
*поскольку этот символ зарезервирован здешними функциями, и его использование вами чревато потерей данных.

*6) Функции не проверяют и не выдают особых сообщений насчет соответствия ваших входящих данных
*требованиям данной функции (так сделано в интересах быстродействия). Поэтому внимательно ознакомливайтесь
*с требованиями конкретных функций, чтобы избегать ошибок.

*7) Функции можно ставить внутрь LOOP - END LOOP и DO IF - END IF.

*8) Для отключения функции в синтаксисе используйте не *, а /*, например:
/*!func(arg%result).
*или
/*!func(arg%result).*/.

*9) Эти матричные функции рассчитаны на числовые входящие данные. Некоторые из функций, не связанные
*с числовым анализом, могут работать с текстовыми данными (например, подсчет значений, перекодировка,
*переструктурирование), однако не гарантированно, т.к. код писался в расчете на числа.
*Пробуйте и проверяйте текстовые данные сами.

*10) Многие из этих функций требуют заранее, до первой в сессии команды MATRIX, установить предельное число
*циклов на достаточно большое значение, чтобы справляться с потенциально большими данными.
*Я рекомендую делать заранее предустановку на большое число циклов вообще всегда, когда вы собираетесь
*работать со здешними функциями - чтобы не беспокоиться об этом дальше. Например:
*set mxloops 100000000.  /*или 1E8
*matrix.
*etc.

*11) Дисперсия/ковариация и корреляция, когда вычисляются внутри функций, не всегда используют одну и
*ту же вычислительную формулу. Например, /*!KO_variance */ или /*!KO_corr*/ применяют одну формулу
*(дающую более вычислительно устойчивый результат), тогда как другие функции могут калькулировать по
*другой формуле (дающей менее вычислительно устойчивый результат).

*----------------------------------------------------------------------------------------------.


******** НАКОПЛЕННЫЕ СУММЫ В СТОЛБЦАХ ********.
*/*!KO_ccum(mat%dir%name)*/*.
*Version 1.
*Берет матрицу или столбец MAT с числом рядов не меньше 2 и выдает матрицу NAME того же размера;
*элементы в каждом столбце NAME есть накопленные суммы в этом столбце при направлении суммирования
*сверху вниз (если DIR положительно) либо снизу вверх (если DIR неположительно).
*DIR - скаляр.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_ccum(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @dir= !3.
do if @dir.
-loop @i= 2 to nrow(!4).
- comp !4(@i,:)= !4(@i-1,:)+!4(@i,:).
-end loop.
else.
-loop @i= nrow(!4)-1 to 1 by -1.
- comp !4(@i,:)= !4(@i+1,:)+!4(@i,:).
-end loop.
end if.
release @dir,@i.
!enddefine.


******** ПРОИЗВЕДЕНИЕ ЭЛЕМЕНТОВ МАТРИЦЫ ********.
*/*!KO_prod(mat%name)*/*.
*Version 2.
*Перемножает элементы MAT.
*Выдает скаляр NAME.

define !KO_prod(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
do if not all(@mat).
-comp !3= 0.
else.
-comp !3= exp(msum(ln(abs(@mat)))).
-do if mod(msum(@mat<0),2)).
- comp !3= -!3.
-end if.
end if.
release @mat.
!enddefine.


******** СУММА НАДДИАГОНАЛЬНЫХ ИЛИ ПОДДИАГОНАЛЬНЫХ ЭЛЕМЕНТОВ МАТРИЦЫ ********.
*/*!KO_trsum(mat%tr%name)*/*.
*Version 1.
*Суммирует элементы верхнего или нижнего треугольника MAT.
*Выдает скаляр NAME.
*Матрица MAT не обязана быть квадратной, но должна быть размером не меньше 2x2.
*TR - скаляр; если положительный, суммируются значения над главной диагональю;
*если неположительный, суммируются значения под главной диагональю.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_trsum(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @tr= !3.
do if @tr.
-comp @mat= t(@mat).
end if.
comp !4= 0.
loop @i= 2 to nrow(@mat).
-comp !4= !4+rsum(@mat(@i,1:rmin({@i-1,ncol(@mat)}))).
end loop.
release @mat,@tr,@i.
!enddefine.


******** ПРОВЕРКА ТРЕУГОЛЬНОГО НЕРАВЕНСТВА В МАТРИЦЕ РАЗЛИЧИЙ ********.
*/*!KO_trineq(dis%ineq%stop%name1%name2)*/*.
*Version 1.
*Берет квадратную матрицу расстояний (различий) DIS и проверяет, есть ли в ней расстояния,
*нарушающие "аксиому треугольного неравенства". Если такие есть, то расстояния не являются метрическими.
*Функция проверяет только наддиагональные элементы DIS; если вы хотите проверить
*поддиагональные элементы DIS - транспонируйте матрицу.
*DIS - квадратная с неотрицательными значениями, размером не менее 3.
*INEQ - цифра 0 или 1 (не имя и не выражение; цифру можно опционально взять в кавычки или апострофы).
*Если 0, проверяется нестрогое треугольное неравенство: Dab <= Dac+Dbc (где a b c - объекты, т.е.
*ряды/столбцы матрицы). Если 1, проверяется строгое треугольное неравенство: Dab < Dac+Dbc (Dab = Dac+Dbc
*не засчитывается как нарушение).
*STOP - цифра 0 или 1 (не имя и не выражение; цифру можно опционально взять в кавычки или апострофы).
*Если 0, проверяются все наддиагональные элементы: функция работает до конца. Если 1, функция обрывается
*после встречи первого же нарушения треугольного неравенства, если есть.
*Результаты:
*NAME1 - вектор-ряд длиной 2: первое значение - число троек (сочетаний элементов по три), проверенных
*над диагональю матрицы; второе значение - число троек из них с нарушенным треугольным
*неравенством. (При STOP=1 это значение всегда =1.)
*NAME2 - вектор-ряд длиной 4; он содержит максимальную найденную величину нарушения треугольного
*неравенства (1-е значение вектора) и номера трех объектов (рядов/столбцов матрицы), составляющих
*эту тройку. Если нарушений найдено не было, NAME2 состоит из -1.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_trineq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
!let !ie= !unquote(!3)
!let !stop= !unquote(!4)
!if (!stop) !then !let !ifstop= 'if @stop' !ifend
comp !5= 0.
comp !6= make(1,4,-1).
comp @stop= 0.
comp @cnt= 0.
loop @x= 1 to ncol(@d)-2.
-loop @y= @x+1 to ncol(@d)-1.
- comp @dxy= @d(@x,@y).
- loop @z= @y+1 to ncol(@d).
-  comp @cnt= @cnt+1.
-  comp @dxz= @d(@x,@z).
-  comp @dyz= @d(@y,@z).
-  comp @viol= rmax({@dxy-@dxz-@dyz,@dxz-@dyz-@dxy,@dyz-@dxy-@dxz}).
-  do if @viol and @viol<1E-14.
-   comp @viol= 0.
-  end if.
-  do if @viol !if (!ie) !then >=0 !ifend.
-   comp !5= !5+1.
-   do if @viol>!6(1).
-    comp !6= {@viol,@x,@y,@z}.
-   end if.
    !if (!stop) !then
-    comp @stop= 1.
    !ifend
-  end if.
- end loop !ifstop.
-end loop !ifstop.
end loop !ifstop.
comp !5= {@cnt,!5}.
release @d,@stop,@cnt,@viol,@x,@y,@z,@dxy,@dxz,@dyz.
!enddefine.


******** НИЖНЯЯ НЕПОЛНАЯ ГАММА-ФУНКЦИЯ ********.
*/*!KO_ligamma(arg%ubound%name)*/*.
*Version 1.
*Выдает значение нижней неполной гамма функции для аргумента ARG и верхнего предела
*интегрирования UBOUND.
*ARG - положительное число или числа: скаляр, вектор или матрица.
*UBOUND - неотрицательное число или числа: матрица размером как ARG.
*Выдает матрицу NAME размером как ARG.
*Для этой функции требуется предварительная установка предела числу циклов на 100 (или больше)
*командой SET MXLOOPS.

define !KO_ligamma(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @arg= !2.
comp @ubound= !3.
comp @den= 1.
comp !4= 1.
loop @it= 1 to 100.
-comp @den= @den&*(@arg+@it).
-comp !4= !4+@ubound&**@it/@den.
end loop.
comp !4= @ubound&**@arg/exp(@ubound)/@arg&*!4.
release @arg,@ubound,@den,@it.
!enddefine.


******** QR-РАЗЛОЖЕНИЕ МЕТОДОМ ОТРАЖЕНИЙ ХАУСХОЛДЕРА ********.
*/*!KO_qrdc(mat%name1%name2)*/*.
*Version 1.
*Делает QR-разложение матрицы MAT на верхнетреугольную матрицу R (NAME2) и ортонормированную
*матрицу Q (NAME1) методом отражений Хаусхолдера. Q*R=MAT.
*MAT - матрица с числом столбцов не меньше, чем число рядов.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Встроенная функция GSCH (ортонормирование Грама-Шмидта) - тоже QR-разложение, но использующее другой
*метод и дающее иной результат.


define !KO_qrdc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= nrow(!4).
comp !3= ident(@n).
loop @i= 1 to @n.
-comp @ind= @i:@n.
-comp @v= !4(@i:@n,@i).
-comp @v(1)= @v(1)+((@v(1)>=0)-(@v(1)<0))*sqrt(mssq(@v)).
-comp @h= mssq(@v).
-comp @h= ident(nrow(@v))-2*sscp(t(@v))/(@h+(@h=0)).
-comp !4(@ind,@i:ncol(!4))= @h*!4(@ind,@i:ncol(!4)).
-comp !3(:,@ind)= !3(:,@ind)*@h.
end loop.
release @n,@i,@ind,@v,@h.
!enddefine.


******** ХЕССЕНБЕРГОВА МАТРИЦА / ТРИДИАГОНАЛЬНАЯ МАТРИЦА ********.
*/*!KO_hess(mat%name1%name2)*/*.
*Version 1.
*Превращает квадратную матрицу MAT (размером не меньше 3) в верхнюю матрицу Хессенберга NAME2.
*Эта матрица верхнетреугольная с заполненной также верхней поддиагональю, и ее собственные
*числа те же, что у MAT. Выдается также ортонормированная матрица NAME1, такая,
*что NAME1*MAT*t(NAME1)=NAME2 и t(NAME1)*NAME2*NAME1=MAT.
*Если MAT симметричная матрица, то NAME2 получается тридиагональной.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_hess(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= ncol(!4).
comp !3= ident(@n,@n).
loop @i= 1 to @n-2.
-comp @ind= (@i+1):@n.
-comp @v= !4(@ind,@i).
-comp @v1= ((@v(1)<0)-(@v(1)>=0))*sqrt(mssq(@v)).
-comp @u1= sqrt((@v1-@v(1))/(2*@v1+(@v1=0))).
-comp @u= -@v/(2*@u1*@v1+(@v1=0)).
-comp @u(1)= @u1.
-comp @h= ident(nrow(@u))-2*sscp(t(@u)).
-comp !4(@ind,@i:@n)= @h*!4(@ind,@i:@n).
-comp !4(:,@ind)= !4(:,@ind)*@h.
-comp !3(@ind,:)= @h*!3(@ind,:).
end loop.
release @n,@i,@ind,@v,@v1,@u,@u1,@h.
!enddefine.


******** СОБСТВЕННОЧИСЛОВОЕ РАЗЛОЖЕНИЕ 2x2 МАТРИЦЫ ********.
*/*!KO_eig2x2(mat%name1%name2)*/*.
*Version 1.
*Выдает (вещественные) собственные числа (NAME2) и собственные векторы (столбцы NAME1) квадратной
*матрицы MAT размером 2x2. MAT может быть симметричной или асимметричной; в последнем случае
*собственные векторы могут быть неортогональны (используйте ф-цию /*!KO_schur2x2*/ для получения
*ортогональных собственных векторов).
*Восстановление матрицы: MAT = NAME1*mdiag(NAME2)*inv(NAME1).
*Если собственные числа MAT не вещественные, NAME1 и NAME2 выдаются как нулевые скаляры.

define !KO_eig2x2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @t= @mat(1,1)+@mat(2,2).
comp @d= @mat(1,1)*@mat(2,2)-@mat(1,2)*@mat(2,1).
comp @r= @t**2-4*@d.
do if @r>=0.
-comp !4= make(2,1,0).
-comp @r= sqrt(@r).
-comp !4(1)= (@t+@r)/2.
-comp !4(2)= (@t-@r)/2.
-comp !3= make(2,2,1).
-do if @mat(1,2)=0 and @mat(2,1)=0.
- comp !3= ident(2).
- do if @mat(1,1)<@mat(2,2).
-  comp !3= !3(:,2:1).
- end if.
-else if @mat(1,2)=0.
- comp !3(1,1)= (!4(1)-@mat(2,2))/@mat(2,1).
- comp !3(1,2)= (!4(2)-@mat(2,2))/@mat(2,1).
-else.
- comp !3(2,1)= (!4(1)-@mat(1,1))/@mat(1,2).
- comp !3(2,2)= (!4(2)-@mat(1,1))/@mat(1,2).
-end if.
-comp !3= !3/sqrt(cssq(!3))({1,1},:).
else.
-comp !4= 0.
-comp !3= 0.
end if.
release @mat,@t,@d,@r.
!enddefine.


******** СОБСТВЕННОЧИСЛОВОЕ РАЗЛОЖЕНИЕ 2x2 МАТРИЦЫ (ВАРИАНТ ШУРА) ********.
*/*!KO_schur2x2(mat%name1%name2)*/*.
*Version 1.
*Выдает (вещественные) собственные числа и собственные векторы квадратной матрицы MAT размером 2x2.
*Собственные числа находятся на диагонали верхнетреугольной матрицы NAME2, а собственные векторы
*это столбцы NAME1. Это ортогональные собственные векторы и NAME1 это унитарная матрица.
*Восстановление матрицы: MAT = NAME1*NAME2*t(NAME1).
*Если MAT симметрична, то NAME2 диагональная матрица.
*Если собственные числа MAT не вещественные, NAME1 и NAME2 выдаются как нулевые скаляры.
*Для шуровского разложения матрицы большего размера используйте /*!KO_qreig*/.

define !KO_schur2x2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
do if @mat(2,1)=0.
-comp !4= @mat.
-comp !3= ident(2).
else.
-comp @t= @mat(1,1)+@mat(2,2).
-comp @d= @mat(1,1)*@mat(2,2)-@mat(1,2)*@mat(2,1).
-comp @r= @t**2-4*@d.
-do if @r>=0.
- comp !4= (@t+sqrt(@r))/2.
- comp !3= make(2,1,1).
- do if @mat(1,2)=0.
-  comp !3(1,1)= (!4-@mat(2,2))/@mat(2,1).
- else.
-  comp !3(2,1)= (!4-@mat(1,1))/@mat(1,2).
- end if.
- comp !3= !3/sqrt(cssq(!3)).
- comp !3= {!3,{-!3(2);!3(1)}}.
- comp !4= t(!3)*@mat*!3.
-else.
- comp !4= 0.
- comp !3= 0.
-end if.
-release @t,@d,@r.
end if.
release @mat.
!enddefine.


******** СОБСТВЕННОЧИСЛОВОЕ РАЗЛОЖЕНИЕ (ВАРИАНТ ШУРА) ********.
*/*!KO_qreig(mat%needvec%conv%name1%name2%name3)*/*.
*Version 1.
*Делает собственночисловое разложение квадратной матрицы MAT (симметричной или асимметричной).
*Используемый алгоритм - Implicit QR iterations on recursively splitted Hessenberg matrix with
*double (Francis) shifts. Функция выдает верхнетреугольную матрицу NAME2 (с собственными числами на
*диагонали) и унитарную матрицу NAME1, столбцы которой - взаимоортогональные собственные векторы.
*Порядок собственных векторов отвечает порядку собственных чисел, но собственные числа не обязательно
*сортированы по величине. Если MAT симметричная, то NAME2 диагональна.
*Восстановление матрицы: MAT = NAME1*NAME2*t(NAME1).
*Аргумент NEEDVEC - цифра (не имя и не выражение) 1 или 0 (цифру можно взять в кавычки или апострофы).
*Если "1", то функция выдаст собственные числа NAME2 и собственные векторы NAME1. Если "0", то
*функция выдаст только собственные числа NAME2, не тратя время на получение собственных векторов
*(NAME1 выйдет как скаляр 1).
*Аргумент CONV - скаляр; это задаваемая точность схождения при вычислении собственных чисел;
*рекомендуемое значение: от 1E-8 до 1E-12.
*NAME3 - этот скаляр на выходе есть максимальное потребовавшееся число итераций для вычисления одного
*собственного числа.
*Несхождение/неудача. Асимметричные матрицы MAT иногда имеют комплексные (невещественные) собственные
*числа, связанные в пары. Если некий поддиагональный элемент в NAME2 оказался ненулевой, а NAME3 не
*более 100 (предельное число итераций в функции), то налицо пара комплексных собственных чисел.
*Вещественные части этих двух значений находятся над и справа от несошедшегося к нулю поддиагонального
*элемента, однако они не обязательно вычислены точно. Если поддиагональный элемент в NAME2 оказался
*ненулевой и NAME3=101 (т.е. исчерпаны все итерации), то налицо неудача алгоритма - это сравнительно
*редкий случай, связанный с особой дефектностью данной MAT.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если MAT размером 2x2, функция /*!KO_schur2x2*/ делает то же разложение, что и эта функция.
*Если MAT симметрична, встроенная функция CALL EIGEN делает то же разложение, что и эта функция.

define !KO_qreig(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend(')'))
comp !6= !2.
!let !q= !unquote(!3)
comp @conv= !4.
comp @n= ncol(!6).
!if (!q) !then
-comp !5= ident(@n).
!else
-comp !5= 1.
!ifend
do if @n>2.
-loop @i= 1 to @n-2.
- comp @ind= (@i+1):@n.
- comp @v= !6(@ind,@i).
- comp @v1= ((@v(1)<0)-(@v(1)>=0))*sqrt(mssq(@v)).
- comp @u1= sqrt((@v1-@v(1))/(2*@v1+(@v1=0))).
- comp @u= -@v/(2*@u1*@v1+(@v1=0)).
- comp @u(1)= @u1.
- comp @h= ident(nrow(@u))-2*sscp(t(@u)).
- comp !6(@ind,@i:@n)= @h*!6(@ind,@i:@n).
- comp !6(:,@ind)= !6(:,@ind)*@h.
  !if (!q) !then
-  comp !5(@ind,:)= @h*!5(@ind,:).
  !ifend
-end loop.
-release @v,@v1,@u,@u1,@h.
else.
-comp @v= 0.
-comp @l= 0.
end if.
!if (!q) !then
-comp !5= t(!5).
!ifend
comp @stv= make(1,@n,0).
comp @env= make(1,@n,0).
comp @stv(1)= 1.
comp @env(1)= @n.
comp !7= 1.
comp @wr= 1.
loop @re= 1 to @n if @stv(@re).
-comp @st= @stv(@re).
-comp @en= @env(@re).
-loop @m= 1 to @n if @st<@en.
- do if @en-@st=1.
-  do if abs(!6(@en,@st))<=@conv*rmax(abs({!6(@st,@st),!6(@en,@en)})).
-   comp !6(@en,@st)= 0.
-  else.
-   comp @l= !6(@st,@st)+!6(@en,@en).
-   comp @v= !6(@st,@st)*!6(@en,@en)-!6(@st,@en)*!6(@en,@st).
-   comp @v= @l**2-4*@v.
-   do if @v>=0.
-    comp @l= (@l+sqrt(@v))/2.
-    comp @v= make(2,1,1).
-    do if !6(@st,@en)=0.
-     comp @v(1,1)= (@l-!6(@en,@en))/!6(@en,@st).
-    else.
-     comp @v(2,1)= (@l-!6(@st,@st))/!6(@st,@en).
-    end if.
-    comp @v= @v/sqrt(mssq(@v)).
-    comp @v= {@v,{-@v(2);@v(1)}}.
-    comp !6(@st:@en,:)= t(@v)*!6(@st:@en,:).
-    comp !6(:,@st:@en)= !6(:,@st:@en)*@v.
     !if (!q) !then
-     comp !5(:,@st:@en)= !5(:,@st:@en)*@v.
     !ifend
-   end if.
-  end if.
-  break.
- else.
-  loop @it= 1 to 100.
-   comp @en1= @en-1.
-   comp @st1= @st+1.
-   loop @i= @st to @en-2.
-    comp @ind= @i:(@i+2).
-    do if @i=@st.
-     comp @k1= !6(@en1,@en1)+!6(@en,@en).
-     comp @k2= !6(@en1,@en1)*!6(@en,@en)-!6(@en1,@en)*!6(@en,@en1).
-     comp @v= {!6(@st,@st)**2-!6(@st,@st)*@k1+@k2+!6(@st,@st1)*!6(@st1,@st);!6(@st1,@st)*(!6(@st,@st)+!6(@st1,@st1)-@k1);!6(@st1,@st)*!6(@st+2,@st1)}.
-    else.
-     comp @v= !6(@ind,@i-1).
-    end if.
-    comp @s= ((@v(1)>=0)-(@v(1)<0))*sqrt(mssq(@v)).
-    do if @s<>0.
-     comp @v(1)= @v(1)+@s.
-     comp @v= @v*(t(@v)/(@v(1)*@s)).
-     comp @v= ident(3)-@v.
-     comp !6(@ind,:)= @v*!6(@ind,:).
-     comp !6(:,@ind)= !6(:,@ind)*@v.
      !if (!q) !then
-      comp !5(:,@ind)= !5(:,@ind)*@v.
      !ifend
-    end if.
-   end loop.
-   comp @b= !6(@en,@en-2).
-   comp @r= !6(@en1,@en-2).
-   comp @h= sqrt(@r**2+@b**2).
-   do if @h.
-    comp @c= abs(@r)/@h.
-    comp @s= ((@r>=0)-(@r<0))*@b/@h.
-    comp @v= {@c,-@s;@s,@c}.
-    comp !6(@en1:@en,:)= t(@v)*!6(@en1:@en,:).
-    comp !6(:,@en1:@en)= !6(:,@en1:@en)*@v.
     !if (!q) !then
-     comp !5(:,@en1:@en)= !5(:,@en1:@en)*@v.
     !ifend
-   end if.
-   comp @dg= abs(diag(!6)).
-   comp @az= abs(diag(!6(@st1:@en,@st:@en1)))<=@conv*rmax({@dg(@st1:@en),@dg(@st:@en1)}).
-   do if any(@az).
-    comp @p= cmax(@az&*t(@st:@en1)).
-    comp !6(@p+1,@p)= 0.
-    break.
-   end if.
-  end loop.
-  comp @wr= @wr+1.
-  comp @stv(@wr)= @st.
-  comp @env(@wr)= @p.
-  comp @st= @p+1.
-  comp !7= rmax({!7,@it}).
- end if.
-end loop.
end loop.
do if @n>2.
-release @i,@ind,@h,@it,@en1,@st1,@k1,@k2,@s,@b,@r,@c,@dg,@az,@p.
end if.
release @conv,@n,@v,@l,@stv,@env,@wr,@re,@st,@en,@m.
!enddefine.


******** ВЕКТОР СРЕДНИХ АРИФМЕТИЧЕСКИХ ********.
*/*!KO_mean(data%name)*/*.
*Version 1.
*Вычисляет средние в столбцах DATA.
*Выдает вектор-ряд NAME.

define !KO_mean(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= csum(!3)/nrow(!3).
!enddefine.


******** ВЕКТОР ДИСПЕРСИЙ ********.
*/*!KO_variance(data%name)*/*.
*Version 1.
*Вычисляет дисперсии в столбцах DATA.
*Выдает вектор-ряд NAME.

define !KO_variance(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= cssq(!3)/(nrow(!3)-1).
!enddefine.


******** ВЕКТОР ДИСПЕРСИЙ (DF=N) ********.
*/*!KO_variance2(data%name)*/*.
*Version 1.
*Вычисляет смещенные (посчитанные на "df=n", а не "df=n-1") дисперсии в столбцах DATA.
*Выдает вектор-ряд NAME.

define !KO_variance2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= cssq(!3)/nrow(!3).
!enddefine.


******** ВЕКТОР АСИММЕТРИЙ ********.
*/*!KO_skewness(data%name)*/*.
*Version 1.
*Вычисляет скошенности в столбцах DATA.
*Выдает вектор-ряд NAME.

define !KO_skewness(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @n= nrow(!3).
comp !3= !3-(csum(!3)/@n)(make(1,@n,1),:).
comp @m2= cssq(!3).
comp @sd= sqrt(@m2/(@n-1)).
comp @m3= csum(!3&*!3&*!3).
comp !3= @n*@m3/((@n-1)*(@n-2)*@sd&**3).
release @n,@m2,@m3,@sd.
!enddefine.


******** ВЕКТОР ЭКСЦЕССОВ ********.
*/*!KO_kurtosis(data%name)*/*.
*Version 1.
*Вычисляет заостренности в столбцах DATA. В DATA должно быть не менее трех рядов.
*Выдает вектор-ряд NAME.

define !KO_kurtosis(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @n= nrow(!3).
comp !3= !3-(csum(!3)/@n)(make(1,@n,1),:).
comp @m2= cssq(!3).
comp @sd= sqrt(@m2/(@n-1)).
comp @m4= csum(!3&*!3&*!3&*!3).
comp !3= (@n*(@n+1)*@m4-3*@m2&**2*(@n-1))/((@n-1)*(@n-2)*(@n-3)*@sd&**4).
release @n,@m2,@m4,@sd.
!enddefine.


******** ВЕКТОР МЕДИАН ********.
*/*!KO_median(data%name)*/*.
*Version 1.
*Вычисляет медианы в столбцах DATA.
*Выдает вектор-ряд NAME.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_median(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
loop @i= 1 to ncol(@data).
-comp @data(grade(@data(:,@i)),@i)= @data(:,@i).
end loop.
do if mod(nrow(@data),2).
-comp !3= @data(nrow(@data)/2+.5,:).
else.
-comp !3= (@data(nrow(@data)/2,:)+@data(nrow(@data)/2+1,:))/2.
end if.
release @data,@i.
!enddefine.


******** ПРОЦЕНТИЛИ ********.
*/*!KO_ptile(data%sort%p%method%name)*/*.
*Version 1.
*Вычисляет заказанные процентили в столбцах DATA.
*Выдает матрицу NAME с рядами, отвечающими процентильным уровням.
*SORT - цифра (не имя и не выражение) 1 или 0 (опционально в кавычках или апострофах). Используйте 0
*лишь в том особом случае, когда значения в каждом столбце DATA уже сортированы по возрастанию; эта
*опция позволяет не тратить время на сортировку.
*P - скаляр или вектор (ряд или столбец), содержащий проценты (процентные уровни, процентильные ранги),
*для которых вам нужно определить значения процентией в переменных (каждом столбце DATA).
*Значения в P должны быть больше 0 и меньше 100. Они не обязаны следовать по возрастающей.
*METHOD - метод вычисления процентилей, заглавное ключевое слово (опционально в кавычках или
*апострофах):
*"WAVER" - Weighted Average Definition 2
*"HAVER" - Weighted Average Definition 1
*"HAVER2" - Weighted Average Definition 1
*"EMPIR" - Empirical Distribution Function
*"AEMPIR" - Empirical Distribution Function with Averaging
*"ROUND" - Observation closest aka Rounded
*Это те же методы, что находятся в SPSS-процедуре EXAMINE (Explore). HAVER и HAVER2 это одно и то же,
*только HAVER2 не вычисляет "слишком высокие" для входящих данных процентили (заменяя их в NAME на
*условное число -999), так же поступают SPSS-процедуры Explore и Frequencies для дефолтного
*метода "HAVERAGE". HAVER вместо "-999" ставит максимальное в переменной значение.
*Наиболее популярные и рекомендуемые методы - HAVER/HAVER2 и AEMPIR. 50-й процентиль этих методов
*совпадает с медианой ф-ции /*!KO_median*/.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_ptile(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !meth= !unquote(!5) !let !sort= !unquote(!3)
comp @data= !2.
comp @p= !4/100.
do if nrow(@p)=1.
-comp @p= t(@p).
end if.
comp !6= make(nrow(@p),ncol(@data),0).
loop @v= 1 to ncol(@data).
-comp @var= @data(:,@v).
 !if (!sort) !then
- comp @var(grade(@var))= @var.
 !ifend
 !if (!meth='ROUND') !then
- comp @t= nrow(@var)*@p+.5.
- comp @i= trunc(@t).
- comp @i= @i+(not @i).
- comp @f= 0.
- comp !6(:,@v)= @var(@i).
 !else !if (!meth='EMPIR' !or !meth='AEMPIR') !then
- comp @t= nrow(@var)*@p.
- comp @i= trunc(@t).
- comp @f= @t<>@i.
  !if (!meth='EMPIR') !then
-  comp !6(:,@v)= @var(@i+@f).
  !else
-  comp !6(:,@v)= (@var(@i+@f)+@var(@i+@f+(not @f)))/2.
  !ifend
 !else
  !if (!meth<>'WAVER') !then !let !p_= '+1' !let !n_= '-(@i=nrow(@var))' !ifend
- comp @t= (nrow(@var) !p_)*@p.
- comp @i= trunc(@t).
- comp @f= mod(@t,1).
- comp !6(:,@v)= (1-@f)&*@var(@i +(@i=0) )+@f&*@var(@i+1 !n_).
  !if (!meth='HAVER2') !then
-  comp @f= @i=nrow(@var) and @f.
-  comp !6(:,@v)= !6(:,@v)&*(not @f)-999&*@f.
  !ifend
 !ifend !ifend
end loop.
release @data,@var,@p,@v,@t,@i,@f.
!enddefine.


******** МИНИМУМ, МАКСИМУМ, СРЕДНЯЯ И ДИСПЕРСИЯ, ПО ГРУППАМ ********.
*/*!KO_gdescr(col%bin%name1%name2%name3%name4)*/*.
*Version 2.
*Берет столбец COL (анализируемая переменная) и одну или более двоичных (1 vs 0)
*переменных, являющихся столбцами матрицы BIN с числом рядов как в COL.
*Каждый столбец BIN означает группу, и значение 1 означает, что наблюдение (ряд) принадлежит ей.
*Группы могут пересекаться или непересекаться составом наблюдений (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей, в этом случае группой
*выступит весь массив, все ряды: вычислится статистика "в целом".
*Функция выдает 4 столбца по-группных однопеременных статистик, с рядами, соответствующими группам, столбцам BIN:
*наименьшее значение (NAME1), наибольшее значение (NAME2), средняя арифметическая (NAME3),
*дисперсия (на "df=n-1", NAME4). NAME4 не будет вычислено (с сообщением об ошибке деления на 0),
*если какой-л столбец BIN содержит менее двух единиц.
*Если ваше группирование задается категориальной переменной, создайте из нее фиктивные двоичные
*переменные BIN с помощью функции /*!KO_freq*/.

define !KO_gdescr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @var= !2.
comp @bin= !3.
comp @n= t(csum(@bin)).
comp @vars= @var*make(1,ncol(@bin),1).
comp !4= t(cmin(@vars&*@bin+(not @bin)*mmax(@var))).
comp !5= t(cmax(@vars&*@bin+(not @bin)*mmin(@var))).
comp @bin= t(@bin).
comp @sum= @bin*@var.
comp !6= @sum/@n.
comp !7= (@bin*(@var&*@var)-@sum&*!6)/(@n-1).
release @var,@bin,@n,@vars,@sum.
!enddefine.


******** РАЗНЫЕ ОДНОПЕРЕМЕННЫЕ СТАТИСТИКИ, ПО ГРУППАМ ********.
*/*!KO_aggr(data%bin%stat%name)*/*.
*Version 1.
*Вычисляет требуемую описательную статистику в столбцах DATA, в целом и/или по группам
*наблюдений (рядов).
*Эта функция напоминает внематричную команду AGGREGATE.
*Берет данные DATA (матрица с любым числом столбов, переменных) и одну или более двоичных (1 vs 0)
*переменных, являющихся столбцами матрицы BIN с числом рядов как в DATA.
*Каждый столбец BIN означает группу, и значение 1 означает, что наблюдение (ряд) принадлежит ей.
*Группы могут пересекаться или непересекаться составом наблюдений (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей, в этом случае группой
*выступит весь массив, все ряды: вычислится статистика "в целом".
*Результат NAME - матрица со столбцами, отвечающими столбцам DATA, т.е. переменным, и рядами,
*отвечающими столбцам BIN, т.е. группам. NAME содержит значения вычисленной статистики для каждой
*переменной в каждой группе.
*STAT - требуемая статистика. Это ключевое слово заглавными буквами. Кавычки или апострофы вокруг
*слова - опциональны. Следующие статистики возможны:
*"SUM" - сумма
*"MEAN" - средняя арифметическая
*"SSDEV" - сумма квадратов отклонений
*"VARIANCE" - дисперсия [сумма в каждом столбце BIN должна быть >1]
*"VARIANCE2" - дисперсия, посчитанная на "df=n", а не "df=n-1"
*"SEMEAN" - ст. ошибка средней арифметической [сумма в каждом столбце BIN должна быть >1]
*"MIN" - наименьшее значение
*"MAX" - наибольшее значение
*"FIRST" - первое значение
*"LAST" - последнее значение
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваше группирование задается категориальной переменной, создайте из нее фиктивные двоичные
*переменные BIN с помощью функции /*!KO_freq*/.
*Если хотите заместить вычисленной статистикой значения в DATA, используйте функцию /*!KO_aggrv*/.

define !KO_aggr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @bin= !3.
!let !stat= !unquote(!4)
!if (!index('SUM FIRST LAST MIN MAX',!stat)=0) !then !let !n= ',@n'
-comp @n= csum(@bin).
!ifend
!if (!stat='SUM') !then
-comp !5= t(@bin)*@data.
!else !if (!stat='MEAN') !then
-comp !5= t(@bin)*@data/(t(@n)*make(1,ncol(@data),1)).
!else !if (!stat='SSDEV' !or !stat='VARIANCE' !or !stat='VARIANCE2' !or !stat='SEMEAN') !then
-comp @sum= t(@bin)*@data.
-comp @cssq= t(@bin)*(@data&*@data).
-comp @n= t(@n)*make(1,ncol(@data),1).
-comp !5= @cssq-@sum&*@sum/@n.
 !if (!stat<>'SSDEV') !then
- comp !5= !5/(@n !if (!stat<>'VARIANCE2') !then -1 !ifend).
  !if (!stat='SEMEAN') !then
-  comp !5= sqrt(!5/@n).
  !ifend
 !ifend
-release @sum,@cssq.
!else !if (!stat='MIN') !then
-comp @ones= make(1,ncol(@data),1).
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @bigv= 1E300.
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !5(@i,:)= cmin(@data&*@bini+(not @bini)*@bigv).
-end loop.
-release @ones,@bigv,@i,@bini.
!else !if (!stat='MAX') !then
-comp @ones= make(1,ncol(@data),1).
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @smallv= mmin(@data).
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !5(@i,:)= cmax(@data&*@bini+(not @bini)*@smallv).
-end loop.
-release @ones,@smallv,@i,@bini.
!else !if (!stat='LAST') !then
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(1:nrow(@data)).
-loop @i= 1 to ncol(@bin).
- comp @ind= cmax(@bin(:,@i)&*@rnums).
- comp !5(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!else !if (!stat='FIRST') !then
-comp !5= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(nrow(@data):1).
-loop @i= 1 to ncol(@bin).
- comp @ind= nrow(@data)-cmax(@bin(:,@i)&*@rnums)+1.
- comp !5(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!ifend !ifend !ifend !ifend !ifend !ifend !ifend
release @data,@bin !n.
!enddefine.


******** ЗНАЧЕНИЕ, БЛИЖАЙШЕЕ К ЗАДАННОМУ СНИЗУ ИЛИ СВЕРХУ ********.
*/*!KO_closest(data%ineq%val%name)*/*.
*Version 1.
*В каждом столбце данных DATA выясняет наблюдаемое значение, ближайшее к значению VAL по величине;
*а условие INEQ оговаривает, должно оно быть меньше или больше, чем VAL.
*INEQ это оператор неравенства, ключевое слово (можно взять в кавычки или апострофы)
*на выбор: >, >=, <, <= (или буквами: GT, GE, LT, LE). Например, <= означает выяснить наибольшее
*наблюдаемое значение, которое <=VAL, т.е. прилегает к нему "снизу" и либо меньше, либо равно ему.
*VAL - скаляр или вектор-ряд длиной как число столбцов DATA. Вектор означает задание отдельного числа
*для каждого столбца DATA.
*Результат - ряд NAME, содержащий искомые значения. Значение 1E+300 либо 1E-300 в NAME говорит о том,
*что в данном столбце DATA не нашлось значений, удовлетворяющих условию INEQ VAL.

define !KO_closest(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !ineq= !upcase(!unquote(!3))
comp @data= !2.
comp @val= !4.
!if (!index(!ineq,'>')>0 !or !index(!ineq,'G')>0) !then
 !let !plus= '+' !let !max= 'cmin' !else !let !plus= '-' !let !max= 'cmax'
!ifend
do if ncol(@val)>1.
-comp @val= @val(make(1,nrow(@data),1),:).
end if.
comp @flag= @data !ineq @val.
comp !5= !max(@flag&*@data !plus (not @flag)*1E300).
release @data,@val,@flag.
!enddefine.


******** КОВАРИАЦИОННАЯ МАТРИЦА ********.
*/*!KO_cov(data%name)*/*.
*Version 1.
*Вычисляет ковариации между столбцами DATA.
*Выдает квадратную симметрическую матрицу NAME.

define !KO_cov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= sscp(!3)/(nrow(!3)-1).
!enddefine.


******** КОВАРИАЦИОННАЯ МАТРИЦА (DF=N) ********.
*/*!KO_cov2(data%name)*/*.
*Version 1.
*Вычисляет ковариации между столбцами DATA. Значения посчитаны на "df=n" (а не "df=n-1"). 
*Выдает квадратную симметрическую матрицу NAME.

define !KO_cov2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= sscp(!3)/nrow(!3).
!enddefine.


******** КОРРЕЛЯЦИОННАЯ МАТРИЦА ********.
*/*!KO_corr(data%name)*/*.
*Version 1.
*Вычисляет корреляции Пирсона между столбцами DATA.
*Выдает квадратную симметрическую матрицу NAME.

define !KO_corr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:)).
comp @rssdev= sqrt(diag(!3)).
comp !3= !3/(@rssdev*t(@rssdev)).
call setdiag(!3,1).
release @rssdev.
!enddefine.


******** КОСИНУСНАЯ МАТРИЦА ********.
*/*!KO_cosine(data%name)*/*.
*Version 1.
*Вычисляет косинусы-сходства (коэффициент пропорциональности, или конгруентности Такера)
*между столбцами DATA.
*Выдает квадратную симметрическую матрицу NAME.

define !KO_cosine(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3).
comp @rss= sqrt(diag(!3)).
comp !3= !3/(@rss*t(@rss)).
call setdiag(!3,1).
release @rss.
!enddefine.


******** МАТРИЦА КОЭФФИЦИЕНТОВ ТОЖДЕСТВЕННОСТИ ********.
*/*!KO_idc(data%name)*/*.
*Version 1.
*Вычисляет сходства-коэффициенты тождественности (идентичности) [Zegers, ten Berge, 1985]
*между столбцами DATA.
*Выдает квадратную симметрическую матрицу NAME.

define !KO_idc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3).
comp @ss2= (diag(!3)/2)*make(1,ncol(!3),1).
comp !3= !3/(@ss2+t(@ss2)).
release @ss2.
!enddefine.


******** МАТРИЦА ОТНОШЕНИЙ ПОДОБИЯ ********.
*/*!KO_simr(data%name)*/*.
*Version 1.
*Вычисляет отношения подобия (similarity ratio), известные также как сходства Kohonen,
*между столбцами DATA.
*Выдает квадратную симметрическую матрицу NAME.

define !KO_simr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= sscp(!3).
comp @ss= diag(!3)*make(1,ncol(!3),1).
comp !3= !3/(@ss+t(@ss)-!3).
release @ss.
!enddefine.


******** КОВАРИАЦИЯ, КОРРЕЛЯЦИЯ, КОСИНУС, КОЭФФИЦИЕНТ ТОЖДЕСТВЕННОСТИ, ОТНОШЕНИЕ ПОДОБИЯ ********.
*/*!KO_biv(col1%col2%name)*/*.
*Version 2.
*Для двух векторов-столбцов COL1 и COL2 (одинаковой длины) вычисляет ряд из пяти статистик сходства:
*коэффициент ковариации (1-й элемент NAME), коэффициент корреляции Пирсона (2-й элемент), коэффициент
*пропорциональности (косинус, 3-й элемент), коэффициент тождественности (4-й элемент),
*отношение подобия (5-й элемент).

define !KO_biv(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @var1= !2.
comp @var2= !3.
comp @vars= {@var1,@var2}.
comp !4= make(1,5,0).
comp @ss= cssq(@vars).
comp @scp= csum(@var1&*@var2).
comp !4(3)= @scp/sqrt(@ss(1)*@ss(2)).
comp !4(4)= 2*@scp/(@ss(1)+@ss(2)).
comp !4(5)= @scp/(@ss(1)+@ss(2)-@scp).
comp @vars= @vars-(csum(@vars)/nrow(@vars))(make(1,nrow(@vars),1),:).
comp @ss= cssq(@vars).
comp @scp= csum(@vars(:,1)&*@vars(:,2)).
comp !4(1)= @scp/(nrow(@vars)-1)).
comp !4(2)= @scp/sqrt(@ss(1)*@ss(2)).
release @var1,@var2,@vars,@ss,@scp.
!enddefine.


******** ПРЯМОУГОЛЬНАЯ МАТРИЦА КОВАРИАЦИЙ/КОРРЕЛЯЦИЙ/КОСИНУСОВ ********.
*/*!KO_rect(vars1%vars2%type%name)*/*.
*Version 1.
*Вычисляет заказанный коэффициент ассоциации между столбцами VARS1 и столбцами VARS2.
*Эти две матрицы должны иметь одинаковое число рядов (ряды это наблюдения, а столбцы это переменные).
*Выдает прямоугольную матрицу NAME размером число_столбцов_VARS1 x число_столбцов_VARS2.
*TYPE - заказанная мера связи, заглавное кл. слово (опционально можно в кавычках или апострофах):
*"COV" - коэффициент ковариации (на "df=n-1")
*"COV2" - коэффициент ковариации (на "df=n")
*"CORR" - коэффициент корреляции Пирсона
*"COSINE" - косинус-сходство (коэффициент конгруэнтности Такера).

define !KO_rect(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !type= !unquote(!4)
comp @v1= !2.
comp @v2= !3.
comp @n= nrow(@v1).
!if (!type<>'COSINE') !then
-comp @v1= @v1-(csum(@v1)/@n)(make(1,@n,1),:).
-comp @v2= @v2-(csum(@v2)/@n)(make(1,@n,1),:).
!ifend
comp !5= t(@v1)*@v2.
!if (!type='COV') !then
-comp !5= !5/(@n-1).
!else !if (!type='COV2') !then
-comp !5= !5/@n.
!else
-comp @v1= sqrt(cssq(@v1)).
-comp @v2= sqrt(cssq(@v2)).
-comp !5= !5/(t(@v1)*(@v2)).
!ifend !ifend
release @v1,@v2,@n.
!enddefine.


******** КОВАРИАЦИОННАЯ/КОРРЕЛЯЦИОННАЯ МАТРИЦА, ПО ГРУППАМ И УСРЕДНЕННАЯ ********.
*/*!KO_gcov(data%bin%type%out%name)*/*.
*Version 2.
*Вычисляет рассеянную (scatter), ковариационную или корреляционную матрицу переменных,
*по группам наблюдений, а также усредненную (объединенную, pooled).
*Берет данные DATA (матрица со столбцами-переменными) и одну или более двоичных (1 vs 0)
*переменных, являющихся столбцами матрицы BIN с числом рядов как в DATA.
*Каждый столбец BIN означает группу, и значение 1 означает, что наблюдение (ряд) принадлежит ей.
*Группы могут пересекаться или непересекаться составом наблюдений (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей, в этом случае группой
*выступит весь массив, все ряды: вычислится матрица для всей выборки в целом.
*Аргументы TYPE и OUT это ключевое слово заглавными буквами. Кавычки или апострофы вокруг
*слова - опциональны.
*TYPE - требуемый тип матрицы. Следующие типы возможны:
*"SCAT" - матрица рассеяния
*"COV" - ковариационная матрица [внимание: сумма во всяком столбце BIN должна быть >1]
*"COV2" - ковариационная матрица, посчитанная на "df=n", а не "df=n-1"
*"CORR" - корреляционная матрица [внимание: сумма во всяком столбце BIN должна быть >1].
*OUT - этот аргумент определяет состав выдаваемого результата NAME:
*"BYGR" - матрицы, своя для каждой группы, столбца BIN; матрицы соединены в стопку одна под другой,
*их порядок отвечает порядку столбцов BIN
*"POOL" - объединенная (pooled) матрица групп
*"BOTH" - и то и другое (pooled матрица в стопке будет верхней).
*Что есть объединенная матрица в зависимости от типа матрицы:
*- объединенная матрица рассеяния это суммированная матрица рассеяния групп;
*- объединенная ковариационная матрица это взвешенно усредненная ковариационная матрица групп,
*где весом выступает число степеней свободы ("n-1" при COV и "n" при COV2, где n численность в группе);
*- объединенная корреляционная матрица получается из объединенной ковариационной матрицы
*(или из объединенной матрицы рассеяния), а не усреднением корреляционных матриц групп.
*При вычислении объединенной матрицы любого типа функция за совокупную численность выборки принимает
*сумму численностей в группах, т е число единиц в BIN, - а не число рядов BIN (ведь в BIN некоторые
*ряды могут быть пусты или могут входить более чем в одну группу).
*Поскольку, если DATA имеет лишь один столбец, COV (или COV2) матрица обращается в скаляр, дисперсию,
*то вы можете применять данную функцию и для вычисления усредненных дисперсий групп.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваше группирование задается категориальной переменной, создайте из нее фиктивные двоичные
*переменные BIN с помощью функции /*!KO_freq*/.

define !KO_gcov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @bin= !3.
!let !type= !unquote(!4)
!let !out= !unquote(!5)
comp @n= csum(@bin).
comp !6= make(!if (!out='BYGR') !then 1 !else ncol(@data) !ifend,ncol(@data),0).
comp @m= @bin.
loop @i= 2 to nrow(@m).
-comp @m(@i,:)= @m(@i-1,:)+@m(@i,:).
end loop.
comp @m= @m&*@bin.
comp @inds= make(nrow(@m)+1,ncol(@m),0).
comp @nums= t(1:nrow(@m)).
comp @m= @m+(not @m)*nrow(@inds).
loop @i= 1 to ncol(@m).
-comp @inds(@m(:,@i),@i)= @nums.
end loop.
comp @inds= @inds(1:(nrow(@inds)-1),:).
release @m,@nums.
loop @i= 1 to ncol(@bin).
-comp @ni= @n(@i).
-comp @datai= @data(@inds(1:@ni,@i),:).
-comp @sumi= csum(@datai).
-comp @datai= (sscp(@datai)-t(@sumi)*(@sumi/@ni)).
 !if (!out<>'BYGR') !then
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)+@datai.
 !ifend
 !if (!out<>'POOL') !then
  !if (!type='COV') !then
-  comp @datai= @datai/(@ni-1).
  !else !if (!type='COV2') !then
-  comp @datai= @datai/@ni.
  !else !if (!type='CORR') !then
-  comp @sumi= diag(@datai).
-  comp @datai= @datai/sqrt(@sumi*t(@sumi)).
  !ifend !ifend !ifend
- comp !6= {!6;@datai}.
 !ifend
end loop.
!if (!out<>'BYGR' !and !type<>'SCAT') !then
 !if (!type='COV') !then
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)/(rsum(@n)-ncol(@bin)).
 !else !if (!type='COV2') !then
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)/rsum(@n).
 !else !if (!type='CORR') !then
- comp @sumi= diag(!6).
- comp !6(1:ncol(@data),:)= !6(1:ncol(@data),:)/sqrt(@sumi*t(@sumi)).
 !ifend !ifend !ifend
!ifend
!if (!out='BYGR') !then
-comp !6= !6(2:nrow(!6),:).
!ifend
release @data,@bin,@n,@i,@inds,@ni,@sumi,@datai.
!enddefine.


******** МЕЖГРУППОВАЯ И ОБЪЕДИНЕННАЯ ВНУТРИГРУППОВАЯ МАТРИЦЫ РАССЕЯНИЯ ********.
*/*!KO_bwscat(data%dummy%name1%name2)*/*.
*Version 1.
*Берет данные DATA и группы DUMMY и выдает межгрупповую матрицу рассеяния (between-group
*scatter matrix) NAME1 и объединенную (суммативную) внутригрупповую матрицу рассеяния
*(pooled within-group scatter matrix) NAME2.
*(SSbetween = trace(NAME1) и SSwithin = trace(NAME1).)
*DATA - данные с любым числом столбцов (переменных).
*NAME1 (внимание!) зависит от того, центрованы или нет
*были столбцы DATA, но NAME2 не зависит (за исключением вычислительной устойчивости, ради лучше
*при центрованных данных).
*DUMMY - двоичные фиктивные переменные, помечающие группы (каждая переменная соответствует группе,
*значение 1 = наблюдение принадлежит ей, 0 = не принадлежит ей).
*Всякий столбец DUMMY должен содержать по меньшей мере одну единицу.
*Группы должны быть непересекающиеся составом: сумма в рядах DUMMY не может превосходить единицу.
*Если в DUMMY есть ряды, полные нулей (наблюдения вне групп), то NAME2 не будет корректна.
*Выдаваемые матрицы складываются в sscp-матрицу переменных всего массива: NAME1+NAME2=sscp(DATA).
*Фиктивные переменные DUMMY вы можете создать из категориальной группирующей переменной
*с помощью функции /*!KO_freq*/.
*Если вам нужна объединенная внутригрупповая матрица рассеяния от групп, пересекающихся составом,
*используйте ф-цию /*!KO_gcov*/.

define !KO_bwscat(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @des= !3.
comp @n= csum(@des).
comp !4= t(@des)*@data/(t(@n)*make(1,ncol(@data),1)).
comp !4= sscp(@des*!4).
comp !5= sscp(@data)-!4.
release @data,@des,@n.
!enddefine.


******** МАТРИЦА ОБРАЗОВ ИЛИ АНТИОБРАЗОВ ********.
*/*!KO_image(cov%type%name)*/*.
*Version 1.
*Берет ковариационную или корреляционную матрицу COV и выдает как NAME ковариационную
*матрицу образов (если TYPE положительный скаляр) или антиобразов (если TYPE неположительный скаляр).
*COV должна быть квадратной симметрической невырожденной, в норме - положительно определенной.
*TYPE - скаляр.
*На диагонали матрицы образов находятся "образы", равные Rsq*Ssq, где Rsq это квадрат множественной
*корреляции для данной переменной как зависимой от остальных, а Ssq диагональный элемент COV.
*Внедиагональные элементы матрицы антиобразов это (с обратным знаком) коэффициенты частной ковариации.
*Если обратить знак внедиагональных элементов матрицы антиобразов и сложить ее с матрицей образов,
*восстановится исходная матрица COV.
*Эту функцию можно использовать для любой матрицы sscp-типа: корреляционной, ковариационной, косинусной,
*сырой sscp.

define !KO_image(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cov= !2.
comp @type= !3.
comp @icov= inv(@cov).
comp @diag= diag(@icov).
comp !4= @icov/(@diag*t(@diag)).
comp !4= (!4+t(!4))/2.
do if @type.
-call setdiag(!4,-diag(!4)).
-comp !4= !4+@cov.
end if.
release @cov,@type,@icov,@diag.
!enddefine.


******** МАТРИЦА ВЛИЯНИЯ (HAT-МАТРИЦА) ********.
*/*!KO_hat(data%name)*/*.
*Version 1.
*Берет матрицу данных DATA, которая должна быть с неколлинеарными столбцами.
*При необходимости прежде центруйте или стандартизуйте столбцы DATA.
*Выдает матрицу влияния (hat-матрицу, проекционную матрицу) NAME.
*Значения на ее диагонали известны как рычаги (leverages).
*Если DATA (имеющая n рядов) имеет центрованные столбцы, то NAME*(n-1) есть двойной
*центрат (см. ф-цию /*!KO_dcenter*/) матрицы квадратных расстояний Махаланобиса между рядами DATA,
*а leverages*(n-1) есть квадратные расстояния Махаланобиса до центроида DATA.

define !KO_hat(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3*solve(sscp(!3),t(!3)).
comp !3= (!3+t(!3))/2.
!enddefine.


******** МАТРИЦЫ СОЧАСТОТ ДЛЯ ДВОИЧНЫХ ДАННЫХ ********.
*/*!KO_bincnt(data%name1%name2)*/*.
*Version 1.
*Берет матрицу двоичных (значения 0 и 1) данных DATA и выдает две квадратные матрицы с сочастотами
*между столбцами DATA:
*NAME1 - симметричная матрица, содержащая в ячейке ij число рядов DATA, в которых столбец i
*содержит 1 и столбец j содержит 1. Эти частоты известны как частоты A (признак есть в обоих столбцах).
*NAME2 - асимметричная матрица, содержащия в ячейке ij число рядов DATA, в которых столбец i
*содержит 0 и столбец j содержит 1. Эти частоты известны как частоты B (признак есть в одном и
*отсутствует в другом столбце).
*Частоты C есть транспонат B.
*Частоты D (признака нет в обоих столбцах) = число_рядов_DATA-A-B-C.
*Частоты A, B, C, D полезны для вычисления разнообразных мер близости, придуманных для двоичных данных.

define !KO_bincnt(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp !3= sscp(!4).
comp !4= csum(!4)(make(1,ncol(!4),1),:)-!3.
!enddefine.


******** МАТРИЦА КВАДРАТНЫХ ЕВКЛИДОВЫХ РАССТОЯНИЙ ********.
*/*!KO_seuclid(data%name)*/*.
*Version 2.
*Вычисляет квадратные евклидовы расстояния между рядами DATA.
*Выдает квадратную симметрическую матрицу NAME.
*Если число столбцов DATA >1, функция вычисляет расстояния через скалярные произведения;
*это более быстрый, но калькуляторно потенциально менее точный способ, чем "прямой" способ
*через возведение разниц в квадрат (если вам нужен "прямой способ", используте более
*общую функцию /*!KO_pwmink*/).

define !KO_seuclid(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
do if ncol(!3)=1.
-comp !3= t(!3)(make(1,nrow(!3),1),:).
-comp !3= !3-t(!3).
-comp !3= !3&*!3.
else.
-comp !3= sscp(t(!3)).
-comp @hsq= t(diag(!3))(make(1,ncol(!3),1),:).
-comp !3= @hsq+t(@hsq)-2*!3.
-comp !3= !3&*(!3>0).
-release @hsq.
end if.
!enddefine.


******** ПРЯМОУГОЛЬНАЯ МАТРИЦА КВАДРАТНЫХ ЕВКЛИДОВЫХ РАССТОЯНИЙ ********.
*/*!KO_seuclidr(cases1%cases2%name)*/*.
*Version 2.
*Вычисляет квадратные евклидовы расстояния между рядами CASES1 и рядами CASES2.
*Эти две матрицы должны иметь одинаковое число столбцов (столбцы это измерения данных).
*Выдает прямоугольную матрицу NAME размером число_рядов_CASES1 x число_рядов_CASES2.
*Если число столбцов данных >1, функция вычисляет расстояния через скалярные произведения;
*это более быстрый, но калькуляторно потенциально менее точный способ, чем "прямой" способ
*через возведение разниц в квадрат.

define !KO_seuclidr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @g1= !2.
comp @g2= !3.
do if ncol(@g1)=1.
-comp !4= (@g1*make(1,nrow(@g2),1))-t(@g2)(make(1,nrow(@g1),1),:).
-comp !4= !4&*!4.
else.
-comp @g2= t(@g2).
-comp !4= @g1*@g2.
-comp !4= rssq(@g1)*make(1,ncol(@g2),1)+cssq(@g2)(make(1,nrow(@g1),1),:)-2*!4.
-comp !4= !4&*(!4>0).
end if.
release @g1,@g2.
!enddefine.


******** МАТРИЦА СТЕПЕННЫХ ВЗВЕШЕННЫХ РАССТОЯНИЙ МИНКОВСКОГО ********.
*/*!KO_pwmink(data%power%weights%name)*/*.
*Version 1.
*Вычисляет расстояния Минковского (возведенные в степень POWER) между рядами DATA.
*Выдает квадратную симметрическую матрицу NAME.
*POWER - скаляр (типично >=1); это степень для расстояния.
*WEIGHTS - вектор (ряд или столбец) длиной как число столбцов DATA, содержащий положительные числа;
*это веса для столбцов DATA. Только соотношение величин весов важно, т.к. функция нормирует их.
*Если все веса одинаковы, получится невзвешенное расстояние.
*Расстояние между рядами a и b =  Sum(w_i*|a_i-b_i|^p), где a_i, b_i значения в столбце i,
*p степень, w_i вес столбца i, Sum сумма по всем столбцам.
*Если POWER=2, функция выдаст (взвешенные) квадратные евклидовы расстояния.
*Если POWER=1, функция выдаст (взвешенные) манхэттенские расстояния.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_pwmink(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @p= !3.
comp @w= !4.
comp @w= @w/msum(@w)*rmax({nrow(@w),ncol(@w)}).
comp !5= make(nrow(@data),nrow(@data),0).
comp @ones= make(1,nrow(@data),1).
loop @i= 1 to ncol(@data).
-comp @kroni= @data(:,@i)*@ones.
-comp !5= !5+@w(@i)&*abs(@kroni-t(@kroni))&**@p.
end loop.
release @data,@p,@w,@kroni,@i,@ones.
!enddefine.


******** ПРЯМОУГОЛЬНАЯ МАТРИЦА СТЕПЕННЫХ ВЗВЕШЕННЫХ РАССТОЯНИЙ МИНКОВСКОГО ********.
*/*!KO_pwminkr(cases1%cases2%power%weights%name)*/*.
*Version 1.
*Вычисляет расстояния Минковского (возведенные в степень POWER) между рядами CASES1 и рядами CASES2.
*Эти две матрицы должны иметь одинаковое число столбцов (столбцы это измерения данных).
*Выдает прямоугольную матрицу NAME размером число_рядов_CASES1 x число_рядов_CASES2.
*Аргументы POWER и WEIGHTS - то же, что в функции /*!KO_pwmink*/.
*Если POWER=2, функция выдаст (взвешенные) квадратные евклидовы расстояния.
*Если POWER=1, функция выдаст (взвешенные) манхэттенские расстояния.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_pwminkr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @g1= !2.
comp @g2= !3.
comp @p= !4.
comp @w= !5.
comp @w= @w/msum(@w)*rmax({nrow(@w),ncol(@w)}).
comp !6= make(nrow(@g1),nrow(@g2),0).
comp @ones1= make(1,nrow(@g1),1).
comp @ones2= make(1,nrow(@g2),1).
loop @i= 1 to ncol(@g1).
-comp @kroni1= @g1(:,@i)*@ones2.
-comp @kroni2= t(@g2(:,@i))(@ones1,:).
-comp !6= !6+@w(@i)&*abs(@kroni1-@kroni2)&**@p.
end loop.
release @g1,@g2,@p,@w,@kroni1,@kroni2,@i,@ones1,@ones2.
!enddefine.


******** МАТРИЦА КВАДРАТНЫХ РАССТОЯНИЙ МАХАЛАНОБИСА ********.
*/*!KO_smahal(data%cov%name)*/*.
*Version 1.
*Вычисляет квадратные расстояния Махаланобиса между рядами DATA (матрица n x p, p>=1).
*Выдает квадратную симметрическую матрицу NAME.
*Аргумент COV - либо квадратная p x p симметричная невырожденная матрица, являющаяся
*пользовательской ковариационной матрицей между переменными, либо скаляр 0 - тогда матрица COV
*будет принята равной наблюдаемой матрице ковариаций между p переменными DATA. Тогда DATA должна
*иметь больше рядов, чем столбцов, и столбцы должны быть неколлинеарны.
*Формула расстояния: d^2 = (x-y)'inv(COV)(x-y).
*Если ковариационная матрица - единичная матрица, то результат - квадратные евклидовы расстояния.
*Если вводимая COV это корреляционная матрица, получатся квадратные махаланобисовы расстояния
*без масштабирования.
*Если вы имеете пользовательскую inv(COV) вместо COV, то можете использовать /*!KO_sbutler*/
*для получения махаланобисовых расстояний.

define !KO_smahal(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @cov= !3.
do if nrow(@cov)*ncol(@cov)=1 and @cov(1,1)=0.
-comp !4= !4-(csum(!4)/nrow(!4))(make(1,nrow(!4),1),:).
-comp !4= !4*solve(sscp(!4),t(!4))*(nrow(!4)-1).
else.
-comp !4= !4*inv(@cov)*t(!4).
end if.
comp !4= (!4+t(!4))/2.
comp @hsq= diag(!4)*make(1,ncol(!4),1).
comp !4= @hsq+t(@hsq)-2*!4.
comp !4= !4&*(!4>0).
release @cov,@hsq.
!enddefine.


******** ВЕКТОР КВАДРАТНЫХ РАССТОЯНИЙ МАХАЛАНОБИСА ДО ЦЕНТРОИДА ********.
*/*!KO_smahalc(data%mean%cov%name)*/*.
*Version 1.
*Вычисляет квадратные расстояния Махаланобиса между рядами DATA и центроидом.
*Выдает вектор-столбец NAME с числом рядов как у DATA.
*DATA должна иметь p (p>=1) столбцов и любое число рядов. Это координаты точек в p-мерном
*пространстве.
*Если вы хотите получить расстояния до центроида самого наблюдаемого облака DATA, укажите
*аргумент COV как нулевой скаляр. DATA тогда должна иметь минимум два ряда и неколлинеарные
*столбцы. Аргумент MEAN будет проигнорирован.
*Если вы хотите получить расстояния точек DATA до центроида некоторого другого облака в том же
*пространстве, необходимо тогда указать вектор средних MEAN и ковариационную матрицу COV
*того облака: MEAN должна иметь p столбцов и 1 ряд. Это координаты центра того облака. COV должна
*быть p X p, симметричной и невырожденной. Это ковариации в том облаке.
*Если вы хотите получить расстояние между двумя группами из одной популяции, то введите ряд
*DATA = центроид одной группы, ряд MEAN = центроид другой группы, COV = усредненная
*внутригрупповая ковариационная матрица (последнюю вы можете получить ф-цией /*!KO_gcov*/).
*Если COV - единичная матрица, то результат - квадратные евклидовы расстояния от точек DATA до
*центроида MEAN.

define !KO_smahalc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @mean= !3.
comp @cov= !4.
do if nrow(@cov)*ncol(@cov)=1 and @cov(1,1)=0.
-comp @mean= csum(@data)/nrow(@data).
-comp @data= @data-(csum(@data)/nrow(@data))(make(1,nrow(@data),1),:).
-comp !5= @data*solve(sscp(@data)/(nrow(@data)-1),t(@data)).
else.
-comp @data= @data-@mean(make(1,nrow(@data),1),:).
-comp !5= @data*inv(@cov)*t(@data).
end if.
comp !5= diag(!5).
release @data,@mean,@cov.
!enddefine.


******** МАТРИЦА КВАДРАТНЫХ БАТЛЕРОВЫХ РАССТОЯНИЙ ********.
*/*!KO_sbutler(data%cos%name)*/*.
*Version 1.
*Вычисляет квадратные расстояния Батлера между рядами DATA (матрица n x p, p>=1).
*Выдает квадратную симметрическую матрицу NAME.
*"Батлеровы" расстояния - это евклидовы расстояния не в декартовой (ортогональной), а в
*косоугольной системе, где косинусы углов между осями равны корреляциям между переменными или
*заданы пользователем; входящие же данные принимаются за косые (контравариантные) координаты
*в этой системе. Формула: d^2 = (x-y)'COS(x-y).
*Аргумент COS - либо квадратная p x p симметричная матрица косинусов углов между осями
*(это задание углов пользователем), либо скаляр 0 - тогда матрица косинусов углов COS будет
*принята равной наблюдаемой матрице корреляций между p переменными DATA.
*Если матрица косинусов углов - единичная матрица, то результат - квадратные евклидовы расстояния.
*Если в качестве COS ввести обращенную матрицу ковариаций, то результат - квадратные махаланобисовы
*расстояния, а если обращенную матрицу корреляций - то немасштабированные квадратные махаланобисовы
*расстояния.

define !KO_sbutler(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @cos= !3.
do if nrow(@cos)*ncol(@cos)=1 and @cos(1,1)=0.
-comp !4= !4-(csum(!4)/nrow(!4))(make(1,nrow(!4),1),:).
-comp @cos= sscp(!4).
-comp @h= sqrt(diag(@cos)).
-comp @cos= @cos/(@h*t(@h)).
end if.
comp !4= !4*@cos*t(!4).
comp !4= (!4+t(!4))/2.
comp @h= diag(!4)*make(1,ncol(!4),1).
comp !4= @h+t(@h)-2*!4.
comp !4= !4&*(!4>0).
release @cos,@h.
!enddefine.


******** СПИСОК УНИКАЛЬНЫХ ЗНАЧЕНИЙ В ВЕКТОРЕ И ПОЗИЦИЙ ИХ ПЕРВОГО ПОЯВЛЕНИЯ ********.
*/*!KO_unique(vec%ignore%name1%name2)*/*.
*Version 3.
*Берет вектор (столбец или ряд) VEC и выдает вектор-ряд NAME1, являющийся списком
*уникальных значений VEC; значения идут в порядке их первого появления в VEC.
*Позиции первого появления выдаются как соответствующий ряд NAME2.
*Аргумент IGNORE - цифра (не имя и не выражение, опционально в кавычках или апострофах): если 1,
*значение 0 в VEC игнорируются (считается невалидным); если 0, то нуль считается валидным значением.
*Результат NAME2 в виде скаляра 0 в случае IGNORE=1 означает, что VEC состоял из нулей.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_unique(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @vec= !2.
!let !ignore= !unquote(!3)
!if (!ignore) !then !let !st= 1
-comp !4= 0.
-comp !5= 0.
!else !let !st= 2
-comp !4= @vec(1).
-comp !5= 1.
!ifend
loop @i= !st to rmax({nrow(@vec),ncol(@vec)}).
-do if not any(@vec(@i)=!4).
- comp !4= {!4,@vec(@i)}.
- comp !5= {!5,@i}.
-end if.
end loop.
!if (!ignore) !then
-do if ncol(!4)>1.
- comp !4= !4(2:ncol(!4)).
- comp !5= !5(2:ncol(!5)).
-end if.
!ifend
release @vec,@i.
!enddefine.


******** ЧАСТОТЫ ЗАДАННЫХ ЗНАЧЕНИЙ В СТОЛБЦАХ ********.
*/*!KO_freqval(data%vals%name)*/*.
*Version 1.
*Берет матрицу DATA, состоящую из одного или более столбцов, и список значений VALS (скаляр или вектор),
*и выдает частоту каждого из этих значений в каждом столбце DATA. Результат NAME это матрица этих частот;
*в ней ряды соответствуют значениям VALS и столбцы соответствуют столбцам DATA.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_freqval(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @vals= !3.
comp !4= make(nrow(@vals)*ncol(@vals),ncol(@data),0).
loop @i= 1 to nrow(@vals)*ncol(@vals).
-comp !4(@i,:)= csum(@data=@vals(@i)).
end loop.
release @data,@vals,@i.
!enddefine.


******** ДВУМЕРНАЯ ЧАСТОТНАЯ ПЕРЕКРЕСТНАЯ ТАБЛИЦА ДЛЯ ЗАДАННЫХ ЗНАЧЕНИЙ ********.
*/*!KO_ctabval(rcol%ccol%rvals%cvals%name)*/*.
*Version 1.
*Берет векторы-столбцы RCOL и CCOL, которые должны быть одинаковой длины, и списки значений: RVALS (для RCOL)
*и CVALS (для CCOL). RVALS и CVALS это векторы (или скаляры), их длина может быть разной.
*Выдает частотную кросс-таблицу NAME размером r x c, с рядами, задаваемыми r значениями RVALS и столбцами,
*задаваемыми c значениями CVALS.
*Частоты в таблице - число наблюдений с соответствующим сочетанием значений между RCOL и CCOL.

define !KO_ctabval(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @rcol= !2.
comp @ccol= !3.
comp @rvals= !4.
comp @cvals= !5.
do if ncol(@rvals)>1.
-comp @rvals= t(@rvals).
end if.
do if nrow(@cvals)>1.
-comp @cvals= t(@cvals).
end if.
comp !6= make(nrow(@rcol),1,1).
comp !6= (t(@rcol)(make(nrow(@rvals),1,1),:)=@rvals*t(!6))*(@ccol*make(1,ncol(@cvals),1)=@cvals(!6,:)).
release @rcol,@ccol,@rvals,@cvals.
!enddefine.


******** ФИКТИВНЫЕ ПЕРЕМЕННЫЕ И ЧАСТОТЫ ЗНАЧЕНИЙ В СТОЛБЦЕ ********.
*/*!KO_freq(col%sort%name1%name2%name3)*/*.
*Version 2.
*Берет вектор-столбец COL, который должен не быть константой, и выдает:
*матрицу NAME1 фиктивных (dummy) переменных, по одному столбцу на каждое уникальное значение COL;
*вектор-столбец NAME2 частот уникальных значений COL (это суммы в столбцах NAME1);
*вектор-столбец NAME3 - список уникальных значений COL.
*Аргумент SORT (скаляр) - сортировать ли результаты по возрастанию значений: если положительный,
*то столбцы в NAME1 и ряды в NAME2 и NAME3 будут сортированы в возрастающем порядке уникальных значений COL;
*если же аргумент неположительный, то порядок рядов и столбцов в результатах определяется очередностью
*встречи уникальных значений COL (при просмотре COL сверху вниз).
*Эта функция может требовать от компьютера много оперативной памяти при массивных данных.

define !KO_freq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @var= !2.
comp @sort= !3.
comp !4= design(@var).
comp !5= csum(!4).
comp !6= @var*make(1,ncol(!4),1)&*!4.
comp !6= cmax(!6)+cmin(!6).
do if @sort.
-comp @grd= grade(!6).
-comp !4(:,@grd)= !4.
-comp !5(@grd)= !5.
-comp !6(@grd)= !6.
-release @grd.
end if.
comp !5= t(!5).
comp !6= t(!6).
release @var,@sort.
!enddefine.


******** ДВУМЕРНАЯ ЧАСТОТНАЯ ПЕРЕКРЕСТНАЯ ТАБЛИЦА ********.
*/*!KO_crosstab(rcol%ccol%sort%name1%name2%name3)*/*.
*Version 2.
*Берет векторы-столбцы RCOL и CCOL, которые должны не быть константами и должны быть одинаковой длины.
*Выдает частотную кросс-таблицу NAME1, в которой ряды заданы уникальными значениями RCOL
*и столбцы заданы уникальными значениями CCOL. Списки самих уникальных значений выдаются как вектор-столбец NAME2
*(значения RCOL) и вектор-ряд NAME3 (значения CCOL).
*Аргумент SORT (скаляр) - сортировать ли результаты по возрастанию значений: если положительный,
*то ряды в NAME1 и NAME2 будут сортированы в возрастающем порядке уникальных значений RCOL и столбцы в
*NAME1 и NAME3 будут сортированы в возрастающем порядке уникальных значений CCOL;
*если же аргумент неположительный, то порядок рядов и столбцов в результатах определяется очередностью
*встречи уникальных значений в RCOL и CCOL, соответственно (при просмотре этих столбцов сверху вниз).
*Эта функция может требовать от компьютера много оперативной памяти при массивных данных.

define !KO_crosstab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @rcol= !2.
comp @ccol= !3.
comp @sort= !4.
comp !6= design(@rcol).
comp !7= design(@ccol).
comp !5= t(!6)*!7.
comp !6= @rcol*make(1,ncol(!6),1)&*!6.
comp !6= t(cmax(!6)+cmin(!6)).
comp !7= @ccol*make(1,ncol(!7),1)&*!7.
comp !7= cmax(!7)+cmin(!7).
do if @sort.
-comp @rgrd= grade(!6).
-comp @cgrd= grade(!7).
-comp !5(@rgrd,@cgrd)= !5.
-comp !6(@rgrd)= !6.
-comp !7(@cgrd)= !7.
-release @rgrd,@cgrd.
end if.
release @rcol,@ccol,@sort.
!enddefine.


******** ФИКТИВНЫЕ ПЕРЕМЕННЫЕ ВЗАИМОДЕЙСТВИЯ И МНОГОМЕРНАЯ ЧАСТОТНАЯ АГРЕГАЦИЯ ********.
*/*!KO_aggrtab(cols%sort%empty%name1%name2%name3)*/*.
*Version 1.
*Делает частотную кросстабуляцию для любого числа переменных.
*Выдает результаты в виде агрегированного массива (а не в виде кросс-таблиц).
*(Подобным образом частоты выдает вне-матричная команда AGGREGATE.)
*Функция также выдает двоичный массив фиктивных переменных, соответствующих взаимодействию
*высшего порядка.
*COLS - категориальные данные, где переменные - столбцы; столбцов должно быть минимум два и они
*должны не быть константами.
*Функция выдает частотный вектор-столбец NAME2, где каждый элемент соответствует одному сочетанию значений
*всех переменных, т.е. это частота в ячейке многомерной таблицы, задаваемой переменными.
*Сочетания самих уникальных значений переменных содержатся в NAME3, где ряды отвечают рядам NAME2,
*а столбцы отвечают переменным, столбцам COLS.
*Вектор NAME2 это суммы в столбцах выдаваемого массива NAME1, который есть фиктивные переменные: они
*соответствуют вышеозначенным сочетаниям уникальных значений. Т.е. столбцы NAME1 соответственны рядам
*NAME2 и NAME3.
*Аргумент SORT (скаляр) - сортировать ли результаты по возрастанию значений: если положительный,
*то столбцы в NAME1 и ряды в NAME2 и NAME3 будут сортированы в возрастающем порядке уникальных значений
*переменных; если же аргумент неположительный, то упомянутый порядок в результатах определяется очередностью
*встречи уникальных значений в данных (при просмотре столбцов COLS сверху вниз).
*Аргумент EMPTY (скаляр) - если положительный, из результатов будут выпущены столбцы (в NAME1) и ряды
*(в NAME2 и NAME3) с частотой 0 в NAME1. Если же аргумент неположительный, результаты будут
*выданы полными.
*Эта функция может требовать от компьютера много оперативной памяти при массивных данных.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_aggrtab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cols= !2.
comp @sort= !3.
comp @empty= !4.
comp @grd= 0.
comp !5= design(@cols(:,ncol(@cols))).
comp !7= @cols(:,ncol(@cols))*make(1,ncol(!5),1)&*!5.
comp !7= cmax(!7)+cmin(!7).
do if @sort.
-comp @grd= grade(!7).
-comp !5(:,@grd)= !5.
-comp !7(@grd)= !7.
end if.
loop @i= ncol(@cols)-1 to 1 by -1.
-comp @desi= design(@cols(:,@i)).
-comp @codesi= @cols(:,@i)*make(1,ncol(@desi),1)&*@desi.
-comp @codesi= cmax(@codesi)+cmin(@codesi).
-do if @sort.
- comp @grd= grade(@codesi).
- comp @desi(:,@grd)= @desi.
- comp @codesi(@grd)= @codesi.
-end if.
-comp !5= kroneker(@desi,make(1,ncol(!5),1))&*kroneker(make(1,ncol(@desi),1),!5).
-comp !7= {kroneker(@codesi,make(1,ncol(!7),1));kroneker(make(1,ncol(@codesi),1),!7)}.
end loop.
comp !6= t(csum(!5)).
comp !7= t(!7).
do if @empty.
-comp @nz= !6>0.
-comp @nnz= csum(@nz).
-do if @nnz<nrow(@nz).
- do if @nnz/nrow(@nz)>.8.
-  comp @grd= make(1,@nnz+1,0).
-  comp @c= 1.
-  loop @i= 1 to nrow(@nz).
-   comp @grd(@c)= @i.
-   comp @c= @c+@nz(@i).
-  end loop.
-  comp @grd= @grd(1:@nnz).
-  release @c.
- else if @nnz/nrow(@nz)<.08.
-  comp @grd= make(1,@nnz,0).
-  comp @s= 1.
-  loop @c= 1 to @nnz.
-   loop @i= @s to nrow(@nz).
-   end loop if @nz(@i).
-   comp @grd(@c)= @i.
-   comp @s= @i+1.
-  end loop.
-  release @s,@c.
- else.
-  comp @grd= make(1,@nnz,0).
-  comp @c= 0.
-  loop @i= 1 to nrow(@nz).
-   do if @nz(@i).
-    comp @c= @c+1.
-    comp @grd(@c)= @i.
-   end if.
-  end loop.
-  release @c.
- end if.
- comp !5= !5(:,@grd).
- comp !6= !6(@grd).
- comp !7= !7(@grd,:).
-end if.
-release @nz,@nnz.
end if.
release @cols,@sort,@empty,@grd,@desi,@codesi,@i.
!enddefine.


******** ФИКТИВНЫЕ ПЕРЕМЕННЫЕ ВЗАИМОДЕЙСТВИЯ И МНОГОМЕРНАЯ ЧАСТОТНАЯ АГРЕГАЦИЯ (ДИХОТОМИЧЕСКИЕ ДАННЫЕ) ********.
*/*!KO_baggrtab(cols%empty%name1%name2%name3)*/*.
*Version 1.
*Делает частотную кросстабуляцию для любого числа дихотомических переменных.
*Переменные должны быть кодированы двоично (1 vs 0); если у вас другие значения, сделайте их двоичными.
*Данная функция эквивалентна /*!aggrtab*/, но рассчитана только на все переменные COLS
*двоичные (0 vs 1). Она быстрее, чем /*!aggrtab*/, и бережет память.
*Аргумент EMPTY и результаты NAME1, NAME2, NAME3 - см. в описании /*!aggrtab*/.
*Порядок значений в результатах всегда сортирован так: 0 1, поэтому аргумента SORT нет.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_baggrtab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cols= !2.
comp @empty= !3.
comp !4= @cols(:,ncol(@cols)).
comp !4= {not !4,!4}.
comp !6= {0,1}.
comp @codesi= {0,1}.
loop @i= ncol(@cols)-1 to 1 by -1.
-comp @desi= @cols(:,@i).
-comp @desi= {not @desi,@desi}.
-comp !4= kroneker(@desi,make(1,ncol(!4),1))&*kroneker(make(1,ncol(@desi),1),!4).
-comp !6= {kroneker(@codesi,make(1,ncol(!6),1));kroneker(make(1,ncol(@codesi),1),!6)}.
end loop.
comp !5= t(csum(!4)).
comp !6= t(!6).
do if @empty.
-comp @nz= !5>0.
-comp @nnz= csum(@nz).
-do if @nnz<nrow(@nz).
- do if @nnz/nrow(@nz)>.8.
-  comp @grd= make(1,@nnz+1,0).
-  comp @c= 1.
-  loop @i= 1 to nrow(@nz).
-   comp @grd(@c)= @i.
-   comp @c= @c+@nz(@i).
-  end loop.
-  comp @grd= @grd(1:@nnz).
-  release @c.
- else if @nnz/nrow(@nz)<.08.
-  comp @grd= make(1,@nnz,0).
-  comp @s= 1.
-  loop @c= 1 to @nnz.
-   loop @i= @s to nrow(@nz).
-   end loop if @nz(@i).
-   comp @grd(@c)= @i.
-   comp @s= @i+1.
-  end loop.
-  release @s,@c.
- else.
-  comp @grd= make(1,@nnz,0).
-  comp @c= 0.
-  loop @i= 1 to nrow(@nz).
-   do if @nz(@i).
-    comp @c= @c+1.
-    comp @grd(@c)= @i.
-   end if.
-  end loop.
-  release @c.
- end if.
- comp !4= !4(:,@grd).
- comp !5= !5(@grd).
- comp !6= !6(@grd,:).
- release @grd.
-end if.
-release @nz,@nnz.
end if.
release @cols,@empty,@desi,@codesi,@i.
!enddefine.


******** ЭЛЕМЕНТАРНЫЕ ПЕРЕМЕННЫЕ И ЧАСТОТЫ ЗНАЧЕНИЙ В КАТЕГОРИАЛЬНОМ НАБОРЕ МНОЖЕСТВЕННОГО ОТВЕТА ********.
*/*!KO_mrfreq(mrc%dupl%sort%name1%name2%name3)*/*.
*Version 1.
*Эта функция подобна /*!KO_freq*/ - выдает двоичные переменные и частоты значений - но в качестве
*входящих принимает не одну категориальную переменную, а набор таковых; у этих переменных общий
*пул значений. Это "категориальный набор множественного ответа". Если каждое наблюдение это респондент,
*он мог, каким образом, выбрать несколько, а не один ответ.
*MRC - входящие данные, столбцы которых категориальные переменные, составляющие упомянутый набор.
*Нуль в MRC считается "наполнителем", т.е. пустой ячейкой. Всякое же иное значение есть валидное.
*Допускается, чтобы некоторые ряды (наблюдения) или некоторые столбцы (переменные) были пусты.
*Аргумент DUPL (скаляр) - если положительный, то каждое уникальное значение засчитывается в ряду MRC
*не более раза, т.е. повторение одного и того же ответа одним респондентом, если есть, игнорируется;
*а если аргумент неположительный, то все повторения ответов респондентом засчитываются как валидные.
*Функция выдает:
*NAME1 - матрица, по одному столбцу на каждое уникальное валидное значение MRC. Значение
*(i,j) в NAME1 это сколько раз респондент i дал ответ j (если DUPL неположителен), либо 0 или 1
*(если DUPL положителен); таким образом, при положительном DUPL NAME1 есть "двоичный набор
*множественного ответа".
*NAME2 - вектор-столбец частот уникальных значений MRC (это суммы в столбцах NAME1);
*NAME3 - вектор-столбец, список уникальных значений MRC.
*Аргумент SORT (скаляр) - сортировать ли результаты по возрастанию значений: если положительный,
*то столбцы в NAME1 и ряды в NAME2 и NAME3 будут сортированы в возрастающем порядке уникальных значений MRC;
*если же аргумент неположительный, то порядок рядов и столбцов в результатах определяется очередностью
*встречи уникальных значений MRC (при просмотре MRC сверху вниз, сначала 1-й столбец, потом 2-й...).
*Эта функция может требовать от компьютера много оперативной памяти при массивных данных.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_mrfreq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @dupl= !3.
comp @sort= !4.
comp @n= nrow(!5).
comp @p= ncol(!5).
comp !5= reshape(t(!5),@n*@p,1).
comp @des= design({0;!5}).
comp @des= @des(2:nrow(@des),2:ncol(@des)).
comp !6= csum(@des).
comp !7= !5*make(1,ncol(@des),1)&*@des.
comp !7= cmax(!7)+cmin(!7).
comp !5= @des(1:@n,:).
comp @nums= 1:@n.
loop @i= 1 to @p-1.
-comp !5= !5+@des(@nums+@i*@n,:).
end loop.
do if @dupl.
-comp !5= !5>0.
-comp !6= csum(!5).
end if.
do if @sort.
-comp @grd= grade(!7).
-comp !5(:,@grd)= !5.
-comp !6(@grd)= !6.
-comp !7(@grd)= !7.
-release @grd.
end if.
comp !6= t(!6).
comp !7= t(!7).
release @dupl,@sort,@n,@p,@des,@nums,@i.
!enddefine.


******** ЧАСТОТЫ КОНКОРДАНТНЫХ, ДИСКОРДАНТНЫХ И СВЯЗАННЫХ ПАР ********.
*/*!KO_concdisc(col1%col2%name1%name2%name3%name4%name5)*/*.
*Version 1.
*Берет векторы-столбцы COL1 и COL2, которые должны не быть константами и должны быть одинаковой длины
*(это две мерные или порядковые переменные) и выдает 5 скалярных величин, необходимых для вычисления
*таких коэффициентов связи между порядковыми переменными, как Гамма Гудмана-Краскела, коэффициент Сомера,
*корреляция Кендалла, и т.д.
*Эти величины:
*NAME1 - число конкордантных пар (P);
*NAME2 - число дискордантных пар (Q);
*NAME3 - число пар со связью (ties) по первой переменной, COL1 (Tx);
*NAME4 - число пар со связью (ties) по второй переменной, COL2 (Ty);
*NAME5 - число пар со связью (ties) по обеим переменным сразу (Txy).
*Если COL1 и COL2 категориальные (не континуальные) данные, другая ф-ция, /*!KO_concdisct*/, вычисляющая
*те же величины, может оказаться более скоростной.

define !KO_concdisc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                   /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @x= !2.
comp @y= !3.
comp @n= nrow(@x).
comp @x= @x*make(1,@n,1).
comp @y= @y*make(1,@n,1)).
comp @x= @x-t(@x).
comp @y= @y-t(@y).
comp @xy= @x&*@y.
comp !4= msum(@xy>0)/2.
comp !5= msum(@xy<0)/2.
comp !6= msum(@x=0)-@n.
comp !7= msum(@y=0)-@n.
comp !8= (msum(@x=0 and @y=0)-@n)/2.
comp !6= !6/2-!8.
comp !7= !7/2-!8.
release @x,@y,@n,@xy.
!enddefine.


******** ЧАСТОТЫ КОНКОРДАНТНЫХ, ДИСКОРДАНТНЫХ И СВЯЗАННЫХ ПАР (ИЗ ТАБЛИЦЫ) ********.
*/*!KO_concdisct(tab%name1%name2%name3%name4%name5)*/*.
*Version 2.
*Берет частотную кросс-таблицу TAB, полученную для двух мерных или порядковых переменных
*(т.е. ряды и столбцы в таблице упорядочены по возрастанию уникальных значений переменных)
*и выдает 5 скалярных величин, необходимых для вычисления таких коэффициентов связи между
*порядковыми переменными, как Гамма Гудмана-Краскела, коэффициент Сомера, корреляция Кендалла, и т.д.
*Эти величины:
*NAME1 - число конкордантных пар (P);
*NAME2 - число дискордантных пар (Q);
*NAME3 - число пар со связью (ties) по первой переменной, образующей ряды таблицы (Tx);
*NAME4 - число пар со связью (ties) по второй переменной, образующей столбцы таблицы (Ty);
*NAME5 - число пар со связью (ties) по обеим переменным сразу (Txy).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_concdisct(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
comp @c= make(nrow(@tab),ncol(@tab),0).
comp @c(1:(nrow(@tab)-1),1:(ncol(@tab)-1))= @tab(2:nrow(@tab),2:ncol(@tab)).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp @c= t(@c).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp !3= msum(@tab&*t(@c)). /*P
comp @c= make(nrow(@tab),ncol(@tab),0).
comp @c(1:(nrow(@tab)-1),2:ncol(@tab))= @tab(2:nrow(@tab),1:(ncol(@tab)-1)).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp @c= t(@c).
loop @i= 2 to nrow(@c).
-comp @c(@i,:)= @c(@i,:)+@c((@i-1),:).
end loop.
comp !4= msum(@tab&*t(@c)). /*Q
comp @c= make(ncol(@tab),nrow(@tab),0).
comp @c(1:(ncol(@tab)-1),:)= t(@tab(:,2:ncol(@tab))).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp !5= msum(@tab&*t(@c)). /*Tx
comp @c= make(nrow(@tab),ncol(@tab),0).
comp @c(1:(nrow(@tab)-1),:)= @tab(2:nrow(@tab),:).
loop @i= nrow(@c)-1 to 1 by -1.
-comp @c(@i,:)= @c(@i,:)+@c((@i+1),:).
end loop.
comp !6= msum(@tab&*@c). /*Ty
comp @n= msum(@tab).
comp !7= (@n**2-@n)/2-!3-!4-!5-!6. /*Txy
release @tab,@c,@i,@n.
!enddefine.


******** МАТРИЦА ПУТАНОСТИ СОЧЛЕНСТВА ДЛЯ ДВУХ РАЗБИЕНИЙ ********.
*/*!KO_ccm(part1%part2%name)*/*.
*Version 1.
*Матрица (или таблица) путаности сочленства (comembership confusion matrix/table) это такая 2x2 матрица
*путаности (confusion matrix), где элементом счета выступает пара объектов (а не объект). Эта матрица
*является основой для оценки сходства двух классифицаций (разбиений) одних и тех же объектов.
*Пусть есть n объектов, которые были классифицированы двумя способами, I и II. Тогда в матрице путаницы
*сочленства ячейка (1,1) содержит число пар, в которых оба объекта совстречаются в одном классе
*как в классификации I, так и классийикации II; ячейка (1,2) содержит число пар, в которых оба объекта
*совстречаются в классификации I, но не в классификации II; ячейка (2,1) содержит число пар, в которых
*оба объекта совстречаются в классификации II, но не в классификации I; ячейка (2,2) содержит число пар,
*в которых оба объекта не совстречаются в одном классе ни в той, ни в другой классификации.
*Сумма в такой таблице равна числу пар объектов, n(n-1)/2.
*Входящие PART1 и PART2 - два массива с одинаковым числом рядов (ряды - это объекты), являющие собой
*разбиения/классификации I и II соответственно. Массив может быть либо столбцом с любыми значениями, либо
*несколькими столбцами с двоичными (1 и 0) значениями. Когда классификация задана одним столбцом,
*то функция принимает его за категориальную переменную: каждое уникальное значение обозначает класс.
*Когда классификация задана более чем одним столбцом, функция принимает их за набор двоичных переменных,
*каждая обозначает класс, а значения обозначают принадлежность ему (1=принадлежит, 0=не принадлежит).
*Двоичный набор может быть любой, например содержать ряды или столбцы, состоящие из одних нулей или из
*одних единиц; т.е. двоичным набором можно передать не только классификацию, где классы непересекаемы
*составом объектов, но любую, в том числе с пустыми классами или объектами, не принадлежащими классам.
*В том случае, если ваша классификация представляет только один класс, то задайте ее как два столбца:
*один двоичный (принадлежит - не принадлежит), а другой пустой, из нулей.
*Результат NAME - матрица (или таблица) путаности сочленства.

define !KO_ccm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @p1= !2.
comp @p2= !3.
do if ncol(@p1)=1.
-comp @p1= @p1*make(1,nrow(@p1),1).
-comp @p1= @p1=t(@p1).
else.
-comp @p1= sscp(t(@p1)).
-call setdiag(@p1,1).
end if.
do if ncol(@p2)=1.
-comp @p2= @p2*make(1,nrow(@p2),1).
-comp @p2= @p2=t(@p2).
else.
-comp @p2= sscp(t(@p2)).
-call setdiag(@p2,1).
end if.
comp !4= make(2,2,0).
comp !4(1,1)= msum(@p1 and @p2)-ncol(@p1).
comp @p2= not @p2.
comp !4(1,2)= msum(@p1 and @p2).
comp @p1= not @p1.
comp !4(2,2)= msum(@p1&*@p2).
comp !4(2,1)= ncol(@p1)*(ncol(@p1)-1)-msum(!4).
comp !4= !4/2.
release @p1,@p2.
!enddefine.


******** СЧЕТ ЗНАЧЕНИЙ В РЯДАХ (ТОЧНОЕ СОВПАДЕНИЕ) ********.
*/*!KO_count1(data%vals%name)*/*.
*Version 2.
*Подсчитывает в рядах DATA число элементов, совпадающих с любым из значений списка VALS.
*Выдает столбец NAME длиной как число рядов DATA.
*Эта функция подобна внематричной команде COUNT, в которой значения заданы списком.
*VALS - вектор (столбец или ряд) любой длины. Значения в списке могут повторяться: это не влияет на
*результат, но замедляет работу.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_count1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @vals= !3.
comp !4= @data=@vals(1).
loop @i= 2 to nrow(@vals)*ncol(@vals).
-comp !4= !4+(@data=@vals(@i)).
end loop.
comp !4= rsum(!4>0).
release @data,@vals,@i.
!enddefine.


******** СЧЕТ ЗНАЧЕНИЙ В РЯДАХ (ПОПАДАНИЕ В ДИАПАЗОН) ********.
*/*!KO_count2(data%low%high%name)*/*.
*Version 2.
*Подсчитывает в рядах DATA число элементов, попадающих в любой из диапазонов от LOW до HIGH (включительно).
*Выдает столбец NAME длиной как число рядов DATA.
*Эта функция подобна внематричной команде COUNT, в которой значения заданы диапазоном(ами) от - до.
*LOW и HIGH - два скаляра или два вектора одинаковой длины. Каждая пара их соответствующих элементов
*LOW(i) и HIGH(i) есть диапазон; HIGH(i) в норме должно быть >= LOW(i). Диапазоны могут пересекаться.
*Чтобы подсчитать все значения <=HIGH, установите LOW на заведомо слишком малое число.
*Чтобы подсчитать все значения >=LOW, установите HIGH на заведомо слишком большое число.
*Если аргументы LOW и HIGH тождественны, функция даст тот же результат, что /*!KO_count1*/.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_count2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @lo= !3.
comp @hi= !4.
comp !5= @data>=@lo(1) and @data<=@hi(1).
loop @i= 2 to nrow(@lo)*ncol(@lo).
-comp !5= !5+(@data>=@lo(@i))&*(@data<=@hi(@i)).
end loop.
comp !5= rsum(!5>0).
release @data,@lo,@hi,@i.
!enddefine.


******** ХИ-КВАДРАТ (ПИРСОНА И ОТНОШЕНИЯ ПРАВДОПОДОБИЯ) ДВУМЕРНОЙ ТАБЛИЦЫ СОПРЯЖЕННОСТИ ********.
*/*!KO_chitab(tab%type%name)*/*.
*Version 1.
*Вычисляет статистику хи-квадрат, говорящую о связи рядов и столбцов в двумерной
*таблице сопряженности. Выдает скаляр NAME.
*Аргумент TYPE (скаляр) - какую версию статистики выдать: хи-квадрат Пирсона (TYPE неположителен)
*или хи-квадрат отношения правдоподобия (TYPE положителен).
*Таблица TAB должна содержать неотрицательные числа, и суммы в ее рядах и ее столбцах
*должны быть все ненулевые.

define !KO_chitab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @type= !3.
comp @n= msum(!4).
comp @exp= rsum(!4)*csum(!4)/@n.
do if @type.
-comp !4= 2*msum(!4&*ln(!4/@exp+(!4=0))).
else.
-comp !4= msum(!4&*!4/@exp)-@n.
end if.
release @type,@n,@exp.
!enddefine.


******** ДОЛИ И ОСТАТКИ В ДВУМЕРНОЙ ТАБЛИЦЕ СОПРЯЖЕННОСТИ ********.
*/*!KO_cells(tab%stat%name)*/*.
*Version 1.
*Берет двумерную таблицу сопряженности (например частотную перекрестную таблицу) TAB с
*(в норме) неотрицательными элементами и (обязательно) ненулевыми суммами в рядах и столбцах.
*Выдает таблицу этого же размера NAME, содержащую в ячейках требуемую статистику.
*Требуемая статистика - аргумент STAT, ключевое заглавными буквами:
*"TABPR" - доли от суммы таблицы
*"ROWPR" - доли от суммы в рядах
*"COLPR" - доли от суммы в столбцах
*"EXPECT" - ожидаемые (при независимости рядов и столбцов) значения 
*"RESID" - остатки
*"SRESID" - стандартизованные остатки
*"ASRESID" - приведенные стандартизованные остатки.
*Кавычки или апострофы вокруг ключевого слова необязательны.

define !KO_cells(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
!let !stat= !unquote(!3)
!if (!stat='TABPR') !then
-comp !4= @tab/msum(@tab).
!else !if (!stat='ROWPR') !then
-comp !4= @tab/(rsum(@tab)*make(1,ncol(@tab),1)).
!else !if (!stat='COLPR') !then
-comp !4= @tab/(make(nrow(@tab),1,1)*csum(@tab)).
!else
-comp @rsum= rsum(@tab).
-comp @csum= csum(@tab).
-comp @msum= msum(@tab).
-comp @exp= @rsum*@csum/@msum.
 !if (!stat='EXPECT') !then
- comp !4= @exp.
 !else !if (!stat='RESID') !then
- comp !4= @tab-@exp.
 !else !if (!stat='SRESID') !then
- comp !4= (@tab-@exp)/sqrt(@exp).
 !else !if (!stat='ASRESID') !then
- comp !4= (@tab-@exp)/sqrt(@exp&*((1-@rsum/@msum)*(1-@csum/@msum))).
 !ifend !ifend !ifend !ifend
-release @rsum,@csum,@msum,@exp.
!ifend !ifend !ifend
!enddefine.


******** МАТРИЦА КВАДРАТНЫХ ХИ-КВАДРАТ-РАССТОЯНИЙ В ТАБЛИЦЕ СОПРЯЖЕННОСТИ ********.
*/*!KO_schitab(tab%name)*/*.
*Version 1.
*Вычисляет квадратные хи-квадрат-расстояния между рядами таблицы сопряженности TAB.
*Выдает квадратную симметрическую матрицу NAME.
*TAB это матрица с неотрицательными элементами, и суммы в ее рядах и ее столбцах должны
*быть все ненулевые.
*Хи-квадрат расстояние между рядами это взвешенное евклидово расстояние между рядными долями,
*где весами служат обращенные суммы в столбцах. Расстояние (квадратное) между рядами i и i' есть:
*N * Sum_c[1/Nj * (Nij/Ni-Ni'j/Ni')^2], где Nij значение в ячейке, Ni сумма в ряду i, Nj сумма в
*столбце j, N общая сумма, Sum_c сумма сквозь все столбцы.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_schitab(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
comp @w= msum(@tab)/csum(@tab).
comp @tab= @tab/(rsum(@tab)*make(1,ncol(@tab),1)).
comp !3= make(nrow(@tab),nrow(@tab),0).
loop @i= 1 to ncol(@tab).
-comp @coli= @tab(:,@i)*make(1,nrow(@tab),1).
-comp @coli= @coli-t(@coli).
-comp !3= !3+@w(@i)&*@coli&*@coli.
end loop.
release @tab,@w,@i,@coli.
!enddefine.


******** МАТРИЦА ХИ-КВАДРАТОВ МЕЖДУ РЯДЯМИ ЧАСТОТНЫХ ДАННЫХ ПОПАРНО ********.
*/*!KO_schi2c(data%type%name)*/*.
*Version 1.
*Берет данные DATA размером не менее 2x2 с неотрицательными значениями, которые рассматривает как частоты.
*Суммы в рядах и столбцах должны быть больше нуля.
*Для каждых двух рядов DATA попарно вычленяет табличку 2xNC (NC - число столбцов DATA) и вычисляет
*статистику хи-квадрат (или фи-квадрат) Пирсона этой таблички. Эта статистика может считаться мерой различия
*между двумя данными рядями: хи-квадрат расстояние (квадратное).
*Выдается квадратная симметричная матрица NAME таких расстояний между всеми рядями DATA.
*Это те же расстояния, только в квадрате, что выдает команда PROXIMITIES /MEASURE=CHISQ (или /MEASURE=PH2).
*Не следует путать эти "расстояния" с хи-квадрат расстояниями, выдаваемыми ф-цией /*!KO_schitab*/.
*TYPE - скаляр: если неположительный, вычисляется хи-квадрат; если положительный, вычисляется фи-квадрат
*(хи-квадрат делится на общую частоту 2xNC таблички), фи-квадрат нечувствителен к неодинаковости сумм
*в рядах DATA.
*Если в какой-либо 2xNC табличке встретится пустой столбец, функция игнорирует его и пытается вычислить
*на основании оставшихся столбцов.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_schi2c(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @type= !3.
comp @ones= make(1,nrow(@data),1).
comp @r= rsum(@data)*@ones.
comp @tot= @r+t(@r).
comp !4= make(nrow(@data),nrow(@data),0).
loop @i= 1 to ncol(@data).
-comp @c= @data(:,@i)*@ones.
-comp @cc= @c+t(@c).
-comp @e= @cc&*@r/@tot.
-comp @e= @e+(@e=0).
-comp @cc= @c&*@c/@e.
-comp !4= !4+@cc+t(@cc).
end loop.
comp !4= !4-@tot.
do if @type.
-comp !4= !4/@tot.
end if.
release @data,@type,@ones,@r,@tot,@i,@c,@cc,@e.
!enddefine.


******** РАССТОЯНИЯ ДО ЦЕНТРОИДОВ ГРУПП ********.
*/*!KO_dtoc(data%bin%name)*/*.
*Version 1.
*Берет данные DATA с любым числом рядов (точек) и столбцов (переменных).
*И берет одну или более двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DATA.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME - матрица точки x группы, содержащая квадратные евклидовы расстояния между точками
*и центроидами групп.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_dtoc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp !4= !3.
comp @mean= t(!4)*@data/(t(csum(!4))*make(1,ncol(@data),1)).
comp @ones= make(1,nrow(@data),1).
loop @i= 1 to ncol(!4).
-comp !4(:,@i)= rssq(@data-@mean(@i,:)(@ones,:)).
end loop.
release @data,@mean,@ones,@i.
!enddefine.


******** РАССТОЯНИЯ ДО ЦЕНТРОИДОВ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАССТОЯНИЙ) ********.
*/*!KO_dtocfrd(dis%bin%self%name)*/*.
*Version 2.
*Берет квадратную симметричную матрицу расстояний между точками DIS. Расстояния должны быть
*квадратные евклидовы; во всяком случае, ф-ция воспримет их как квадратные евклидовы.
*И берет одну или более двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME - матрица точки x группы, содержащая квадратные евклидовы расстояния между точками
*и центроидами групп.
*Аргумент SELF - цифра 0 или 1 (не имя и не выражение; цифру можно взять в кавычки/апострофы).
*Если 1, то при вычислении расстояния между точкой i и центроидом группы, которой она принадлежит
*(и которая кроме нее содержит еще точки), точка i участвует в определении центроида. Если 0, тогда
*точка i не участвует в определении центроида: она рассматривается в этот момент как внешняя точка.

define !KO_dtocfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @bin= !3.
!let !self= !unquote(!4)
comp @n= csum(@bin).
comp @ones= make(1,nrow(@bin),1).
comp @pgs= !5*@bin.
comp !5= csum(@pgs&*@bin)/2.
!if (!self) !then
-comp !5= (@pgs-(!5/@n)(@ones,:))/@n(@ones,:).
!else
-comp !5= !5(@ones,:)-@pgs&*@bin.
-comp @n= @n(@ones,:)-@bin.
-comp @n= @n+(@n=0).
-comp !5= (@pgs-!5/@n)/@n.
!ifend
release @bin,@n,@ones,@pgs.
!enddefine.


******** РАССТОЯНИЯ МЕЖДУ ЦЕНТРОИДАМИ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАССТОЯНИЙ) ********.
*/*!KO_dbwcfrd(dis%bin%name)*/*.
*Version 1.
*Берет квадратную симметричную матрицу расстояний между точками DIS. Расстояния должны быть
*квадратные евклидовы; во всяком случае, ф-ция воспримет их как квадратные евклидовы.
*И берет не менее двух двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME - квадратная симметричная матрица квадратных евклидовых расстояний между центроидами
*групп точек.

define !KO_dbwcfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @bin= !3.
comp @n= csum(@bin).
comp @nsq= @n&*@n.
comp @n= sscp(@n).
comp @bgs= t(@bin)*!4*@bin.
comp !4= diag(@bgs)/2*@nsq.
comp !4= (@bgs-(!4+t(!4))/@n)/@n.
comp !4= (!4+t(!4))/2.
call setdiag(!4,0).
release @bin,@n,@nsq,@bgs.
!enddefine.


******** УСРЕДНЕННЫЕ РАССТОЯНИЯ ДО ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_datofrd(dis%bin%self%name)*/*.
*Version 1.
*Среднее расстояние от каждой точки до точек каждой из предложенных групп.
*Берет квадратную симметричную матрицу расстояний (различий) между точками DIS.
*Это должна быть неотрицательная матрица с нулевой диагональю.
*И берет одну или более двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME - матрица точки x группы: элемент ij в ней это усредненное расстояние между
*точкой i и точками группы j.
*Аргумент SELF - цифра 0 или 1 (не имя и не выражение; цифру можно взять в кавычки/апострофы).
*Если 1, то расстояние точки до самой себя (нуль), когда она принадлежит группе, считается, т.е. входит
*в усреднение. Если 0, то расстояние точки до самой себя не считается, т.е. не входит в усреднение,
*кроме случая, когда точка - единственный член группы.
*Если ваша матрица DIS это сходства, переведите их сначала в различия любым способом, каким хотите.

define !KO_datofrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @bin= !3.
!let !self= !unquote(!4)
comp !5= !5*@bin.
!if (!self) !then
-comp @n= csum(@bin)(make(1,nrow(@bin),1),:).
!else
-comp @n= csum(@bin)(make(1,nrow(@bin),1),:)-@bin.
-comp @n= @n+(not @n).
!ifend
comp !5= !5/@n.
release @bin,@n.
!enddefine.


******** УСРЕДНЕННЫЕ РАССТОЯНИЯ МЕЖДУ ГРУППАМИ (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_dabwfrd(dis%bin%self%name)*/*.
*Version 1.
*Расстояния межгрупповой средней связи, т.е. средние расстояния между точками от каждых двух групп.
*Берет квадратную симметричную матрицу расстояний (различий) между точками DIS.
*Это должна быть неотрицательная матрица с нулевой диагональю.
*И берет не менее двух двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME - квадратная симметричная матрица группы x группы, содержащая усредненные
*расстояния между точками разных групп. На диагонали стоят усредненные расстояния внутри групп.
*Аргумент SELF - цифра 0 или 1 (не имя и не выражение; цифру можно взять в кавычки/апострофы).
*Если 1, то расстояние точки до самой себя (нуль), когда она принадлежит группе (или обеим группам),
*считается, т.е. входит в усреднение. Если 0, то расстояние точки до самой себя не считается, т.е.
*не входит в усреднение, кроме случая, когда точка - единственный член группы (или единственный член
*в обеих группах, когде речь идет о разных группах).
*Если ваша матрица DIS это сходства, переведите их сначала в различия любым способом, каким хотите.

define !KO_dabwfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @bin= !3.
!let !self= !unquote(!4)
comp !5= t(@bin)*!5*@bin.
comp @n= csum(@bin).
comp @nn= sscp(@n).
!if (!self) !then
-call setdiag(@nn,@n&*(@n+1)).
!else
-comp @nn= @nn-sscp(@bin).
-call setdiag(@nn,@n&*(@n-1)).
-comp @nn= @nn+(not @nn).
!ifend
comp !5= !5/@nn.
release @bin,@n,@nn.
!enddefine.


******** РАССТОЯНИЯ ДО ДАЛЬНИХ СОСЕДЕЙ ИЗ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_dtoffrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Расстояние от каждой точки до ее дальнего соседа каждой из предложенных групп.
*Берет квадратную симметричную матрицу расстояний (различий) между точками DIS.
*Это должна быть неотрицательная матрица с нулевой диагональю.
*И берет одну или более двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Аргумент ALSOIND - цифра 0 или 1 (не имя и не выражение; цифру можно взять в кавычки/апострофы).
*Если вам нужны только расстояния до дальних соседей, укажите 0. Результат NAME будет матрица
*точки x группы: элемент ij в ней это расстояние (значение из DIS) между точкой i и ее дальним соседом
*из группы j. Если вы также хотите знать номера этих дальних соседей (их индексы рядов в DIS и BIN),
*укажите ALSOIND как 1. Тогда в матрице NAME будет вдвое больше столбцов: к матрице с расстояниями до
*дальних соседей будет справа пришита матрица, содержащая номера этих точек, являющихся дальними соседями.
*Если у точки i существует более одного дальнего соседа в группе j, показывается номер последнего из них
*(т.е. у кого номер больше).
*Только если группа состоит из одной точки, ее сосед в этой группе - она сама.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваша матрица DIS это сходства, переведите их сначала в различия любым способом, каким хотите.

define !KO_dtoffrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
!let !ind= !unquote(!4)
comp !5= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
-comp @d= @d+1.
-call setdiag(@d,0).
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp !5(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
!if (!ind) !then
-comp !5= {!5-(!5>0),@ind}.
-release @ind,@nums,@numsm,@n_1.
!ifend
release @d,@bin,@ones,@i,@dbini,@dneii.
!enddefine.


******** РАССТОЯНИЯ МЕЖДУ ДАЛЬНИМИ СОСЕДЯМИ ИЗ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_dbwffrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Расстояния полной связи между группами, т.е. наибольшие расстояния между точками каждых двух групп.
*Берет квадратную симметричную матрицу расстояний (различий) между точками DIS.
*Это должна быть неотрицательная матрица с нулевой диагональю.
*И берет не менее двух двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Аргумент ALSOIND - цифра 0 или 1 (не имя и не выражение; цифру можно взять в кавычки/апострофы).
*Если вам нужны только расстояния между дальними соседями, укажите 0. Результат NAME будет квадратная
*симметричная матрица группы x группы: элемент ij в ней это расстояние (значение из DIS) между самыми
*дальними друг от друга точками, одна из группы i, другая из группы j. На диагонали - расстояния
*между самыми разудаленными точками внутри каждой группы.
*Если вы также хотите знать номера точек, являющихся теми дальними соседями (их индексы рядов в DIS и BIN),
*укажите ALSOIND как 1. Тогда в матрице NAME столбцов будет вдвое больше плюс еще один: к матрице с
*расстояниями между дальними соседями будет справа пришита матрица, содержащая номера этих точек, являющихся
*дальними соседями. В этой матрице (к-рая будет пришита), и которая асимметрична, элемент ij это номер точки,
*являющейся дальним соседом со стороны группы i (т.е. группы-ряда). Например, элементы (2,4)=10 и (4,2)=6
*означают, что максимальное расстояние между группами 2 и 4 это расстояние между точкой 10 (она из группы 2)
*и точкой 6 (она из группы 4). Что касается номеров точек-дальних соседей внутри каждой группы, то номер
*одного из двух стоит на диагонали данной пришитой матрицы с номерами, а номер второго вынесен в отдельный
*столбец, пришитый справа всего. Таким образом, если k это число столбцов BIN (число групп), то при ALSOIND=1
*NAME имеет 2*k+1 столбцов.
*Точка может выйти своим собственным соседом только в том случае, если в обоих сопоставляемых группах она
*единственный член группы.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваша матрица DIS это сходства, переведите их сначала в различия любым способом, каким хотите.

define !KO_dbwffrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
comp @k= ncol(@bin).
!let !ind= !unquote(!4)
comp @dtog= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
-comp @d= @d+1.
-call setdiag(@d,0).
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp @dtog(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
comp !5= make(@k,@k,0).
!if (!ind) !then
-comp @numsm= make(@k,@k,0).
-comp @2nd= make(@k,1,0).
-comp @dtog= @dtog+1.
!ifend
loop @i= 1 to @k.
-comp @d= make(1,@k-@i+1,1).
-comp @dbini= @bin(:,@i)*@d&*@dtog(:,@i:@k).
-comp @dneii= cmax(@dbini).
-comp !5(@i,@i:@k)= @dneii.
 !if (!ind) !then
- comp @dbini= @dbini=@dneii(@ones,:).
- comp @n_1= cmax(@nums*@d&*@dbini).
- comp @numsm(@i,@i:@k)= @n_1.
- comp @2nd(@i)= @n_1(1).
- comp @indi= @ind(:,@i:@k).
- comp @numsm(@i:@k,@i)= t(cmax((@indi<>@n_1(@ones,:))&*@dbini&*@indi)).
 !ifend
end loop.
comp !5= !5+t(!5)-mdiag(diag(!5)).
!if (!ind) !then
-comp @numsm= @numsm+t(@numsm)&*(not @numsm).
-comp @dg= diag(@numsm).
-call setdiag(@numsm,@dg+@2nd&*(not @dg)).
-comp !5= {!5-1-(!5>1),@numsm,@2nd}.
-release @ind,@indi,@nums,@numsm,@n_1,@2nd,@dg.
!ifend
release @d,@bin,@k,@dtog,@ones,@i,@bini,@dbini,@dneii.
!enddefine.


******** РАССТОЯНИЯ ДО БЛИЖНИХ СОСЕДЕЙ ИЗ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_dtonfrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Расстояние от каждой точки до ее ближнего соседа каждой из предложенных групп.
*Эта функция такая же, как функция /*!KO_dtoffrd*/, но выдает расстояния до ближних, а не до дальних соседей.

define !KO_dtonfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
!let !ind= !unquote(!4)
comp @maxd1= mmax(@d)+1.
comp @d= @maxd1-@d.
call setdiag(@d,0).
comp !5= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp !5(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
comp !5= @maxd1&*(!5>0)-!5.
!if (!ind) !then
-comp !5= {!5,@ind}.
-release @ind,@nums,@numsm,@n_1.
!ifend
release @d,@bin,@ones,@i,@dbini,@dneii,@maxd1.
!enddefine.


******** РАССТОЯНИЯ МЕЖДУ БЛИЖНИМИ СОСЕДЯМИ ИЗ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_dbwnfrd(dis%bin%alsoind%name)*/*.
*Version 1.
*Расстояния единичной связи между группами, т.е. наименьшие расстояния между точками каждых двух групп.
*Эта функция такая же, как функция /*!KO_dbwffrd*/, но выдает расстояния между ближними,
*а не между дальними соседями.

define !KO_dbwnfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
comp @k= ncol(@bin).
!let !ind= !unquote(!4)
comp @maxd1= mmax(@d)+1.
comp @d= @maxd1-@d.
call setdiag(@d,0).
comp @dtog= make(nrow(@bin),ncol(@bin),0).
comp @ones= make(nrow(@bin),1,1).
!if (!ind) !then
-comp @ind= make(nrow(@bin),ncol(@bin),0).
-comp @nums= t(1:nrow(@bin)).
-comp @numsm= (1:nrow(@bin))(@ones,:).
-comp @n_1= csum(@bin)=1.
!ifend
loop @i= 1 to ncol(@bin).
-comp @bini= @bin(:,@i).
-comp @dbini= @d&*t(@bini)(@ones,:).
-comp @dneii= rmax(@dbini).
-comp @dtog(:,@i)= @dneii.
 !if (!ind) !then
- do if @n_1(@i).
-  comp @ind(:,@i)= make(nrow(@bin),1,cmax(@bini&*@nums)).
- else.
-  comp @ind(:,@i)= rmax((@dbini=(@dneii*t(@ones)))&*@numsm).
- end if.
 !ifend
end loop.
comp !5= make(@k,@k,0).
!if (!ind) !then
-comp @numsm= make(@k,@k,0).
-comp @2nd= make(@k,1,0).
-comp @dtog= @dtog+1.
!ifend
loop @i= 1 to @k.
-comp @d= make(1,@k-@i+1,1).
-comp @dbini= @bin(:,@i)*@d&*@dtog(:,@i:@k).
-comp @dneii= cmax(@dbini).
-comp !5(@i,@i:@k)= @dneii.
 !if (!ind) !then
- comp @dbini= @dbini=@dneii(@ones,:).
- comp @n_1= cmax(@nums*@d&*@dbini).
- comp @numsm(@i,@i:@k)= @n_1.
- comp @2nd(@i)= @n_1(1).
- comp @indi= @ind(:,@i:@k).
- comp @numsm(@i:@k,@i)= t(cmax((@indi<>@n_1(@ones,:))&*@dbini&*@indi)).
 !ifend
end loop.
comp !5= !5+t(!5)-mdiag(diag(!5))-!ind.
comp !5= @maxd1&*(!5>0)-!5.
!if (!ind) !then
-comp @numsm= @numsm+t(@numsm)&*(not @numsm).
-comp @dg= diag(@numsm).
-call setdiag(@numsm,@dg+@2nd&*(not @dg)).
-comp !5= {!5,@numsm,@2nd}.
-release @ind,@indi,@nums,@numsm,@n_1,@2nd,@dg.
!ifend
release @d,@bin,@k,@dtog,@ones,@i,@bini,@dbini,@dneii,@maxd1.
!enddefine.


******** РАССТОЯНИЯ ДО МЕДОИДОВ ГРУПП (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАЗЛИЧИЙ) ********.
*/*!KO_dtomfrd(dis%bin%name1%name2)*/*.
*Version 1.
*Расстояние от каждой точки до медоида каждой из предложенных групп.
*(Медоид это та точка в группе точек, сумма расстояний от которой до точек группы минимальна.)
*Берет квадратную симметричную матрицу расстояний (различий) между точками DIS.
*Это должна быть неотрицательная матрица с нулевой диагональю.
*И берет одну или более двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME1 - матрица точки x группы, содержащая расстояния между точками и медоидами групп.
*NAME2 это ряд, показывающий номера точек, являющихся медоидами в группах.
*Если в группе более одной точки являются медоидами (как, например, в группе из двух точек),
*то функция назначает точку с наибольшим индексом быть медоидом группы.
*Матрица расстояний между медоидами групп есть DIS(NAME2,NAME2).
*Если ваша матрица DIS это сходства, переведите их сначала в различия любым способом, каким хотите.

define !KO_dtomfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @d= !2.
comp @bin= !3.
comp !4= @d*@bin.
comp !4= !4+(not @bin)*1E300.
comp !4= !4=cmin(!4)(make(nrow(@bin),1,1),:).
comp !5= cmax(t(1:nrow(!4))*make(1,ncol(!4),1)&*!4).
comp !4= @d(:,!5).
release @d,@bin.
!enddefine.


******** ВНУТРИГРУППОВЫЕ СУММЫ КВАДРАТОВ ОТКЛОНЕНИЙ (ВЫЧИСЛЕНИЕ ИЗ МАТРИЦЫ РАССТОЯНИЙ) ********.
*/*!KO_sswfrd(dis%bin%name)*/*.
*Version 1.
*Берет квадратную симметричную матрицу расстояний между точками DIS. Расстояния должны быть
*квадратные евклидовы; во всяком случае, ф-ция воспримет их как квадратные евклидовы.
*И берет одну или более двоичных (1 vs 0) переменных, образующих матрицу BIN с числом рядов как в DIS.
*Каждый столбец BIN означает группу, и значение 1 означает, что точка принадлежит ей.
*В целом, группы могут пересекаться или непересекаться составом точек (в последнем случае BIN
*является набором фиктивных, dummy, переменных). Всякий столбец BIN должен содержать по меньшей
*мере одну единицу. Столбец может, как частный случай, не содержать нулей.
*Результат NAME - столбец длиной как число столбцов BIN, показывающий сумму квадратов отклонений
*от центроида (SSW) в каждой из этих групп.

define !KO_sswfrd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @m= !2.
comp @bin= !3.
comp !4= diag(t(@bin)*@m*@bin)/(2*t(csum(@bin))).
release @m,@bin.
!enddefine.


******** ВНУТРИКЛАССОВЫЙ КОЭФФИЦИЕНТ КОРРЕЛЯЦИИ (ОДНОФАКТОРНЫЙ) ********.
*/*!KO_iccow(col%dummy%type%name)*/*.
*Version1.
*Вычисляет внутриклассовую корреляцию (ICC) в ситуации однофакторного случайного дизайна:
*скаляр NAME. Классы могут быть неравночисленными (unbalanced).
*COL - столбец количественных данных.
*DUMMY - двоичные фиктивные переменные (минимум две), помечающие классы (группы): каждая переменная
*соответствует группе, значение 1 = наблюдение принадлежит ей, 0 = не принадлежит ей.
*В каждом ряду DUMMY должна быть ровно одна единица.
*В каждом столбце DUMMY должно быть минимум две единицы (при TYPE=PAIR2 или PAIR3), либо минимум одна
*единица, и хотя бы в одном столбце больше одной (при остальных TYPE).
*TYPE - версия оценивателя (estimator) ICC, ключевое слово заглавными буквами (опционально в кавычках
*или апострофах) [см. Donner, 1986: A review of inference procedures...; Shieh, 2012: A comparison
*of two indices...]:
*"ANOVA" - ANOVA-оцениватель.
*"CETASQ" - оцениватель Corrected Eta-square.
*"PAIR1" - оцениватель Combinatorial pairing r Pearson без взвешивания классов, классы большего
*размера будут информационно слишком превалировать.
*"PAIR2" - то же, но с взвешиванием классов: классам придаются равные веса важности, несмотря на
*разночисленность.
*"PAIR3" - то же с промежуточным, компромиссным вариантом взвешивания: большие классы будут умеренно
*превалировать.
*Если все классы (группы) одного размера (т.е. balanced), то PAIR1=PAIR2=PAIR3=CETASQ.
*NAME в данных иногда бывает отрицательным, хотя ICC в популяции неотрицателен; вы можете
*воспользоваться советом занулять отрицательные значения.
*Фиктивные переменные DUMMY вы можете создать из категориальной группирующей переменной
*с помощью функции /*!KO_freq*/.

define !KO_iccow(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @col= !2.
comp @des= !3.
!let !type= !unquote(!4)
comp @n= csum(@des).
!if (!type='ANOVA' !or !type='CETASQ') !then
-comp @sum= t(@col)*@des.
-comp @ssq= t(@col&*@col)*@des.
-comp @ssw= rsum(@ssq-@sum&*@sum/@n).
-comp @sst= cssq(@col)-csum(@col)&**2/nrow(@des).
-comp @ssb= @sst-@ssw.
-comp @n_= (nrow(@des)-rssq(@n)/nrow(@des))/(ncol(@des)-1).
 !if (!type='ANOVA') !then
- comp @ssw= @ssw/(nrow(@des)-ncol(@des)).
- comp @ssb= @ssb/(ncol(@des)-1).
- comp !5= (@ssb-@ssw)/(@ssb+(@n_-1)*@ssw).
 !else
- comp !5= (@ssb/@sst-1/@n_)/(1-1/@n_).
 !ifend
-release @sum,@ssq,@ssw,@ssb,@sst,@n_.
!else
 !if (!type='PAIR1') !then
- comp @w= 1/rsum(@n&*(@n-1)).
 !else !if (!type='PAIR2') !then
- comp @w= 1/(ncol(@des)*@n&*(@n-1)).
 !else
- comp @w= 1/(nrow(@des)*(@n-1)).
 !ifend !ifend
-comp @gsum= t(@col)*@des.
-comp @gmean= @gsum/@n.
-comp @mean= rsum(@gsum&*(@n-1)&*@w).
-comp @col= @col-@mean.
-comp @wss= t(@col&*@col)*@des&*@w.
-comp !5= (rsum((@gmean-@mean)&**2&*@n&**2&*@w)-rsum(@wss))/rsum(@wss&*(@n-1)).
-release @w,@gsum,@gmean,@mean,@wss.
!ifend
release @col,@des,@n.
!enddefine.


******** ВНУТРИКЛАССОВЫЙ КОЭФФИЦИЕНТ КОРРЕЛЯЦИИ (ОДНОФАКТОРНЫЙ, СБАЛАНСИРОВАННЫЕ КЛАССЫ) ********.
*/*!KO_iccowb(data%type%name)*/*.
*Version1.
*Эта функция вычисляет тот же ICC, что ф-ция /*!KO_iccow*/, но рассчитана на частный случай
*равночисленных (balanced) классов. Входящие тут должны быть оформлены в широком (wide) формате:
*DATA это матрица данных (не меньше 2x2) классы x замеры, т.е. каждый класс это ряд, все ряды
*содержат одно и то же число значений. Не следует придавать столбцам DATA смысл регулярных
*оценщиков, поскольку в данном случае принимается однофакторный случайный дизайн.
*TYPE - версия оценивателя (estimator) ICC, ключевое слово заглавными буквами (опционально в кавычках
*или апострофах) [см. Donner, 1986: A review of inference procedures...]
*"ANOVA" - ANOVA-оцениватель.
*"PAIR" - оцениватель Combinatorial pairing r Pearson.
*Т.к. группы одного размера, pairing оцениватель тождествен оценивателю Corrected Eta-square и
*оценивателю максимального правдоподобия, и он несмещенный, в отличие от ANOVA-оценивателя.
*NAME в данных иногда бывает отрицательным, хотя ICC в популяции неотрицателен; вы можете
*воспользоваться советом занулять отрицательные значения.

define !KO_iccowb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
!let !type= !unquote(!3)
comp @n= ncol(@data).
comp @sum= msum(@data).
comp @mean= @sum/(nrow(@data)*@n).
comp @sst= (mssq(@data)-@sum&*@mean).
comp @gmean= rsum(@data)/@n.
comp @ssb= cssq(@gmean-@mean)*@n.
!if (!type='PAIR') !then
-comp !4= 1/(@n-1)*(@n*@ssb/@sst-1).
!else
-comp @ssw= @sst-@ssb.
-comp @ssw= @ssw/(nrow(@data)*@n-nrow(@data)).
-comp @ssb= @ssb/(nrow(@data)-1).
-comp !4= (@ssb-@ssw)/(@ssb+(@n-1)*@ssw).
-release @ssw.
!ifend
release @data,@n,@sum,@mean,@gmean,@sst,@ssb.
!enddefine.


******** КОЭФФИЦИЕНТ ДИАГОНАЛЬНОСТИ МАТРИЦЫ ********.
*/*!KO_diagns(mat%name)*/*.
*Version 1.
*Один из возможных коэффициентов, измеряющий степень концентрации больших (по абсолютной величине)
*значений матрицы вблизи ее главной диагонали. Колеблется в пределах от 0 до 1.
*Пусть MAT произвольная матрица, тогда коэффициент есть 1-D/Dmax, где D это взвешенное манхэттенское
*физическое расстояние элементов от главной диагонали, равное Sum[abs(MAT)&*OFF], где OFF это
*матрица размером с MAT, главная диагональ в которой состоит из 0, пара прилегающих к ней диагоналей
*(т.е. отстоящих от нее на порядок 1) состоят из 1, пара диагоналей, отстоящих от нее на порядок 2
*состоят из 2, и так далее. (&* это перемножение соответствующих элементов, а Sum это сумма элементов.)
*Dmax это максимально возможное для данной MAT расстояние, т.е. наблюдаемое, когда все большие элементы
*MAT концентрируются у лево-нижнего и право-верхнего углов матрицы; Dmax вычисляется так же, как D,
*но после сортировки всех значений MAT по абс. величине и сортировке всех значений OFF по их величине.
*Если MAT не квадратна, то не одна, а несколько диагоналей в ней претендуют называться "главной диагональю".
*Это все диагонали (идущие слева-сверху вправо-вниз) одинаковой максимальной длины. Например,
*в матрице размером 4x6 это три диагонали, начинающиеся соответственно с позиций (1,1), (1,2) и (1,3).
*Функция выдаст, соответственно, три коэффициента, по одному в отношении каждой диагонали: результат NAME
*это вектор-столбец, содержащий столько коэффициентов, сколько "главных диагоналей" помещается в матрицу,
*и 1-й коэффициент относится к диагонали, начинающейся с позиции (1,1).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_diagns(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= abs(!2).
do if nrow(@mat)>ncol(@mat).
-comp @mat= t(@mat).
end if.
comp @matv= reshape(@mat,1,nrow(@mat)*ncol(@mat)).
comp @matv(grade(@matv))= @matv.
comp @off= make(nrow(@mat),1,1)*(1:ncol(@mat))-t(1:nrow(@mat))*make(1,ncol(@mat),1).
comp !3= make(abs(nrow(@mat)-ncol(@mat))+1,1,0).
loop @i= 1 to nrow(!3).
-comp @offi= abs(@off-@i+1).
-comp @offiv= reshape(@offi,1,nrow(@off)*ncol(@off)).
-comp @offiv(grade(@offiv))= @offiv.
-comp !3(@i)= msum(@mat&*@offi)/rsum(@matv&*@offiv).
end loop.
comp !3= 1-!3.
release @mat,@matv,@off,@offi,@offiv,@i.
!enddefine.


******** ЦЕНТРАЦИЯ СТОЛБЦОВ ДАННЫХ ********.
*/*!KO_center(data%name)*/*.
*Version 1.
*Центрует столбцы DATA (приводит среднее в них к 0).
*Выдает данные NAME.

define !KO_center(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
!enddefine.


******** Z-СТАНДАРТИЗАЦИЯ СТОЛБЦОВ ДАННЫХ ********.
*/*!KO_zscore(data%name)*/*.
*Version 1.
*Стандартизует столбцы DATA (приводит среднее в них к 0, а дисперсию к 1).
*Выдает данные NAME.

define !KO_zscore(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= !3/sqrt(cssq(!3)/(nrow(!3)-1))(make(1,nrow(!3),1),:).
!enddefine.


******** Z-СТАНДАРТИЗАЦИЯ СТОЛБЦОВ ДАННЫХ (DF=N) ********.
*/*!KO_zscore2(data%name)*/*.
*Version 1.
*Стандартизует столбцы DATA (приводит среднее в них к 0, а дисперсию к 1).
*Стандартное отклонение, стандартизующее данные, посчитано на "df=n" (а не "df=n-1").
*Выдает данные NAME.

define !KO_zscore2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3-(csum(!3)/nrow(!3))(make(1,nrow(!3),1),:).
comp !3= !3/sqrt(cssq(!3)/nrow(!3))(make(1,nrow(!3),1),:).
!enddefine.


******** МАСШТАБИРОВАНИЕ СТОЛБЦОВ ДАННЫХ К SS=1 ********.
*/*!KO_scale(data%name)*/*.
*Version 1.
*Масштабирует столбцы DATA (приводит сумму квадратов значений, SS, к 1).
*Выдает данные NAME.
*Если надо сделать SS=число_рядов DATA, умножьте результат на sqrt(число_рядов).

define !KO_scale(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3/sqrt(cssq(!3))(make(1,nrow(!3),1),:).
!enddefine.


******** ЦЕНТРАЦИЯ СТОЛБЦОВ ДАННЫХ, ПО ГРУППАМ ********.
*/*!KO_gcenter(data%dummy%name)*/*.
*Version 1.
*Центрует столбцы DATA (приводит среднее в них к 0), по группам наблюдений.
*Выдает данные NAME.
*DUMMY - двоичные фиктивные переменные, помечающие группы (каждая переменная соответствует группе,
*значение 1 = наблюдение принадлежит ей, 0 = не принадлежит ей).
*Всякий столбец DUMMY должен содержать по меньшей мере одну единицу.
*Группы должны быть непересекающиеся составом: сумма в рядах DUMMY не может превосходить единицу.
*Если в DUMMY есть ряды, полные нулей, т.е. наблюдения вне групп, то данные для них останутся не тронуты.
*Фиктивные переменные DUMMY вы можете создать из категориальной группирующей переменной
*с помощью функции /*!KO_freq*/.

define !KO_gcenter(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @des= !3.
comp @m= t(@des)*!4/(t(csum(@des))*make(1,ncol(!4),1)).
comp !4= !4-@des*@m.
release @des,@m.
!enddefine.


******** Z-СТАНДАРТИЗАЦИЯ СТОЛБЦОВ ДАННЫХ, ПО ГРУППАМ ********.
*/*!KO_gzscore(data%dummy%df%name)*/*.
*Version 2.
*Стандартизует столбцы DATA (приводит среднее в них к 0, а дисперсию к 1), по группам наблюдений.
*Выдает данные NAME.
*DUMMY - двоичные фиктивные переменные, помечающие группы (каждая переменная соответствует группе,
*значение 1 = наблюдение принадлежит ей, 0 = не принадлежит ей).
*Группы должны быть непересекающиеся составом, и каждое наблюдение должно принадлежать группе,
*т.е. сумма в каждом ряду DUMMY должна равняться единице.
*Всякий столбец DUMMY должен содержать по меньшей мере две единицы.
*DF - скаляр. Если положительный, стандартизующее стандартное отклонение будет посчитано на "df=n-1"
*(тогда в каждом столбце DUMMY должно быть не менее 2 единиц); если неположительный, стандартное
*отклонение будет посчитано на "df=n" (в каждом столбце DUMMY должно быть не менее 1 единицы).
*Фиктивные переменные DUMMY вы можете создать из категориальной группирующей переменной
*с помощью функции /*!KO_freq*/.

define !KO_gzscore(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @des= !3.
comp @df= !4.
comp @n= t(csum(@des))*make(1,ncol(!5),1).
comp @m= t(@des)*!5/@n.
comp !5= !5-@des*@m.
do if @df.
-comp @sd= sqrt(t(@des)*(!5&*!5)/(@n-1)).
else.
-comp @sd= sqrt(t(@des)*(!5&*!5)/@n).
end if.
comp !5= !5/(@des*@sd).
release @des,@df,@n,@m,@sd.
!enddefine.


******** МАСШТАБИРОВАНИЕ СТОЛБЦОВ ДАННЫХ, ПО ГРУППАМ ********.
*/*!KO_gscale(data%dummy%scale%name)*/*.
*Version 1.
*Масштабирует столбцы DATA, по группам наблюдений.
*Выдает данные NAME.
*SCALE - скаляр. Если неположительный, то сумма квадратов значений переменной в каждой группе
*будет приведена к 1; а если положительный, то - к численности в данной группе (т.е. средний квадрат будет 1).
*DUMMY - двоичные фиктивные переменные, помечающие группы (каждая переменная соответствует группе,
*значение 1 = наблюдение принадлежит ей, 0 = не принадлежит ей).
*Всякий столбец DUMMY должен содержать по меньшей мере одну единицу.
*Группы должны быть непересекающиеся составом, и каждое наблюдение должно принадлежать группе,
*т.е. сумма в каждом ряду DUMMY должна равняться единице.
*Фиктивные переменные DUMMY вы можете создать из категориальной группирующей переменной
*с помощью функции /*!KO_freq*/.

define !KO_gscale(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @des= !3.
comp @sc= !4.
comp @sqrn= 1.
do if @sc.
-comp @sqrn= t(sqrt(csum(@des)))*make(1,ncol(!5),1).
end if.
comp @norm= sqrt(t(@des)*(!5&*!5)).
comp !5= !5&*(@des*(@sqrn/@norm)).
release @des,@sc,@sqrn,@norm.
!enddefine.


******** ЛИНЕЙНОЕ ПЕРЕШКАЛИРОВАНИЕ ********.
*/*!KO_rescale(data%type%val1%val2%name)*/*.
*Version 1.
*Линейно перешкалирует переменные данных (столбцы DATA) к требуемым величинам статистик в этих данных,
*таких как минимум, максимум, средняя, ст. отклонение.
*Данные DATA должны состоять не менее чем из двух рядов и переменные (столбцы) должны не быть
*константами.
*Каждая переменная перешкалируется независимо; результат NAME в каждом столбце будет иметь
*линейно преобразованные данные, точно удовлетворяющие двум заданным значениям VAL1 и VAL2
*двух соответственных, согласно TYPE, статистик.
*VAL1 и VAL2 - скаляры либо векторы-ряды длиной как число столбцов DATA; вектор(ы) означает, что
*переменные надо привести каждую к своему заданному значению статистики.
*TYPE (ключевое слово заглавными буквами, опционально в кавычках или апострофах) задает,
*к каким двум статистикам привести данные в переменных:
*"MIN_MAX" - к заданным минимуму (VAL1) и максимуму (VAL2), (VAL1<VAL2)
*"MIN_MEAN" - к заданным минимуму (VAL1) и средней (VAL2), (VAL1<VAL2)
*"MAX_MEAN" - к заданным максимуму (VAL1) и средней (VAL2), (VAL1>VAL2)
*"MIN_SD" - к заданным минимуму (VAL1) и стандартному отклонению (VAL2), (VAL2 положительное)
*"MAX_SD" - к заданным максимуму (VAL1) и стандартному отклонению (VAL2), (VAL2 положительное)
*"MEAN_SD" - к заданным средней (VAL1) и стандартному отклонению (VAL2), (VAL2 положительное).

define !KO_rescale(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')')) 
!let !type= !unquote(!3)
comp !6= !2.
comp @val1= !4.
comp @val2= !5.
comp @ones= make(1,nrow(!6),1).
do if ncol(@val1)=1.
-comp @val1= make(1,ncol(!6),@val1).
end if.
do if ncol(@val2)=1.
-comp @val2= make(1,ncol(!6),@val2).
end if.
!if (!type='MIN_MAX') !then
-comp @max= cmax(!6).
-comp !6= ((@val2-@val1)/(@max-cmin(!6)))(@ones,:)&*(!6-@max(@ones,:))+@val2(@ones,:).
-release @max.
!else !if (!type='MIN_MEAN') !then
-comp !6= !6-(csum(!6)/nrow(!6))(@ones,:).
-comp !6= ((@val1-@val2)/cmin(!6))(@ones,:)&*!6+@val2(@ones,:).
!else !if (!type='MAX_MEAN') !then
-comp !6= !6-(csum(!6)/nrow(!6))(@ones,:).
-comp !6= ((@val1-@val2)/cmax(!6))(@ones,:)&*!6+@val2(@ones,:).
!else
-comp !6= !6-(csum(!6)/nrow(!6))(@ones,:).
-comp !6= !6/(sqrt(cssq(!6)/(nrow(!6)-1))(@ones,:)).
 !if (!type='MIN_SD') !then
- comp !6= (!6-cmin(!6)(@ones,:))&*@val2(@ones,:)+@val1(@ones,:).
 !else !if (!type='MAX_SD') !then
- comp !6= (!6-cmax(!6)(@ones,:))&*@val2(@ones,:)+@val1(@ones,:).
 !else
- comp !6= !6&*@val2(@ones,:)+@val1(@ones,:).
 !ifend !ifend
!ifend !ifend !ifend
release @val1,@val2,@ones.
!enddefine.


******** КОНВЕРТАЦИЯ КОВАРИАЦИОННОЙ МАТРИЦЫ В КОРРЕЛЯЦИОННУЮ ********.
*/*!KO_covcorr(cov%name)*/.
*Version 1.
*Конвертирует ковариационную матрицу в корреляционную (или sscp матрицу в косинусную).
*Полученная матрица NAME имеет единицы на диагонали.
*Входящая матрица COV должна иметь положительные диагональные элементы.

define !KO_covcorr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @vrnc= diag(!3).
comp !3= !3/sqrt(@vrnc*t(@vrnc)).
release @vrnc.
!enddefine.


******** КОНВЕРТАЦИЯ КОРРЕЛЯЦИОННОЙ МАТРИЦЫ В КОВАРИАЦИОННУЮ ********.
*/*!KO_corrcov(corr%dg%name)*/.
*Version 1.
*Конвертирует корреляционную матрицу в ковариационную (или косинусную матрицу в sscp матрицу).
*Полученная матрица NAME имеет на диагонали вектор DG.
*Входящая матрица CORR должна иметь единицы на диагонали.
*DG - вектор-столбец положительных значений, длиной как число рядов/столбцов CORR.

define !KO_corrcov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @vrnc= !3.
comp !4= !4&*sqrt(@vrnc*t(@vrnc)).
release @vrnc.
!enddefine.


******** ДВОЙНАЯ ЦЕНТРАЦИЯ МАТРИЦЫ КВАДРАТНЫХ РАССТОЯНИЙ ********.
*/*!KO_dcenter(dis%name)*/*.
*Version 1.
*Переводит матрицу квадратных различий DIS в матрицу сходств NAME, который суть скалярные произведения.
*Если DIS это квадратные различия между точками облака, то на диагонали NAME - квадратные евклидовы
*расстояния от точек до центроида облака, а внедиагональные элементы - скалярные произведения
*между векторами, исходящими из центроида и кончающимися на точках.
*Если различия это квадратные евклидовы расстояния, то данная операция есть "геометрически корректный"
*перевод их в сходства. Именно, тогда NAME=X*t(X), где матрица X - любые данные (точки на измерения,
*причем измерения, столбцы, центрованы), имеющие между точками кв. евклидовы расстояния, равные DIS. 
*В матрице-двойном центрате NAME рядные и столбцовые суммы равны 0.
*DIS должна быть квадратной симметрической с неотрицательными значениями, с 0 на диагонали.
*Для обратной конвертации используйте противоположную функцию /*!KO_sdcosth*/.

define !KO_dcenter(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @rmean= (rsum(!3)/ncol(!3))*make(1,ncol(!3),1).
comp !3= (@rmean+t(@rmean)-!3-msum(!3)/ncol(!3)**2)/2.
release @rmean.
!enddefine.


******** КОНВЕРТАЦИЯ МАТРИЦЫ УГЛОВЫХ СХОДСТВ В КВАДРАТНЫЕ РАССТОЯНИЯ ПО ТЕОРЕМЕ КОСИНУСОВ ********.
*/*!KO_sdcosth(sim%name)*/*.
*Version 1.
*Переводит матрицу сходств SIM, которые принимает за скалярные произведения, в матрицу квадратных
*различий NAME. Диагональные значения SIM используются как квадраты расстояний от точек до точки начала
*пространства. Если SIM это скалярные произвеления и SIM положительно полуопределена, то различия NAME это
*квадратные евклидовы расстояния. Если SIM не положительно полуопределенная, NAME иногда могут оказаться
*квадратными евклидовыми (сходиться в евклидовом пространстве).
*Данная операция - "геометрически корректная" конвертация ковариаций, корреляций или косинусов в
*соответствующие им квадратные евклидовы расстояния.
*SIM должна быть квадратной симметричной с положительными значениями на диагонали.
*Эта функция противоположна функциям /*!KO_dcenter*/ и /*!KO_dscosth*/.

define !KO_sdcosth(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')')) 
comp !3= !2.
comp @hsq= diag(!3)*make(1,ncol(!3),1).
comp !3= @hsq+t(@hsq)-2*!3.
release @hsq.
!enddefine.


******** ПЕРЕКЛЮЧЕНИЕ МАТРИЦЫ СКАЛЯРНЫХ ПРОИЗВЕДЕНИЙ НА НОВУЮ ДИАГОНАЛЬ (СОХРАНЯЯ УГОЛ) ********.
*/*!KO_swdiag1(sim%diag%name)*/.
*Version 1.
*Пересчитывает элементы матрицы SSCP-типа (= матрицы скалярных произведений; к ней относятся,
*в частности, ковариационная, корреляционная, косинусная матрицы) на заданные диагональные значения.
*При этом неизменными удерживаются углы (косинусы) между векторами.
*Этой функцией можно, например, превратить ковариационную матрицу в корреляционную или наоборот.
*SIM должна квадратной симметричной, с положительными элементами на диагонали. Матрица трактуется
*как матрица близостей SSCP-типа (матрица скалярных произведений).
*DIAG - вектор (ряд или столбец) длиной как число рядов/столбцов SIM, содержащий положительные элементы;
*это новая диагональ для SIM.
*Результат NAME это SIM с диагональю DIAG и пересчитанными внедиагональными элементами.
*Если входящая SIM грамова (все собственные числа неотрицательны), NAME тоже грамова.

define !KO_swdiag1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @diag= !3.
do if nrow(@diag)=1.
-comp @diag= t(@diag).
end if.
comp @coef= sqrt(@diag/diag(!4)).
comp !4= !4&*(@coef*t(@coef)).
call setdiag(!4,@diag).
release @diag,@coef.
!enddefine.


******** ПЕРЕКЛЮЧЕНИЕ МАТРИЦЫ СКАЛЯРНЫХ ПРОИЗВЕДЕНИЙ НА НОВУЮ ДИАГОНАЛЬ (СОХРАНЯЯ РАССТОЯНИЕ) ********.
*/*!KO_swdiag2(sim%diag%name)*/.
*Version 1.
*Пересчитывает элементы матрицы SSCP-типа (= матрицы скалярных произведений; к ней относятся,
*в частности, ковариационная, корреляционная, косинусная матрицы) на заданные диагональные значения.
*При этом неизменными удерживаются евклидовы расстояния между концами векторов.
*SIM должна квадратной симметричной, с положительными элементами на диагонали. Матрица трактуется
*как матрица близостей SSCP-типа (матрица скалярных произведений).
*DIAG - вектор (ряд или столбец) длиной как число рядов/столбцов SIM, содержащий в норме положительные
*элементы; это новая диагональ для SIM.
*Результат NAME это SIM с диагональю DIAG и пересчитанными внедиагональными элементами.
*В зависимости от DIAG (и, конечно, от входящей SIM), NAME может оказаться грамовой
*(все собственные числа неотрицательны) или неграмовой.

define !KO_swdiag2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @diag= !3.
do if nrow(@diag)=1.
-comp @diag= t(@diag).
end if.
comp @diff= (@diag-diag(!4))*make(1,ncol(!4),1).
comp !4= (@diff+t(@diff)+2*!4)/2.
call setdiag(!4,@diag).
release @diag,@diff.
!enddefine.


******** ПЕРЕКЛЮЧЕНИЕ MSCP-МАТРИЦЫ НА НОВЫЙ ЦЕНТРОИД ДАННЫХ ********.
*/*!KO_swcentr(sim%oldmean%newmean%name)*/.
*Version 1.
*Пересчитывает элементы MSCP матрицы, заданной столбцами неких данных, если тем данным сдвигом поменять
*центроид (многомерную среднюю). Функция не нуждается в самих данных. Допустим, что между некоторыми
*переменными была вычислена MSCP-матрица (это SSCP-матрица - т.е. матрица t(данные)*данные, - деленная на
*число наблюдений в данных); а также вычислен вектор средних у переменных. Затем данные утеряны. Какова
*была бы новая MSCP-матрица после того, как переменным поменять произвольно средние путем прибавления к
*каждой переменной некоторой константы? Функция отвечает на этот вопрос в отсутствие самих данных.
*Например, с помощью этой Функции можно восстановить MSCP-матрицу исходных данных из ковариационной
*матрицы (которая, как известно, тоже является MSCP-матрицей), если изначальные средние у переменных
*известны.
*SIM - квадратная симметрическая MSCP матрица между столбцами некоторых данных; средние в тех столбцах
*записаны как вектор (ряд или столбец) OLDMEAN. После сдвига этого центроида на новые координаты NEWMEAN
*новая MSCP-матрица, отвечающая сдвинутым данным, и есть матрица NAME, которая выдается.
*Учтите, что результатная матрица не обязана быть грамовой (положительно [полу]-определенной) -
*грамовость будет зависеть от входящей матрицы SIM и реалистичности OLDMEAN и NEWMEAN. Функция
*не следит за реалистичностью входящих. Когда входящие статистики происходят от реальных переменных,
*проблем с грамовостью должно не быть.

define !KO_swcentr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @omean= !3.
comp @nmean= !4.
do if ncol(@omean)=1.
-comp @omean= t(@omean).
end if.
do if ncol(@nmean)=1.
-comp @nmean= t(@nmean).
end if.
comp !5= !5+t(@nmean)*@nmean-t(@omean)*@omean.
release @omean,@nmean.
!enddefine.


******** ЭКСПОНЕНЦИАЛЬНОЕ НОРМИРОВАНИЕ (SOFTMAX-ФУНКЦИЯ) ********.
*/*!KO_nef(data%name)*/*.
*Version 1.
*Экспоненциально нормирует элементы в столбцах DATA.
*Результат NAME - в каждом столбце значения преобразованы экспоненциально так, что
*они лежат в диапазоне [0,1] и их сумма равна 1.

define !KO_nef(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @epv= exp(!3).
comp !3= @epv/csum(@epv)(make(1,nrow(!3),1),:).
release @epv.
!enddefine.


******** ЛОГ-ОТНОШЕНЧЕСКАЯ ЦЕНТРАЦИЯ (CENTERED LOGRATIO TRANSFORM) ********.
*/*!KO_clr(data%name)*/*.
*Version 1.
*Делает centered logratio преобразование элементов в столбцах DATA. Значения в DATA
*должны быть положительными.
*Преобразование заключается в приведении сначала суммы в столбце к 1, затем делении
*на геометрическую среднюю в столбце, и взятии логарифма.
*Результат NAME - в каждом столбце значения имеют нулевую среднюю.

define !KO_clr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp !3= !3/csum(!3)(make(1,nrow(!3),1),:).
comp @gmean= exp(csum(ln(!3))/nrow(!3)).
comp !3= ln(!3/@gmean(make(1,nrow(!3),1),:)).
release @gmean.
!enddefine.


******** ПРЕОБРАЗОВАНИЕ БОКСА-КОКСА ********.
*/*!KO_boxcox(col%grid%name1%name2)*/*.
*Version 1.
*Преобразует количественную переменную COL монотонно степенно так, чтобы распределение в ней стало
*ближе к нормальному. Поиск оптимальной величины степенного параметра лямбда делается поиском по
*решетке (перебором) так же, как это делает SPSS-команда ADP, осуществляющая такое преобразование.
*COL - входящая переменная, столбец положительных значений. Если имеются неположительрные значения,
*прибавьте предварительно константу, чтобы сделать все значения положительными (как вариант, можно
*всегда, независимо от наличия или отсутствия неположительных значений, сделать
*COMPUTE COL= COL-cmin(COL)+1, так поступает, например, команда ADP).
*GRID - вектор (ряд или столбец) из трех значений - это решетка для перебора значений лямбды:
*{min,max,step}. Max должно быть > Min, Step положительно, и (Max-Min)/Step должно быть целым.
*Например, часто используют решетку {-3,3,0.5}.
*Результат:
*NAME2 - выбранная степень лямбда преобразования Бокса-Кокса.
*NAME1 - преобразованные ею данные COL. (Команда ADP еще дополнительно z-стандартизует ее,
*но данная ф-ция этого не делает).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_boxcox(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @p= !3.
comp @b_= (@p(2)-@p(1))/@p(3).
comp @p= (@p(2)-@p(1))/@b_*(0:@b_)+@p(1).
comp @n= nrow(!4).
comp @lny= ln(!4).
comp !4= !4*make(1,ncol(@p),1).
loop @j= 1 to ncol(@p).
-comp @pj= @p(@j).
-do if @pj=0.
- comp !4(:,@j)= @lny.
-else.
- comp !4(:,@j)= (!4(:,@j)&**@pj-1)/@pj.
-end if.
end loop.
comp @vrnc= !4-(csum(!4)/nrow(!4))(make(1,nrow(!4),1),:).
comp @vrnc= cssq(@vrnc)/(nrow(@vrnc)-1).
comp @loglik= -@n/2*ln((1-1/@n)*@vrnc)+(@p-1)&*csum(@lny).
comp @j= rmax((@loglik=rmax(@loglik))&*(ncol(@loglik):1)).
comp @j= (ncol(@loglik)+1)*(@j>0)-@j.
comp !5= @p(@j).
comp !4= !4(:,@j).
release @p,@b_,@n,@lny,@j,@pj,@vrnc,@loglik.
!enddefine.


******** МАТРИЦА ОТБЕЛИВАНИЯ (СФЕРИЗАЦИИ) ********.
*/*!KO_white(cov%method%name)*/*.
*Version 1.
*Отбеливанием или сферизацией называется преобразование многомерных данных в "сферическое" облако.
*Функция берет ковариационную или корреляционную матрицу COV и вычисляет матрицу отбеливания
*(сферизации) NAME такую, что если COV соответствуют некоторые переменные X, то переменные Z=X*NAME
*будут иметь единичную ковариационную матрицу. Таким образом, отбеленные данные это декореллированные
*и стандартизованные переменные, а функция выдает матрицу преобразования для этого.
*COV - квадратная p x p симметричная матрица ковариаций или корреляций.
*METHOD - метод отбеливания, ключевое заглавное слово (можно взять в кавычки или апострофы):
*"ZCA" - нулевофазный компонентный метод; выдает NAME (симметричную при этом методе), минимизирующую
*квадратные различия между значениями Z и X. Употребляйте этот метод, если нужно, чтобы переменные Z
*минимально отличались от соответствующих им переменных X.
*"PCA" - главнокомпонентный метод; выдает NAME, максимизирующую учет дисперсии X в первых столбцах Z;
*Z есть ни что иное, как главные компоненты X, и значения Z это стандартизованные главнокомпонентные
*оценки. Употребляйте этот метод, если пожелаете снизить размерность (число столбцов) Z по сравнению
*с X, с наименьшими информационными потерями.
*"CHOL" - метод Холецкого; выдает NAME (верхнетреугольную при этом методе), обеспечивающую полную
*корреляцию Z1 с X1 и ортогональность Zp со всеми X, кроме Xp. Употребляйте, если нужно, чтобы Z1
*была связана со всеми X, особенно с X1; Z2 не была связана с X1; Z3 не была связана с X1 и X2, и т.д.
*При методах "ZCA" и "PCA" отбеленные переменные Z и их отношения с исходными переменными X будут
*зависеть от того, ковариационной или корреляционной является COV (поэтому существенно, ковариационную
*или корреляционную COV вы введете в функцию). Пусть матрица R есть ковариации (если COV ковариационная)
*или корреляции (если COV корреляционная) между X-переменными (ряды R) и Z-переменными (столбцы R).
*Тогда:
*-метод "ZCA" максимизирует диагональные элементы R (они положительны).
*-при методе "PCA" R есть матрица главнокомпонентных нагрузок, так что данный метод максимизирует
*в первую очередь сумму квадратов элементов 1-го столбца R (т.е. ковариации/корреляции Z1 со всеми X),
*во вторую очередь сумму квадратов элементов ее 2-го столбца (т.е. Z2 со всеми X), и т.д.
*При методе "CHOL" отбеленные переменные Z будут одинаковы что при ковариационной, что при
*корреляционной COV (поэтому не важно, корреляционную или ковариационную матрицу вы введете в функцию).
*Этот метод ведет к корреляционной матрице R такой, что она есть нижняя треугольная с положительной
*диагональю и элементом (1,1)=1, в последнем же ее столбце элементы кроме (p,p) - нули.
*Важно отметить, что для всех методов должно соблюдаться правило: если COV - корреляционная, то
*при отбеливании Z=X*NAME столбцы X должны быть стандартизованными переменными (а не сырыми или
*центрованными).
*Методы "ZCA" и "PCA" связаны между собой через поворот данных так, что ZCA-отбеливание эквивалентно
*PCA-отбеливанию с последующим прокрустовым поворотом Z обратно к X, и матрица такого поворота есть
*собственные векторы COV.

define !KO_white(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @cov= !2.
!let !method= !unquote(!3)
!if (!method='CHOL') !then
-comp !4= inv(chol(@cov)).
!else
-call eigen(@cov,@eivec,@eival).
 !if (!method='PCA') !then
- comp !4= @eivec/sqrt(t(@eival))(make(1,ncol(@cov),1),:).
- comp @cov= t(diag(!4)>=0).
- comp !4= !4&*(@cov-(not @cov))(make(1,ncol(!4),1),:).
 !else
- comp !4= @eivec/sqrt(t(@eival))(make(1,ncol(@cov),1),:)*t(@eivec).
 !ifend
-release @eivec,@eival.
!ifend
release @cov.
!enddefine.


******** ПРЕОБРАЗОВАТЬ ПЕРЕМЕННЫЕ ТОЧНО К ЗАДАННЫМ КОВАРИАЦИЯМ ********.
*/*!KO_tocov(ingot%type%cov%method%name)*/*.
*Version 2.
*Преобразование многомерных количественных данных INGOT такое, чтобы наблюдаемые ковариации (или
*корреляции или другие sscp-меры связи) между этими переменными стали точно равны заданным пользователем.
*Аргумент INGOT - n наблюдений x p переменных (n>p) матрица данных со значеними, служащими "сырьем" для
*получения вариат, связи между которыми станут удовлетворять матрице COV. INGOT могут быть случайные
*порожденные данные, которые вы ходите привести к нужной коррелированности между ними, или INGOT могут
*быть существующие анализируемые данные, которые вы хотите преобразовать минимально, заставив их
*воспроизвести заданную коррелированность (см. METHOD ниже). Столбцы INGOT должны быть неколлинеарны.
*Аргумент TYPE - ключевое слово заглавными буквами (опционально можно в кавычках или апострофах).
*Обозначьте, какого типа целевую матрицу вы задаете как аргумент COV: 
*"COV" - ковариационная матрица (на "df=n-1");
*"COV2" - ковариационная матрица (на "df=n") или mscp матрица;
*"CORR" - корреляционная матрица или матрица косинус-близости.
*(Все перечисленные матрицы родственны: они являются матрицами вида sscp(переменные)/"df".)
*Аргумент COV - p x p квадратная симметричная положительно определенная матрица ассоциаций (ее тип
*вы обозначили в TYPE). Это есть та целевая матрица, для которой надо получить из INGOT переменные,
*удовлетворяющие этой матрице.
*Результат NAME - данные n x p, столбцы которых это вариаты, точно воспроизводящие заданные
*коэффициенты COV.
*Пользуйтесь следующими правилами задания входящих:
*-Если целевая матрица COV ковариационная (на "df=n-1"), то type="COV", столбцы INGOT должны быть
* центрованы (центрировать переменные можно ф-цией /*!KO_center*/).
*-Если целевая матрица COV ковариационная (на "df=n"), то type="COV2", столбцы INGOT должны быть центрованы.
*-Если целевая матрица COV это средние квадраты и кросспроизведения (mscp), то type="COV2", INGOT это
* сырые данные как есть.
*-Если целевая матрица COV корреляционная (единица должна быть на диагонали), то type="CORR", столбцы INGOT
* должны быть центрованы. На выходе переменные сохранят свой начальный масштаб (дисперсию).
*-Если целевая матрица COV это косинус-близости (единица должна быть на диагонали), то type="CORR", столбцы
* INGOT это сырые данные как есть. На выходе переменные сохранят свой начальный масштаб (средний квадрат).
*Аргумент METHOD - скаляр, определяющий алгоритм подгонки. Если он положительный, то при преобразовании
*данных будет применены компонентные нагрузки целевой матрицы, прокрустово повернутые к нагрузкам
*данных INGOT. Используйте этот вариант, если хотите, чтобы значения NAME минимально отличались, в смысле
*суммарного квадрата разниц от значений INGOT: mssq(INGOT-NAME)=min. Другими словами, если стоит задача -
*"подправить" INGOT, доведя наличные данные до нужного коррелирования.
*Если аргумент METHOD неположительный, то при преобразовании данных будет применен корень Холецкого
*из целевой матрицы. Используйте этот более быстрый вариант, если хотите просто создать данные, т.е. если
*для вас INGOT - лишь "болванка" случайных значений, которые заранее порождены. Этот вариант не преследует
*задачу минимального изменения INGOT.
*Замечание: если INGOT это значения из нормального распределения, то NAME это данные тоже с
*многомерно-нормальным распределением; в ином же случае распределение в NAME не обязано соотноситься
*с распределением в INGOT.

define !KO_tocov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
!let !type= !unquote(!3)
comp @mx= !4.
comp @method= !5.
comp @mscp= sscp(@data)/(nrow(@data)-!eval(!type='COV')).
comp @s= diag(@mscp).
!if (!type='CORR') !then
-comp @mx= @mx&*sqrt(@s*t(@s)).
!ifend
call eigen(@mscp,@v,@s).
comp @s= t(sqrt(@s))(make(1,ncol(@data),1),:).
comp @u= @data*(@v/@s).
do if @method.
-call eigen(@mx,@eivec,@eival).
-comp @mx= @eivec&*t(sqrt(@eival))(make(1,ncol(@data),1),:).
-release @eivec,@eival.
-call svd(t(@v&*@s)*@mx,@u_,@s_,@v_).
-comp @mx= t(@mx*@v_*t(@u_)).
-release @u_,@s_,@v_.
else.
-comp @mx= chol(@mx).
end if.
comp !6= @u*@mx.
release @data,@mx,@method,@mscp,@s,@v,@u.
!enddefine.


******** ПРЕОБРАЗОВАТЬ ПЕРЕМЕННУЮ ТОЧНО К ЗАДАННЫМ КОВАРИАЦИЯМ С ДРУГИМИ ПЕРЕМЕННЫМИ ********.
*/*!KO_ytocov(yingot%xvars%type%cov%name)*/*.
*Version 1.
*Преобразует переменную YINGOT массива так, чтобы ее наблюдаемые ковариации (или корреляции или другие
*sscp-меры связи) с переменными XVARS данных стали точно равны заданным пользователем. Результат NAME -
*преобразованная YINGOT, измененная минимально в смысле наименьших квадратов: mssq(YINGOT-NAME)=min.
*XVARS не изменятся.
*Аргумент INGOT - столбец n наблюдений, это количественная переменная-"сырье" для получения переменной
*NAME. Вы можете породить YINGOT как случайные значения или это могут быть сушествующие анализируемые
*данные.
*Аргумент XVARS - одна или более количественных переменных массива, n x p; эти p столбцов должны быть
*неколлинеарны.
*Аргумент TYPE - ключевое слово заглавными буквами (опционально можно в кавычках или апострофах).
*Обозначьте, какого типа целевой вектор вы задаете как аргумент COV:
*"COV" - ковариации (на "df=n-1");
*"COV2" - ковариации (на "df=n") или средние кросспроизведения;
*"CORR" - корреляции или косинус-близости.
*(Все перечисленные виды родственны: они являются коэффициентами типа sscp(переменные)/"df".)
*Аргумент COV - вектор-ряд собственно целевых коэффициентов (их тип вы обозначили в TYPE). Это
*те коэффициенты связи с XVARS, ради которых будет преобразована YINGOT в NAME так, чтобы точно
*удовлетворять им. Длина ряда COV равна p+1. Первые p значений порядково отвечают p столбцам
*(переменным) XVARS, а последнее, добавочное значение это масштаб для переменной NAME (см. ниже).
*Пользуйтесь следующими правилами задания входящих:
*-Если ваши целевые коэффициенты COV это ковариации (на "df=n-1"), то type="COV", последним
* элементом вектора COV должна быть желаемая дисперсия NAME (положительное значение), YINGOT и столбцы
* XVARS должны быть центрованы (центрировать переменные можно ф-цией /*!KO_center*/).
*-Если ваши целевые коэффициенты COV это ковариации (смещенные, на "df=n"), то type="COV2", последним
* элементом вектора COV должна быть желаемая смещенная дисперсия NAME (положительное значение), YINGOT и
* столбцы XVARS должны быть центрованы.
*-Если ваши целевые коэффициенты COV это средние кросспроизведения, то type="COV2", последним
* элементом вектора COV должен быть желаемый средний квадрат NAME (положительное значение), YINGOT и
* XVARS это сырые данные как есть.
*-Если ваши целевые коэффициенты COV это корреляции, то type="CORR", последним элементом вектора COV должна
* быть единица, YINGOT и столбцы XVARS должны быть центрованы. NAME сохранит начальный масштаб
* (дисперсию) YINGOT.
*-Если ваши целевые коэффициенты COV это косинус-близости, то type="CORR", последним элементом вектора COV
* должна быть единица, YINGOT и XVARS это сырые данные как есть. NAME сохранит начальный масштаб
* (средний квадрат) YINGOT.
*Замечание1: Если заданные коэффициенты COV своей величиной не допускают создание корректной
*переменной NAME из переменной YINGOT (по причине возникшей неграмовой, не положительно-определенной
*конфигурации), то функция выдаст NAME= скаляр 0.
*Замечание2: если YINGOT это значения из нормального распределения, то NAME это данные тоже с
*нормальным распределением; в ином же случае распределение в NAME не обязано соотноситься
*с распределением в YINGOT.

define !KO_ytocov(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !type= !unquote(!4)
comp @y= !2.
comp @x= !3.
comp @mx= !5.
comp @p= ncol(@x).
comp @df= nrow(@x)-!eval(!type='COV').
comp @mscp= sscp({@x,@y})/@df.
comp @ms= diag(@mscp).
!if (!type='CORR') !then
-comp @mx= @mx&*t(sqrt(@ms*@ms(nrow(@ms)))).
!ifend
comp @res= @y-@x*solve(@mscp(1:@p,1:@p),@mscp(1:@p,ncol(@mscp))).
call eigen(@mscp(1:@p,1:@p),@v,@s).
comp @s= sqrt(@s).
comp @u= @x*(@v/t(@s)(make(1,@p,1),:)).
comp @xdual= @u/t(@s)(make(1,nrow(@x),1),:)*t(@v).
release @u,@s,@v,@x.
comp @mscp= sscp(@xdual)/@df.
comp @mxy= @mx(ncol(@mx)).
comp @mx= @mx(1:@p).
comp @s2= (@mxy-@mx*@mscp*t(@mx))/(cssq(@res)/@df).
do if @s2.
-comp !6= @xdual*t(@mx)+sqrt(@s2)*@res.
else.
-comp !6= 0.
end if.
release @y,@mx,@p,@df,@mscp,@ms,@res,@xdual,@s2.
!enddefine.


******** НОРМАЛЬНОЕ ОБЛАКО В РАВНОМЕРНЫЙ ШАР ********.
*/*!KO_unifball(data%alpha%name1%name2)*/*.
*Version 1.
*Берет одно- или многомерные данные DATA, которые функция ожидает быть из
*стандартного нормального распределения. Функция уменьшает эксцесс данных, поджимая
*периферию (хвосты) сильнее, чем околоцентральные области. При параметре ALPHA=1
*нормальное облако превращается в равномерное, в (гипер)шар.
*DATA - данные (наблюдения x переменные) из нормального распределения со средней 0 и
*дисперсией 1; если переменных более одной, функция ожидает, что они независимы, некоррелируют.
*(Вы можете породить случайные стандартные ортогональные нормальные данные функцией /*!KO_normal*/,
*а стандартизовать переменные, если необходимо - функцией /*!KO_zscore*/.)
*ALPHA - скаляр, параметр степени поджатия периферии (хвостов) данных. Чем он выше, тем сильнее
*функция уменьшит эксцесс в данных. При параметре 1 нормальное облако превратится в равномерное,
*причем если данные многомерны, это будет равномерный (гипер)шар. При параметре 0 данные останутся
*как были на входе. При параметре между 0 и 1 вы получите данные с эксцессом промежуточным между
*нормальным и равномерным. Параметр выше 1 уже делает данные бимодальными (вогнутыми в центре).
*Формула преобразования данных следующая:
*Y_i = X_i * coef,
*coef = 1 + alpha * (sqrt(p+2)*cdf.chisq(d^2,p)^(1/p)/d - 1),
*где Y_i и X_i - значения точки i данных на выходе (NAME1) и на входе (DATA);
*коэффициент поджатия coef выдается как NAME2;
*p - размерность, число столбцов DATA;
*d - евклидово расстояние от точки i в DATA до начала координат, т.е. до 0 (согласно сказанному
*выше, центроид DATA должен лежать в 0 или стремиться к 0);
*cdf.chisq(val,df) - ф-ция кумулятивного распределения хи-квадрат;
*фактор sqrt(p+2) сохраняет исходную ожидаемую изменчивость (т.е. 1 для каждого измерения);
*а если ALPHA=1, то гипершар NAME1 ожидает радиус sqrt(p+2).
*(См.: https://stats.stackexchange.com/questions/79919).
*Коррелирующие переменные. Хотя функция уменьшит эксцесс в таком облаке и сделает эллипсоид
*более шарообразным, она не добъется равномерного распределения под ALPHA=1. Функция выполняет
*свою роль, только когда входящие переменные некоррелируют. Если вы хотите коррелирующие переменные
*из нормального распределения превратить к эллипс с равномерным распределением внутри него,
*сделайте сначала отбеливание данных (сферизацию плюс стандартизацию), например, посредством
*анализа главных компонент (ф-ция /*!KO_pcomp*/). Затем примените данную функцию, и в конце сделайте
*преобразование, обратное тому анализу главных компонент - см. ПРИМЕР.

define !KO_unifball(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @alpha= !3.
comp @p= ncol(!4).
comp @d= rssq(!4).
comp !5= 1+@alpha*(chicdf(@d,@p)&**(1/@p)*sqrt(@p+2)/sqrt(@d)-1).
comp !4= !4&*(!5*make(1,@p,1)).
release @alpha,@p,@d.
!enddefine.


******** ЛИНЕЙНАЯ РЕГРЕССИЯ ********.
*/*!KO_regress(dv%ivs%const%name1%name2%name3)*/*.
*Version 1.
*Линейная регрессия наименьших квадратов (без тестов значимости).
*Берет зависимую переменную DV и одну или более независимых переменных IVS (последние должны быть неколлинеарны).
*Аргумент CONST (скаляр) - нужна модель с константой (CONST положителен) или без константы (CONST неположителен).
*Результаты:
*NAME1 - SSregression, SSresiduals, наблюдаемый коэффициент детерминации R-square.
*NAME2 - двухстолбцовая матрица регрессионных коэффициентов: в 1-м столбце сырые коэффициенты; во 2-м столбце
*стандартизованные коэффициенты (beta). Если модель была с константой, то число рядов в NAME2 на 1 больше
*числа столбцов IVS, и константа находится в последнем ряду.
*NAME3 - предсказанные значения (сырые) зависимой переменной.

define !KO_regress(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                  /!pos= !charend('%') /!pos= !charend(')'))
comp @dv= !2.
comp !7= !3.
comp @const= !4.
comp @p= ncol(!7).
comp @p1= @p+1.
do if @const.
-comp @mean= csum({!7,@dv})/nrow(@dv).
-comp @scp= sscp({!7,@dv}-@mean(make(1,nrow(!7),1),:)).
-comp !6= make(@p1,2,0).
else.
-comp @scp= sscp({!7,@dv}).
-comp !6= make(@p,2,0).
end if.
comp !6(1:@p,1)= solve(@scp(1:@p,1:@p),@scp(1:@p,@p1)).
comp @ss= diag(@scp).
comp !6(1:@p,2)= !6(1:@p,1)&*sqrt(@ss(1:@p)/@ss(@p1)).
comp !7= !7*!6(1:@p,1).
do if @const.
-comp !6(@p1,1)= @mean(@p1)-@mean(1:@p)*!6(1:@p,1).
-comp !7= !7+!6(@p1,1).
-comp !5= cssq(!7-@mean(@p1)(make(1,nrow(!7),1),:))/@ss(@p1).
-release @mean.
else.
-comp !5= cssq(!7)/@ss(@p1).
end if.
comp !5= {!5*@ss(@p1),@ss(@p1)-!5*@ss(@p1),!5}.
release @dv,@const,@p,@p1,@scp,@ss.
!enddefine.


******** ГЛАВНЫЕ КОМПОНЕНТЫ ********.
*/*!KO_pcomp(data%df%m%out%name1%name2%name3%name4%name5)*/*.
*Version 3.
*Анализ главных компонент.
*Берет данные DATA (где ряды - точки данных, столбцы - переменные, оси пространства) и выдает главные
*компоненты этих данных.
*Внимание: данные анализируются "как есть", поэтому если нужна предварительная центрация или
*стандартизация, используйте прежде соответствующие функции. (Центрация соответствует анализу ковариаций,
*а стандартизация - анализу корреляций.)
*Аргумент DF (скаляр) - если положительный, разлагаться будет матрица sscp(DATA)/(n-1), а если неположительный,
*то матрица sscp(DATA)/n, где n это число рядов в DATA.
*Аргумент M (скаляр) - сколько первых главных компонент выделять: укажите целое положительное число от 1
*до число_столбцов DATA. Или укажите любое неположительное число, тогда выделены будут все ненулевые компоненты.
*Аргумент OUT (скаляр) управляет формой выдачи результатов.
*Результаты:
*NAME1 содержит в 1-м столбце собственные числа разложенной матрицы,
*и во 2-м столбце - доли объясненной дисперсии (или масштаба), т.е. доли от следа разложенной матрицы.
*NAME2 - собственные векторы (если OUT неположителен) либо нагрузки (если OUT положителен); сумма квадратов
*в столбцах нагрузок = собственные числа. Нагрузки это коэффициенты предсказания переменных масштабированными
*компонентами. Собственные векторы это коэффициенты предсказания переменных сырыми компонентами, или косинусы углов поворота.
*NAME3 - перешкалированные нагрузки (если OUT положителен) или собственные векторы NAME2 (если OUT неположителен).
*Перешкалированная или стандартизованная нагрузка это нагрузка, деленная на масштаб переменной - на корень из
*соответствующего диагонального элемента разложенной матрицы. Этим делением снят эффект неодинаковости масштабов (дисперсий)
*переменных на нагрузки. Перешкалированная нагрузка это косинус-сходство или корреляция между компонентой и переменной.
*Квадраты элементов собственных векторов можно понимать как величину вкладов переменных в компоненты.
*Квадраты элементов перешкалированных нагрузок можно понимать как величину вкладов компонент в переменные.
*NAME4 - регрессионные коэффициенты вычисления главнокомпонентных оценок по переменным. Если OUT
*неположителен, эта матрица тождественна собственным векторам NAME2. Если OUT положителен, эта матрица вычисляется
*из нагрузок NAME2. Матрицу NAME4 вы можете использовать для подсчета главнокомпонентных оценок данных, не участвовавших
*в анализе.
*NAME5 - главнокомпонентные оценки. Они есть DATA*NAME4. Если OUT неположителен, это сырые оценки:
*сумма квадратов в столбцах NAME5 = собственные числа матрицы sscp(DATA)). Если OUT положителен, это
*масштабированные оценки: сумма квадратов в столбцах NAME5 = n-1 либо n, в зависимости от аргумента DF.

define !KO_pcomp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !10= !2.
comp @df= !3.
comp @m= !4.
comp @out= !5.
call eigen(sscp(!10)/(nrow(!10)-(@df>0)),!7,!6).
comp !6= {!6,!6/csum(!6)}.
do if not @m.
-comp @m= csum(!6(:,1)>0).
end if.
comp !6= !6(1:@m,:).
comp !7= !7(:,1:@m).
comp @eipos= !6(:,1)>0.
comp !6(:,1)= !6(:,1)&*@eipos.
comp @ones= make(1,nrow(!7),1).
comp !7= !7&*t(@eipos)(@ones,:).
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(@ones,:).
do if @out.
-comp !7= !7&*t(sqrt(!6(:,1)))(@ones,:).
-comp !8= !7/(t(sqrt(cssq(!10)/(nrow(!10)-(@df>0))))*make(1,@m,1)).
-comp !9= !7/t(!6(:,1)+(not @eipos))(@ones,:).
else.
-comp !8= !7.
-comp !9= !7.
end if.
comp !10= !10*!9.
release @df,@m,@out,@eipos,@sign,@ones.
!enddefine.


******** ГЛАВНЫЕ КОМПОНЕНТЫ (ДЛЯ N<P) ********.
*/*!KO_pcomp2(data%df%m%out%name1%name2%name3%name4%name5)*/*.
*Version 1.
*Анализ главных компонент.
*Эта функция тождественна ф-ции /*!KO_pcomp*/, но быстрее ее, когда число столбцов в данных больше
*чем число рядов. И медленнее ее, когда число столбцов в данных меньше числа рядов.

define !KO_pcomp2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @df= !3.
comp @m= !4.
comp @out= !5.
comp @df= nrow(@data)-(@df>0).
call eigen(sscp(t(@data))/@df,@u,!6).
comp !6= {!6,!6/csum(!6)}.
do if not @m.
-comp @m= csum(!6(:,1)>0).
end if.
comp !6= !6(1:@m,:).
comp @u= @u(:,1:@m).
comp @eipos= !6(:,1)>0.
comp !6(:,1)= !6(:,1)&*@eipos.
comp @u= @u&*t(@eipos)(make(1,nrow(@u),1),:).
comp !7= t(@data)*@u/sqrt(@df).
do if @out.
-comp !10= @u*sqrt(@df).
-comp !8= !7/(t(sqrt(cssq(@data)/@df))*make(1,@m,1)).
-comp !9= !7/t(!6(:,1)+(not @eipos))(make(1,nrow(!7),1),:).
else.
-comp !10= @u*sqrt(@df)&*t(sqrt(!6(:,1)))(make(1,nrow(@data),1),:).
-comp !7= !7/t(sqrt(!6(:,1)))(make(1,nrow(!7),1),:).
-comp !8= !7.
-comp !9= !7.
end if.
comp @sign= csum(!7)>=0.
comp @sign= @sign-(not @sign).
comp !7= !7&*@sign(make(1,nrow(!7),1),:).
comp !8= !8&*@sign(make(1,nrow(!8),1),:).
comp !10= !10&*@sign(make(1,nrow(!10),1),:).
comp !9= !9&*@sign(make(1,nrow(!9),1),:).
release @data,@df,@m,@out,@u,@eipos,@sign.
!enddefine.


******** ГЛАВНЫЕ КОМПОНЕНТЫ (ВХОДЯЩИЕ - МАТРИЦА) ********.
*/*!KO_pca(cov%w%m%out%name1%name2%name3%name4)*/*.
*Version 1.
*Анализ главных компонент (с опцией взвешивания переменных).
*Эта функция, в отличие от /*!KO_pcomp*/, берет в качестве входящих не данные, а уже готовую матрицу
*(ковариаций и т.п.), и она сама не вычисляет главнокомпонентных оценок.
*Есть опция придать неодинаковые веса (важности) переменным, т.е. исполнить взвешенный PCA.
*Входящие:
*COV - квадратная симметричная матрица коэффициентов sscp-типа - т.е. корреляционная, ковариационная,
*косинусная или сырая sscp-матрица коэффициентов связи между переменными.
*W - вектор длиной как число рядов/столбцов COV, содержащий веса переменным; либо неположительный скаляр
*(что значит - не давать веса, результаты будут тождественны с /*!KO_pcomp*/). Веса должны быть положительными
*числами, находящимися в нужном вам соотношении (только соотношение величин весов играет роль).
*Вес ровно 0 не разрешен, но вы можете придать очень близкий к нулю положительный вес, если хотите
*сделать некую переменную "пассивной", supplementary. 
*Аргументы M и OUT - те же, что в функции /*!KO_pcomp*/.
*Результаты:
*NAME1, NAME2, NAME3, NAME4 - аналогичны тем, что в функции /*!KO_pcomp*/.
*Матрицу коэффициентов компонентных оценок NAME4 вы можете использовать для вычисления главнокомпонентных
*оценок, каковые будут равны DATA*NAME4, где DATA - данные наблюдения x переменные (если COV это
*ковариационная матрица, переменные должны быть центрованы, если COV это корреляционная матрица,
*переменные должны быть z-стандартизованы, и т.п., т.е. должно быть соответствие с матрицей).
*Ввод неграмовой COV:
*Если среди M выделяемых компонент случится таковая с отрицательным собственным числом в NAME1,
*т.е. мнимая, то соответствующие ей результаты в NAME2, NAME3, NAME4 будут посчитаны, словно эта компонента
*реальная (с положительным собственным числом), но эти ее результаты следует считать мнимыми.

define !KO_pca(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @w= !3.
comp @m= !4.
comp @out= !5.
comp @dg= diag(@r).
do if nrow(@w)>ncol(@w).
-comp @w= t(@w).
end if.
do if ncol(@w)=1 and not @w(1).
-comp @w= make(1,ncol(@r),1).
else.
-comp @w= @w/rsum(@w)*ncol(@w).
-comp @r= @r&*sscp(sqrt(@w)).
end if.
call eigen(@r,!7,!6).
comp @einzero= abs(!6)>1E-13.
comp !6= !6&*@einzero.
comp !6= {!6,!6/csum(!6)}.
do if not @m.
-comp @m= csum(!6(:,1)>0).
end if.
comp !6= !6(1:@m,:).
comp @einzero= @einzero(1:@m).
comp !7= !7(:,1:@m).
comp @eival= abs(!6(1:@m,1)).
comp @ones= make(1,nrow(!7),1).
comp !7= !7&*t(@einzero)(@ones,:).
comp !7= !7/(t(sqrt(@w))*make(1,ncol(!7),1)).
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(@ones,:).
comp @w= t(@w)*make(1,ncol(!7),1).
do if @out.
-comp !7= !7&*t(sqrt(@eival))(@ones,:).
-comp !8= !7/(sqrt(@dg)*make(1,ncol(!7),1)).
-comp !9= !7&*@w/t(@eival+(not @einzero))(@ones,:).
else.
-comp !8= !7.
-comp !9= !7&*@w.
end if.
release @r,@w,@m,@out,@dg,@einzero,@eival,@sign,@ones.
!enddefine.


******** ОРТОГОНАЛЬНОЕ ПРОКРУСТОВО ВРАЩЕНИЕ ********.
*/*!KO_procr(fit%targ%refl%iso%name1%name2)*/*.
*Version 1.
*Осуществляет ортогональный прокрустов поворот подгоночной конфигурации FIT в целевую
*конфигурацию TARG. Подогнанная (повернутая) подгоночная конфигурация выдается как NAME2,
*а матрица поворота как NAME1.
*FIT и TARG - матрицы одинакового размера. Ряды в них - это точки данных, а столбцы - измерения (оси)
*пространства. Один и тот же ряд в двух матрицах есть пара взаимосоответственных точек.
*Измерения пространства у FIT и TARG у каждой свои. Прокрустово вращение это такой поворот осей FIT
*в оси TARG, который обеспечивает максимальную совмещенность взаимосоответственных точек.
*NAME2 это координаты точек FIT в пространстве осей TARG после поворота.
*Если число измерений в FIT и TARG неодинаково, добавьте к одной из этих матриц заполненные нулем
*столбцы, т.к. функция требует, чтобы число столбцов было там и там одинаково.
*В матрице поворота NAME1 ряды отвечают осям FIT, а столбцы отвечают осям TARG.
*REFL - скаляр; если положительный, повороту разрешено быть с отражением (если это улучшит подгонку);
*если неположительный, отражение запрещено (подгонка может оказаться хуже).
*ISO - скаляр; если положительный, после поворота осуществляется дополнительная операция
*изомасштабирования для лучшей подгонки; если неположительный, изомасштабирования не делается. Этот
*аргумент не влияет на матрицу поворота NAME1.
*Функция берет данные FIT и TARGET "как есть". Если требуется совмещение центроидов и/или уравнивание
*масштабов двух этих конфигураций до операции прокрустова поворота, сделайте эти преобразования заранее.

define !KO_procr(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @a= !2.
comp @b= !3.
comp @refl= !4.
comp @iso= !5.
call svd(t(@b)*@a,@u,@s,@v).
comp @s= diag(@s).
comp !6= @v*t(@u).
do if not @refl and not det(!6).
-comp @v(:,ncol(@v))= -@v(:,ncol(@v)).
-comp @s(nrow(@s))= -@s(nrow(@s)).
-comp !6= @v*t(@u).
end if.
release @b,@u,@v.
comp !7= @a*!6.
do if @iso.
-comp !7= !7*(csum(@s)/trace(sscp(@a))).
end if.
release @a,@refl,@iso,@s.
!enddefine.


******** ЛИНЕЙНЫЕ ДИСКРИМИНАНТЫ ********.
*/*!KO_discrim(data%dummy%m%out%name1%name2%name3%name4%name5)*/*.
*Version 2.
*Осуществляет линейный дискриминантный анализ (без классификации и тестов значимости): выдает дискриминанты.
*Берет данные DATA (ряды - точки, столбцы - оси пространства) и группирование DUMMY.
*Столбцы DATA (переменные) должны быть уже центрованы (или стандартизованы). Столбцы DATA должны быть
*не коллинеарны.
*DUMMY - двоичные фиктивные переменные, помечающие группы (каждая переменная соответствует группе,
*значение 1 = наблюдение принадлежит ей, 0 = не принадлежит ей).
*Всякий столбец DUMMY должен содержать по меньшей мере одну единицу.
*Группы должны быть непересекающиеся составом, и каждое наблюдение должно принадлежать группе,
*т.е. сумма в каждом ряду DUMMY должна равняться единице.
*Аргумент M (скаляр) - сколько дискриминант выделять. Укажите целое положительное число от 1 до
*min(число_групп-1,число_переменных). Либо укажите любое неположительное число: тогда выделятся все
*ненулевые дискриминанты.
*Аргумент OUT (скаляр) управляет формой выдачи результатов.
*Результаты:
*NAME1 - двухстолбцовая матрица: собственные числа в 1-м столбце, канонические корреляции во 2-м.
*NAME2 - собственные векторы с нормированными столбцами (если OUT неположителен) либо сырые (нестандартизованные)
*канонические дискриминантные коэффициенты (если OUT положителен).
*NAME3 - стандартизованные канонические дискриминантные коэффициенты (веса переменных в дискриминантах).
*NAME4 - корреляции между переменными и дискриминантами (матрица структуры).
*NAME5 - дискриминантные оценки, они есть DATA*NAME2, т.е посчитаны с помощью нормированных собственных
*векторов (если OUT неположителен) либо с помощью сырых дискриминантных коэффициентов (если OUT положителен).
*Если к матрице нестандартизованных коэфициентов необходим вектор констант (для получения нецентрованных
*дискриминант), то он равен -csum(mdiag(MEAN)*NAME2), где MEAN это вектор средних исходных переменных
*и NAME2 это нестандартизованные канонические дискриминантные коэффициенты.
*Фиктивные переменные DUMMY вы можете создать из категориальной группирующей переменной
*с помощью функции /*!KO_freq*/.

define !KO_discrim(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !10= !2.
comp @des= !3.
comp @k= ncol(@des).
comp @m= !4.
comp @out= !5.
comp @b= t(@des)*!10/(t(csum(@des))*make(1,ncol(!10),1)).
comp @b= sscp(@des*@b).
comp @w= sscp(!10)-@b.
comp @invc= inv(chol(@w)). 
comp @invcb= t(@invc)*@b*@invc.
comp @invcb= (@invcb+t(@invcb))/2.
call eigen(@invcb,!7,@eival).
do if not @m.
-comp @m= rmin({@k-1,ncol(!10),csum(@eival>0)}). 
end if.
comp !6= make(@m,2,0).
comp @eival= @eival(1:@m).
comp !6(:,1)= @eival.
comp !6(:,2)= sqrt(!6(:,1)/(!6(:,1)+1)).
comp !7= (@invc*!7)(:,1:@m).
comp !7= !7&*t(@eival>0)(make(1,nrow(!7),1),:).
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(make(1,nrow(!7),1),:).
release @b,@invc,@invcb,@eival,@sign.
comp @dgw= sqrt(diag(@w))*make(1,@m,1).
comp !8= !7&*@dgw.
comp !9= @w*!7/@dgw.
do if @out.
-comp !7= sqrt(nrow(!10)-@k)*!7.
else.
-comp @dgw= cssq(!7).
-comp !7= !7/(sqrt(@dgw)+(@dgw=0))(make(1,ncol(!10),1),:).
end if.
comp !10= !10*!7.
release @k,@out,@m,@w,@dgw.
!enddefine.


******** ГЛАВНЫЕ КООРДИНАТЫ ********.
*/*!KO_pcoord(dis%m%method%name1%name2)*/*.
*Version 1.
*Анализ главных координат (= метрическое многомерное шкалирование Торгерсона).
*Берет матрицу различий DIS, которые принимает за квадратные (sic!) расстояния, и укладывает эти
*объекты в евклидово пространство, выдавая координаты объектов как NAME2.
*Измерения (столбцы NAME2) являются некоррелированными переменными.
*DIS долна быть квадратной симметричной, число_объектов x число_объектов.
*M (целый скаляр) - нужная размерность евклидового пространста: целое число от 1 до число_объектов,
*либо любое неположительное число (тогда M будет = все реально существующие
*нелишние измерения).
*Аргумент METHOD (скаляр) это метод разложения матрицы-двойного центрата, промежуточного продукта DIS:
*eigen-разложение (при METHOD положительно) или svd-разложение (если METHOD неположительно).
*Все собственные или сингулярные числа от этих разложений выдаются как вектор NAME1.
*Различие этих методов состоит в следующием: возможные отрицательные собственные числа двойного
*центрата обнаруживаются eigen-разложением, и функция не использует их далее, но корректирует ими сумму
*положительных собственных чисел. Отрицательные собственные числа не распознаются svd-разложением и выходят
*как младшие положительные сингулярные числа. С точки зрения точности репродукции итоговыми евклидовыми
*расстояниями (расстояниями между объектами в полученной конфигурации NAME2) входящих различий DIS вариант
*с eigen-разложением как правило предпочтителен.
*Если DIS это квадратные евклидовы расстояния, тогда оба метода одинаковы, причем координаты NAME2 тогда
*тождественны сырым комнонентным оценкам, которые были бы получены анализом главных компонент тех данных
*(центрованных), из которых можно вычислить матрицу расстояний DIS. (Это свидетельствует от тесной
*родственной связи между анализом главных координат и анализом главных компонент.).

define !KO_pcoord(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @m= !3.
comp @method= !4.
comp @rmean= rsum(@mat)/ncol(@mat)*make(1,ncol(@mat),1).
comp @mat= (@rmean+t(@rmean)-@mat-msum(@mat)/ncol(@mat)**2)/2.
release @rmean.
do if @method.
-call eigen(@mat,@v,!5).
-comp !5= !5&*(abs(!5)>1E-13).
-comp @eival= !5&*(!5>0).
-comp @eival= @eival*(trace(@mat)/csum(@eival)).
else.
-call svd(@mat,@u,@eival,@v).
-release @u.
-comp @eival= diag(@eival).
-comp @eival= @eival&*(@eival>1E-13).
-comp !5= @eival.
end if.
comp @sign= csum(@v)>=0.
comp @v= @v&*(@sign-(not @sign))(make(1,nrow(@v),1),:).
do if not @m.
-comp @m= csum(@eival>0).
end if.
comp !6= @v(:,1:@m)*mdiag(sqrt(@eival(1:@m))).
release @mat,@m,@method,@eival,@v,@sign.
!enddefine.


******** БИПЛОТ ********.
*/*!KO_biplot(data%rw%cw%norm%m%name1%name2%name3%name4)*/*.
*Version 1.
*Биплот с опцией взвешивания рядов и столбцов.
*Выдает координаты точек-рядов и точек-столбцов, и другие статистики.
*Эта функция также может считаться делающей взвешенный анализ главных компонент (см. ниже).
*DATA - произвольная матрица данных (не менее двух рядов и столбцов). Анализ берет данные как есть, поэтому при
*необходимости заранее сделайте нужную стандартизацию.
*RW - вес для каждого ряда DATA (вектор-столбец) либо неположительный скаляр (что значит нет взвешивания,
*программа придаст равный вес всем рядам).
*CW - вес для каждого столбца DATA (вектор-ряд) либо неположительный скаляр (что значит нет взвешивания,
*программа придаст равный вес всем столбцам).
*Задаваемые вами веса могут быть любыми неотрицательными числами, только соотношение величин весов играет роль.
*Заданный вес 0 для ряда/столбца означает, что данный ряд/столбец должен быть пассивным (добавочным,
*supplementary). Всякий положительный вес означает активный ряд/столбец.
*NORM - вектор из двух чисел от 0 до 1. Это коэффиенты нормирования, или наделения инерцией, координат
*точек-рядов (первое число) и точек-столбов (второе число). Коэффициент 0 означает "стандартное нормирование",
*когда масштаб координат стандартизован к 1. Коэффициент 1 означает "главное нормирование", когда весь масштаб
*собственных чисел измерений отдан масштабу координат. Например, {0,1} есть стандартное нормирование для
*точек-рядов и главное нормирование для точек-столбцов; {1,1} есть главное нормирование для тех и других;
*{.5,.5} есть так называемое "симметричное нормирование".
*M - сколько измерений (главных осей) выделять: целое число от 1 до
*min(число_активных_рядов,число_активных_столбцов) DATA. Либо укажите любое неположительное число - тогда
*выделены будут все ненулевые измерения.
*Результаты:
*NAME1 - двухстолбцовая матрица: собственные числа (масштаб, инерция) измерений (1-й столбец) и доля
*объясненной суммарной инерции в DATA (2-й столбец).
*В NAME2, NAME3, NAME4 ряды это точки биплота: сначала идут точки-ряды DATA, потом, под ними, точки-столбцы DATA.
*NAME2 - двухстолбцовая матрица: масса точки (1-й столбец) и инерция точки (2-й столбец).
*Масса активного ряда или столбца пропорциональна его весу. Масса пассивного ряда или столбца пропорциональна
*усредненному весу активных рядов или столбцов, соответственно.
*Инерция точки это ее масштаб в нормированной (взвешенной) матрице DATA. Если взвешивания не было или веса всех
*активных рядов/столбцов были равными, тогда инерция ряда/столбца есть его сумма квадратов в DATA, деленная на r*c
*(r число активных рядов, c число активных столбцов).
*NAME3 - координаты точек рядов и столбцов в пространстве M измерений. Характер нормирования инерцией
*(аргумент NORM) сказывается на координатах.
*NAME4 - вклады точек в инерцию измерений (первые M столбцов) и вклады измерений в инерцию точек
*(последние M столбцов). Вклады точек в измерения это квадраты эл-тов собственных векторов. Точки с высоким
*вкладом в инерцию измерения это точки, кому преимущественно обязано измерение своим появлением. У пассивных
*точек вклад 0, т.к. они не влияют на формирование измерений. Вклад измерения в инерцию точки это показатель
*полноты объяснения данной точки данным измерением; высокий вклад означает, что точка хорошо репрезентируется
*данным измерением и его достаточно для ее "объяснения".
*ОТНОШЕНИЕ к анализу главных компонент (PCA). Биплот и PCA почти тождественны: единственное различие состоит
*в том, что биплот приводит сумму весов столбцов входящих данных к 1, а PCA - к числу (активных) столбцов.
*Это делает результаты двух методов пропорционально тождественными, и данную функцию можно использовать для
*исполнения анализа главных компонент, в том числе PCA с взвешиванием важности столбцов и/или рядов данных.
*Конкретно, пусть C это число активных столбцов, тогда верны следующие отношения для  пересчета:
*1) собс. числа PCA = собс. числа биплота * C;
*2) нагрузки = координаты столбцов под "главным нормированием" столбцов;
*3) стандартизованные компонентные оценки = координаты рядов под "стандартным нормированием" рядов;
*4) собственные векторы PCA = координаты столбцов под "стандартным нормированием" столбцов / sqrt(C);
*5) сырые компонентные оценки = координаты рядов под "главным нормированием" рядов * sqrt(C).

define !KO_biplot(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @rw= !3.
comp @cw= !4.
comp @norm= !5.
comp @m= !6.
do if nrow(@rw)=1 and not @rw(1).
-comp @rw= make(nrow(@data),1,1).
end if.
do if ncol(@cw)=1 and not @cw(1).
-comp @cw= make(1,ncol(@data),1).
end if.
comp @rw= @rw/csum(@rw).
comp @cw= @cw/rsum(@cw).
comp @posrw= @rw>0.
comp @poscw= @cw>0.
comp @rm= @rw+(not @posrw)/csum(@posrw).
comp @cm= @cw+(not @poscw)/rsum(@poscw).
comp @dat= @data&*sqrt(@rm*@cm).
comp @ri= rssq(@dat&*@poscw(make(1,nrow(@dat),1),:)).
comp @ci= cssq(@dat&*(@posrw*make(1,ncol(@dat),1))).
comp !8= {@rm,@ri;t(@cm),t(@ci)}.
comp @dat= @data&*sqrt(@rw*@cw).
call svd(@dat,@u,@s,@v).
release @dat.
comp @s= diag(@s)(1:rmin({csum(@posrw),rsum(@poscw)})).
comp @s= @s&*(@s>1E-13).
comp !7= @s&**2.
comp !7= {!7,!7/csum(!7)}.
comp @m_= csum(!7(:,1)>0).
do if not @m.
-comp @m= @m_.
end if.
comp !7= !7(1:@m,:).
comp @mm_= @m-@m_.
comp @m= rmin({@m,@m_}).
comp @s= @s(1:@m).
comp @u= @u(:,1:@m).
comp @v= @v(:,1:@m).
comp !10= {@u&**2;@v&**2}.
comp @u= @u/((sqrt(@rw)+(not @posrw))*make(1,ncol(@u),1)).
comp @v= @v/(t(sqrt(@cw)+(not @poscw))*make(1,ncol(@v),1)).
comp @sign= csum({@u;@v})>=0.
comp @sign= @sign-(not @sign).
comp @u= @u&*@sign(make(1,nrow(@u),1),:).
comp @v= @v&*@sign(make(1,nrow(@v),1),:).
comp !9= {@data&*@cw(make(1,nrow(@data),1),:)*@v&*(t(@s)&**(@norm(1)-1))(make(1,nrow(@data),1),:);
        t(@data&*(@rw*make(1,ncol(@data),1)))*@u&*(make(ncol(@data),1,1)*t(@s)&**(@norm(2)-1))}.
comp !10= {!10,{!9(1:nrow(@data),:)&**2&*(@rm/@ri*t(@s&**(2-2*@norm(1))));
                !9((nrow(@data)+1):nrow(!9),:)&**2&*t(@s&**(2-2*@norm(2))*(@cm/@ci))}}.
do if @mm_.
-comp @mm_= make(nrow(!9),@mm_,0).
-comp !9= {!9,@mm_}.
-comp !10= {!10(:,1:@m),@mm_,!10(:,(@m+1):ncol(!10)),@mm_}.
end if.
release @data,@rw,@cw,@norm,@m,@m_,@mm_,@posrw,@poscw,@rm,@cm,@ri,@ci,@s,@u,@v,@sign.
!enddefine.


******** ПРОСТОЙ АНАЛИЗ СООТВЕТСТВИЙ ********.
*/*!KO_corresp(tab%sta%rsup%csup%norm%m%name1%name2%name3%name4)*/*.
*Version 1.
*Простой (двувходовый) анализ соответствий. Является приложением идеи биплота к анализу таблицы сопряженности.
*Выдает координаты точек-рядов и точек-столбцов, и другие статистики.
*TAB - таблица сопряженности (не менее двух рядов и столбцов). Это любая таблица с неотрицательными
*элементами, в которой значение в ячейке имеет смысл сопряженности, близости между данным рядом и данным
*столбцом. В TAB суммы в рядах от активных столбцов должны быть ненулевыми и суммы с столбцах от активных
*рядов должны быть ненулевыми (активные ряды/столбцы это не пассивные).
*STA - нужная стандартизация для таблицы, целое число от 1 до 6:
*1 - двухвходовая центрация и взвешенное нормирование
*2 - двухвходовая центрация и простое нормирование
*3 - центрация рядов и простое нормирование
*4 - центрация столбцов и простое нормирование
*5 - уравнивание рядных сумм, центрация рядов и простое нормирование
*6 - уравнивание столбцовых сумм, центрация столбцов и простое нормирование
*Опция (1) есть хи-квадратная модель анализа, она оптимальна, если TAB это частотная таблица. Остальные пять
*опций на выбор это евклидова модель анализа, которая подходит, если TAB это не частоты (а, например, суммы
*или средние). Хи-квадратная модель подразумевает биплот со взвешиванием.
*RSUP - вектор номеров рядов, которые надо считать пассивными (добавочными, supplementary), либо любой
*неположительный скаляр, если пассивные не нужны.
*CSUP - аналогично, вектор номеров столбцов, которые надо считать пассивными, либо любой неположительный
*скаляр. Ряды и столбцы TAB, не являющиеся пассивными, считаются активными.
*NORM - вектор из двух чисел от 0 до 1. Это коэффиенты нормирования, или наделения инерцией, координат
*точек-рядов (первое число) и точек-столбов (второе число). Коэффициент 0 означает "стандартное нормирование",
*когда масштаб координат стандартизован к 1. Коэффициент 1 означает "главное нормирование", когда весь масштаб
*собственных чисел измерений отдан масштабу координат. Например, {0,1} есть стандартное нормирование для
*точек-рядов и главное нормирование для точек-столбцов; {1,1} есть главное нормирование для тех и других;
*{.5,.5} есть так называемое "симметричное нормирование".
*M - сколько измерений выделять: целое число от 1 до min(число_активных_рядов-1,число_активных_столбцов-1) TAB.
*Либо укажите любое неположительное число - тогда выделены будут все ненулевые измерения.
*Результаты (они того же вида, что в биплоте /*!KO_biplot*/):
*NAME1 - двухстолбцовая матрица: собственные числа (масштаб, инерция) измерений (1-й столбец) и доля
*объясненной суммарной инерции в преобразованной стандартизацией TAB (2-й столбец).
*В NAME2, NAME3, NAME4 ряды это точки биплота: сначала идут точки-ряды TAB, потом, под ними, точки-столбцы TAB.
*NAME2 - двухстолбцовая матрица: масса точки (1-й столбец) и инерция точки (2-й столбец).
*Масса активного ряда или столбца пропорциональна его весу. Масса пассивного ряда или столбца пропорциональна
*усредненному весу активных рядов или столбцов, соответственно.
*Инерция точки это ее масштаб (сумма квадратов) в стандартизованной TAB.
*NAME3 - координаты точек рядов и столбцов в пространстве M измерений. Характер нормирования инерцией
*(аргумент NORM) сказывается на координатах.
*NAME4 - вклады точек в инерцию измерений (первые M столбцов) и вклады измерений в инерцию точек (последние M
*столбцов). Вклады точек в измерения это квадраты эл-тов собственных векторов. Точки с высоким вкладом в
*инерцию измерения это точки, кому преимущественно обязано измерение своим появлением. У пассивных точек
*вклад 0, т.к. они не влияют на формирование измерений. Вклад измерения в инерцию точки это показатель полноты
*объяснения данной точки данным измерением; высокий вклад означает, что точка хорошо репрезентируется данным
*измерением и его достаточно для ее "объяснения".

define !KO_corresp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
               /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @tab= !2.
comp @sta= !3.
comp @rsup= !4.
comp @csup= !5.
comp @norm= !6.
comp @m= !7.
comp @nar= nrow(@tab).
comp @nac= ncol(@tab).
comp @rn= rsum(@tab).
comp @cn= csum(@tab).
comp @n= rsum(@cn).
do if @rsup(1).
-comp @nar= @nar-nrow(@rsup)*ncol(@rsup).
-comp @cn= @cn-csum(@tab(@rsup,:)).
-comp @n= rsum(@cn).
end if.
do if @csup(1).
-comp @nac= @nac-nrow(@csup)*ncol(@csup).
-comp @rn= @rn-rsum(@tab(:,@csup)).
-comp @n= csum(@rn).
end if.
do if @rsup(1) and @csup(1).
-comp @n= @n-csum(@rn(@rsup)).
end if.
do if @sta=1.
-comp @rm= @rn/@n.
-comp @cm= @cn/@n.
-comp @tb= @tab-@rn*@cn/@n.
-comp @tab= @tb.
-comp @tb= @tb/sqrt(@rn*@cn).
else.
-comp @rm= make(nrow(@tab),1,1/@nar).
-comp @cm= make(1,ncol(@tab),1/@nac).
-do if @sta=2.
- comp @tb= @tab-@rn*@cn/@n.
-else if @sta=3.
- comp @tb= @tab-(@rn/@nac)*make(1,ncol(@tab),1).
-else if @sta=4.
- comp @tb= @tab-(@cn/@nar)(make(1,nrow(@tab),1),:).
-else if @sta=5.
- comp @tb= @tab&*((@n/(@nar*@rn))*make(1,ncol(@tab),1)).
- comp @tb= @tb-@n/@nar/@nac.
-else if @sta=6.
- comp @tb= @tab&*(@n/(@nac*@cn))(make(1,nrow(@tab),1),:).
- comp @tb= @tb-@n/@nac/@nar.
-end if.
-comp @tb= @tb*@nar*@nac/@n.
-comp @tab= @tb.
-comp @tb= @tb/sqrt(@nar*@nac).
end if.
comp @ri= rssq(@tb).
comp @ci= cssq(@tb).
do if @rsup(1).
-comp @ci= @ci-cssq(@tb(@rsup,:)).
end if.
do if @csup(1).
-comp @ri= @ri-rssq(@tb(:,@csup)).
end if.
comp !9= {@rm,@ri;t(@cm),t(@ci)}.
do if @rsup(1).
-comp @tb(@rsup,:)= make(nrow(@rsup)*ncol(@rsup),ncol(@tb),0).
end if.
do if @csup(1).
-comp @tb(:,@csup)= make(nrow(@tb),nrow(@csup)*ncol(@csup),0).
end if.
call svd(@tb,@u,@s,@v).
release @tb.
comp @s= diag(@s).
comp @s= @s(1:(rmin({@nar,@nac})-1)).
comp @s= @s&*(@s>1E-13).
comp !8= @s&**2.
comp !8= {!8,!8/csum(!8)}.
comp @m_= csum(!8(:,1)>0).
do if not @m.
-comp @m= @m_.
end if.
comp !8= !8(1:@m,:).
comp @mm_= @m-@m_.
comp @m= rmin({@m,@m_}).
comp @s= @s(1:@m).
comp @u= @u(:,1:@m).
comp @v= @v(:,1:@m).
comp !11= {@u&**2;@v&**2}.
do if @sta=1.
-comp @u= @u/(sqrt(@rn/@n)*make(1,ncol(@u),1)).
-comp @v= @v/(t(sqrt(@cn/@n))*make(1,ncol(@v),1)).
-comp @denr= @rn*make(1,@m,1).
-comp @denc= t(@cn)*make(1,@m,1).
else.
-comp @u= @u*sqrt(@nar).
-comp @v= @v*sqrt(@nac).
-comp @denr= @nac.
-comp @denc= @nar.
end if.
comp @sign= csum({@u;@v})>=0.
comp @sign= @sign-(not @sign).
comp @u= @u&*@sign(make(1,nrow(@u),1),:).
comp @v= @v&*@sign(make(1,nrow(@v),1),:).
comp !10= {@tab*@v&*(t(@s)&**(@norm(1)-1))(make(1,nrow(@tab),1),:)/@denr; 
         t(@tab)*@u&*(t(@s)&**(@norm(2)-1))(make(1,ncol(@tab),1),:)/@denc}.
comp !11= {!11,{!10(1:nrow(@tab),:)&**2&*(@rm/@ri*t(@s&**(2-2*@norm(1))));
                !10((nrow(@tab)+1):nrow(!10),:)&**2&*t(@s&**(2-2*@norm(2))*(@cm/@ci))}}.
do if @mm_.
-comp @mm_= make(nrow(!10),@mm_,0).
-comp !10= {!10,@mm_}.
-comp !11= {!11(:,1:@m),@mm_,!11(:,(@m+1):ncol(!11)),@mm_}.
end if.
release @tab,@rsup,@csup,@norm,@m,@m_,@mm_,@nar,@nac,@rm,@cm,@ri,@ci,@s,@u,@v,@sign,@denr,@denc.
!enddefine.


******** ФАКТОРЫ МЕТОДОМ ГЛАВНЫХ ОСЕЙ ********.
*/*!KO_paf(cov%comm%m%conv%name1%name2%name3%name4%name5)*/*.
*Version 1.
*Линейный факторный анализ (только извлечение факторных нагрузок из матрицы; функция не делает
*вращений факторов и вычисления факторных оценок). Метод главных осей применяется для экстракции.
*Входящие:
*COV - квадратная p x p симметричная матрица коэффициентов sscp-типа - т.е. корреляционная, ковариационная,
*косинусная или сырая sscp-матрица коэффициентов связи между переменными.
*COMM - вектор-столбец начальных оценок общностей, либо неположительный скаляр. Если вектор, то он должен
*быть длиной как диагональ COV (т.е. равен числу переменных p) и содержать положительные числа, не
*превышающие соответствующие значения этой диагонали (т е масштабы, дисперсии переменных).
*Если COMM неположительный скаляр, функция будет использовать в качестве начальных общностей образы
*(images) матрицы COV - они как правило являются лучшими кандидатами на начальные общности. Эта опция
*требует положительно определенную COV.
*M - сколько факторов выделять, укажите целое положительное число, меньшее чем число переменных
*(вы должны заранее решить, используя известные методы оценки, сколько выделять факторов).
*CONV - параметры итеративного схождения: вектор из двух чисел, например {.001,25}. Первое это порог
*изменения общностей - неотрицательное число; если изменение будет меньше, общности считаются
*стабилизировавшимися и итерации прекращаются. Второе это максимально позволенное число итераций -
*задайте целое положительное число.
*Результаты:
*NAME1 - M x 4 матрица масштабов (дисперсий) выделенных факторов и долей объясненного ими полного масштаба.
*1-й столбец это масштабы (дисперсии) факторов как есть, равные M первым собственным числам
*редуцированной COV на последней итерации. 2-й столбец это доля этих масштабов от полного масштаба,
*равного следу входящей COV. 3-й столбец это перешкалированные масштабы (дисперсии), т.е. такие, из
*которых удален эффект неодинаковости масштабов (дисперсий) переменных. 4-й столбец относится к 3-му как
*2-й относится к 1-му.
*Если в NAME1 окажется фактор с отрицательной дисперсией, то он мнимый; однако увеличение числа итераций
*часто переводит его в реальный; либо уменьшите M.
*NAME2 - p x M матрица факторных нагрузок, как они есть. Их столбцовые суммы квадратов равны масштабам
*факторов, а рядные суммы квадратов - итоговым общностям.
*NAME3 - перешкалированные нагрузки. Это нагрузки, из которых удален эффект неодинаковости масштабов
*(дисперсий) переменных; они суть корреляции (или косинусы) между факторами и переменными и удобны для
*интерпретации факторов.
*NAME4 - начальные и итоговые оценки общностей, p x 4. Столбцы слева направо: начальные как есть,
*итоговые как есть, начальные перешкалированные, итоговые перешкалированные. Перешкалированная общность
*это общность, деленная на масштаб (дисперсию) переменной, и является долей масштаба переменной,
*объясняемой общими факторами.
*Если COV это корреляционная или косинусная матрица, то "сырые" и перешкалированные результаты
*в NAME1, в NAME2/NAME3, в NAME4 будут тождественны.
*NAME5 - отчет о схождении, вектор из двух чисел, который можно сравнить с аргументом CONV:
*первое число это максимальное изменение общностей, случившееся на последней проделанной итерации и
*которое, ожидается, должно быть меньше порога в CONV. Второе число - число проделанных итераций; если
*это число больше на 1 чем заданное в CONV, значит, было проделано максимальное число итераций, но
*заказанная точность схождения так и не была достигнута.
*Случай Хейвуда. Если на какой-либо итерации общность переменной превысит масштаб переменной, итерации
*обрываются, результаты выдаются существующие на момент обрыва. О возникшем случае Хейвуда видно по
*значению, превышающему единицу, в последнем столбце NAME4; кроме того, функция ради внимания
*даст второму числу в NAME5 отрицательный знак. Результаты в случае Хейвуда следует признать
*несостоявшимися.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_paf(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
           /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @inic= !3.
comp @m= !4.
comp @conv= !5.
comp @dg= diag(@r).
do if nrow(@inic)*ncol(@inic)=1 and not @inic(1).
-comp @inic= @dg-1/diag(inv(@r)).
end if.
comp !9= @inic.
loop @i= 1 to @conv(2).
-comp @rc= !9/@dg.
-call setdiag(@r,!9).
-call eigen(@r,!7,!6).
-comp @eival= abs(!6(1:@m)).
-comp !7= !7(:,1:@m)&*t(sqrt(@eival))(make(1,nrow(!7),1),:).
-comp !9= rssq(!7).
-comp !10= cmax(abs(!9/@dg-@rc)).
end loop if !10<@conv(1) or any(!9>@dg).
comp !10= {!10,@i}.
do if any(!9>@dg).
-comp !10(2)= -!10(2).
end if.
comp @sign= csum(!7)>=0.
comp !7= !7&*(@sign-(not @sign))(make(1,nrow(!7),1),:).
comp !8= !7/(sqrt(@dg)*make(1,@m,1)).
comp @ssload= {!6(1:@m),t(cssq(!8))}.
comp @ssload(:,2)= @ssload(:,2)&*((@ssload(:,1)>=0)-(@ssload(:,1)<0)).
comp !6= {@ssload(:,1),@ssload(:,1)/csum(@dg),@ssload(:,2),@ssload(:,2)/nrow(!8)}.
comp !9= {@inic,!9,@inic/@dg,!9/@dg}.
release @r,@inic,@m,@conv,@dg,@i,@rc,@eival,@sign,@ssload.
!enddefine.


******** ОРТОГОНАЛЬНЫЕ АНАЛИТИЧЕСКИЕ ВРАЩЕНИЯ ********.
*/*!KO_ortrot(load%kaiser%method%conv%reord%name1%name2%name3)*/*.
*Version 1.
*Делает разные виды ортогональных аналитических вращений, обобщенно известных как "ортомакс",
*которые обычно употребляются для вращения экстрагированных факторов в факторном анализе.
*LOAD - входящая матрица нагрузок ортогональных факторов (p переменных x m факторов,
*p и m >1), которую надо вращать (результат вращения - матрица NAME1). LOAD это не обязательно
*факторные нагрузки, это могут быть просто координаты каких-то p точек в пространстве m
*перпендикулярных осей.
*KAISER - скаляр. Если положительный, то перед вращением LOAD подвергнется нормированию Кайзера
*(а после вращения результат подвергнется обратному разнормированию). Нормирование Кайзера это
*приведение сумм квадратов в рядах LOAD к единице. Если скаляр неположительный, то
*нормирования Кайзера не будет.
*METHOD - метод вращения, кл. слово заглавными буквами (можно взять в кавычки или апострофы).
*Это критерий, который максимизируется при данном методе:
*"QUARTIMAX" - квартимакс; он равен сумме итоговых нагрузок (итоговых координат), возведенных
*в 4-ю степень: msum(NAME1^4).
*Остальные методы имеют формулу критерия p*Q-c*V, где Q это критерий квартимакс, а V это сумма
*квадратов дисперсий итоговых факторов, столбцов NAME1: rsum(cssq(NAME1)^2).
*Коэффициент c определяет метод:
*"VARIMAX" - варимакс: c=1
*"EQUAMAX" - эквамакс: c=m/2
*"PARSIMAX" - парзимакс: c=p*(m-1)/(p+m-2)
*"FACPARS" - факпарз (factor parsimony): c=p
*"CUSTOM c" - произвольное значение c: задайте ему значение (число или имя скаляра). Значение может
*быть любым. Чем выше c, тем дальше конфигурация будет от "генерального фактора".
*С приближением c к +бесконечности дисперсии факторов, cssq(NAME1), будут равными.
*С приближением c к -бесконечности дисперсии факторов будут равны таковым, как если подвергнуть LOAD
*вращению в свои главные компоненты (без предварительной центрации).
*CONV - параметры итеративного схождения: вектор из двух чисел, например {.0001,50}. Первое это
*порог изменения критерия - неотрицательное число; если изменение будет меньше порога, критерий
*считается стабилизировавшимся и итерации прекращаются (см. NAME3). Второе - это максимально
*позволенное число итераций - задайте целое положительное число.
*REORD - скаляр. Если положительный, то факторы после вращения, столбцы NAME1, будут упорядочены
*по убыванию их дисперсий (т.е. сумм квадратов в столбцах), а также в каждом столбце знак нагрузкам
*будет назначен так, чтобы сумма нагрузок была в нем положительна. Если скаляр неположительный,
*эти две послевращательные операции делаться не будут.
*Результаты:
*NAME1 - итоговая матрица нагрузок (координат).
*NAME2 - матрица ортогонального поворота, так что NAME1=LOAD*NAME2.
*NAME3 - история итераций, это вектор значений критерия на выходе из итераций. 1-е значение в
*векторе это величина критерия для входящей LOAD; 2-е значение это величина критерия после 1-й
*итерации вращения, и т.д. Последнее значение в векторе это величина критерия на выходе из последней
*потребовавшейся итерации. Всего итераций, таким образом, nrow(NAME3)-1.
*Значения критерия растут от итерации к итерации вплоть до стабилизации или исчерпания предела
*итераций. У квартимакса и варимакса значения всегда положительные, но у остальных методов они могут
*быть и отрицательными, это зависит от входящих данных.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Замечание: если факторный/главнокомпонентный анализ делался на ковариациях (а не корреляциях), то
*нагрузки бывают двух видов – сырые и перешкалированные (последние получаются делением первых на
*ст. отклонения переменных). Больше смысла вращать к "простой структуре" – именно перешкалированные
*нагрузки, т.к. интерпретируют обычно их. Впрочем, если вы делаете вращение с нормированием Кайзера, то
*результаты вращения перешкалированных и сырых нагрузок будут эквивалентны, т.к. NAME2 тогда одна и та же.

define !KO_ortrot(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                 /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !method= !unquote(!4)
comp !7= !2.
comp @kais= !3.
comp @conv= !5.
comp @reord= !6.
comp @p= nrow(!7).
comp @m= ncol(!7).
comp @h= 0.
do if @kais.
-comp @h= sqrt(rssq(!7)).
-comp !7= !7/(@h*make(1,@m,1)).
end if.
comp @a2= !7&*!7.
!if (!method='QUARTIMAX') !then !let !crit= 'mssq(@a2)'
!else !if (!method='VARIMAX') !then !let !crit= '@p*mssq(@a2)-rssq(csum(@a2))' !let !c1= '2/@p' !let !c2= '1/@p'
!else !if (!method='EQUAMAX') !then !let !crit= '@p*mssq(@a2)-@m/2*rssq(csum(@a2))' !let !c1= '@m/@p' !let !c2= '@m/(2*@p)'
!else !if (!method='PARSIMAX') !then !let !crit= '@p*(mssq(@a2)-(@m-1)/(@p+@m-2)*rssq(csum(@a2)))' !let !c1= '2*(@m-1)/(@p+@m-2)' !let !c2= '(@m-1)/(@p+@m-2)'
!else !if (!method='FACPARS') !then !let !crit= '@p*(mssq(@a2)-rssq(csum(@a2)))' !let !c1= '2' !let !c2= '1'
!else !let !crit= '@p*mssq(@a2)-@c*rssq(csum(@a2))' !let !c1= '2*@c/@p' !let !c2= '@c/@p'
-comp @c= !tail(!method).
!ifend !ifend !ifend !ifend !ifend
comp !9= !crit.
comp !8= ident(@m).
loop @it= 1 to @conv(2).
-loop @i= 1 to @m-1.
- loop @j= @i+1 to @m.
-  comp @ij= {@i,@j}.
-  comp @ai= !7(:,@i).
-  comp @aj= !7(:,@j).
-  comp @u= @ai&*@ai-@aj&*@aj.
-  comp @v= 2*@ai&*@aj.
-  comp @num= 2*csum(@u&*@v).
-  comp @den= csum(@u&*@u-@v&*@v).
   !if (!method<>'QUARTIMAX') !then
-   comp @us= csum(@u).
-   comp @vs= csum(@v).
-   comp @num= @num-!c1*@us*@vs.
-   comp @den= @den-!c2*(@us*@us-@vs*@vs).
   !ifend
-  comp @den= @den+(@den=0)*1E-14.
-  comp @phi= artan(@num/@den).
-  do if @den.
-   comp @phi= @phi/4.
-  else if @num.
-   comp @phi= (3.14159265358979+@phi)/4.
-  else.
-   comp @phi= (@phi-3.14159265358979)/4.
-  end if.
-  comp @sinphi= sin(@phi).
-  do if abs(@sinphi)>1E-14.
-   comp @cosphi= cos(@phi).
-   comp @qij= {@cosphi,-@sinphi;@sinphi,@cosphi}.
-   comp !7(:,@ij)= {@ai,@aj}*@qij.
-   comp !8(:,@ij)= !8(:,@ij)*@qij.
-  end if.
- end loop.
-end loop.
-comp @a2= !7&*!7.
-comp !9= {!9;!crit}.
end loop if abs(!9(nrow(!9))-!9(@it))<=@conv(1).
do if @kais.
-comp !7= !7&*(@h*make(1,@m,1)).
end if.
do if @reord.
-comp @h= csum(!7)>=0.
-comp !7= !7&*(@h-(not @h))(make(1,@p,1),:).
-comp @h= grade(-cssq(!7)).
-comp !7(:,@h)= !7.
-comp !8(:,@h)= !8.
end if.
release @kais,@conv,@p,@m,@reord,@h,@a2,@it,@i,@j,@ij,@ai,@aj,@u,@v,@num,@den,@phi,@sinphi,@cosphi, @qij
 !if (!method<>'QUARTIMAX') !then ,@us,@vs !ifend !if (!head(!method)='CUSTOM') !then ,@c !ifend.
!enddefine.


******** КОСОУГОЛЬНОЕ ВРАЩЕНИЕ ПРОМАКС И ПРОМАЙ ********.
*/*!KO_promax(ort%kaiser%k%name1%name2%name3)*/*.
*Version 1.
*Делает вращение промакс или промай ортогональной конфигурации ORT, которую функция принимает за
*"уже близкую к простой структуре". ORT это матрица факторных нагрузок, уже повернутая
*к "простой структуре" ортогонально; в классическом варианте - методом варимакс (используйте для
*этого ф-цию */*!KO_ortrot*/). Промакс рассматривает ORT^K как целевую конфигурацию - более идеальную
*"простую структуру", чем достигнуто в ORT (возведение элементов ORT в степень K>1 с сохранением их
*знака сближает малые нагрузки с нулем скорее, чем большие нагрузки, "заостряя" паттерн) и пытается
*приблизить ORT к ORT^K, ослабив ортогональность осей (факторов) неортогональным прокрустовым
*вращением, причем ORT^K принимается за целевое отображение (нагрузки, а не за корреляции).
*Итог: косоугольное отображение NAME1, полученное из ортогонального ORT.
*Промай похож на промакс, но создает целевую для вращения конфигурацию не возведением в степень K,
*а другим способом (см. Mulaik, Foundations of factor analysis, 2010).
*Входящие:
*ORT - p x m (p>=m) матрица нагрузок или каких-то координат в ортогональной (декартовой) системе.
*ORT должна быть полноранговой, т.е. t(ORT)*ORT обратимой. ORT рассматривается как близкая к
*"простой структуре", но еще ортогональная.
*KAISER - скаляр. Если положительный, то целевая конфигурация будет основана на ORT, подвергнутой
*нормированию Кайзера. Это рекомендованный вариант (нормирование Кайзера это приведение сумм
*квадратов в рядах к единице). Если скаляр неположительный, то целевая конфигурация будет основана
*на ORT как она есть. В последнем случае предпочтительно, чтобы значения в ORT не превышали 1 по
*абс. величине.
*K - скаляр, промаксный параметр, степень для элементов ORT. В промаксе смысл имеют только
*степени >1. Чаще используют K от 2 до 4.
*Если вы укажете неположительное значение для K, то ф-ция осуществит метод промай вместо промакса.
*Конкретное значение K тогда не играет роли, т.к. промай не возводит элементы в степень K.
*Результаты:
*NAME3 - матрица промаксного (или промайного) преобразования (де-ортогонализации осей).
*NAME1 - результатная матрица факторного отображения, = ORT*NAME3.
*NAME2 - матрица итоговых корреляций между факторами (косинусы углов между неортогнальными осями
*в конфигурации NAME1); она равна inv(sscp(NAME3)).
*Матрица факторной структуры есть NAME1*NAME2.

define !KO_promax(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @a= !2.
comp @kais= !3.
comp @k= !4.
comp @ones= make(1,ncol(@a),1)).
do if @kais.
-comp @b= @a/(sqrt(rssq(@a))*@ones).
else.
-comp @b= @a.
end if.
do if @k.
-comp @b= abs(@b)&**(@k+1)/(@b+(@b=0)).
else.
-comp @b2= @b&*@b.
-comp @flag= @b2<=(csum(@b2)/nrow(@b2))(make(1,nrow(@b2),1),:).
-comp @z= cmax(@flag&*@b2)(make(1,nrow(@b2),1),:).
-comp @b= abs(@b)/(@b+(@b=0))&*sqrt((@b2-@z)&*(not @flag)).
-release @b2,@flag,@z.
end if.
comp @q= solve(sscp(@a),t(@a))*@b.
comp @q= @q/sqrt(cssq(@q))(@ones,:).
comp @ic= inv(sscp(@q)).
comp @d= t(sqrt(diag(@ic)))).
comp !7= @q&*@d(@ones,:).
comp !5= @a*!7.
comp !6= @ic/sscp(@d).
release @a,@kais,@k,@ones,@b,@q,@ic,@d.
!enddefine.


******** КОЭФФИЦИЕНТЫ ФАКТОРНЫХ БАЛЛОВ ********.
*/*!KO_fsc(cov%load%fcorr%method%name1%name2)*/*.
*Version 1.
*Вычисляет матрицу коэффициентов факторных баллов (factor scores), с помощью которой можно получить
*оценки значений факторов, выделенных в факторном анализе, как переменных. Имеется несколько методов
*вычисления ее.
*Входящие, необходимые для всех методов METHOD:
*COV - анализировавшаяся в факторном анализе корреляционная или ковариационная симметричная матрица
*размером p x p.
*LOAD - p x m (1<=m<p) матрица факторных нагрузок. Если факторы подверглись косоугольному вращению,
*то это должна быть матрица факторного отображения. Если COV это ковариации (не корреляции), LOAD это
*"сырые" (не "перешкалированные") нагрузки.
*Входящее, необходимое не для всех методов METHOD:
*FCORR - m x m матрица корреляций между факторами. Если факторы ортогональны или если m=1, вы можете
*указать вместо матрицы произвольный скаляр.
*Метод вычисления факторных баллов METHOD (ключевое заглавное слово, опционально можно в кавычках
*или апострофах):
*"REG" - регрессионный метод Терстоуна (Томпсона); максимизирует валидность баллов.
*"REG2" - тот же метод, но в качестве внедиагональных элементов COV будет использовать воссозданные
*факторами значения.
*"REG3" - метод Хорста, или "идеализированных переменных"; факторные баллы вычисляются точно так, как
*компонентные баллы в главнокомпонентном анализе; эквивалентен методу "REG" с использованием вместо COV
*воссозданной редуцированной матрицы. Не нуждается в FCORR (укажите любой скаляр).
*"BART" - метод Бартлета (Bartlett); минимизирует смещенность баллов. Не нуждается в FCORR
*(укажите любой скаляр).
*Следующие три метода называются корреляционно-сохраняющими, потому что добиваются корреляций между
*факторными баллами таких же, какие между факторами:
*"AR" - метод Андерсона-Рубина (Anderson-Rubin). Не нуждается в FCORR (укажите любой скаляр).
*Этот метод всегда дает некоррелированные баллы, даже для косоугольных факторов.
*"MAR" - метод Макдональда-Андерсона-Рубина (McDonald-Anderson-Rubin) является обобщением предыдущего,
*годящимся и для косоугольных факторов.
*"GREEN" - метод Грина (Green). Его можно рассматривать как приближение к "MAR", которому он становится
*тождествен, если все общности у переменных одинаковы.
*Для методов REG, MAR, GREEN необходима невырожденная COV.
*Результаты:
*NAME1 - коэффициенты для вычисления факторных баллов; последние есть DATA*NAME1, где DATA -
*анализировавшиеся факторным анализом переменные (центрованные, если факторный анализ базировался на
*ковариациях, или стандартизованные, если он базировался на корреляциях).
*NAME2 - перешкалированные (стандартизованные) коэффициенты, т.е. такие, из которых устранен эффект
*неодинаковости дисперсий переменных. Если COV это корреляции, то NAME2=NAME1.
*Если вместо собственно факторного анализа применялся анализ главных компонент, то компонентные баллы,
*одни и те же, выдаются методами REG, REG3, GREEN.

define !KO_fsc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @p= !3.
comp @c= !4.
!let !method= !unquote(!5)
do if nrow(@c)*ncol(@c)=1.
-comp @c= 1.
end if.
comp @u2_1= 1/diag(@r-@p*@c*t(@p)).
!if (!method='REG') !then
-comp !6= solve(@r,@p*@c).
!else !if (!method='REG2') !then
-comp @r#= @p*@c*t(@p).
-call setdiag(@r#,diag(@r)).
-comp !6= solve(@r#,@p*@c).
-release @r#.
!else !if (!method='REG3') !then
-comp !6= t(solve(sscp(@p),t(@p))).
!else !if (!method='BART') !then
-comp @tpu= t(@p&*(@u2_1*make(1,ncol(@p),1))).
-comp !6= t(solve(@tpu*@p,@tpu)).
-release @tpu.
!else !if (!method='AR') !then
-comp @tpu= t(@p&*(@u2_1*make(1,ncol(@p),1))).
-call eigen(@tpu*@r*t(@tpu),@eivec,@eival).
-do if ncol(@p)=1.
- comp !6= t(@eivec*sqrt(1/@eival)*t(@eivec)*@tpu).
-else.
- comp !6= t(@eivec*mdiag(sqrt(1/@eival))*t(@eivec)*@tpu).
-end if.
-release @tpu,@eivec,@eival.
!else !if (!method='MAR') !then
-call eigen(@r,@eivec,@eival).
-comp @r12= @eivec*mdiag(sqrt(@eival))*t(@eivec).
-comp @r_12= inv(@r12).
-call eigen(@c,@eivec,@eival).
-do if ncol(@p)=1.
- comp @c12= @eivec*sqrt(@eival)*t(@eivec).
-else.
- comp @c12= @eivec*mdiag(sqrt(@eival))*t(@eivec).
-end if.
-call svd(@r12&*t(@u2_1)(make(1,nrow(@r12),1),:)*@p*@c12,@g,@d,@m).
-comp !6= @r_12*(@g(:,1:ncol(@p))*t(@m))*@c12.
-release @r12,@r_12,@c12,@eivec,@eival,@g,@d,@m.
!else
-call eigen(@r,@eivec,@eival).
-comp @r_12= @eivec*mdiag(1/sqrt(@eival))*t(@eivec).
-call eigen(@c,@eivec,@eival).
-do if ncol(@p)=1.
- comp @c12= @eivec*sqrt(@eival)*t(@eivec).
-else.
- comp @c12= @eivec*mdiag(sqrt(@eival))*t(@eivec).
-end if.
-call svd(@r_12*@p*(@c*@c12),@g,@d,@m).
-comp !6= @r_12*(@g(:,1:ncol(@p))*t(@m))*@c12.
-release @r_12,@c12,@eivec,@eival,@g,@d,@m.
!ifend !ifend !ifend !ifend !ifend !ifend
comp !7= sqrt(diag(@r))*make(1,ncol(!6),1)&*!6.
release @r,@p,@c,@u2_1.
!enddefine.


******** КЛАСТЕРИЗАЦИЯ МЕТОДОМ K-СРЕДНИХ ********.
*/*!KO_kmeans(data%ini%iter%empty%name1%name2%name3%name4)*/*.
*Version 1.
*Кластерный анализ методом k-средних.
*Берет мерные данные DATA (n наблюдений x p переменных) и кластеризует наблюдения в заданное число
*кластеров k (k<n). Инициальные центры пользователь должен предложить.
*INI - матрица инициальных центров (средних) размером k кластеров x p переменных (столбцов DATA).
*ITER - проделать столько итераций (неотрицательный целый скаляр, например 10). ITER=0 соответствует
*классификации без итераций, т.е. приписыванию наблюдений к инициальным центрам без поправления последних.
*EMPTY - цифра (не имя и не выражение, цифру можно опционально взять в кавычки или апострофы) 0 или 1.
*Это задание способа решения проблемы пустых кластеров, если таковые возникнут:
*"1" - избавиться от пустого кластера, уменьшив k; k уже не сможет вернуться к исходному значению.
*"0" - оставить пустому кластеру его предшествующий центр, в надежде что в дальнейшем кластер наполнится
*(чего может и не произойти).
*Пустые кластеры бывают при плохих инициальных центрах, а также сильно дискретных данных.
*Результаты:
*NAME1 - получившееся кластерное членство, в виде n x k двоичной матрицы, т е k фиктивных переменных,
*соответствующих кластерам. Каждое наблюдение приписано одному кластеру: в этом столбце стоит 1.
*Вы можете создать единую категориальную переменную кластерного членства так: NAME1*t(1:ncol(NAME1)).
*NAME2 - k x p матрица окончательных, перевычисленных средних: это центроиды кластеров NAME1.
*NAME3 - k x p матрица предокончательных средних: это последние центры, к которым были классифицированы
*наблюдения.
*NAME4 - n x 1 столбец квадратных евклидовых расстояний от наблюдений до центров NAME3, к которым
*наблюдения были в последний раз приписаны. Например, если наблюдение согласно NAME1 оказалось
*зачислено в кластер 3, то его расстояние NAME4 до центра этого кластера есть расстояние до центра,
*к-рый есть 3-й ряд в NAME3, именно на основании этого расстояния наблюдение было к нему приписано.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_kmeans(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp !7= !3.
comp @k= nrow(!7).
comp @iter= !4.
!let !empty= !unquote(!5)
comp @ind= 0.
comp @cones= make(nrow(@data),1,1).
comp @rones= make(1,ncol(@data),1).
comp !6= make(nrow(@data),@k,0).
loop @it= 0 to @iter.
-comp !8= !7.
-loop @i= 1 to @k.
- comp !6(:,@i)= rssq(@data-!8(@i,:)(@cones,:)).
-end loop.
-comp !9= rmin(!6).
-comp @ass= make(nrow(@data),1,0).
-loop @i= 1 to @k.
- comp !6(:,@i)= (!6(:,@i)=!9)&*(not @ass).
- comp @ass= @ass+!6(:,@i).
-end loop.
-comp @n= t(csum(!6)).
-do if all(@n).
- comp !7= t(!6)*@data/(@n*@rones).
-else.
  !if (!empty) !then
-  comp @ass= @n>0.
-  comp @nnz= csum(@ass).
-  comp @ind= make(1,@nnz+1,0).
-  comp @c= 1.
-  loop @i= 1 to @k.
-   comp @ind(@c)= @i.
-   comp @c= @c+@ass(@i).
-  end loop.
-  comp @ind= @ind(1:@nnz).
-  comp @n= @n(@ind).
-  comp !6= !6(:,@ind).
-  comp @k= @nnz.
-  comp !7= t(!6)*@data/(@n*@rones).
-  release @nnz,@c.
  !else
-  comp @ass= @n=0.
-  comp @ind= @ass*@rones&*!8.
-  comp @n= @n+@ass.
-  comp !7= t(!6)*@data/(@n*@rones)+@ind.
  !ifend
-end if.
end loop.
release @data,@k,@iter,@cones,@rones,@it,@i,@ass,@n,@ind.
!enddefine.


******** БЛОК-ДИАГОНАЛИЗАЦИЯ МЕТОДОМ VAT/IVAT ********.
*/*!KO_vat(mat%method%priority%name1%name2)*/*.
*Version 1.
*Переупорядочивает ряды/столбцы квадратной матрицы расстояний MAT методом VAT (и IVAT).
*Это один из способов (Bezdek, Hathaway, "VAT: a tool for...", 2002; Havens, Bezdek, "An efficient
*formulation of...", 2012) блок-диагонализации матрицы, целью которой является сосредоточить малые
*значения расстояний вблизи диагонали матрицы. Если блок-диагонализованную матрицу отобразить на
*графике "тепловая карта" (heatmap), то можно получить представление о присутствии кластерной
*структуры в данных и примерно оценить число кластеров.
*Матрица после блок-диагонализации (переупорядочивания рядов/столбцов) - результат NAME1.
*Входящие:
*MAT - квадратная симметричная неотрицательная матрица расстояний (или стоимостей) с нулевой
*диагональю.
*METHOD - кл. слово заглавными буквами (опционально в кавычках или апострофах):
*"VAT" - сделать переупорядочивание VAT-алгоритмом. Он тесно связан с алгоритмом построения
*минимального остовного дерева Прима. Порядок будет записан в векторе NAME2.
*"IVAT" - сделав переупорядочивание VAT, затем сделать iVAT-операцию. Она определенным образом
*заменяет некоторые расстояния в матрице другими ее расстояниями, уменьшая разнообразие расстояний
*в матрице. Эффект iVAT-замены тот, что на "тепловой карте" (1) контраст между междукластерными и
*внутрикластерными расстояниями усилится, помогая визуально обнаружить кластеры; (2) повысится
*обнаруживаемость кластеров цепочечной структуры (включая сильно вытянутые, древовидные, кольцевидные).
*iVAT-операция по своему эффекту тождественна обработке матрицы, полученной в VAT, алгоритмом
*Флойда-Уоршалла в варианте "определение легчайших проходов".
*Таким образом, method=IVAT возвращает матрицу с переупорядоченными+измененными (более дискретными)
*значениями, нежели были в MAT.
*PRIORITY - цифра (не имя и не выражение) 0 или 1 (опционально в кавычках или апострофах).
*Это технический параметр, потенциально сказывающийся только если в MAT есть одинаковые значения
*(ties). Он определяет порядок поиска/выбора элемента. 0 это поиск "от меньшего к большему": элемент
*с меньшим индексом будет предпочтен. 1 это поиск "от большего к меньшему": элемент с большим индексом
*будет предпочтен (этот вариант чуть быстрее). PRIORITY не меняет валидности результата, но
*результаты в условиях ties могут различаться, вследствие того, что порядок (NAME2) может стать другим.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_vat(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !method= !unquote(!3)
!let !pr= !unquote(!4)
comp !5= !2.
comp @n= ncol(!5).
!if (!pr) !then !let !nums1= '1:ncol(@m)' !let !nums2= '1:nrow(@m)'
!else
 !let !nums1= 'ncol(@m):1' !let !nums2= 'nrow(@m):1'
 !let !comp1= 'comp @ci= ncol(@m)+1-@ci' !let !comp2= 'comp @ri= nrow(@m)+1-@ri'
!ifend
call setdiag(!5,1E300).
comp !6= mmax((!5=mmin(!5))&*(1:@n)(make(1,@n,1),:)).
call setdiag(!5,0).
do if !6=1.
-comp @out= 2:@n.
else if !6=@n.
-comp @out= 1:(@n-1).
else.
-comp @out= {1:(!6-1),(!6+1):@n}.
end if.
!if (!method='IVAT') !then
-comp @rilst= 0.
!ifend
loop @i= 2 to @n.
-comp @m= !5(!6,@out).
-comp @val= mmin(@m).
-comp @m= @m=@val.
-comp @ci= mmax(@m&*(!nums1)(make(1,nrow(@m),1),:)).
-!comp1.
-comp !6= {!6,@out(@ci)}.
!if (!method='IVAT') !then
-comp @ri= mmax(@m(:,@ci)&*t(!nums2)).
-!comp2.
-comp @rilst= {@rilst,@ri}.
!ifend
-do if @i<@n.
- do if @ci=1.
-  comp @out= @out(2:ncol(@out)).
- else if @ci=ncol(@out).
-  comp @out= @out(1:(ncol(@out)-1)).
- else.
-  comp @out= @out({1:(@ci-1),(@ci+1):ncol(@out)}).
- end if.
-end if.
end loop.
comp !5= !5(!6,!6).
release @out,@ci.
!if (!method='IVAT') !then
-comp @m= make(@n,@n,0).
-loop @i= 2 to @n.
- comp @j= @rilst(@i).
- comp @val= !5(@i,@j).
- comp @m(@i,1:(@i-1))= make(1,@i-1,@val).
- loop @c= 1 to @j-1.
-  comp @m(@i,@c)= mmax({@val,@m(@j,@c)}).
- end loop.
-end loop.
-comp !5= @m+t(@m).
-release @rilst,@ri,@j,@c.
!ifend
release @n,@i,@val,@m.
!enddefine.


******** СЛУЧАЙНЫЕ ЗНАЧЕНИЯ ИЗ СТАНДАРТНОГО НОРМАЛЬНОГО РАСПРЕДЕЛЕНИЯ ********.
*/*!KO_normal(nr%nc%name)*/*.
*Version 1.
*Порождает матрицу NAME размером NR x NC, элементы которой - случайные
*значения из стандартного нормального распределения.
*NR, NC - целые положительные скаляры.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для генерации данных из популяции с коррелирующими нормальными переменными используйте
*более общую функцию /*!KO_mvnorm*/.

define !KO_normal(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @nc= !3.
comp !4= sqrt(-2*ln(uniform(@nr,@nc)))&*cos(6.28318530717959*uniform(@nr,@nc)).
release @nr,@nc.
!enddefine.


******** СЛУЧАЙНЫЕ ЗНАЧЕНИЯ ИЗ НОРМАЛЬНОГО РАСПРЕДЕЛЕНИЯ С ЗАДАННЫМИ ПАРАМЕТРАМИ ********.
*/*!KO_rvnorm(mean%sd%name)*/*.
*Version 1.
*Порождает случайное число из нормального распределения со средней MEAN и стандартным отклонением SD.
*MEAN и SD - матрицы любого, одинакового размера. Результат NAME - того же размера.
*Элемент ij NAME содержит случайное значение из нормального распределения с параметрами ij MEAN и ij SD.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).

define !KO_rvnorm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mean= !2.
comp @sd= !3.
comp !4= sqrt(-2*ln(uniform(nrow(@mean),ncol(@mean))))&*cos(6.28318530717959*uniform(nrow(@mean),ncol(@mean))).
comp !4= !4&*@sd+@mean.
release @mean,@sd.
!enddefine.


******** СЛУЧАЙНЫЕ ЗНАЧЕНИЯ ИЗ РАВНОМЕРНОГО РАСПРЕДЕЛЕНИЯ С ЗАДАННЫМИ ПАРАМЕТРАМИ ********.
*/*!KO_rvunif(min%max%name)*/*.
*Version 1.
*Порождает случайное число из равномерного распределения с минимумом MIN и максимумом MAX.
*MIN и MAX - матрицы любого, одинакового размера. Результат NAME - того же размера.
*Элемент ij NAME содержит случайное значение из равномерного распределения с параметрами ij MIN и ij MAX
*(ij MIN должен быть не больше ij MAX).
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).

define !KO_rvunif(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @min= !2.
comp @max= !3.
comp !4= uniform(nrow(@min),ncol(@min)).
comp !4= (@max-@min)&*!4+@min.
release @min,@max.
!enddefine.


******** СЛУЧАЙНЫЕ ЗНАЧЕНИЯ ИЗ ХИ-КВАДРАТ РАСПРЕДЕЛЕНИЯ ********.
*/*!KO_chisq(nr%nc%df%name)*/*.
*Version 1.
*Порождает матрицу NAME размером NR x NC, элементы которой - случайные
*значения из хи-квадрат распределения с числом степеней свободы DF.
*NR, NC, DF - целые положительные скаляры.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).

define !KO_chisq(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @nc= !3.
comp @df= !4.
comp !5= sqrt(-2*ln(uniform(@df,@nr*@nc)))&*cos(6.28318530717959*uniform(@df,@nr*@nc)). 
comp !5= reshape(cssq(!5),@nr,@nc).
release @nr,@nc,@df.
!enddefine.


******** СЛУЧАЙНЫЕ ЗНАЧЕНИЯ ИЗ БИНОМИАЛЬНОГО РАСПРЕДЕЛЕНИЯ ********.
*/*!KO_binom(nr%ntrials%prob%name)*/*.
*Version 1.
*Порождает вектор-столбец NAME длиной NR, элементы которого - случайные значения из
*биномиального распределения с параметрами число испытаний NTRIALS и вероятностью успеха PROB.
*NTRIALS (целое положительное) и PROB (между 0 и 1) - либо два скаляра, либо два вектора длиной NR.
*Если это векторы, значит, разные значения NAME можно породить под разными значениями параметров.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_binom(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @n= !3.
comp @p= !4.
do if nrow(@n)*ncol(@n)*nrow(@p)*ncol(@p)=1.
-comp !5= rsum(uniform(@nr,@n)<=@p).
else.
-comp !5= make(@nr,1,0).
-loop @i= 1 to @nr.
- comp !5(@i)= rsum(uniform(1,@n(@i))<=@p(@i)).
-end loop.
-release @i.
end if.
release @nr,@n,@p.
!enddefine.


******** СЛУЧАЙНЫЕ ЗНАЧЕНИЯ ИЗ КАТЕГОРИАЛЬНОГО РАСПРЕДЕЛЕНИЯ ********.
*/*!KO_categ(nr%nc%kprob%name)*/*.
*Version 2.
*Порождает матрицу NAME размером NR x NC, элементы которой - случайные
*значения (коды от 1 до k) из k-категориального распределения. (Категориальное распределение
*это мультиномиальное распределение с числом испытаний 1. Распределение Бернулли это 2-категориальное
*распределение.)
*NR, NC - целые положительные скаляры.
*KPROB - либо вектор длиной k, содержащий вероятности для k категорий, либо целый скаляр, равный k.
*Если KPROB - вектор вероятностей, это должны быть неотрицательные числа, хотя бы одно положительное.
*Числа не обязаны давать в сумме 1: функция сама приведет их сумму к 1; достаточно, чтобы числа
*находились в соотношениях тех же, как вероятности.
*Если KPROB - скаляр, он должен быть целым положительным (дробное значение будет обрезано до целого).
*Он есть k, и k вероятностей полагаются равными. Когда цель - равновероятное распределение, выгоднее
*такое задание KPROB, нежели заданее вектором вероятностей, т.к. будет быстрее.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_categ(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nr= !2.
comp @nc= !3.
comp @kprob= !4.
comp @k= nrow(@kprob)*ncol(@kprob).
do if @k=1.
-comp !5= trunc(uniform(!2,!3)*trunc(@kprob))+1.
else.
-comp @kprob= @kprob/msum(@kprob).
-comp !5= uniform(!2,!3).
-comp @cump= 0.
-loop @i= 1 to @k.
- comp @cump= @cump+@kprob(@i).
- comp !5= !5+(!5<=@cump)*@i.
-end loop.
-comp !5= trunc(!5).
-release @cump,@i.
end if.
release @nr,@nc,@kprob,@k.
!enddefine.


******** ВЫБОРКА ИЗ КАТЕГОРИАЛЬНОГО РАСПРЕДЕЛЕНИЯ БЕЗ ВОЗВРАЩЕНИЯ ********.
*/*!KO_catwor(n%kprob%stock%method%name1%name2)*/*.
*Version 1.
*Набирает случайную выборку N объектов из ограниченной совокупности объектов, принадлежащих
*разным категориям. Эта функция выдает случайные значения из k-категориального распределения
*(k - число категорий объектов), как и функция /*!KO_categ*/, но она делает выборку без возвращения
*из ограниченной популяции, тогда как /*!KO_categ*/ делает выборку с возвращением.
*Т.к. объекты извлекаются один за другим и не возвращаются "на склад" (в ограниченную популяцию),
*вероятности выбора ("реализации") для разных категорий меняются по ходу процесса. Эти вероятности
*определяются, с одной стороны, заданными вероятностями "спроса", а с другой стороны, текущими
*вероятностями "наличия на складе", так сказать.
*Входящие:
*N - целый положительный скаляр: требуемый размер выборки.
*KPROB - вектор длиной k, содержащий вероятности для k категорий объектов. Это должны быть
*неотрицательные числа, хотя бы одно положительное. Числа не обязаны давать в сумме 1: функция
*сама приведет их сумму к 1; достаточно, чтобы числа находились в соотношениях тех же,
*как вероятности. KPROB надо понимать как исходные вероятности "спроса"; т.е. это вероятности
*выбора, если бы популяция была бесконечной.
*STOCK - вектор длиной k и той же ориентации что KPROB, содержащий частоты для k категорий
*объектов - это наблюдаемое распределение нашей ограниченной совокупности ("склада"), откуда будут
*изыматься объекты в выборку. Частоты должны быть неотрицательными целыми числами, хотя бы одно
*положительное. Если вы введете дробные значения, функция сперва обрежет их до целых.
*Если веса KPROB задают категориям вероятности "спроса", то частоты STOCK задает им начальные
*вероятности "наличия".
*METHOD - цифра (не переменная и не выражение; цифру опционально можно закавычить или заапострофить).
*Это задание метода определения текущих вероятностей выбора (реализации):
*"1" - метод "Исходный спрос, если осталось". Вероятность PROB(i) выбрать сейчас объект категории i
*пропорциональна KPROB(i)*LEFT(i), где LEFT(i)=1, если объекты категории i еще остались в популяции,
*иначе LEFT(i)=0. Если частоты STOCK все достаточно велики, чтобы не исчерпаться до окончания
*процедуры выборки, то этот метод эквивалентен применению ф-ции /*!KO_categ*/.
*"2" - метод "Исходный спрос * сколько осталось". Вероятность PROB(i) выбрать сейчас объект категории i
*пропорциональна KPROB(i)*FLEFT(i), где FLEFT(i) это доля категории i среди оставшихся пока в
*популяции объектов. Это классический случайный выбор без возвращения.
*"3" - метод "Плавающий спрос * сколько осталось". Вероятность PROB(i) выбрать сейчас объект
*категории i пропорциональна PROB_(i)*FLEFT(i), где PROB_(i) это PROB(i), вычисленное перед выбором
*предыдущего объекта в выборку (перед выбором 1-го объекта PROB_(i)=KPROB(i)). Это "заостряющий" метод,
*ведущий к быстрому по ходу выбора смещению предпочтения в сторону одной категории, как правило с
*большей представленностью "на складе" ограниченной популяции.
*"4" - метод "Плавающий не-спрос * сколько осталось". Вероятность PROB(i) выбрать сейчас объект
*категории i пропорциональна (1-PROB_(i))*FLEFT(i) (перед выбором 1-го объекта 1-PROB_(i)=KPROB(i)).
*Этот метод может быть интересен как "сглаживающая" альтернатива методу "2": выбор из малопредставленных
*категорий будет немного повышен, а выбор из многопредставленных категорий будет несколько понижен, по
*сравнению с тем методом.
*Результаты:
*NAME1 - вектор-столбец длиной N, содержащий коды категорий (номера из последовательности 1,2,...,k)
*созданной случайной выборки (и последовательность тоже случайна). В том случае, если объектов в
*популяции не хватило для выбора без возвращения N объектов, длина NAME1 будет меньше N. Если не
*удалось взять ни одного объекта, NAME1=0.
*NAME2 - вектор-столбец длиной k, содержащий оставшиеся в STOCK частоты.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если KPROB это равные вероятности, а STOCK это вектор из одних единиц или двоичный (1 и 0), то эквивалентно
*будет воспользоваться функцией /*!KO_sample*/, которая быстрее.

define !KO_catwor(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
comp @n= !2.
comp @kprob= !3.
comp !7= trunc(!4).
!let !method= !unquote(!5)
comp @k= nrow(@kprob)*ncol(@kprob).
comp @tstock= msum(!7).
comp !6= uniform(@n,1).
!if (!method='2') !then !let !kprob= '@kprob'
!else !if (!method='3') !then !let !kprob= '@prob'
-comp @prob= @kprob.
!else !if (!method='4') !then !let !kprob= '(1-@prob)'
-comp @prob= 1-@kprob/msum(@kprob).
!ifend !ifend !ifend
loop @c= 1 to @n.
 !if (!method='1') !then
- comp @prob= @kprob&*(!7>0).
 !else
- comp @prob= !kprob&*(!7/@tstock).
 !ifend
-comp @psum= msum(@prob).
-do if @psum.
- comp @prob= @prob/@psum.
- comp @objc= !6(@c).
- comp @cump= 0.
- loop @i= 1 to @k.
-  comp @cump= @cump+@prob(@i).
-  do if @objc<=@cump.
-   comp !6(@c)= @i.
-   comp !7(@i)= !7(@i)-1.
-   comp @tstock= @tstock-1.
-   break.
-  end if.
- end loop.
-else.
- break.
-end if.
end loop if not @tstock.
do if @c<=@n.
-comp @c= @c-(not @psum).
-do if @c.
- comp !6= !6(1:@c).
-else.
- comp !6= 0.
-end if.
end if.
do if nrow(!7)=1.
-comp !7= t(!7).
end if.
release @n,@kprob,@k,@tstock,@c,@prob,@psum,@objc,@cump,@i.
!enddefine.


******** СЛУЧАЙНО ВЫБРАТЬ N ЭЛЕМЕНТОВ ВЕКТОРА ********.
*/*!KO_sample(n%cases%name1%name2)*/*.
*Version 1.
*Случайно выбирает N разных элементов из последовательности номеров 1,2,...,k.
*Эта функция удобна для формирования случайной подвыборки ("без возвращения") из массива,
*а также для случайной сортировки наблюдений массива.
*Входящие:
*N - целый положительный скаляр: сколько разных элементов выбрать.
*CASES - "популяция" элементов: либо целый положительный скаляр, либо двоичный (0 или 1) вектор.
*Если скаляр, то он есть k, и выбор будет из чисел 1,2,...,k. Если двоичный вектор, то его
*длина есть k и номера элементов соответствуют числам 1,2,...,k; значение 1 у элемента в векторе
*разрешает выбрать номер этого элемента, а значение 0 - запрещает. Таким образом, вектор
*CASES является фильтром. Вектор, состоящий из одних единиц, эквивалентен заданию CASES скаляром.
*Вектор должен содержать минимум одну единицу.
*Результаты:
*NAME1 - вектор-столбец длиной N, содержащий случайно выбранные номера, и в случайном порядке.
*NAME2 - двоичный вектор-столбец длиной k, где 1 помечает не выбранный (оставшийся) элемент CASES,
*а 0 помечает выбранный (изъятый) элемент.
*Если N>=k, то длина NAME1 будет k, а NAME2 будет состоять из одних нулей.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваша популяция CASES взвешенная, т.е. состоит из частот (а не 0 или 1), или если ее элементы
*неравновероятны к выбору, используйте функцию /*!KO_catwor*/.

define !KO_sample(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')')) 
comp @n= !2.
comp !5= !3.
do if nrow(!5)*ncol(!5)=1.
-comp !5= make(!5,1,1).
else if nrow(!5)=1.
-comp !5= t(!5).
end if.
comp @ns= csum(!5).
comp @n_= rmin({@n,@ns}).
comp @lnk= ln(nrow(!5)).
comp @ns= ln(@ns/nrow(!5)).
do if nrow(!5)<15000 and ln(@n)>22.189-147.7272/@lnk+.78529*@ns+268.29/@lnk**2+.044*@ns**2+3.266*@ns/@lnk.
-comp !4= {uniform(nrow(!5),1)&*!5,t(1:nrow(!5))}.
-comp !4(grade(-!4(:,1)),:)= !4.
-comp !4= !4(1:@n_,2).
-comp !5(!4)= make(@n_,1,0).
else.
-comp @g= 0.
-comp @val= 0.
-comp !4= trunc(uniform(@n_,1)*nrow(!5))+1.
-loop @i= 1 to @n_.
- do if !5(!4(@i)).
-  comp !5(!4(@i))= 0.
- else.
-  loop @g= 1 to 1E12.
-   comp @val= trunc(uniform(1,1)*nrow(!5))+1.
-  end loop if !5(@val).
-  comp !4(@i)= @val.
-  comp !5(@val)= 0.
- end if.
-end loop.
-release @i,@g,@val.
end if.
release @n,@n_,@ns,@lnk.
!enddefine.


******** СЛУЧАЙНАЯ МАТРИЦА ИЗ РАСПРЕДЕЛЕНИЯ УИШАРТА ********.
*/*!KO_wishart(chl%n%name)*/*.
*Version 1.
*Выдает случайную матрицу NAME из распределения Уишарта с параметром масштаба C и степенями свободы N.
*C это популяционная ковариационная матрица (в популяции предполагается многомерное нормальное
*распределение). Аргументом CHL функции выступает не C, а ее корень Холецкого: верхняя треугольная
*матрица, выдаваемая встроенной матричной функцией chol(C).
*N (целый положительный скаляр) это размер выборки для порождения случайной матрицы NAME.
*Последняя является матрицей рассеяния (scatter); чтобы перевести ее в ковариационную матрицу,
*поделите ее на N.
*Если в качестве CHL указать скаляр (положительный), то он считается популяционным стандартным
*отклонением, и NAME/N будет случайным значением дисперсии.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_wishart(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= !3.
comp @a= sqrt(-2*ln(uniform(ncol(!4),ncol(!4))))&*cos(6.28318530717959*uniform(ncol(!4),ncol(!4))).
comp @v= sqrt(-2*ln(uniform(1,@n)))&*cos(6.28318530717959*uniform(1,@n)).
comp @a(1,1)= sqrt(rssq(@v)).
loop @i= 2 to nrow(@a).
-comp @v= sqrt(-2*ln(uniform(1,@n-@i+1)))&*cos(6.28318530717959*uniform(1,@n-@i+1)).
-comp @a(@i,@i)= sqrt(rssq(@v)).
-comp @a(@i,1:(@i-1))= make(1,@i-1,0).
end loop.
comp !4= @a*!4.
comp !4= sscp(!4).
release @n,@a,@v,@i.
!enddefine.


******** СЛУЧАЙНЫЕ ДАННЫЕ ИЗ НОРМАЛЬНОЙ ПОПУЛЯЦИИ С ЗАДАННЫМИ КОВАРИАЦИЯМИ ********.
*/*!KO_mvnorm(n%chl%mean%name)*/*.
*Version 1.
*Порождает данные NAME - случайную выборку размером N из нормально распределенной популяции, имеющей
*ковариации C и центроид MEAN.
*Аргумент CHL функции это не сама ковариационная матрица C (размером p x p) между переменными,
*а ее корень Холецкого: верхняя треугольная матрица (того же размера), выдаваемая встроенной матричной
*функцией chol(C).
*N - число наблюдений для порождения, целый положительный скаляр.
*MEAN - скаляр либо вектор-ряд длиной p: это средние переменных в популяции.
*Результат NAME матрица случайных данных n наблюдений x p переменных из нормальной популяции с указанными
*ковариациями между переменными и средними переменных.
*Как частный случай, C может быть размером 1 x 1; тогда C это дисперсия (а CHL это ст. отклонение).
*Если матрица C это корреляции, а не ковариации, аргумент MEAN держите как 0, полученные переменные NAME
*вы можете затем умножить на популяционные стандартные отклонения и потом прибавить популяционные средние;
*но более удобный путь - конвертируйте корреляционную матрицу в ковариационную функцией /*!KO_corrcov*/,
*после чего примените /*!KO_mvnorm*/ с заданием популяционных средних как MEAN в ней.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).

define !KO_mvnorm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @n= !2.
comp !5= !3.
comp @mean= !4.
comp !5= sqrt(-2*ln(uniform(@n,ncol(!5))))&*cos(6.28318530717959*uniform(@n,ncol(!5)))*!5.
do if ncol(@mean)=1.
-comp !5= !5+@mean.
else.
-comp !5= !5+@mean(make(1,@n,1),:).
end if.
release @n,@mean.
!enddefine.


******** ПЕРЕКОДИРОВКА ЗНАЧЕНИЙ (ТОЧНОЕ СОВПАДЕНИЕ) ********.
*/*!KO_recode1(data%filter%old%new%name)*/*.
*Version 1.
*В матрице данных DATA любого размера заменяет значения списка OLD на соответствующие значения списка NEW.
*Выдает матрицу NAME, являющуюся матрицей DATA с перекодированными значениями.
*Эта функция подобна внематричной команде RECODE, в которой значения для перекодировки заданы списком.
*OLD и NEW - векторы-ряды. i-е значение OLD, если найдено в DATA, перекодируется в i-е значение NEW.
*Оба эти вектора - одинаковой длины, либо NEW может быть на один элемент длиннее; в последнем случае все прочие
*кроме OLD значения данных будут перекодированы в это последнее значение NEW.
*Если некое значение встречается в OLD больше раза, оно перекодируется в NEW-значение, соответствующее первой
*(самой левой) его встрече в векторе OLD.
*Аргумент FILTER позволяет запретить перекодировать значения некоторых рядов DATA. FILTER либо скаляр,
*либо столбец длиной как число рядов DATA. Если ненулевой скаляр - никакого запрета нет; если нулевой
*скаляр - то полный запрет (функция не сработает, NAME будет равна DATA). Если столбец, то 0 в нем запрещает
*перекодировку в данном ряду DATA, а иное значение - разрешает.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_recode1(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @filt= !3.
comp @old= !4.
comp @new= !5.
do if any(@filt).
-comp @nflagm= make(nrow(@data),ncol(@data),1).
-comp @newm= make(nrow(@data),ncol(@data),0).
-loop @i= 1 to ncol(@old).
- comp @flagi= (@data=@old(@i))&*@nflagm.
- comp @newm= @newm+@flagi*@new(@i).
- comp @nflagm= @nflagm-@flagi.
-end loop.
-comp !6= @data&*@nflagm+@newm.
-do if ncol(@new)>ncol(@old).
- comp !6= !6&*(not @nflagm)+@nflagm*@new(@i).
-end if.
-do if not all(@filt).
- comp @filt= (@filt<>0)*make(1,ncol(@data),1).
- comp !6= !6&*@filt+@data&*(not @filt).
-end if.
-release @nflagm,@newm,@i,@flagi.
else.
-comp !6= @data.
end if.
release @data,@filt,@old,@new.
!enddefine.


******** ПЕРЕКОДИРОВКА ЗНАЧЕНИЙ (ПОПАДАНИЕ В ДИАПАЗОН) ********.
*/*!KO_recode2(data%filter%low%high%new%name)*/*.
*Version 1.
*В матрице данных DATA любого размера заменяет значения, попадающие в диапазон от LOW до HIGH (включительно),
*на соответствующие значения списка NEW.
*Выдает матрицу NAME, являющуюся матрицей DATA с перекодированными значениями.
*Эта функция подобна внематричной команде RECODE, в которой значения для перекодировки заданы
*диапазоном(ами) от... до....
*LOW, HIGH и NEW - векторы-ряды. LOW(i) и HIGH(i) есть диапазон; HIGH(i) в норме должно быть >= LOW(i).
*Всякое значение DATA, лежащее в диапазоне от LOW(i) до HIGH(i), будет
*перекодировано в значение NEW(i). LOW и HIGH должны быть одинаковой длины. NEW может быть на один элемент
*длиннее; в последнем случае все прочие, кроме перекодированных, значения данных будут перекодированы в
*это последнее значение NEW.
*Если некоторые диапазоны, указанные векторами LOW и HIGH, взаимонакладываются, их общие значения
*перекодируется в NEW-значение, соответствующее самому раннему (левому в векторах) из этих диапазонов.
*Например, если LOW есть {6,5} и HIGH есть {11,10}, то есть заданы пересекающиеся диапазоны 6-11 или 5-10,
*NEW же есть {23,37}, то "спорные" значения данных - от 6 до 10 - будут перекодированы в 23, а не в 37,
*т.к. диапазон 6-11 задан "раньше" (левее в векторах), чем 5-10. (Так же ведет себя вне-матричная команда
*RECODE.)
*Аргумент FILTER позволяет запретить перекодировать значения некоторых рядов DATA. FILTER либо скаляр,
*либо столбец длиной как число рядов DATA. Если ненулевой скаляр - никакого запрета нет; если нулевой
*скаляр - то полный запрет (функция не сработает, NAME будет равна DATA). Если столбец, то 0 в нем запрещает
*перекодировку в данном ряду DATA, а иное значение - разрешает.
*Если векторы LOW и HIGH идентичны, эта функция даст тот же результат, что /*!KO_recode1*/.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_recode2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @filt= !3.
comp @lo= !4.
comp @hi= !5.
comp @new= !6.
do if any(@filt).
-comp @nflagm= make(nrow(@data),ncol(@data),1).
-comp @newm= make(nrow(@data),ncol(@data),0).
-loop @i= 1 to ncol(@lo).
- comp @flagi= (@data>=@lo(@i) and @data<=@hi(@i))&*@nflagm.
- comp @newm= @newm+@flagi*@new(@i).
- comp @nflagm= @nflagm-@flagi.
-end loop.
-comp !7= @data&*@nflagm+@newm.
-do if ncol(@new)>ncol(@lo).
- comp !7= !7&*(not @nflagm)+@nflagm*@new(@i).
-end if.
-do if not all(@filt).
- comp @filt= (@filt<>0)*make(1,ncol(@data),1).
- comp !7= !7&*@filt+@data&*(not @filt).
-end if.
-release @nflagm,@newm,@i,@flagi.
else.
-comp !7= @data.
end if.
release @data,@filt,@lo,@hi,@new.
!enddefine.


******** РАЗНЫЕ ОДНОПЕРЕМЕННЫЕ СТАТИСТИКИ, ПО ГРУППАМ: ЗАМЕНА ИМИ ИСХОДНЫХ ЗНАЧЕНИЙ ********.
*/*!KO_aggrv(data%bin%stat%check%value1%value2%name)*/*.
*Version 1.
*Эта функция вычисляет требуемую описательную статистику по группам - точно так, как функция /*!KO_aggr*/,
*но в качестве результата NAME выдает не таблицу группы X переменные, а массив размером с DATA, в котором
*исходные значения DATA заменены на соответствующие сводные статистики: если ряд (наблюдение)
*принадлежит группе i и только ей, то его значения в NAME это есть ряд i в таблице, выдаваемой
*функцией /*!KO_aggr*/. Таким образом, /*!KO_aggrv()*/ делает то же, что SPSS команда AGGREGATE под опцией
*MODE=ADDVARIABLES.
*Аргументы DATA, BIN, STAT - точно те же, что в функции /*!KO_aggr*/ - см.
*Аргумент CHECK (цифра 0 или 1, не имя и не выражение):
*Если вы знаете, что суммы во всех рядах BIN равны 1, т.е. что каждое наблюдение принадлежит ровно одной
*группе, так что BIN это фиктивные (dummy) переменные, установите CHECK на 0. Если же вы допускаете, что
*суммы в некоторых рядах BIN могут быть не равны 1, установите CHECK на 1.
*Аргументы VALUE1 и VALUE2 игнорируются, если СHECK есть 0.
*VALUE1 (число, не имя и не выражение): задайте значение, котороое вставить в NAME тем наблюдениям, которые
*не принадлежат ни одной группе (т.е. у которых сумма в этом ряду BIN =0).
*VALUE2: задайте значение (число, не имя и не выражение), котороое вставить в NAME тем наблюдениям, которые
*принадлежат более чем одной группе (т.е. у которых сумма в этом ряду BIN >1); либо укажите заглавными
*буквами ключевое слово "AVER" - что значит требование у наблюдения, принадлежащего более чем одной
*группе и значит располагающего более чем одним посчитанным значением статистики, усреднить эти его
*значения статистики.
*Значения аргументов CHECK, VALUE1, VALUE2 можно опционально взять в кавычки или апострофы.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваше группирование задается категориальной переменной, создайте из нее фиктивные двоичные
*переменные BIN с помощью функции /*!KO_freq*/.

define !KO_aggrv(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
             /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @bin= !3.
!let !stat= !unquote(!4)
!if (!index('SUM FIRST LAST MIN MAX',!stat)=0) !then !let !n= ',@n'
-comp @n= csum(@bin).
!ifend
!if (!stat='SUM') !then
-comp !8= t(@bin)*@data.
!else !if (!stat='MEAN') !then
-comp !8= t(@bin)*@data/(t(@n)*make(1,ncol(@data),1)).
!else !if (!stat='SSDEV' !or !stat='VARIANCE' !or !stat='VARIANCE2' !or !stat='SEMEAN') !then
-comp @sum= t(@bin)*@data.
-comp @cssq= t(@bin)*(@data&*@data).
-comp @n= t(@n)*make(1,ncol(@data),1).
-comp !8= @cssq-@sum&*@sum/@n.
 !if (!stat<>'SSDEV') !then
- comp !8= !8/(@n !if (!stat<>'VARIANCE2') !then -1 !ifend).
  !if (!stat='SEMEAN') !then
-  comp !8= sqrt(!8/@n).
  !ifend
 !ifend
-release @sum,@cssq.
!else !if (!stat='MIN') !then
-comp @ones= make(1,ncol(@data),1).
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @bigv= 1E300.
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !8(@i,:)= cmin(@data&*@bini+(not @bini)*@bigv).
-end loop.
-release @ones,@bigv,@i,@bini.
!else !if (!stat='MAX') !then
-comp @ones= make(1,ncol(@data),1).
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @smallv= mmin(@data).
-loop @i= 1 to ncol(@bin).
- comp @bini= @bin(:,@i)*@ones.
- comp !8(@i,:)= cmax(@data&*@bini+(not @bini)*@smallv).
-end loop.
-release @ones,@smallv,@i,@bini.
!else !if (!stat='LAST') !then
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(1:nrow(@data)).
-loop @i= 1 to ncol(@bin).
- comp @ind= cmax(@bin(:,@i)&*@rnums).
- comp !8(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!else !if (!stat='FIRST') !then
-comp !8= make(ncol(@bin),ncol(@data),0).
-comp @rnums= t(nrow(@data):1).
-loop @i= 1 to ncol(@bin).
- comp @ind= nrow(@data)-cmax(@bin(:,@i)&*@rnums)+1.
- comp !8(@i,:)= @data(@ind,:).
-end loop.
-release @rnums,@i,@ind.
!ifend !ifend !ifend !ifend !ifend !ifend !ifend
comp !8= @bin*!8.
!if (!unquote(!5)) !then !let !n= ',@n'
-comp @ones= make(1,ncol(@data),1).
-comp @n= rsum(@bin).
-comp @notn= not @n.
-comp !8= (not @notn)*@ones&*!8+@notn*!unquote(!6)*@ones.
 !if (!unquote(!7)='AVER') !then
- comp !8= !8/((@n+@notn)*@ones).
-!else
- comp @notn= @n>1.
- comp !8= (not @notn)*@ones&*!8+@notn*!unquote(!7)*@ones.
 !ifend
-release @notn,@ones.
!ifend
release @data,@bin !n.
!enddefine.


******** ПОЗИЦИИ НЕНУЛЕВЫХ ЭЛЕМЕНТОВ В ВЕКТОРЕ ********.
*/*!KO_indices(vec%name)*/*.
*Version 2.
*Берет вектор (ряд или столбец) VEC и выдает вектор-ряд NAME, содержащий список индексов
*(порядковых номеров, позиций) ненулевых элементов в VEC. Индексы идут по возрастающей.
*Если ненулевых элементов в VEC нет, ф-ция выдает скаляр 0.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_indices(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @vec= !2.
comp @vec= @vec<>0.
comp @n= nrow(@vec)*ncol(@vec).
comp @nnz= msum(@vec).
do if not @nnz.
-comp !3= 0.
else if @nnz=@n.
-comp !3= 1:@n.
else if @nnz/@n>.8.
-comp !3= make(1,@nnz+1,0).
-comp @c= 1.
-loop @i= 1 to @n.
- comp !3(@c)= @i.
- comp @c= @c+@vec(@i).
-end loop.
-comp !3= !3(1:@nnz).
-release @c,@i.
else if @nnz/@n<.08.
-comp !3= make(1,@nnz,0).
-comp @s= 1.
-loop @c= 1 to @nnz.
- loop @i= @s to @n.
- end loop if @vec(@i).
- comp !3(@c)= @i.
- comp @s= @i+1.
-end loop.
-release @s,@c,@i.
else.
-comp !3= make(1,@nnz,0).
-comp @c= 0.
-loop @i= 1 to @n.
- do if @vec(@i).
-  comp @c= @c+1.
-  comp !3(@c)= @i.
- end if.
-end loop.
-release @c,@i.
end if.
release @vec,@n,@nnz.
!enddefine.


******** ПОЗИЦИИ НЕНУЛЕВЫХ И НУЛЕВЫХ ЭЛЕМЕНТОВ В ВЕКТОРЕ ********.
*/*!KO_indices2(vec%name1%name2)*/*.
*Version 1.
*Берет вектор (ряд или столбец) VEC и выдает вектор-ряд NAME1, содержащий список индексов
*(порядковых номеров, позиций) ненулевых элементов в VEC, и вектор-ряд NAME2, содержащий список
*индексов нулевых элементов в VEC. Индексы идут по возрастающей.
*Если ненулевых элементов в VEC нет, NAME1 будет скаляр 0. Если нулевых элементов в VEC нет,
*NAME2 будет скаляр 0.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_indices2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @vec= !2.
comp @vec= @vec<>0.
comp @n= nrow(@vec)*ncol(@vec).
comp @nnz= msum(@vec).
do if not @nnz.
-comp !3= 0.
-comp !4= 1:@n.
else if @nnz=@n.
-comp !3= 1:@n.
-comp !4= 0.
else if @nnz/@n>.3.
-comp !3= make(1,@nnz,0).
-comp !4= make(1,@n-@nnz,0).
-comp @c= 0.
-comp @c2= 0.
-loop @i= 1 to @n.
- do if @vec(@i).
-  comp @c= @c+1.
-  comp !3(@c)= @i.
- else.
-  comp @c2= @c2+1.
-  comp !4(@c2)= @i.
- end if.
-end loop.
-release @c,@c2,@i.
else.
-do if nrow(@vec)=1.
- comp @vec= {@vec,1}.
-else.
- comp @vec= {@vec;1}.
-end if.
-comp @n= @n+1.
-comp !3= make(1,@nnz+1,0).
-comp !4= 0.
-comp @s= 1.
-loop @c= 1 to @nnz+1.
- loop @i= @s to @n.
- end loop if @vec(@i).
- comp !3(@c)= @i.
- do if @i>@s.
-  comp !4= {!4,@s:(@i-1)}.
- end if.
- comp @s= @i+1.
-end loop.
-release @s,@c,@i.
-comp !3= !3(1:(ncol(!3)-1)).
-comp !4= !4(2:ncol(!4)).
end if.
release @vec,@n,@nnz.
!enddefine.


******** ПОЗИЦИИ НЕНУЛЕВЫХ ЭЛЕМЕНТОВ В МАТРИЦЕ (И САМИ ЭЛЕМЕНТЫ) ********.
*/*!KO_indicesm(mat%ANDVAL%name)*/*.
*Version 1.
*Берет матрицу любого размера MAT и выдает двухстолбцовую матрицу NAME, содержащую список индексов
*(порядковых номеров, позиций) ненулевых элементов в MAT. Индексы идут по возрастающей.
*Номер ряда записан в 1-м столбце NAME, а номер столбца - во 2-м столбце NAME.
*Аргумент ANDVAL позволяет выписать и сами ненулевые значения, в качестве 3-го столбца NAME.
*ANDVAL - цифра (не имя и не выражение) 0 или 1 (цифру можно взять в кавычки или апострофы).
*Если 1, то будет выписка значений.
*Если ненулевых элементов в MAT нет, ф-ция выдает скаляр 0.
*Эта функция работает немного быстрее на широких, чем на высоких матрицах.
*Эта функция близка к /*!KO_nzlist*/, но выдает результат немного в другом виде.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_indicesm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !andval= !unquote(!3)
!if (!andval) !then !let !nc= 3 !let !vals= ',t(@mat(@i,@rami))' !else !let !nc= 2 !ifend
comp @mat= !2.
comp @ram= @mat<>0.
comp @rnnz= rsum(@ram).
do if not any(@rnnz).
-comp !4= 0.
else if mmin(@rnnz)=ncol(@ram).
-comp @rind= t(1:nrow(@ram)).
-comp @cind= t(1:ncol(@ram)).
-comp !4= {kroneker(@rind,make(nrow(@cind),1,1)),kroneker(make(nrow(@rind),1,1),@cind)}.
 !if (!andval) !then
- comp !4= {!4,reshape(@mat,nrow(!4),1)}.
 !ifend
-release @rind,@cind.
else.
-comp @m= @ram.
-loop @i= 2 to ncol(@m).
- comp @m(:,@i)= @m(:,@i-1)+@m(:,@i).
-end loop.
-comp @m= @m&*@ram.
-comp @ram= make(nrow(@m),ncol(@m)+1,0).
-comp @nums= 1:ncol(@m).
-comp @m= @m+(not @m)*ncol(@ram).
-loop @i= 1 to nrow(@m).
- comp @ram(@i,@m(@i,:))= @nums.
-end loop.
-comp !4= make(msum(@rnnz),!nc,0).
-comp @st= 1.
-loop @i= 1 to nrow(@ram).
- comp @n= @rnnz(@i).
- do if @n.
-  comp @en= @st+@n-1.
-  comp @rami= @ram(@i,1:@n).
-  comp !4(@st:@en,:)= {make(@n,1,@i),t(@rami) !vals}.
-  comp @st= @en+1.
- end if.
-end loop.
-release @m,@i,@nums,@st,@en,@n,@rami.
end if. 
release @mat,@ram,@rnnz.
!enddefine.


******** ПОЗИЦИЯ ПЕРВОЙ ВСТРЕЧИ ЗНАЧЕНИЯ В РЯДУ/РЯДАХ ********.
*/*!KO_indx(mat%val%name)*/*.
*Version 1.
*Берет MAT - матрицу либо вектор-ряд. Выдает индекс (позицию) первой (самой левой)
*встречи значения VAL в каждом ряду MAT.
*VAL может быть скаляром или вектором-столбцом длиной как число рядов MAT; в последнем случае
*каждому ряду MAT отвечает свое значение VAL.
*Результат NAME - вектор-столбец с числом рядов как в MAT.
*Позиция 0 означает, что VAL в ряду не найдено.

define !KO_indx(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @val= !3.
do if nrow(@val)>1.
-comp @val= @val*make(1,ncol(@mat),1).
end if.
comp @mat= (@mat=@val)&*(ncol(@mat):1)(make(1,nrow(@mat),1),:).
comp !4= rmax(@mat).
comp !4= (ncol(@mat)+1)*(!4>0)-!4.
release @mat,@val.
!enddefine.


******** ПОЗИЦИЯ ПОСЛЕДНЕЙ ВСТРЕЧИ ЗНАЧЕНИЯ В РЯДУ/РЯДАХ ********.
*/*!KO_rindx(mat%val%name)*/*.
*Version 1.
*Берет MAT - матрицу либо вектор-ряд. Выдает индекс (позицию) последней (самой правой)
*встречи значения VAL в каждом ряду MAT.
*VAL может быть скаляром или вектором-столбцом длиной как число рядов MAT; в последнем случае
*каждому ряду MAT отвечает свое значение VAL.
*Результат NAME - вектор-столбец с числом рядов как в MAT.
*Позиция 0 означает, что VAL в ряду не найдено.

define !KO_rindx(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @val= !3.
do if nrow(@val)>1.
-comp @val= @val*make(1,ncol(@mat),1).
end if.
comp @mat= (@mat=@val)&*(1:ncol(@mat))(make(1,nrow(@mat),1),:).
comp !4= rmax(@mat).
release @mat,@val.
!enddefine.


******** ПОЗИЦИЯ (ПАРА ИНДЕКСОВ) ОДНОГО ЗНАЧЕНИЯ В МАТРИЦЕ ********.
*/*!KO_ij(mat%val%priority%name)*/*.
*Version 2.
*Находит в матрице MAT один элемент, равный значению VAL (скаляр).
*Выдает в качестве NAME позицию (индексы i,j) такого элемента в MAT.
*Если элементов, равных VAL, найдено более одного, то выдается позиция одного из них согласно
*аргументу PRIORITY. Если не найдено ни одного, выдается {0,0}.
*PRIORITY - кл. слово заглавными буквами (можно взять в кавычки или апострофы):
*"TBLR" - поиск сверху вниз, слева направо: выдается позиция в наиболее верхнем ряду
*наиболее левого элемента.
*"LRTB" - поиск слева направо, сверху вниз: выдается позиция в наиболее левом столбце
*наиболее верхнего элемента.
*"BTRL" - поиск снизу вверх, справа налево: выдается позиция в наиболее нижнем ряду
*наиболее правого элемента.
*"RLBT" - поиск справа налево, снизу вверх: выдается позиция в наиболее правом столбце
*наиболее нижнего элемента.
*"RANDOM" - из всех позиций VAL выдать одну по случайному выбору.
*Есть крохотная разница в скорости исполнения; если позиционная приоритетность сама не
*важна вам, но вы жадны на предмет скорости, выберите RLBT как самое быстрое.

define !KO_ij(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !pr= !unquote(!4)
comp @mat= !2.
comp @val= !3.
comp @mat= @mat=@val.
!if (!pr='RLBT' !or !pr='BTRL' !or !pr='RANDOM') !then !let !rnums= '1:nrow(@mat)' !let !cnums= '1:ncol(@mat)'
!else !let !rnums= 'nrow(@mat):1' !let !cnums= 'ncol(@mat):1'
!ifend
!if (!pr='LRTB' !or !pr='TBLR') !then
 !let !rrevert= 'comp @r= (nrow(@mat)+1)*(@r>0)-@r' !let !crevert= 'comp @c= (ncol(@mat)+1)*(@c>0)-@c'
!ifend
!if (!pr='RANDOM') !then
-comp @val= msum(@mat).
-do if @val.
!else
-do if any(@mat).
!ifend
 !if (!pr='RANDOM') !then
- do if @val>1.
-  comp @mat= @mat&*uniform(nrow(@mat),ncol(@mat)).
-  comp @mat= @mat=mmax(@mat).
- end if. 
 !ifend
 !if (!pr='BTRL' !or !pr='TBLR') !then
- comp @r= mmax(@mat&*(t(!rnums)*make(1,ncol(@mat),1))).
- !rrevert.
- comp @c= mmax(@mat(@r,:)&*(!cnums)).
- !crevert.
 !else
- comp @c= mmax(@mat&*(!cnums)(make(1,nrow(@mat),1),:)).
- !crevert.
- comp @r= mmax(@mat(:,@c)&*t(!rnums)).
- !rrevert.
 !ifend
- comp !5= {@r,@c}.
- release @r,@c.
-else.
- comp !5= {0,0}.
-end if.
release @mat,@val.
!enddefine.


******** ПОМЕТКА В РЯДУ И СТОЛБЦЕ НЕ БОЛЕЕ ОДНОГО ЭЛЕМЕНТА, РАВНОГО ЗНАЧЕНИЮ ********.
*/*!KO_prime(mat%val%name)*/*.
*Version 1.
*Берет матрицу или вектор MAT и значение VAL.
*Выдает двоичную матрицу NAME размером с MAT, в которой единицей помечены элементы, равные VAL, но функция
*оставляет в каждом ряду и столбце не более одной такой пометки. Другими словами, единица в NAME
*эксклюзивно спаривает ряды и столбцы: "один ряд - один столбец".
*Это не алгоритм оптимального спаривания: результат может зависеть от того, в каком направлении вести поиск
*элементов, равных VAL. Данная функция просматривает ряды сверху вниз, а элементы в ряду справа налево.
*Если вы хотите другой порядок, просто измените соответственно этому порядок рядов и/или столбцов в MAT.
*Аргумент VAL может быть не только скаляром, но и вектором-столбцом длиной как число рядов MAT.
*В последнем случае в каждом ряду MAT ищется и помечается свое значение, а дальше
*(стирание "лишних" пометок) функция действует так же, как при скаляре.
*Функция работает несколько быстрее с широкой матрицей, чем с высокой матрицей.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_prime(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @val= !3.
do if nrow(@val)>1.
-comp @val= @val*make(1,ncol(!4),1).
end if.
comp !4= (!4=@val)&*(1:ncol(!4))(make(1,nrow(!4),1),:).
comp @rzero= make(1,ncol(!4),0).
comp @czero= make(nrow(!4),1,0).
loop @i= 1 to nrow(!4).
-comp @j= rmax(!4(@i,:)).
-do if @j. 
- comp !4(@i,:)= @rzero.
- comp !4(:,@j)= @czero.
- comp !4(@i,@j)= 1.
-end if.
end loop.
release @val,@rzero,@czero,@i,@j.
!enddefine.


******** ПОМЕТКА ПРЯМЫХ ЦЕПОЧЕК ********.
*/*!KO_runs(mat%maxw%dir%hl%name)*/*.
*Version 2.
*В двоичной матрице MAT помечает прямые цепочки подряд идущих единиц.
*Цепочки какой длины каким значением будут помечены, зависит от аргумента
*"максимальный вес" MAXW (неотрицательный целый скаляр).
*Если MAXW>0, то цепочка длиной >=MAXW будет помечена числом MAXW,
*цепочка длиной MAXW-1 будет помечена числом MAXW-1,
*цепочка длиной MAXW-2 будет помечена числом MAXW-2, и так далее.
*Если MAXW задано неположительное (любое такое число), то цепочка будет помечена числом,
*равным ее длине, какой бы длинной она ни оказалась. Иными словами, неположительное
*MAXW есть отказ от предела "максимальный вес".
*Аргумент DIR (скаляр) - цепочки какого направления интересуют: вертикальные, в столбцах
*(положительный аргумент) или горизонтальные, в рядах (неположительный аргумент).
*Аргумент HL - цифра 1 или 0 (не имя и не выражение, цифру можно взять в кавычки или апострофы).
*Этот аргумент о том, как помечать. Если HL=1, то все члены цепочки будут помечены одним числом;
*если HL=0, то они будут помечены порядковыми номерами.
*Например, если MAXW=0, то цепочка 1 1 1 1 1 будет помечена как 5 5 5 5 5 при HL=1 и
*как 1 2 3 4 5 при HL=0. А если MAXW=3, то она будет помечена как 3 3 3 3 3 при HL=1 и
*как 1 2 3 3 3 при HL=0.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_runs(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !6= !2.
comp @maxw= !3.
comp @dir= !4.
!let !hl= !unquote(!5)
do if not @maxw.
-comp @maxw= 1E300.
end if.
do if not @dir.
-comp !6= t(!6).
end if.
comp @nr= nrow(!6).
comp @r_= !6(1,:).
loop @i= 2 to @nr.
-comp @r= !6(@i,:).
-comp @r_= @r&*(@r_+(@r_<@maxw)).
-comp !6(@i,:)= @r_.
end loop.
!if (!hl) !then
-comp @r_= !6(@nr,:).
-loop @i= @nr-1 to 1 by -1.
- comp @r= !6(@i,:).
- comp @rr_= @r and @r_.
- comp @r_= @r_&*@rr_+@r&*(not @rr_).
- comp !6(@i,:)= @r_.
-end loop.
-release @rr_.
!ifend
do if not @dir.
-comp !6= t(!6).
end if.
release @maxw,@dir,@nr,@i,@r,@r_.
!enddefine.


******** ПОМЕТКА ПРЯМЫХ ЦЕПОЧЕК (ДРУГОЙ АЛГОРИТМ) ********.
*/*!KO_runs2(mat%maxw%dir%name)*/*.
*Version 1.
*В двоичной матрице MAT помечает прямые цепочки подряд идущих единиц.
*Цепочки какой длины каким значением будут помечены, зависит от аргумента
*"максимальный вес" MAXW (положительный скаляр от 2 до
*число_рядов или число_столбцов матрицы - в зависимости от DIR):
*единицы цепочек длиной >=MAXW обратятся в значение MAXW; единицы цепочек длиной MAXW-1
*обратятся в значение MAXW-1; единицы цепочек
*длиной MAXW-2 обратятся в значение MAXW-2, и т.д.
*Аргумент DIR (скаляр) - цепочки какого направления интересуют: вертикальные, в столбцах
*(положительный аргумент) или горизонтальные, в рядах (неположительный аргумент).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Ф-ция /*!KO_runs*/ исполняет ту же работу, что и данная ф-ция, но реализована по-другому. /*!KO_runs2*/
*отличается от /*!KO_runs*/ следующим:
*- аргумент MAXW должен быть положительным;
*- аргумент HL отсутствует;
*- скорость у /*!KO_runs2*/ зависит от MAXW, а у /*!KO_runs*/ - нет.
*- скорость у /*!KO_runs2*/ немного выше, чем у /*!KO_runs*/, если MAXW не выше 3 и MAT
*мала; во всех прочих случаях /*!KO_runs*/ быстрей и поэтому она обычно предпочтительней.

define !KO_runs2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @maxw= !3.
comp @dir= !4.
do if not @dir.
-comp !5= t(!5).
end if.
loop @w= 2 to @maxw.
-comp @w_= @w-1.
-comp @a= !5(1:(nrow(!5)-@w_),:).
-comp @b= @a.
-loop @i= 2 to @w.
- comp @b= @b and !5(@i:(nrow(!5)-@w+@i),:).
-end loop.
-comp !5(1:(nrow(!5)-@w_),:)= @a+@b.
-loop @i= 1 to @w_.
- comp @a= !5(2:nrow(!5),:).
- comp @b= !5(1:(nrow(!5)-1),:).
- comp !5(2:nrow(!5),:)= @a+(@a=@w_)&*(@b=@w).
-end loop.
end loop.
do if not @dir.
-comp !5= t(!5).
end if.
release @maxw,@dir,@w,@w_,@a,@b,@i.
!enddefine.


******** ПОМЕТКА КОСЫХ ЦЕПОЧЕК ********.
*/*!KO_slant(mat%maxw%dir%hl%name)*/*.
*Version 2.
*В двоичной матрице MAT помечает косые цепочки подряд идущих единиц.
*Цепочки какой длины каким значением будут помечены, зависит от аргумента
*"максимальный вес" MAXW (неотрицательный целый скаляр).
*Если MAXW>0, то цепочка длиной >=MAXW будет помечена числом MAXW,
*цепочка длиной MAXW-1 будет помечена числом MAXW-1,
*цепочка длиной MAXW-2 будет помечена числом MAXW-2, и так далее.
*Если MAXW задано неположительное (любое такое число), то цепочка будет помечена числом,
*равным ее длине, какой бы длинной она ни оказалась. Иными словами, неположительное
*MAXW есть отказ от предела "максимальный вес".
*Аргумент DIR (скаляр) - косые цепочки какого направления интересуют: верх-лево/низ-право
*(положительный аргумент) или верх-право/низ-лево (неположительный аргумент).
*Аргумент HL - цифра 1 или 0 (не имя и не выражение, цифру можно взять в кавычки или апострофы).
*Этот аргумент о том, как помечать. Если HL=1, то все члены цепочки будут помечены одним числом;
*если HL=0, то они будут помечены порядковыми номерами.
*Например, если MAXW=0, то цепочка 1 1 1 1 1 будет помечена как 5 5 5 5 5 при HL=1 и
*как 1 2 3 4 5 при HL=0. А если MAXW=3, то она будет помечена как 3 3 3 3 3 при HL=1 и
*как 1 2 3 3 3 при HL=0.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Ф-ция /*!KO_slant2*/ исполняет ту же работу, что и данная ф-ция, но реализована по-другому.

define !KO_slant(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !6= !2.
comp @maxw= !3.
comp @dir= !4.
!let !hl= !unquote(!5)
do if not @maxw.
-comp @maxw= 1E300.
end if.
comp @nr= nrow(!6).
comp @nc= ncol(!6).
do if not @dir.
-comp !6= !6(:,@nc:1).
end if.
loop @i= 2 to @nr.
-comp @r= !6(@i,2:@nc).
-comp @r_= !6(@i-1,1:(@nc-1)).
-comp !6(@i,2:@nc)= @r&*(@r_+(@r_<@maxw)).
end loop.
!if (!hl) !then
-loop @i= @nr-1 to 1 by -1.
- comp @r= !6(@i,1:(@nc-1)).
- comp @r_= !6(@i+1,2:@nc).
- comp @rr_= @r and @r_.
- comp !6(@i,1:(@nc-1))= @r_&*@rr_+@r&*(not @rr_).
-end loop.
-release @rr_.
!ifend
do if not @dir.
-comp !6= !6(:,@nc:1).
end if.
release @maxw,@dir,@nr,@nc,@i,@r,@r_.
!enddefine.


******** ПОМЕТКА КОСЫХ ЦЕПОЧЕК (ДРУГОЙ АЛГОРИТМ) ********.
*/*!KO_slant2(mat%maxw%dir%name)*/*.
*Version 1.
*В двоичной матрице MAT помечает косые цепочки подряд идущих единиц.
*Цепочки какой длины каким значением будут помечены, зависит от аргумента
*"максимальный вес" MAXW (положительный скаляр от 2 до
*min(число_рядов,число_столбцов) матрицы): единицы цепочек длиной >=MAXW обратятся в
*значение MAXW; единицы цепочек длиной MAXW-1 обратятся в значение MAXW-1; единицы цепочек
*длиной MAXW-2 обратятся в значение MAXW-2, и т.д.
*Аргумент DIR (скаляр) - косые цепочки какого направления интересуют: верх-лево/низ-право
*(положительный аргумент) или верх-право/низ-лево (неположительный аргумент).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Ф-ция /*!KO_slant*/ исполняет ту же работу, что и данная ф-ция, но реализована по-другому. /*!KO_slant2*/
*отличается от /*!KO_slant*/ следующим:
*- аргумент MAXW должен быть положительным;
*- аргумент HL отсутствует;
*- скорость у /*!KO_slant2*/ зависит от MAXW, а у /*!KO_slant*/ - нет.
*- скорость у /*!KO_slant2*/ немного выше, чем у /*!KO_slant*/, если MAXW=2, либо если MAXW= 3-4 и MAT
*мала; во всех же прочих случаях /*!KO_slant*/ быстрей и поэтому она обычно предпочтительней.

define !KO_slant2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @maxw= !3.
comp @dir= !4.
do if not @dir.
-comp !5= !5(:,ncol(!5):1).
end if.
loop @w= 2 to @maxw.
-comp @w_= @w-1.
-comp @a= !5(1:(nrow(!5)-@w_),1:(ncol(!5)-@w_)).
-comp @b= @a.
-loop @i= 2 to @w.
- comp @b= @b and !5(@i:(nrow(!5)-@w+@i),@i:(ncol(!5)-@w+@i)).
-end loop.
-comp !5(1:(nrow(!5)-@w_),1:(ncol(!5)-@w_))= @a+@b.
-loop @i= 1 to @w_.
- comp @a= !5(2:nrow(!5),2:ncol(!5)).
- comp @b= !5(1:(nrow(!5)-1),1:(ncol(!5)-1)).
- comp !5(2:nrow(!5),2:ncol(!5))= @a+(@a=@w_)&*(@b=@w).
-end loop.
end loop.
do if not @dir.
-comp !5= !5(:,ncol(!5):1).
end if.
release @maxw,@dir,@w,@w_,@a,@b,@i.
!enddefine.


******** ПОИСК В ГЛУБИНУ В ОДНОДОЛЬНОМ ГРАФЕ ********.
*/*!KO_dfs(adj%start%stop%forest%out%name)*/*.
*Version 1.
*Осуществляет обход/поиск связанных вершин в глубину (Depth-first search) в однодольном невзвешенном
*графе ADJ, стартуя с вершины START. Опциональная вершина STOP - пункт остановки обхода: дойдя до нее,
*посещение вершин прекращается. START и STOP - скаляры, номера вершин, они должны различаться. Если STOP
*задавать не хотите, т.е. обход графа должен идти максимально далеко, установите этот аргумент на
*значение 0.
*Аргумент OUT - цифра 1 или 0 (не имя и не выражение; цифру можно взять в кавычки или апострофы).
*Если 0, результат NAME будет содержать только один столбец, показывающий, сверху вниз, порядок
*посещения вершин. Если 1, будут дополнительные вычисления и NAME будет содержать еще столбцы:
*2-й столбец показывает родителя данной посещенной (указанной в 1-м столбце) вершины. Под родителем
*разумеется тут самая ранопосещенная вершина, указывающая на (т.е. имеющая смежником) нашу вершину.
*3-й столбец показывает расстояние посещенной вершины от вершины START (корневой вершины); это
*расстояние есть длина кратчайшего пути от START до нашей вершины через уже посещенные вершины
*(в DFS это расстояние от START до нашей вершины не обязательно глобально минимально).
*4-й столбец помечает единицей начало новой цепи в посещениях. Цепь это непрерывная последовательность
*вглубь графа, где каждая следующая посещенная вершина является смежной предыдущей вершине и это ребро
*алгоритм прошел (учел). Новая цепь начинается, когда дальше двигаться нельзя (т.к. еще непосещенных
*смежных вершин больше нет); тогда новая цепь начинается с еще не посещенной вершины, лежащей "под"
*некоторой "развилкой" путей; алгоритм возвращается к развилке, чтобы с этой вершины начать новый,
*другой путь вглубь. DFS-алгоритм имеет целью максимизировать длину цепи.
*5-й столбец добавляется, если FOREST=1 (см.). Здесь помечается единицей начало нового дерева.
*Аргумент FOREST - цифра 1 или 0 (не имя и не выражение; цифру можно взять в кавычки или апострофы).
*Если 0, алгоритм ограничится обходом вглубь, стартуя только из одной вершины - заданной как START. NAME
*будет представлять единственное DF-дерево (Depth-first tree), идущее из корня START. Если аргумент 1,
*то - в случае если дерево из корня START не смогло посетить все вершины графа - берется дополнительная
*корневая вершина (из еще непосещенных вершин), и дерево еще непосещенных вершин строится из нее.
*Так повторяется, пока наконец все вершины графа не окажутся посещены. NAME в этом случае будет
*представлять не одно дерево, а коллекцию деревьев, называемую DF-лесом (Depth-first forest).
*Начало (вершина-корень) нового дерева в лесу помечается единицей в 5-м столбце NAME (вы должны задать
*OUT=1, чтобы в NAME было несколько столбцов). Отличие нового дерева от новой цепи то, что дерево
*начинается из нового начала, и расстояния в новом дереве отмеряются уже от нового начала, тогда как
*расстояния в новой цепи продолжают отмеряться от старого начала - от корня того дерева, которому
*принадлежит данная цепь. Т.о., цепи это пути, вложенные в деревья.
*Аргумент ADJ - входящий граф. Граф может быть любым - ненаправленным или направленным, с циклами или
*без, с петлями или без, быть деревом или сетью. Он задается списком смежности (adjacency list) ADJ.
*Каждый ряд ADJ это вершина графа, и ряд содержит номера вершин, которые с ней смежны (при направленном
*графе это те, на которые она указывает); номера должны быть записаны слева направо утрамбованно.
*Например, если из 2-й вершины идут ребра к вершинам 1,4 и 5, то 2-й ряд должен выглядеть как
*{1,4,5,0,...,0}, где пустоты добиты нулями. Порядок номеров может быть любой: {4,5,1,0,...,0},
*например, тоже корректно. Число столбцов ADJ должно быть достаточно для того, чтобы вместить информацию
*о всех ребрах графа. Список смежности соответствует квадратной двоичной матрице, в которой единица в
*ячейке (i,j) обозначают наличие ребра (дуги) из вершины i в вершину j. Список смежности является другим
*способом представить такую матрицу. Если ваши входящие - двоичная матрица, вы можете получить из нее
*список смежности функцией /*!KO_ram*/.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_dfs(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
!let !for= !unquote(!5) !let !out= !unquote(!6)
comp @a= !2.
comp @start= !3.
comp @g= !4.
comp @brk= 0.
comp @clst= 0.
comp @na= rsum(@a<>0).
comp @notd= make(nrow(@a),1,1).
!if (!for) !then
-comp @newt= make(nrow(@a),1,0).
-comp @newt(1)= 1.
!ifend
!if (!out) !then
-comp @p= make(1,nrow(@a),0).
-comp @d= make(1,nrow(@a),0).
-comp !7= make(nrow(@a),4,0).
-comp !7(1,4)= 1.
-comp @not= 0.
!else
-comp !7= make(nrow(@a),1,0).
!ifend
comp @cnt= 0.
!if (!for) !then
-loop.
!ifend
- comp @s= {@start,0}.
- loop if ncol(@s)>1.
-  comp @u= @s(1).
-  comp @s= @s(2:ncol(@s)).
-  do if @notd(@u).
-   comp @notd(@u)= 0.
-   comp @cnt= @cnt+1.
    !if (!out) !then
-    comp !7(@cnt,1:3)= {@u,@p(@u),@d(@u)}.
-    do if @cnt>1.
-     comp !7(@cnt,4)= not any(@a(!7(@cnt-1,1),:)=@u).
-    end if.
    !else
-    comp !7(@cnt)= @u.
    !ifend
-   do if @u=@g.
-    comp @brk= 1.
-    break.
-   end if.
-   do if @na(@u).
-    comp @clst= @a(@u,1:@na(@u)).
-    comp @s= {@clst,@s}.
     !if (!out) !then
-     comp @not= not @p(@clst).
-     comp @p(@clst)= @p(@clst)+@not*@u.
-     comp @d(@clst)= @d(@clst)+@not*(@d(@u)+1).
     !ifend
-   end if.
-  end if.
- end loop.
!if (!for) !then
- do if any(@notd).
-  comp @start= cmax(@notd&*t(1:nrow(@notd))).
-  comp @newt(@cnt+1)= 1.
- else.
-  break.
- end if.
-end loop if @brk.
 !if (!out) !then
- comp !7= {!7,@newt}.
 !ifend
!ifend
comp !7= !7(1:@cnt,:).
release @a,@start,@u,@g,@brk,@na,@notd,@cnt,@s,@clst !if (!for) !then ,@newt !ifend !if (!out) !then ,@not,@p,@d !ifend.
!enddefine.


******** ПОИСК В ШИРИНУ В ОДНОДОЛЬНОМ ГРАФЕ ********.
*/*!KO_bfs(adj%start%stop%forest%out%name)*/*.
*Version 1.
*Осуществляет обход/поиск связанных вершин в ширину (Breadth-first search) в однодольном невзвешенном
*графе ADJ, стартуя с вершины START. Опциональная вершина STOP - пункт остановки обхода: дойдя до нее,
*посещение вершин прекращается. START и STOP - скаляры, номера вершин, они должны различаться. Если STOP
*задавать не хотите, т.е. обход графа должен идти максимально далеко, установите этот аргумент на
*значение 0.
*Аргумент OUT - цифра 1 или 0 (не имя и не выражение; цифру можно взять в кавычки или апострофы).
*Если 0, результат NAME будет содержать только один столбец, показывающий, сверху вниз, порядок
*посещения вершин. Если 1, будут дополнительные вычисления и NAME будет содержать еще столбцы:
*2-й столбец показывает родителя данной посещенной (указанной в 1-м столбце) вершины. Под родителем
*разумеется тут самая ранопосещенная вершина, указывающая на (т.е. имеющая смежником) нашу вершину.
*3-й столбец показывает расстояние посещенной вершины от вершины START (корневой вершины); это
*расстояние есть длина кратчайшего пути от START до нашей вершины через посещенные вершины
*(в BFS это расстояние от START до нашей вершины непременно глобально минимально).
*4-й столбец помечает единицей начало нового веера в посещениях. Веер это непрерывнная
*последовательность вширь графа, т.е. перечисление сибсов (детей одного ранее посещенного родителя).
*Новый веер начинается, когда сибсов больше нет. BFS-алгоритм имеет целью максимизировать размер веера.
*5-й столбец добавляется, если FOREST=1 (см.). Здесь помечается единицей начало нового дерева.
*Аргумент FOREST - цифра 1 или 0 (не имя и не выражение; цифру можно взять в кавычки или апострофы).
*Если 0, алгоритм ограничится обходом вширь, стартуя только из одной вершины - заданной как START. NAME
*будет представлять единственное BF-дерево (Breadth-first tree), идущее из корня START. Если аргумент 1,
*то - в случае если дерево из корня START не смогло посетить все вершины графа - берется дополнительная
*корневая вершина (из еще непосещенных вершин), и дерево еще непосещенных вершин строится из нее.
*Так повторяется, пока наконец все вершины графа не окажутся посещены. NAME в этом случае будет
*представлять не одно дерево, а коллекцию деревьев, называемую BF-лесом (Breadth-first forest).
*Начало (вершина-корень) нового дерева в лесу помечается единицей в 5-м столбце NAME (вы должны задать
*OUT=1, чтобы в NAME было несколько столбцов). Отличие нового дерева от нового веера то, что дерево
*начинается из нового начала, и расстояния в новом дереве отмеряются уже от нового начала, тогда как
*расстояния в новом веере продолжают отмеряться от старого начала - от корня того дерева, которому
*принадлежит данный веер. Т.о., веера это пути, вложенные в деревья.
*Аргумент ADJ - входящий граф. Граф может быть любым - ненаправленным или направленным, с циклами или
*без, с петлями или без, быть деревом или сетью. Он задается списком смежности (adjacency list) ADJ.
*Каждый ряд ADJ это вершина графа, и ряд содержит номера вершин, которые с ней смежны (при направленном
*графе это те, на которые она указывает); номера должны быть записаны слева направо утрамбованно.
*Например, если из 2-й вершины идут ребра к вершинам 1,4 и 5, то 2-й ряд должен выглядеть как
*{1,4,5,0,...,0}, где пустоты добиты нулями. Порядок номеров может быть любой: {4,5,1,0,...,0},
*например, тоже корректно. Число столбцов ADJ должно быть достаточно для того, чтобы вместить информацию
*о всех ребрах графа. Список смежности соответствует квадратной двоичной матрице, в которой единица в
*ячейке (i,j) обозначают наличие ребра (дуги) из вершины i в вершину j. Список смежности является другим
*способом представить такую матрицу. Если ваши входящие - двоичная матрица, вы можете получить из нее
*список смежности функцией /*!KO_ram*/.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_bfs(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
!let !for= !unquote(!5) !let !out= !unquote(!6)
comp @a= !2.
comp @start= !3.
comp @g= !4.
comp @brk= 0.
comp @na= rsum(@a<>0).
comp @notd= make(nrow(@a),1,1).
!if (!for) !then
-comp @newt= make(nrow(@a)+1,1,0).
-comp @newt(1)= 1.
!ifend
!if (!out) !then
-comp !7= make(nrow(@a)+1,4,0).
-comp !7(1,4)= 1.
!else
-comp !7= make(nrow(@a)+1,1,0).
!ifend
comp @st= 1.
comp @en= 1.
comp @v= 0.
comp @v_= 0.
!if (!for) !then
-loop.
- comp @en= @st.
!ifend
- comp @notd(@start)= 0.
- comp !7(@st,1)= @start.
- loop if !7(@st,1).
-  comp @u= !7(@st,1).
-  do if @u=@g.
-   comp @en= @st.
-   comp @brk= 1.
-   break.
-  end if.
-  loop @v= 1 to @na(@u).
-   comp @v_= @a(@u,@v).
-   do if @notd(@v_).
-    comp @notd(@v_)= 0.
-    comp @en= @en+1.
     !if (!out) !then
-     comp !7(@en,1:3)= {@v_,@u,!7(@st,3)+1}.
     !else
-     comp !7(@en)= @v_.
     !ifend
-   end if.
-  end loop.
-  comp @st= @st+1.
- end loop.
!if (!for) !then
- do if any(@notd).
-  comp @start= cmax(@notd&*t(1:nrow(@notd))).
-  comp @newt(@st)= 1.
- else.
-  break.
- end if.
-end loop if @brk.
 !if (!out) !then
- comp !7= {!7,@newt}.
 !ifend
!ifend
comp !7= !7(1:@en,:).
!if (!out) !then
-do if @en>1.
-comp !7(2:@en,4)= !7(2:@en,2)<>!7(1:(@en-1),2) or !7(2:@en,2)=0.
-end if.
!ifend
release @a,@start,@u,@g,@brk,@na,@notd,@st,@en,@v,@v_ !if (!for) !then ,@newt !ifend.
!enddefine.


******** АЛГОРИТМ МАКСИМАЛЬНЫХ КЛИК БРОНА-КЕРБОША (PIVOT-ВЕРСИЯ) ********.
*/*!KO_bronkerb(adj%return%name)*/*.
*Version 1.
*Эта функция исполняет алгоритм Брона-Кербоша для перечисления максимальных клик в ее pivot-версии.
*(Эквивалентную, но более медленную базовую версию алгоритма - см. функцию /*!KO_bronkerb2*/.)
*Входящая матрица ADJ - это ненаправленный невзвешенный граф в виде "списка смежности"
*(adjacency list). Каждый ряд ADJ это вершина графа, и ряд содержит номера вершин, с которыми эта
*вершина смежна (связана). Номера идут слева направо утрамбованно. Например, пусть вершина под
*номером 2 связана с вершинами 1, 4, 5. Тогда 2-й ряд ADJ должен выглядеть так: {1,4,5[,0...]},
*где [,0...] - это дозаполнение ряда нулями, если необходимо. Число столбцов в ADJ может быть любым
*достаточным для того, чтобы разместить список всех смежников в графе. Порядок номеров вершин в ряду
*произволен: например, можно написать тот же ряд как {5,1,4[,0...]}. Вершина в графе не может
*быть смежна себе (т.е., напр. во 2-м ряду не может присутствовать вершина 2). Изолированные вершины
*в графе не допускаются (т.е. ряд не может состоять из нулей - но см. ПРИМЕР). Список смежности ADJ
*соответствует квадратной симметричной двоичной матрице смежности с нулевой диагональю, в которой
*единицы обозначают наличие связи, ADJ является другим способом ее представить. Если ваши входящие -
*такая двоичная матрица, то вы можете получить из нее список смежности функцией /*!KO_ram*/.
*Результат NAME - двухстолбцовая матрица. Ее первый столбец это номера вершин, входящих в максимальные
*клики. Ее 2-й столбец это идентификатор клик - их номера в порядке их обнаружения алгоритмом.
*Аргумент RETURN - кл. слово заглавными буквами (можно также в кавычках или апострофах):
*"ALL" - позволить алгоритму выполниться до конца; NAME будет перечислять все максимальные клики,
*существующие в графе;
*"COUNT n" - выдать n найденных первыми максимальных клик и остановиться; n это целое >0, может быть
*цифрой, именем или выражением (примеры заданий: COUNT 10, COUNT x, COUNT abs(z)*2);
*"SIZEEQ n" - выдать одну максимальную клику, состоящую из n вершини, и остановиться; n это целое >1,
*может быть цифрой, именем или выражением;
*"SIZEGE n" - выдать одну максимальную клику, состоящую из >=n вершин, и остановиться; n это целое >1,
*может быть цифрой, именем или выражением.
*В двух последних случаях - если клик затребованного размера не нашлось, результат будет {0,0}.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_bronkerb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !return= !unquote(!3)
comp @adj= !2.
!if (!head(!return)='COUNT') !then !let !cond= !conc('@cnt=',!tail(!return))
!else !if (!head(!return)='SIZEEQ') !then !let !cond= !conc('ncol(@r)=',!tail(!return))
!else !if (!head(!return)='SIZEGE') !then !let !cond= !conc('ncol(@r)>=',!tail(!return))
!ifend !ifend !ifend
comp @n= rsum(@adj>0).
comp @rs= {0,0}.
comp @ps= {1:nrow(@adj),0}.
comp @xs= {0,0}.
comp @qs= {0,0}.
comp @ri= {1,0}.
comp @pi= {nrow(@adj),0}.
comp @xi= {1,0}.
comp @qi= {1,0}.
comp !4= {0,0}.
comp @cnt= 0.
loop if ncol(@rs)>1.
 !do !w !in ('@r @p @x @q') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
- comp !w= !ws(1:!wi(1)).
- comp !ws= !ws((ncol(!w)+1):ncol(!ws)).
- comp !wi= !wi(2:ncol(!wi)).
 !doend
-do if not (any(@p) or any(@x)).
- comp @cnt= @cnt+1.
- comp !4= {!4;t(@r),make(ncol(@r),1,@cnt)}.
  !if (!return<>'ALL') !then
-  do if !cond.
-   comp @cnt= -1.
-   break.
-  end if.
  !ifend
-else.
- do if not any(@x).
-  comp @#= @p.
- else if not any(@p).
-  comp @#= @x.
- else.
-  comp @#= {@p,@x}.
- end if.
- comp @n#= @n(@#).
- comp @i= mmax((@n#=mmax(@n#))&*t(1:nrow(@n#))).
- comp @q= @#(@i).
- comp @#= 0.
- do if any(@p).
-  comp @nei= @adj(@q,1:@n(@q)).
-  loop @i= 1 to ncol(@p).
-   do if not any(@nei=@p(@i)).
-    comp @#= {@#,@p(@i)}.
-   end if.
-  end loop.
-  do if ncol(@#)>1.
-   comp @#= @#(2:ncol(@#)).
-  end if.
- end if.
- do if any(@#).
-  comp @v= @#(1).
-  comp @rs= {@r,@rs}.
-  comp @ri= {ncol(@r),@ri}.
-  do if ncol(@p)=1.
-   comp @#= 0.
-  else.
-   comp @#= mmax((@p=@v)&*(1:ncol(@p))).
-   do if @#=1.
-    comp @#= @p(2:ncol(@p)).
-   else if @#= ncol(@p).
-    comp @#= @p(1:(ncol(@p)-1)).
-   else.
-    comp @#= @p({1:(@#-1),(@#+1):ncol(@p)}).
-   end if.
-  end if.
-  comp @ps= {@#,@ps}.
-  comp @pi= {ncol(@#),@pi}.
-  do if any(@x).
-   comp @#= {@x,@v}.
-  else.
-   comp @#= @v.
-  end if.
-  comp @xs= {@#,@xs}.
-  comp @xi= {ncol(@#),@xi}.
-  comp @qs= {@q,@qs}.
-  comp @qi= {1,@qi}.
-  do if any(@r).
-   comp @#= {@r,@v}.
-  else.
-   comp @#= @v.
-  end if.
-  comp @rs= {@#,@rs}.
-  comp @ri= {ncol(@#),@ri}.
-  comp @nei= @adj(@v,1:@n(@v)).
   !do !w !in('@p @x') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
-   comp @#= 0.
-   do if any(!w).
-    loop @i= 1 to ncol(!w).
-     do if any(@nei=!w(@i)).
-      comp @#= {@#,!w(@i)}.
-     end if.
-    end loop.
-    do if ncol(@#)>1.
-     comp @#= @#(2:ncol(@#)).
-    end if.
-   end if.
-   comp !ws= {@#,!ws}.
-   comp !wi= {ncol(@#),!wi}.
   !doend
-  comp @qs= {@q,@qs}.
-  comp @qi= {1,@qi}.
- end if.
-end if.
end loop.
do if nrow(!4)>1.
-comp !4= !4(2:nrow(!4),:).
end if.
!if (!head(!return)='SIZEEQ' !or !head(!return)='SIZEGE') !then
-do if @cnt=-1.
- comp !4= !4((nrow(!4)-ncol(@r)+1):nrow(!4),:).
-else.
- comp !4= {0,0}.
-end if.
!ifend
release @adj,@n,@rs,@ps,@xs,@qs,@ri,@pi,@xi,@qi,@cnt,@r,@p,@x,@q,@v,@#,@nei,@i,@n#.
!enddefine.


******** АЛГОРИТМ МАКСИМАЛЬНЫХ КЛИК БРОНА-КЕРБОША (БАЗОВАЯ ВЕРСИЯ) ********.
*/*!KO_bronkerb2(adj%return%name)*/*.
*Version 1.
*Эта функция исполняет алгоритм Брона-Кербоша для перечисления максимальных клик в ее базовой версии
*и она эквивалентна функции /*!KO_bronkerb*/, но медленнее ее, поэтому не предпочтительна - кроме
*случаев очень разреженного графа, когда она бывает немного быстрее. Аргументы - те же,
*что /*!KO_bronkerb*/. Порядок нахождения клик и порядок вершин внутри клик, т.е. порядок рядов
*в результате NAME - обычно будет другой, чем при использовании /*!KO_bronkerb*/, но список и состав
*клик при RETURN=ALL всегда такой же. Порядок вершин в кликах тут всегда - по возрастанию их номеров.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_bronkerb2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !return= !unquote(!3)
comp @adj= !2.
!if (!head(!return)='COUNT') !then !let !cond= !conc('@cnt=',!tail(!return))
!else !if (!head(!return)='SIZEEQ') !then !let !cond= !conc('ncol(@r)=',!tail(!return))
!else !if (!head(!return)='SIZEGE') !then !let !cond= !conc('ncol(@r)>=',!tail(!return))
!ifend !ifend !ifend
comp @n= rsum(@adj>0).
comp @rs= {0,0}.
comp @ps= {1:nrow(@adj),0}.
comp @xs= {0,0}.
comp @ri= {1,0}.
comp @pi= {nrow(@adj),0}.
comp @xi= {1,0}.
comp !4= {0,0}.
comp @cnt= 0.
loop if ncol(@rs)>1.
 !do !w !in ('@r @p @x') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
- comp !w= !ws(1:!wi(1)).
- comp !ws= !ws((ncol(!w)+1):ncol(!ws)).
- comp !wi= !wi(2:ncol(!wi)).
 !doend
-do if not (any(@p) or any(@x)).
- comp @cnt= @cnt+1.
- comp !4= {!4;t(@r),make(ncol(@r),1,@cnt)}.
  !if (!return<>'ALL') !then
-  do if !cond.
-   comp @cnt= -1.
-   break.
-  end if.
  !ifend
-end if.
-do if any(@p).
- comp @v= @p(1).
- comp @rs= {@r,@rs}.
- comp @ri= {ncol(@r),@ri}.
- do if ncol(@p)=1.
-  comp @#= 0.
- else.
-  comp @#= @p(2:ncol(@p)).
- end if.
- comp @ps= {@#,@ps}.
- comp @pi= {ncol(@#),@pi}.
- do if any(@x).
-  comp @#= {@x,@v}.
- else.
-  comp @#= @v.
- end if.
- comp @xs= {@#,@xs}.
- comp @xi= {ncol(@#),@xi}.
- do if any(@r).
-  comp @#= {@r,@v}.
- else.
-  comp @#= @v.
- end if.
- comp @rs= {@#,@rs}.
- comp @ri= {ncol(@#),@ri}.
- comp @nei= @adj(@v,1:@n(@v)).
  !do !w !in('@p @x') !let !ws= !conc(!w,'s') !let !wi= !conc(!w,'i')
-  comp @#= 0.
-  do if any(!w).
-   loop @i= 1 to ncol(!w).
-    do if any(@nei=!w(@i)).
-     comp @#= {@#,!w(@i)}.
-    end if.
-   end loop.
-   do if ncol(@#)>1.
-    comp @#= @#(2:ncol(@#)).
-   end if.
-  end if.
-  comp !ws= {@#,!ws}.
-  comp !wi= {ncol(@#),!wi}.
  !doend
-end if.
end loop.
do if nrow(!4)>1.
-comp !4= !4(2:nrow(!4),:).
end if.
!if (!head(!return)='SIZEEQ' !or !head(!return)='SIZEGE') !then
-do if @cnt=-1.
- comp !4= !4((nrow(!4)-ncol(@r)+1):nrow(!4),:).
-else.
- comp !4= {0,0}.
-end if.
!ifend
release @adj,@n,@rs,@ps,@xs,@ri,@pi,@xi,@cnt,@r,@p,@x,@v,@#,@nei,@i.
!enddefine.


******** ВЫПУКЛАЯ ОБОЛОЧКА: 2D АЛГОРИТМ ДЖАРВИСА ********.
*/*!KO_chjarv(data%name)*/*.
*Version 1.
*Очерчивает выпуклую оболочку (выпуклый многоугольник из всех периферических точек) у облака точек DATA.
*Базовый алгоритм "заворачивания подарка", называемый также алгоритмом Джарвиса, рассчитан на
*2-мерный случай. DATA - n x 2 матрица данных, координат n точек (n>=3) на плоскости.
*Результат NAME - список номеров точек, составляющих выпуклую оболочку. Оболочка
*собирает точки против часовой стрелки.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_chjarv(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @data= !2.
comp @pt= t(1:nrow(@data)).
comp @min= cmin(@data).
comp @fstpt= mmax((@data(:,1)=@min(1))&*@pt).
do if @data(@fstpt,2)=@min(2).
-comp @off= 100.
else.
-comp @off= -100.
end if.
comp @prev= {@min(1)-1,@data(@fstpt,2)-@off}.
comp @orig= {@min(1),@data(@fstpt,2)}.
comp @prev= @prev-@orig.
comp @data= @data-@orig(make(1,nrow(@data),1),:).
comp !3= 0.
comp @ind= 0.
comp @ind2= 0.
loop.
-comp @sp= @data*t(@prev).
-comp @h= sqrt(rssq(@data)).
-comp @hh= @h*sqrt(rssq(@prev)).
-comp @sp= @sp/(@hh+(@hh=0)).
-comp @sp= rnd(@sp*1E14).
-comp @sp= @sp=mmin(@sp).
-comp @nt= msum(@sp).
-do if @nt=nrow(@sp).
- comp @nxtpt= @pt.
- comp !3= {!3;@nxtpt}.
- break.
-else if @nt=1.
- comp @ind= mmax(@sp&*t(1:nrow(@sp))).
- do if @ind=1.
-  comp @ind2= 2:nrow(@sp)).
- else if @ind=nrow(@sp).
-  comp @ind2= 1:(nrow(@sp)-1).
- else.
-  comp @ind2= {1:(@ind-1),(@ind+1):nrow(@sp)}.
- end if.
-else.
- comp @sp= {@sp;1}.
- comp @n= nrow(@sp).
- comp @ind= make(1,@nt+1,0).
- comp @ind2= 0.
- comp @s= 1.
- loop @c= 1 to @nt+1.
-  loop @i= @s to @n.
-  end loop if @sp(@i).
-  comp @ind(@c)= @i.
-  do if @i>@s.
-   comp @ind2= {@ind2,@s:(@i-1)}.
-  end if.
-  comp @s= @i+1.
- end loop.
- release @s,@c,@i.
- comp @ind= @ind(1:(ncol(@ind)-1)).
- comp @ind2= @ind2(2:ncol(@ind2)).
- comp @ind(grade(@h(@ind)))= @ind.
-end if.
-comp @nxtpt= @pt(@ind).
-comp !3= {!3;@nxtpt}.
-comp @orig= @data(@ind(@nt),:).
-comp @data= @data(@ind2,:).
-comp @pt= @pt(@ind2).
-comp @prev= -@orig.
-comp @data= @data-@orig(make(1,nrow(@data),1),:).
end loop if any(@nxtpt=@fstpt).
comp !3= !3(2:nrow(!3)).
do if @off.
-comp !3= !3(nrow(!3):1).
end if.
release @data,@pt,@min,@fstpt,@off,@prev,@orig,@sp,@h,@hh,@nt,@ind,@ind2,@nxtpt.
!enddefine.


******** ВЕЛИЧИНЫ ОТСТОЯНИЙ ЭЛЕМЕНТОВ ОТ ДИАГОНАЛИ МАТРИЦЫ ********.
*/*!KO_diagoff(nr%nc%diag%name)*/*.
*Version 1.
*Данная функция помечает нулем диагональ матрицы размером NR x NC, а остальные
*элементы помечает величиной отстояния от диагонали. Пометки под диагональю имеют (для удобства)
*отрицательный знак.
*Когда матрица неквадратная, можно говорить, что у нее более одной диагонали.
*Аргумент DIAG (слово заглавными буквами, опционально в кавычках/апострофах) позволяет
*выбрать, какую диагональ помечать нулем и отсчитывать от нее отстояния:
*"FIRST" - первая, левая или главная диагональ (начинается в верхнелевом углу матрицы)
*"LAST" - последняя или правая диагональ (кончается в нижнеправом углу матрицы)
*"MIDDLE" - срединная диагональ, она находится посредине между первой и последней;
*в том случае если abs(NR-NC) есть число нечетное, то срединных диагоналей - две.
*Если матрица квадратна (NR=NC), то диагональ единственная, тогда не важно, как задать DIAG.

define !KO_diagoff(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @c= !3.
!let !diag= !unquote(!4)
comp @o1= (1:@c)(make(1,@r,1),:).
comp @o2= t(1:@r)*make(1,@c,1).
comp !5= @o1-@o2.
!if (!diag<>'FIRST') !then
-do if @r<>@c.
- comp @diff= abs(@r-@c).
- comp @shift= !if (!diag='LAST') !then @diff !else trunc(@diff/2) !ifend.
- do if @r<@c.
-  comp !5= !5-@shift.
   !if (!diag='MIDDLE') !then
-   do if mod(@diff,2).
-    comp !5= !5-(!5>0).
-   end if.
   !ifend
- else.
-  comp !5= !5+@shift.
   !if (!diag='MIDDLE') !then
-   do if mod(@diff,2).
-    comp !5= !5+(!5<0).
-   end if.
   !ifend
- end if.
- release @diff,@shift.
-end if.
!ifend
release @r,@c,@o1,@o2.
!enddefine.


******** ВЕЛИЧИНЫ ОТСТОЯНИЙ ЭЛЕМЕНТОВ ОТ ПОЛОСЫ ДИАГОНАЛЕЙ МАТРИЦЫ ********.
*/*!KO_diagboff(nr%nc%name)*/*.
*Version 1.
*Когда матрица неквадратная, можно говорить, что у нее более одной диагонали.
*Первая диагональ это главная или левая диагональ: начинается их верхнелевого угла.
*Последняя диагональ оканчивается в нижнеправом углу. Всего диагоналей abs(NR-NC)+1 штук,
*они образуют полосу диагоналей.
*Данная функция помечает нулем всю полосу диагоналей матрицы размером NR x NC, а остальные
*элементы помечает величиной отстояния от этой полосы. Пометки под полосой имеют (для удобства)
*отрицательный знак.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_diagboff(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @r= !2.
comp @c= !3.
comp @o1= (1:@c)(make(1,@r,1),:).
comp @o2= t(1:@r)*make(1,@c,1).
comp @diff= abs(@r-@c).
comp @i= 0.
do if @r=@c.
-comp !4= @o1-@o2.
else if @r<@c.
-loop @i= 1 to @r.
- comp @o1(@i,1:@diff+@i)= make(1,@diff,@i).
-end loop.
-comp !4= @o1-@o2.
-comp !4= !4-(!4>0)*@diff.
else if @r>@c.
-loop @i= 1 to @c.
- comp @o2(1:@diff+@i,@i)= make(@diff,1,@i).
-end loop.
-comp !4= @o1-@o2.
-comp !4= !4+(!4<0)*@diff.
end if.
release @r,@c,@o1,@o2,@diff,@i.
!enddefine.


******** СЧИТКА ВСТРАИВАЕМЫХ ДАННЫХ ********.
*/*!KO_read(nr nc name) data*/*.
*Version 2.
*Эта функция имеет особую форму задания. Она читает предложенные вами в синтаксисе значения
*и конструирует из них матрицу NAME. Используйте ее для создания небольших матриц.
*Функция эквивалентна явному созданию матрицы из предложенных элементов при помощи
*операторов {,;}. Ее удобство как раз в том, что она не требует написания вами этих
*операторов (на что уходит человеческое время). Таким образом, вы можете
*просто подать массив значений, разделенных лишь пробелами, и функция создаст из них матрицу.
*Аргументы (заметьте, аргументы в скобках идут через пробел, не через %):
*NR - число рядов в создаваемой матрице (положительное целое: число или имя, не выражение)
*NC - число столбцов в создаваемой матрице (положительное целое: число или имя, не выражение).
*Вы вправе также задать один из этих двух аргументов символом ?. Тогда функция сама
*вычислит его, разделив число вводимых значений на величину другого, заданного аргумента
*(деление должно быть без остатка).
*Сами встраиваемые (inline) данные DATA должны следовать после скобок. Не ставьте точку между
*скобкой и DATA. Точку надо поставить после DATA.
*Данные это массив значений через пробел(ы) или табуляцию. Не важно, на сколько рядов и столбцов
*вы разбили массив значений: порядок создаваемой матрицы определяется аргументами NR, NC. Функция
*читает значения как свободный (неструктурированный) текст, слева направо и строки сверху вниз.
*Число значений должно равняться произведению NRxNC.
*Чем может являться считываемое значение?:
*-положительным, нулевым или отрицательным числом (примеры: 2.5 -3 +4 -.001 0.15);
*-дробью или делением (примеры: 1/3 -3/16 3/-16);
*-именем переменной (например: x);
*-выражением в кавычках или апострофах (примеры: 'x*2' "2/3" '-ln(4)').
* Считывая значения, функция снимает с них кавычки/апострофы (используйте функцию /*!KO_read2()*/,
* если желаете сохранить кавычки/апострофы, т.е. считывать значения как текстовые).
*Не применяйте данную функцию для считки большого массива (тысячи элементов), т.к.
*она медленная (быстрее будет взять данные из синтаксиса командой DATA LIST и затем открыть
*их в MATRIX).
*Данная функция удобна для рутинной работы, но не для программирования алгоритмов.

define !KO_read(!pos= !enclose('(',')') /!pos= !cmdend)
!let !mx= !null !let !comma= ','
!do !val !in (!2) !let !val2= !unquote(!val)
 !if (!val2='-' !or !val2='+') !then !let !sign= !val2
 !else !if (!val2='/') !then !let !mx= !conc(!mx,!val2) !let !comma= !null
 !else !let !mx= !conc(!mx,!comma,!sign,!val2) !let !sign= !null !let !comma= ','
 !ifend !ifend
!doend
!let !nr= !head(!1) !let !nc= !head(!tail(!1)) !let !name= !tail(!tail(!1))
comp !name= {!tail(!mx)}.
!if (!nr='?') !then !let !nr= !conc('ncol(',!name,')/',!nc)
!else !if (!nc='?') !then !let !nc= !conc('ncol(',!name,')/',!nr)
!ifend !ifend
comp !name= reshape(!name,!nr,!nc).
!enddefine.


******** СЧИТКА ВСТРАИВАЕМЫХ ДАННЫХ (БЕЗ РАСКАВЫЧКИ) ********.
*/*!KO_read2(nr nc name) data*/*.
*Version 2.
*Эта функция тождественна /*!KO_read()*/ с единственным отличием, что она не снимает
*кавычки/апострофы со считываемого значения. Поэтому если значение взято в кавычки или апострофы,
*то это значение берется как текстовой элемент в матрицу NAME.

define !KO_read2(!pos= !enclose('(',')') /!pos= !cmdend)
!let !mx= !null !let !comma= ','
!do !val !in (!2)
 !if (!val='-' !or !val='+') !then !let !sign= !val
 !else !if (!val='/') !then !let !mx= !conc(!mx,!val) !let !comma= !null
 !else !let !mx= !conc(!mx,!comma,!sign,!val) !let !sign= !null !let !comma= ','
 !ifend !ifend
!doend
!let !nr= !head(!1) !let !nc= !head(!tail(!1)) !let !name= !tail(!tail(!1))
comp !name= {!tail(!mx)}.
!if (!nr='?') !then !let !nr= !conc('ncol(',!name,')/',!nc)
!else !if (!nc='?') !then !let !nc= !conc('ncol(',!name,')/',!nr)
!ifend !ifend
comp !name= reshape(!name,!nr,!nc).
!enddefine.


******** РАЗРЕЖЕННУЮ МАТРИЦУ В СПИСОК НЕНУЛЕВЫХ ЗНАЧЕНИЙ ********.
*/*!KO_nzlist(mat%name1%name2)*/*.
*Version 1.
*Берет матрицу (или вектор) MAT любого размера. Выписывает из MAT все ненулевые элементы в один список.
*Этот список есть 2-й столбец NAME2, а в 1-м столбце NAME2 записаны номера столбцов MAT, в которых
*находились выписанные значения. NAME2 имеет столько рядов, сколько ненулевых элементов в MAT.
*Вектор-столбец NAME1 имеет длину как число рядов MAT, в нем записано, сколько ненулевых элементов
*в каждом ряду MAT.
*Если в MAT нет ненулевых элементов, функция выдает NAME1 и NAME2 как нулевые скаляры.
*Если входящую MAT транспонировать, результаты будут другие, но информационно эквивалентные.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Функция /*!KO_nzunlist*/ противоположна данной и может восстановить MAT из NAME1 и NAME2.

define !KO_nzlist(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @ram= @mat<>0.
comp !3= rsum(@ram).
do if any(!3).
-comp @m= @ram.
-loop @i= 2 to ncol(@m).
- comp @m(:,@i)= @m(:,@i-1)+@m(:,@i).
-end loop.
-comp @m= @m&*@ram.
-comp @ram= make(nrow(@m),ncol(@m)+1,0).
-comp @nums= 1:ncol(@m).
-comp @m= @m+(not @m)*ncol(@ram).
-loop @i= 1 to nrow(@m).
- comp @ram(@i,@m(@i,:))= @nums.
-end loop.
-comp !4= make(msum(!3),2,0).
-comp @st= 1.
-loop @i= 1 to nrow(@ram).
- comp @n= !3(@i).
- do if @n.
-  comp @en= @st+@n-1.
-  comp @rami= @ram(@i,1:@n).
-  comp !4(@st:@en,:)= {t(@rami),t(@mat(@i,@rami))}.
-  comp @st= @en+1.
- end if.
-end loop.
-release @m,@i,@nums,@st,@en,@n,@rami. 
else.
-comp !3= 0.
-comp !4= 0.
end if. 
release @mat,@ram.
!enddefine.


******** СПИСОК В РАЗРЕЖЕННУЮ МАТРИЦУ ********.
*/*!KO_nzunlist(nnz%list%name)*/*.
*Version 1.
*Создает (разреженную) матрицу, вписывая в пустую матрицу значения из списка.
*Берет ненулевой вектор-столбец NNZ, значения которого должны быть целыми неотрицательными.
*Эти значения - количества вписываемых элементов в каждый ряд создаваемой матрицы NAME (NAME будет иметь столько
*рядов, сколько рядов в NNZ).
*Значения, вписываемые в матрицу NAME, функция берет из 2-го столбца двухстолбцовой матрицы LIST,
*а информацию, в какой столбец NAME следует поместить данное вписываемое значение - берет из соответственнго
*элемента 1-го столбца LIST. (Значения в 1-м столбце LIST должны быть все целые положительные. Сами вписываемые
*значения - 2-й столбец LIST - могут быть любые.)
*В 1-й ряд NAME будет вписано первые NNZ(1) значений из LIST, во 2-й ряд NAME будет вписано NNZ(2) следующих 
*значений из LIST, и так далее.
*Создаваемая матрица NAME имеет столько столбцов, каково максимальное значение в 1-м столбце LIST.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Функция /*!KO_nzlist*/ противоположна данной и может выписать значения из NAME, создавая NNZ и LIST, если
*выписываемые значения - это все ненулевые значения, существующие в NAME.

define !KO_nzunlist(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @nnz= !2.
comp @list= t(!3).
comp !4= make(nrow(@nnz),rmax(@list(1,:)),0).
comp @e= 0.
loop @i= 1 to nrow(@nnz).
-do if @nnz(@i).
- comp @s= @e+1.
- comp @e= @e+@nnz(@i).
- comp !4(@i,@list(1,@s:@e))= @list(2,@s:@e).
-end if.
end loop.
release @nnz,@list,@e,@i,@s.
!enddefine.


******** УТРАМБОВКА ПО ГОРИЗОНТАЛИ НЕНУЛЕВЫХ ЭЛЕМЕНТОВ ИЛИ ИХ ИНДЕКСОВ ********.
*/*!KO_ram(mat%indval%lr%rand%name1%name2)*/*.
*Version 2.
*Берет матрицу или вектор MAT и сдвигает в каждом ряду MAT все ненулевые элементы к левому или
*правому краю, уплотняя их (между ними не будет нулей).
*Аргумент INDVAL - ключевое заглавное слово (можно взять в кавычки или апострофы):
*"VAL" - сдвигать сами ненулевые значения
*"IND" - сдвигать индексы ненулевых значений (т.е. номера их столбцов)
*"BOTH" - делать оба задания
*Аргумент LR - скаляр. Если неположительный, сдвиг будет влево; если положительный, то вправо.
*Результат NAME1 это матрица со сделанной утрамбовкой (сдвижкой) значений или их индексов,
*размером как MAT. При INDVAL=BOTH выдается оба результата в стопку: матрица с индексами подшита
*под матрицу со значениями.
*Результат NAME2 это столбец длиной как число рядов MAT, показывающий число ненулевых элементов в
*каждом ряду MAT/NAME1.
*Аргумент RAND - цифра (не имя и не выражение) 0 или 1 (можно в кавычких или апострофах).
*Если 1, то к сдвижке добавляется действие рандомизации: утрамбованные значения (и их индексы)
*будут идти в случайном порядке, своем в каждом ряду NAME1.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_ram(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
              /!pos= !charend('%') /!pos= !charend(')'))
!let !indval= !unquote(!3)
!let !rand= !unquote(!5)
comp @mat= !2.
comp @lr= !4.
comp @ind= 0.
comp @flag= @mat<>0.
comp !7= rsum(@flag).
do if cmax(!7).
-comp @m= @flag.
-loop @i= 2 to ncol(@m).
- comp @m(:,@i)= @m(:,@i-1)+@m(:,@i).
-end loop.
-do if @lr.
- comp @m= @m+ncol(@m)-!7*make(1,ncol(@m),1).
-end if.
-comp @m= @m&*@flag.
-release @flag.
-comp !6= make(nrow(@m),ncol(@m)+1,0).
 !if (!indval='BOTH') !then !let !ind= '@ind'
- comp @ind= make(nrow(@m),ncol(@m)+1,0).
 !else !let !ind= !6
 !ifend
-comp @nums= 1:ncol(@m).
-comp @m= @m+(not @m)*ncol(!6).
-loop @i= 1 to nrow(@m).
  !if (!indval='VAL' !or !indval='BOTH') !then
-  comp !6(@i,@m(@i,:))= @mat(@i,:).
  !ifend
  !if (!indval='IND' !or !indval='BOTH') !then
-  comp !ind(@i,@m(@i,:))= @nums.
  !ifend
-end loop.
-comp !6= !6(:,1:(ncol(!6)-1)).
 !if (!indval='BOTH') !then
- comp @ind= @ind(:,1:(ncol(@ind)-1)).
 !ifend
 !if (!rand) !then
- loop @i= 1 to nrow(!6).
-  do if !7(@i)>1.
-   comp @m= grade(uniform(1,!7(@i))).
-   comp @nums= 1:!7(@i).
-   do if @lr.
-    comp @nums= @nums+ncol(!6)-!7(@i).
-    comp @m= @m+ncol(!6)-!7(@i).
-   end if.
-   comp !6(@i,@m)= !6(@i,@nums).
    !if (!indval='BOTH') !then
-    comp @ind(@i,@m)= @ind(@i,@nums).
    !ifend
-  end if.
- end loop.
 !ifend
-release @m,@i,@nums.
else.
-release @flag.
-comp !6= make(nrow(@mat),ncol(@mat),0).
 !if (!indval='BOTH') !then
- comp @ind= make(nrow(@mat),ncol(@mat),0).
 !ifend
end if.
!if (!indval='BOTH') !then
-comp !6= {!6;@ind}.
!ifend
release @mat,@lr,@ind.
!enddefine.


******** УТРАМБОВКА ПО ВЕРТИКАЛИ НЕНУЛЕВЫХ ЭЛЕМЕНТОВ ИЛИ ИХ ИНДЕКСОВ ********.
*/*!KO_vram(mat%indval%tb%rand%name1%name2)*/*.
*Version 1.
*Эта функция как /*!KO_ram*/, только сдвигает элементы по вертикали, а не горизонтали.
*Берет матрицу или вектор MAT и сдвигает в каждом столбце MAT все ненулевые элементы к верхнему или
*нижнему краю, уплотняя их (между ними не будет нулей).
*Аргумент INDVAL - ключевое заглавное слово (можно взять в кавычки или апострофы):
*"VAL" - сдвигать сами ненулевые значения
*"IND" - сдвигать индексы ненулевых значений (т.е. номера их рядов)
*"BOTH" - делать оба задания
*Аргумент TB - скаляр. Если неположительный, сдвиг будет вверх; если положительный, то вниз.
*Результат NAME1 это матрица со сделанной утрамбовкой (сдвижкой) значений или их индексов,
*размером как MAT. При INDVAL=BOTH выдается оба результата в гармошку: матрица с индексами пришита
*справа к матрице со значениями.
*Результат NAME2 это ряд длиной как число столбцов MAT, показывающий число ненулевых элементов в
*каждом столбце MAT/NAME1.
*Аргумент RAND - цифра (не имя и не выражение) 0 или 1 (можно в кавычких или апострофах).
*Если 1, то к сдвижке добавляется действие рандомизации: утрамбованные значения (и их индексы)
*будут идти в случайном порядке, своем в каждом столбце NAME1.
*Для регулирования зерна случайных чисел используйте внематричный генератор
*случайных чисел (соответствует командам SET MTINDEX или SET SEED).
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_vram(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
               /!pos= !charend('%') /!pos= !charend(')'))
!let !indval= !unquote(!3)
!let !rand= !unquote(!5)
comp @mat= !2.
comp @tb= !4.
comp @ind= 0.
comp @flag= @mat<>0.
comp !7= csum(@flag).
do if rmax(!7).
-comp @m= @flag.
-loop @i= 2 to nrow(@m).
- comp @m(@i,:)= @m(@i-1,:)+@m(@i,:).
-end loop.
-do if @tb.
- comp @m= @m+nrow(@m)-!7(make(1,nrow(@m),1),:).
-end if.
-comp @m= @m&*@flag.
-release @flag.
-comp !6= make(nrow(@m)+1,ncol(@m),0).
 !if (!indval='BOTH') !then !let !ind= '@ind'
- comp @ind= make(nrow(@m)+1,ncol(@m),0).
 !else !let !ind= !6
 !ifend
-comp @nums= t(1:nrow(@m)).
-comp @m= @m+(not @m)*nrow(!6).
-loop @i= 1 to ncol(@m).
  !if (!indval='VAL' !or !indval='BOTH') !then
-  comp !6(@m(:,@i),@i)= @mat(:,@i).
  !ifend
  !if (!indval='IND' !or !indval='BOTH') !then
-  comp !ind(@m(:,@i),@i)= @nums.
  !ifend
-end loop.
-comp !6= !6(1:(nrow(!6)-1),:).
 !if (!indval='BOTH') !then
- comp @ind= @ind(1:(nrow(@ind)-1),:).
 !ifend
 !if (!rand) !then
- loop @i= 1 to ncol(!6).
-  do if !7(@i)>1.
-   comp @m= grade(uniform(1,!7(@i))).
-   comp @nums= 1:!7(@i).
-   do if @tb.
-    comp @nums= @nums+nrow(!6)-!7(@i).
-    comp @m= @m+nrow(!6)-!7(@i).
-   end if.
-   comp !6(@m,@i)= !6(@nums,@i).
    !if (!indval='BOTH') !then
-    comp @ind(@m,@i)= @ind(@nums,@i).
    !ifend
-  end if.
- end loop.
 !ifend
-release @m,@i,@nums.
else.
-release @flag.
-comp !6= make(nrow(@mat),ncol(@mat),0).
 !if (!indval='BOTH') !then
- comp @ind= make(nrow(@mat),ncol(@mat),0).
 !ifend
end if.
!if (!indval='BOTH') !then
-comp !6= {!6,@ind}.
!ifend
release @mat,@tb,@ind.
!enddefine.


******** РАЗМНОЖЕНИЕ РЯДОВ ********.
*/*!KO_propag(mat%times%name)*/*.
*Version 1.
*Берет матрицу или вектор любого размера MAT. Размножает ее ряды в количество копий,
*указанное в TIMES. В выходящей матрице NAME копии каждого ряда идут смежно.
*TIMES - либо положительный целый скаляр, либо вектор длиной как число рядов MAT, в котором указано
*количество копий каждого ряда отдельно. Значения в векторе - неотрицательные целые, хотя бы одно
*значение больше нуля. Нулевое значение означает не брать данный ряд MAT в NAME вообще.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_propag(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @times= !3.
do if nrow(@times)*ncol(@times)=1.
-comp !4= kroneker(!4,make(@times,1,1)).
else.
-comp @vec= 1:nrow(!4).
-comp @vec2= make(msum(@times),1,0).
-comp @end= 0.
-do if all(@times).
- loop @i= 1 to nrow(!4).
-  comp @start= @end+1.
-  comp @end= @start+@times(@i)-1.
-  comp @vec2(@start:@end)= make(@times(@i),1,@vec(@i)).
- end loop.
-else.
- loop @i= 1 to nrow(!4).
-  do if @times(@i).
-   comp @start= @end+1.
-   comp @end= @start+@times(@i)-1.
-   comp @vec2(@start:@end)= make(@times(@i),1,@vec(@i)).
-  end if.
- end loop.
-end if.
-comp !4= !4(@vec2,:).
-release @vec,@vec2,@start,@end,@i.
end if.
release @times.
!enddefine.


******** ПРИБАВЛЕНИЕ КОНСТАНТЫ К ЭЛЕМЕНТАМ ТРЕУГОЛЬНИКА МАТРИЦЫ ********.
*/*!KO_tradd(mat%tr%val%name)*/*.
*Version 1.
*Прибавляет число VAL (скаляр) к элементам верхнего или нижнего треугольника MAT.
*Матрица MAT не обязана быть квадратной, но должна быть размером не меньше 2x2.
*TR - скаляр; если положительный, VAL прибавляется к элементам над главной диагональю;
*если неположительный, то прибавляется к элементам под главной диагональю.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_tradd(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @tr= !3.
comp @val= !4.
do if @tr.
-do if nrow(!5)>=ncol(!5).
- comp @m= ncol(!5)-1.
-else.
- comp @m= nrow(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):ncol(!5).
- comp !5(@i,@ind)= !5(@i,@ind)+@val.
-end loop.
else.
-do if ncol(!5)>=nrow(!5).
- comp @m= nrow(!5)-1.
-else.
- comp @m= ncol(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):nrow(!5).
- comp !5(@ind,@i)= !5(@ind,@i)+@val.
-end loop.
end if.
release @tr,@val,@m,@i,@ind.
!enddefine.


******** УМНОЖЕНИЕ НА КОНСТАНТУ ЭЛЕМЕНТОВ ТРЕУГОЛЬНИКА МАТРИЦЫ ********.
*/*!KO_trmult(mat%tr%val%name)*/*.
*Version 1.
*Умножает элементы верхнего или нижнего треугольника MAT на число VAL (скаляр).
*Матрица MAT не обязана быть квадратной, но должна быть размером не меньше 2x2.
*TR - скаляр; если положительный, на VAL умножаются элементы над главной диагональю;
*если неположительный, то умножаются элементы под главной диагональю.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_trmult(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @tr= !3.
comp @val= !4.
do if @tr.
-do if nrow(!5)>=ncol(!5).
- comp @m= ncol(!5)-1.
-else.
- comp @m= nrow(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):ncol(!5).
- comp !5(@i,@ind)= !5(@i,@ind)*@val.
-end loop.
else.
-do if ncol(!5)>=nrow(!5).
- comp @m= nrow(!5)-1.
-else.
- comp @m= ncol(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):nrow(!5).
- comp !5(@ind,@i)= !5(@ind,@i)*@val.
-end loop.
end if.
release @tr,@val,@m,@i,@ind.
!enddefine.


******** ЭКСПОНЕНЦИИРОВАНИЕ КОНСТАНТОЙ ЭЛЕМЕНТОВ ТРЕУГОЛЬНИКА МАТРИЦЫ ********.
*/*!KO_trexp(mat%tr%power%name)*/*.
*Version 1.
*Экспоненциирует (возводит в степень) элементы верхнего или нижнего треугольника MAT.
*Степень - POWER (скаляр). Ограничения:
*Если в треугольнике MAT есть отрицательные значения, POWER должно быть целым числом.
*Если в треугольнике MAT есть нулевые значения, POWER должно быть положительным числом.
*Матрица MAT не обязана быть квадратной, но должна быть размером не меньше 2x2.
*TR - скаляр; если положительный, то экспоненциируются элементы над главной диагональю;
*если неположительный, то экспоненциируются элементы под главной диагональю.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_trexp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @tr= !3.
comp @val= !4.
do if @tr.
-do if nrow(!5)>=ncol(!5).
- comp @m= ncol(!5)-1.
-else.
- comp @m= nrow(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):ncol(!5).
- comp !5(@i,@ind)= !5(@i,@ind)&**@val.
-end loop.
else.
-do if ncol(!5)>=nrow(!5).
- comp @m= nrow(!5)-1.
-else.
- comp @m= ncol(!5).
-end if.
-loop @i= 1 to @m.
- comp @ind= (@i+1):nrow(!5).
- comp !5(@ind,@i)= !5(@ind,@i)&**@val.
-end loop.
end if.
release @tr,@val,@m,@i,@ind.
!enddefine.


******** РАЗВЕРТКА ТРЕУГОЛЬНИКОВ КВАДРАТНОЙ МАТРИЦЫ В ВЕКТОРЫ ********.
*/*!KO_unftri(mat%name1%name2%name3)*/*.
*Version 2.
*Берет матрицу MAT, которая должна быть квадратной, и выдает 2-столбцовую матрицу NAME1,
*в которой 1-й столбец есть развертка под-диагонального (нижне-левого) треугольника MAT
*и 2-й столбец есть развертка над-диагонального (верхне-правого) треугольника MAT.
*Пара элементов в одном ряду NAME1 это симметричные по положению элементы в MAT.
*NAME2 это соответствующие NAME1 индексы (позиции) внедиагональных элементов в векторе,
*который есть развертка (векторизация) матрицы MAT. 1-й ряд NAME2 относится к элементам
*под-диагонального треугольника, а 2-й ряд - к элементам над-диагонального треугольника.
*С помощью NAME2 вы быстро можете развернуть треугольники любой матрицы размером с MAT,
*не прибегая вторично к данной функции.
*NAME3 содержит памятку - номера рядов и столбцов MAT, откуда взяты элементы. Так, 1-я пара
*индексов в NAME3 есть 2 и 1, что напоминает, что 1-я пара значений в NAME1 это: 2-й эл-т
*из 1-го столбца и 2-й элемент из 1-го ряда.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Функция /*!KO_foltri*/ противоположна данной: сворачивает пару векторов NAME1 в треугольники
*квадратной матрицы MAT.

define !KO_unftri(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !3= !2.
comp @n= ncol(!3).
comp @n1= @n+1.
comp @1n= 1-@n.
comp @nn1= @n*(@n-1).
comp !3= reshape(!3,@n**2,1).
comp @uind= make(1,@nn1/2,0).
comp @lind= make(1,@nn1/2,0).
comp @ind1= make(1,@nn1/2,0).
comp @ind2= make(1,@nn1/2,0).
loop @i= 1 to @n-1.
-comp @st= @1n+@i*(@n-@i/2+.5).
-comp @ind= (@i*@n1+@1n):(@i*@n).
-comp @pos= @st:(@st+ncol(@ind)-1).
-comp @uind(@pos)= @ind.
-comp @ind= (@i*@n1):(@i+@nn1):@n.
-comp @lind(@pos)= @ind.
-comp @ind1(@pos)= make(1,ncol(@ind),@i).
-comp @ind2(@pos)= (@i+1):@n.
end loop.
comp !3= {!3(@lind),!3(@uind)}.
comp !4= {@lind;@uind}.
comp !5= {@ind2;@ind1}.
release @n,@n1,@1n,@nn1,@uind,@lind,@ind1,@ind2,@i,@st,@ind,@pos.
!enddefine.


******** СВЕРТКА ДВУХ ВЕКТОРОВ В ТРЕУГОЛЬНИКИ КВАДРАТНОЙ МАТРИЦЫ ********.
*/*!KO_foltri(vecs%name1%name2%name3)*/*.
*Version 2.
*Берет матрицу VECS, в которой должно быть 2 стобца (векторы), и выдает квадратную
*матрицу NAME1, в которой под-диагональный (нижне-левый) треугольник есть свертка 1-го вектора
*и над-диагональный (верхне-правый) треугольник есть свертка 2-го вектора.
*Пара элементов в одном ряду VECS становится симметричными по положению элементами NAME1.
*Диагональ NAME1 состоит из нулей.
*NAME2 это соответствующие NAME1 индексы (позиции) внедиагональных элементов в векторе,
*который есть развертка (векторизация) матрицы NAME1. 1-й ряд NAME2 относится к элементам
*под-диагонального треугольника, а 2-й ряд - к элементам над-диагонального треугольника.
*NAME3 это скаляр, размер матрицы NAME1.
*Если число рядов VECS не позволяет создать квадратную матрицу NAME1, то NAME3 окажется
*дробным числом, и NAME1 и NAME2 будут выданы как нулевые скаляры.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Функция /*!KO_unftri*/ противоположна данной: разворачивает треугольники квадратной матрицы NAME1
*в пару векторов VECS.

define !KO_foltri(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @vecs= !2.
comp !5= .5+sqrt(.25+2*nrow(@vecs)).
do if trunc(!5)=!5.
-comp @n1= !5+1.
-comp @1n= 1-!5.
-comp @nn1= !5*(!5-1).
-comp @uind= make(1,nrow(@vecs),0).
-comp @lind= make(1,nrow(@vecs),0).
-loop @i= 1 to !5-1.
- comp @st= @1n+@i*(!5-@i/2+.5).
- comp @ind= (@i*@n1+@1n):(@i*!5).
- comp @uind(@st:(@st+ncol(@ind)-1))= @ind.
- comp @ind= (@i*@n1):(@i+@nn1):!5.
- comp @lind(@st:(@st+ncol(@ind)-1))= @ind.
-end loop.
-comp !3= make(!5**2,1,0).
-comp !3(@lind)= @vecs(:,1).
-comp !3(@uind)= @vecs(:,2).
-comp !3= reshape(!3,!5,!5).
-comp !4= {@lind;@uind}.
-release @vecs,@n1,@1n,@nn1,@uind,@lind,@i,@st,@ind.
else.
-comp !3= 0.
-comp !4= 0.
end if.
!enddefine.


******** СИММЕТРИЗАЦИЯ КВАДРАТНОЙ МАТРИЦЫ: ЗАМЕНА БОЛЬШЕГО ЭЛЕМЕНТА МЕНЬШИМ ИЛИ МЕНЬШЕГО БОЛЬШИМ ********.
*/*!KO_symtri2(mat%el%name)*/*.
*Version 2.
*Берет матрицу MAT, которая должна быть квадратной, и заменяет в ней каждый внедиагональный элемент
*симметричным ему по положению, если последний больше (или меньше) него;
*выдает квадратную симметрическую матрицу NAME.
*Аргумент EL заказывает, когда делать замещение:
*EL неположительный скаляр => заместить элемент симметричным, если последний меньше него;
*EL положительный скаляр => заместить элемент симметричным, если последний больше него.

define !KO_symtri2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @el= !3.
comp @t= t(!4).
do if @el.
-comp @gv= !4&*(!4>@t).
else.
-comp @gv= !4&*(!4<@t).
end if.
comp @ev= !4&*(!4=@t).
comp !4= @gv+t(@gv)+@ev.
release @el,@t,@gv,@ev.
!enddefine.


******** СИММЕТРИЗАЦИЯ КВАДРАТНОЙ МАТРИЦЫ: ЗАМЕНА БОЛЬШЕГО ЭЛЕМЕНТА МЕНЬШИМ ИЛИ МЕНЬШЕГО БОЛЬШИМ ********.
*/*!KO_symtri2(mat%el%name)*/*.
*Version 1.
*Берет матрицу MAT, которая должна быть квадратной, и заменяет в ней каждый внедиагональный элемент
*симметричным ему по положению, если последний больше (или меньше) него;
*выдает квадратную симметрическую матрицу NAME.
*Аргумент EL заказывает, когда делать замещение:
*EL неположительный скаляр => заместить элемент симметричным, если последний больше него;
*EL положительный скаляр => заместить элемент симметричным, если последний меньше него.

define !KO_symtri2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @el= !3.
comp @el= (@el>0)-(@el<=0).
comp @delta= !4-t(!4).
comp @delta= not(not(@el*@delta)).
comp @delta= @delta&*!4.
comp @delta= @delta+t(@delta).
comp !4= @delta+!4&*(not @delta).
release @el,@delta.
!enddefine.


******** СОРТИРОВКА ВЕКТОРА ********.
*/*!KO_sort(vec%ord%name1%name2)*/*.
*Version 2.
*Сортирует элементы в векторе (ряд или столбец) VEC по величине их значений.
*Выдает сортированный вектор NAME1.
*А также вектор NAME2, который есть номера элементов NAME1, расположенные в последовательности,
*в какой элементы были до сортировки (с помощью NAME2 можно восстановить исходный порядок элементов).
*NAME2 есть целочисленные ранги значений VEC.
*Аргумент ORD заказывает порядок сортировки:
*ORD неположительный скаляр => по убыванию; ORD положительный скаляр => по возрастанию.
*Если хотите затем переупорядочить ряды внешнего массива так же, как оказались сортированы элементы в VEC,
*используйте NAME2 в качестве COL в /*!KO_rsortc*/ и задайте ORD как 0.

define !KO_sort(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @ord= !3.
comp @ord= (@ord>0)-(@ord<=0).
comp !5= grade(@ord&*!4).
comp !4(!5)= !4.
release @ord.
!enddefine.


******** ПРОСТАЯ СОРТИРОВКА РЯДОВ В МАТРИЦЕ ********.
*/*!KO_rsort(mat%cn%ord%name1%name2)*/*.
*Version 2.
*Сортирует ряды в матрице MAT по величине значений в ее столбце CN.
*Выдает матрицу NAME1.
*А также столбец NAME2, который есть номера рядов NAME1, расположенные в последовательности,
*в какой ряды были до сортировки (с помощью NAME2 можно восстановить исходный порядок рядов).
*NAME2 есть целочисленные ранги значений столбца CN (до сортировки).
*CN это номер столбца, положительный целый скаляр.
*Аргумент ORD заказывает порядок сортировки:
*ORD неположительный скаляр => по убыванию; ORD положительный скаляр => по возрастанию.
*Если хотите затем переупорядочить ряды внешнего массива так же, как оказались сортированы ряды в MAT,
*используйте NAME2 в качестве COL в /*!KO_rsortc*/ и задайте ORD как 0.

define !KO_rsort(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @cn= !3.
comp @ord= !4.
comp @ord= (@ord>0)-(@ord<=0).
comp !6= grade(@ord&*!5(:,@cn)).
comp !5(!6,:)= !5.
release @cn,@ord.
!enddefine.


******** СОРТИРОВКА РЯДОВ В МАТРИЦЕ ПО ЗНАЧЕНИЯМ ВНЕШНЕГО СТОЛБЦА ********.
*/*!KO_rsortc(mat%col%ord%name1%name2)*/*.
*Version 2.
*Сортирует ряды в матрице MAT по величине значений столбца COL.
*COL это вектор-столбец длиной как число рядов MAT.
*Выдает матрицу NAME1.
*А также столбец NAME2, который есть номера рядов NAME1, расположенные в последовательности,
*в какой ряды были до сортировки (с помощью NAME2 можно восстановить исходный порядок рядов).
*NAME2 есть целочисленные ранги значений COL.
*Аргумент ORD заказывает порядок сортировки:
*ORD отрицательный скаляр => по убыванию; ORD положительный скаляр => по возрастанию.
*Если ORD есть 0, то COL должен содержать целочисленные ранги и порядок сортировки определяется
*уже ими. При этом функция не тратит время на ранжирование. Этот вариант задания выгоден в смысле скорости,
*когда вы хотите отсортировать ряды в нескольких матрицах (с одним и тем же числом рядов) в идентичном
*порядке, но не хотите ради этого временно объединять матрицы. Тогда вы можете использовать
*NAME2, полученный в сортировке первой матрицы, в качестве COL в сортировках остальных матриц.

define !KO_rsortc(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @col= !3.
comp @ord= !4.
do if @ord=0.
-comp !6= @col.
else.
-comp @ord= (@ord>0)-(@ord<0).
-comp !6= grade(@ord&*@col).
end if.
-comp !5(!6,:)= !5.
release @col,@ord.
!enddefine.


******** ИЕРАРХИЧЕСКАЯ СОРТИРОВКА РЯДОВ В МАТРИЦЕ ********.
*/*!KO_hiesort(mat%ord%name1%name2)*/*.
*Version 2.
*Сортирует ряды в матрице MAT по величине значений во всех ее столбцах.
*Ряды сортируются иерархически: по значениям 1-го столбца; внутри одинаковых
*его значений - по значениям 2-го столбца; внутри одинаковых его значений - по
*значениям 3-го столбца; и т.д. Эта функция подобна внематричной команде SORT CASES,
*когда в ней указываешь несколько переменных.
*Выдает матрицу NAME1.
*А также столбец NAME2, который есть номера рядов NAME1, расположенные в последовательности,
*в какой ряды были до сортировки (с помощью NAME2 можно восстановить исходный порядок рядов).
*Аргумент ORD это вектор длиной как число столбцов MAT либо скаляр; он заказывает порядок сортировки
*в каждом отдельном столбце:
*n-й элемент ORD неположительный => сортировка в n-м столбце по убыванию;
*n-й элемент ORD положительный => сортировка в n-м столбце по возрастанию.
*ORD скаляр значит один и тот же порядок во всех столбцах.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Функция относительно медленная.
*Если хотите затем переупорядочить ряды внешнего массива так же, как оказались сортированы ряды в MAT,
*используйте NAME2 в качестве COL в /*!KO_rsortc*/ и задайте ORD как 0.

define !KO_hiesort(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @ord= !3.
comp @ord= (@ord>0)-(@ord<=0).
do if nrow(@ord)*ncol(@ord)=1.
-comp @ord= make(1,ncol(!4),@ord).
end if.
comp !5= t(1:nrow(!4)).
comp !4= {!4,!5}.
comp !4(grade(@ord(1)*!4(:,1)),:)= !4.
do if ncol(!4)>2.
-loop @i= 2 to ncol(!4)-1.
- comp @lst= 0.
- loop. 
-  comp @fst= @lst+1.
-  loop @lst= @fst to nrow(!4)-1.
-  end loop if any(!4(@lst+1,1:(@i-1))<>!4(@lst,1:(@i-1))).
-  comp !4(grade(@ord(@i)*!4(@fst:@lst,@i))+@fst-1,:)= !4(@fst:@lst,:).
- end loop if @lst=nrow(!4).
-end loop.
-release @i,@lst,@fst.
end if.
comp !5(!4(:,ncol(!4)))= !5.
comp !4= !4(:,1:(ncol(!4)-1)).
release @ord.
!enddefine.


******** РАСТАСКИВАНИЕ МАТРИЦЫ НА ДВЕ ********.
*/*!KO_split(whole%dim%ind%name1%name2)*/*.
*Version 2.
*Из матрицы WHOLE вырезаются ее ряды либо столбцы под номерами IND,
*образуя матрицу-вырезку NAME1. Оставшиеся ряды (столбцы) матрицы WHOLE образуют
*матрицу-остаток NAME2. Функция выдает обе матрицы NAME1 и NAME2.
*Аргумент DIM - ключевое слово (опционально в кавычках или апострофах):
*"ROWS" - вырезать ряды (слово "меры" ниже - это значит ряды);
*"COLS" - вырезать столбцы (слово "меры" ниже - это значит столбцы).
*WHOLE должна иметь число мер более одной.
*IND это вектор (ряд или столбец, все равно) длиной меньше, чем есть мер в WHOLE,
*и содержащий целые положительные числа: номера мер в WHOLE, которые надо вырезать.
*Эти числа не обязаны идти по возрастающей; их последовательность задает последовательность
*мер, какая она будет в матрице-вырезке NAME1.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно
*большое число командой SET MXLOOPS.
*Функция /*!KO_merge*/ противоположна данной и может вставить врезку NAME1 в основу NAME2
*с восстановлением WHOLE.

define !KO_split(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @whole= !2.
!let !dim= !unquote(!3)
comp @ind= !4.
comp !5= @whole(!if (!dim='COLS') !then !let !nd= 'ncol' :,@ind !else !let !nd= 'nrow' @ind,: !ifend).
comp @lenind= rmax({nrow(@ind),ncol(@ind)}).
comp @notind= make(1,!nd(@whole)-@lenind,0).
comp @c= 0.
loop @i= 1 to !nd(@whole).
-do if not any(@ind=@i).
- comp @c= @c+1.
- comp @notind(@c)= @i.
-end if.
end loop.
comp !6= @whole(!if (!dim='COLS') !then :,@notind !else @notind,: !ifend).
release @whole,@ind,@lenind,@notind,@c,@i.
!enddefine.


******** СМЕШЕНИЕ ДВУХ МАТРИЦ ********.
*/*!KO_merge(base%dim%slice%ind%name)*/*.
*Version 2.
*В матрицу-основу BASE добавляются все ряды или столбцы матрицы-врезки SLICE, образуя
*объединенную матрицу NAME, которая и выдается. Ряды (столбцы), составляющие SLICE, займут в NAME
*позиции, заказанные в IND.
*Аргумент DIM - ключевое слово (опционально в кавычках или апострофах):
*"ROWS" - вставлять SLICE как ряды в BASE (слово "меры" ниже - это значит ряды), число столбцов
*в BASE и SLICE должно быть одинаково;
*"COLS" - вставлять SLICE как столбцы в BASE (слово "меры" ниже - это значит столбцы), число рядов
*в BASE и SLICE должно быть одинаково.
*IND это вектор (ряд или столбец, все равно) длиной как число мер в SLICE, и содержащий
*целые положительные числа: номера мер в NAME, на которые надо поставить меры матрицы-врезки SLICE.
*Эти числа не обязаны идти по возрастающей; их последовательность задает последовательность врезки
*мер SLICE между мерами BASE. Максимальное значение в IND не может превышать сумму числа мер BASE и SLICE.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно
*большое число командой SET MXLOOPS.
*Функция /*!KO_split*/ противоположна данной и может расщепить NAME обратно на
*вырезку SLICE и остаток BASE.

define !KO_merge(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @base= !2.
!let !dim= !unquote(!3)
comp @slice= !4.
comp @ind= !5.
comp !6= make(!if (!dim='COLS') !then !let !nd= 'ncol' nrow(@base),ncol(@base)+ncol(@slice) !else !let !nd= 'nrow' nrow(@base)+nrow(@slice),ncol(@base) !ifend,0).
comp !6(!if (!dim='COLS') !then :,@ind !else @ind,: !ifend)= @slice.
comp @r= 0.
loop @i= 1 to !nd(!6).
-do if not any(@ind=@i).
- comp @r= @r+1.
- comp !if (!dim='COLS') !then !6(:,@i)= @base(:,@r) !else !6(@i,:)= @base(@r,:) !ifend.
-end if.
end loop.
release @base,@slice,@ind,@r,@i.
!enddefine.


******** РАЗДЕЛЕНИЕ МАТРИЦЫ ГОРИЗОНТАЛЬНО НА НЕСКОЛЬКО, ПО ГРУППАМ ********.
*/*!KO_msplit(mat%group%'maxcode')*/*.
*Version 1.
*Берет матрицу MAT (некоторое имя, но не выражение) и группирующую переменную GROUP (сортированную).
*Расщепляет MAT горизонтально на матрицы, соответствующие группам в GROUP. Имена результатных
*матриц определяются автоматически.
*Функция отчасти похожа на команду SPLIT FILE.
*Подробно:
*GROUP - группирующая наблюдения переменная, столбец с целочисленными кодами групп: 1, 2, 3,...
*(Дробные коды или меньше единицы - недопустимы.) Наблюдения (ряды) должны быть уже сортированы по
*возрастанию кодов.
*MAT - матрица (данные) с числом рядов как в GROUP, уже сортированными в соответствие с GROUP, которую
*требуется расщепить горизонтально: одна матрица-результат на каждый код из GROUP.
*Имена будут образованы из имени MAT и кода группы: MAT1, MAT2,...
*'MAXCODE' - код последней наличной или интересующей вас группы (максимум 999). Это должно быть число
*(не переменная и не выражение), его можно взять в кавычки или апострофы.
*Если некоторые коды между 1 и maxcode отсутствуют в GROUP, матрицы для этих отсутствующих групп не будут
*созданы.
*Данная функция удобна для рутинной работы, но не для программирования алгоритмов.

define !KO_msplit(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @gr= !3.
comp @ind= t(1:nrow(!2)).
comp @lst= 0.
!do !i= 1 !to !unquote(!4)
-comp @flag= (@gr=!i)&*@ind.
-comp @fst= @lst+1.
-comp @last= cmax(@flag).
-do if @last.
- comp @lst= @last.
- comp !conc(!2,!i)= !2(@fst:@lst,:).
-end if.
!doend
release @gr,@ind,@flag,@fst,@lst,@last.
!enddefine.


******** ПЕРЕСТРУКТУРИРОВАНИЕ "ПЕРЕМЕННЫЕ В НАБЛЮДЕНИЯ" ********.
*/*!KO_vartocas(data%k%ord%name1%name2)*/*.
*Version 1.
*Переструктурирует каждую группу из K переменных (столбцов) DATA в единый стобец.
*Таким образом каждое наблюдение (ряд) разбивается на K рядов.
*Выдает переструктурированные данные NAME1.
*Функция подобна SPSS команде VARSTOCASES.
*DATA - матрица с числом столбцов, кратным K.
*K - целый положительный скаляр; каждые K столбцов в DATA (каждая "группа") дадут один столбец в NAME1.
*Скаляр ORD указывает, в каком порядке расположены в DATA переменные, дающие единый столбец:
*расположены смежно (ORD положителен) или расположены с шагом число_столбцов_DATA/K (ORD неположителен).
*Например, если число_столбцов_DATA=6 и K=2, то смежное расположение значит, что столбцы
*1 и 2 дадут столбец, 3 и 4 дадут столбец, 5 и 6 дадут столбец. А при расположении через шаг 3
*столбцы 1 и 4 дадут столбец, 2 и 5 дадут столбец, 3 и 6 дадут столбец.
*Двухстолбцовая матрица NAME2 отражает этот порядок: она будет содержать
*номера наблюдений DATA в своем 1-м столбце и порядковый индекс
*переменной в "группе" (1,2,...,K) переменных в своем 2-м столбце.
*Функция */*!KO_castovar*/* противоположна данной и переструктурирует NAME1 обратно в DATA.

define !KO_vartocas(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @k= !3.
comp @ord= !4.
comp !6= t({kroneker(1:nrow(!5),make(1,@k,1));kroneker(make(1,nrow(!5),1),1:@k)}).
do if @ord.
-comp !5= !5(:,kroneker(make(1,@k,1),1:ncol(!5):@k)+kroneker(0:(@k-1),make(1,ncol(!5)/@k,1))).
end if.
comp !5= reshape(!5,nrow(!5)*@k,ncol(!5)/@k).
release @k,@ord.
!enddefine.


******** ПЕРЕСТРУКТУРИРОВАНИЕ "НАБЛЮДЕНИЯ В ПЕРЕМЕННЫЕ" ********.
*/*!KO_castovar(data%k%ord%name1%name2)*/*.
*Version 1.
*Переструктурирует каждую группу из K наблюдений (рядов) DATA в единый ряд.
*Таким образом каждая переменная (столбец) разбивается на K столбцов.
*Выдает переструктурированные данные NAME.
*Функция подобна SPSS команде CASESTOVARS.
*DATA - матрица с числом рядов, кратным K.
*K - целый положительный скаляр; каждые K рядов в DATA (каждая "группа") дадут один ряд в NAME1.
*Ряды в DATA должны быть заранее сортированы так, чтобы K рядов каждой группы были смежны.
*Скаляр ORD задает, в каком порядке надо расположить в NAME1 переменные, образованные из
*одной переменной DATA: смежно (ORD положителен) или же с шагом число_столбцов_DATA
*(ORD неположителен). Например, если число_столбцов_DATA=3 и K=2, то смежное расположение значит,
*что столбец 1 даст столбцы 1 и 2, столбец 2 даст столбцы 3 и 4, столбец 3 даст столбцы 5 и 6.
*А при расположении через шаг 3 столбец 1 даст столбцы 1 и 4, столбец 2 даст столбцы 2 и 5,
*столбец 3 даст столбцы 3 и 6.
*Двухрядная матрица NAME2 отражает этот порядок: она будет содержать номера переменных DATA
*в своем 1-м ряду и порядковый индекс наблюдения в "группе" (1,2,...,K) наблюдений в своем 2-м ряду.
*Функция */*!KO_vartocas*/* противоположна данной и переструктурирует NAME1 обратно в DATA.

define !KO_castovar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @k= !3.
comp @ord= !4.
comp @ind= kroneker(make(1,ncol(!5),1),1:(ncol(!5)*@k):ncol(!5))+kroneker(0:(ncol(!5)-1),make(1,@k,1)).
comp !6= {kroneker(make(1,@k,1),1:ncol(!5));kroneker(1:@k,make(1,ncol(!5),1))}.
comp !5= reshape(!5,nrow(!5)/@k,ncol(!5)*@k).
do if @ord.
-comp !6= !6(:,@ind).
-comp !5= !5(:,@ind).
end if.
release @k,@ord,@ind.
!enddefine.


******** СМЕЩЕНИЕ РЯДОВ ПО ГОРИЗОНТАЛИ ********.
*/*!KO_shift(mat%rel%displ%name)*/*.
*Version 1.
*Берет матрицу MAT с не менее чем двумя рядами. Выдает матрицу NAME, в которой ряды
*смещены по горизонтали - вправо или влево на величины (число элементов), указанные в векторе-столбце
*DISPL. DISPL должен быть длиной на 1 меньше, чем число рядов MAT. Его значения - целые числа.
*Отрицательное значение это смещение на столько-то элементов влево, положительное значение это
*смещение на столько-то элементов вправо. Первый элемент это смещение ряда 2, 2-й элемент это смещение
*ряда 3, и т.д.
*Аргумент REL (скаляр) поясняет, относительно чего заданы смещения в DISPL. Если REL неположителен,
*смещение каждого ряда это смещение относительно 1-го ряда. Если REL положителен, смещение каждого ряда
*это смещение относительно предыдущего, вышележащего ряда.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_shift(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @rel= !3.
comp @shift= !4.
do if @rel.
-loop @i= 2 to nrow(@shift).
- comp @shift(@i)= @shift(@i-1)+@shift(@i).
-end loop.
end if.
comp @marg= cmax((@shift<0)&*(-@shift)).
comp !5= make(nrow(@mat),ncol(@mat)+cmax((@shift>0)&*@shift)+@marg,0).
comp @shift= {0;@shift}.
loop @i= 1 to nrow(@shift).
-comp !5(@i,1:ncol(@mat)+@marg+@shift(@i))= @mat(@i,:).
end loop.
release @mat,@rel,@shift,@i,@marg.
!enddefine.


******** БЛОК-ВРЕЗКА ********.
*/*!KO_blockic(mat1%mat2%pos%name)*/*.
*Version 1.
*Берет матрицы MAT1 и MAT2 любого размера и соединяет их, врезая MAT2 снизу-справа
*в MAT1, пристыковывая ее к позиции POS в MAT1. Конкретно, сначала матрица NAME создается как
*блок матрица block(MAT1(1:I,1:J),MAT2). Потом, если в NAME остались ячейки, еще не
*получившие значений из соответствующих им ячеек MAT1, они их получают.
*POS - вектор длиной 2, содержащий позиции I и J в MAT1. I может быть от 1 до число_рядов MAT1.
*J может быть от 1 до число_столбцов MAT1. Когда I и J оба максимальны, функция превращается
*в функцию block().
*Если вы хотите "перевернуть" задачу и врезать MAT2 сверху-слева в MAT1, то сначала
*поменяйте порядок рядов и столбцоа в обеих матрицах на обратный, а в конце поменяйте порядок
*рядов и столбцов в NAME на обратный.

define !KO_blockic(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @m1= !2.
comp @m2= !3.
comp @cr= !4.
comp @rs= @cr(1).
comp @cs= @cr(2).
comp !5= block(@m1(1:@rs,1:@cs),@m2).
do if @cs<ncol(@m1).
-comp @cr= mmin({ncol(@m1),ncol(!5)}).
-comp !5(1:@rs,(@cs+1):@cr)= @m1(1:@rs,(@cs+1):@cr).
end if.
do if @rs<nrow(@m1).
-comp @cr= mmin({nrow(@m1),nrow(!5)}).
-comp !5((@rs+1):@cr,1:@cs)= @m1((@rs+1):@cr,1:@cs).
end if.
release @m1,@m2,@cr,@rs,@cs.
!enddefine.


******** ДВУМЕСТНЫЕ ОПЕРАЦИИ С МНОЖЕСТВАМИ ********.
*/*!KO_setdo(set1%set2%oper%name)*/*.
*Version 1.
*Берет два множества SET1 и SET2 (произвольные матрицы) и выдает их объединение, пересечение,
*разность или симметричную разность. Значение 0 в SET1 и SET2 зарезервировано как пустота, т.е.
*не считается элементом множества. Всякое прочее значение является элементом и учитывается
*однократно в результатном векторе-ряде NAME (т.е., например, пересечение между {1,2,2,2,3}
*и {1,2,2,3,4} есть {1,2,3}, а не {1,2,2,3} или {1,2,2,2,3}).
*OPER - кл. слово заглавными буквами (можно в кавычках/апострофах):
*"UNION" - объединение двух множеств (значения, присутствующие хотя бы в одном из множеств);
*"INTER" - пересечение двух множеств (значения, присутствующие в обоих множествах);
*"DIFF" - разность двух множеств (значения, присутствующие в SET1, но не в SET2);
*"SYMDIFF" - симметричная разность двух множеств (значения, присутствующие в SET1, но не в SET2,
*либо в SET2, но не в SET1).
*NAME = скаляр 0 означает пустое множество на выходе.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_setdo(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @set1= !2.
comp @set2= !3.
!let !oper= !unquote(!4)
!if (!oper='DIFF') !then !let !n= 'not' !ifend
comp @i= 1.
comp !5= 0.
loop @i= 1 to nrow(@set1).
-loop @j= 1 to ncol(@set1).
- comp @val= @set1(@i,@j).
- do if @val<>0.
   !if (!oper='INTER' !or !oper='DIFF') !then
-   do if !n any(@set2=@val).
-    comp !5= {!5,@val}.
-    comp @set1= @set1&*(@set1<>@val).
-   end if.
   !else !if (!oper='UNION') !then
-   comp !5= {!5,@val}.
-   comp @set1= @set1&*(@set1<>@val).
-   comp @set2= @set2&*(@set2<>@val).
   !else
-   do if any(@set2=@val).
-    comp @set2= @set2&*(@set2<>@val).
-   else.
-    comp !5= {!5,@val}.
-   end if.
-   comp @set1= @set1&*(@set1<>@val).
   !ifend !ifend
- end if.
-end loop.
end loop.
!if (!oper='UNION' !or !oper='SYMDIFF') !then
-loop @i= 1 to nrow(@set2).
- loop @j= 1 to ncol(@set2).
-  comp @val= @set2(@i,@j).
-  do if @val<>0.
-   comp !5= {!5,@val}.
-   comp @set2= @set2&*(@set2<>@val).
-  end if.
- end loop.
-end loop.
!ifend
do if ncol(!5)>1.
-comp !5= !5(2:ncol(!5)).
end if.
release @set1,@set2,@i,@j,@val.
!enddefine.


******** ДЕКАРТОВО ПРОИЗВЕДЕНИЕ ДВУХ МНОЖЕСТВ ********.
*/*!KO_cartp(set1%set2%ii%name)*/*.
*Version 1.
*Комбинирует элементы между двумя множествами, создавая декартово произведение. Есть опция не спаривать
*элементы с одинаковыми порядковыми номерами.
*Множества SET1 и SET2 это матрицы одного или разного размера с числом рядов >1.
*Их ряды это есть элементы множеств.
*Результат NAME - матрица, каждый ряд это уникальная пара элементов, первый - из SET1,
*второй - из SET2. Декартово произведение это все возможные парные сочетания.
*Столбцы NAME это объединение столбцов входящих матриц.
*Аргумент II - скаляр. Если он неположителен, выдается обычное, полное декартово произведение.
*Если II положителен, то пары, представляющие элементы с одинаковыми порядковыми номерами, исключены из результата.
*Например, если SET1 это A,B,C и SET2 это 1,2,3,4, то будут опущены пары A1, B2 и C3.

define !KO_cartp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @1= !2.
comp @2= !3.
comp @ii= !4.
do if @ii and nrow(@1)>nrow(@2).
-comp @1_= @1((nrow(@2)+1):nrow(@1),:).
-comp @1= @1(1:nrow(@2),:).
-comp !5= {kroneker(@1,make(nrow(@2),1,1)),kroneker(make(nrow(@1),1,1),@2)}.
-comp @ind= kroneker(make(1,nrow(@1)-1,1),1:nrow(@2)+1)+kroneker(0:(nrow(@1)-2),make(1,nrow(@2),nrow(@2)+1)).
-comp !5= {!5(@ind,:);{kroneker(@1_,make(nrow(@2),1,1)),kroneker(make(nrow(@1_),1,1),@2)}}.
-release @1_,@ind.
else.
-comp !5= {kroneker(@1,make(nrow(@2),1,1)),kroneker(make(nrow(@1),1,1),@2)}.
-do if @ii.
- comp @ind= kroneker(make(1,nrow(@1)-1,1),1:nrow(@2)+1)+kroneker(0:(nrow(@1)-2),make(1,nrow(@2),nrow(@2)+1)).
- do if nrow(@1)<nrow(@2).
-  comp @ind= {@ind,(@ind(ncol(@ind))+2):nrow(!5)}.
- end if.
- comp !5= !5(@ind,:).
- release @ind.
-end if.
end if.
release @1,@2,@ii.
!enddefine.


******** ДЕКАРТОВО ПРОИЗВЕДЕНИЕ N МНОЖЕСТВ ********.
*/*!KO_ncartp(setlist%name)*/*.
*Version 1.
*Комбинирует элементы между двумя или более множествами, создавая декартово произведение.
*Множества это матрицы любого размера, ряды матрицы это элементы множества.
*SETLIST это список (через пробел) имен матриц. Список может опционально быть в кавычках/апострофах.
*Выдает матрицу NAME, ряды которой это элементы декартова произведения этих матриц, а столбцы это
*объединение столбцов этих матриц.

define !KO_ncartp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
!let !sets= !unquote(!2) !let !sts= !null
!do !w !in (!sets) !let !sts= !conc(!w,' ',!sts) !doend
comp !3= !head(!sts).
!do !w !in (!tail(!sts))
-comp !3= {kroneker(!w,make(nrow(!3),1,1)),kroneker(make(nrow(!w),1,1),!3)}.
!doend
!enddefine.


******** СОЧЕТАНИЯ ПО K ЭЛЕМЕНТОВ ********.
*/*!KO_combk(set%k%name)*/*.
*Version 1.
*Образует сочетания по k элементов из множества n элементов.
*Множество SET это столбец или матрица с n рядами: элементом множества выступает ряд.
*K (целый скаляр от 2 до n) - число элементов в сочетании.
*Результат NAME - матрица, каждый ряд которой есть сочетание элементов.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_combk(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @v= !2.
comp @maxk= !3.
comp @n= nrow(@v).
comp @next= 1:@n+1.
comp !4= @v.
loop @k= 2 to @maxk.
-comp @nextnew= 0.
-comp @vecnew= make(1,ncol(!4)+ncol(@v),0).
-loop @i= 1 to nrow(!4).
- do if @next(@i)<=@n.
-  comp @ind= @next(@i):@n.
-  comp @vecnew= {@vecnew;{!4(@i,:)(make(1,ncol(@ind),1),:),@v(@ind,:)}}.
-  comp @nextnew= {@nextnew,@ind+1}.
- end if.
-end loop.
-comp @next= @nextnew(2:ncol(@nextnew)).
-comp !4= @vecnew(2:nrow(@vecnew),:).
end loop.
release @v,@maxk,@n,@next,@k,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** ВСЕ СОЧЕТАНИЯ ПО K=2,3... ЭЛЕМЕНТОВ ********.
*/*!KO_allcomb(set%maxk%prefix)*/*.
*Version 1.
*Образует, из множества n элементов, сочетания по k=2, k=3, ..., k=MAXK элементов.
*Выдает все эти сочетания соответственно как матрицы PREFIX2, PREFIX3, и т.д.
*Т.е. эта функция подобна */*!KO_combk*/*, но выдает сразу все сочетания с разными k.
*Множество SET это столбец или матрица с n рядами: элементом множества выступает ряд.
*MAXK - максимальное число элементов в сочетании. Это должно быть целое число (не имя и не выражение)
*от 2 до n. Опционально его можно взять в кавычки или апострофы.
*PREFIX - приставка (опционально в кавычках или апострофах) в имена выдаваемых матриц - наборов сочетаний.
*В выдаваемых матрицах каждый ряд есть сочетание элементов.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Совет: если вы хотите установить MAXK=n, но число n вам заранее не известно, задайте MAXK числом,
*заведомо превышающим возможное n. Матрицы сверх PREFIXn не смогут быть созданы и ошибки не будет.

define !KO_allcomb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!3) !let !pref= !unquote(!4)
comp @v= !2.
comp @n= nrow(@v).
comp @next= 1:@n+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=@n.
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+ncol(@v),0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=@n.
-   comp @ind= @next(@i):@n.
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind,:)}}.
-   comp @nextnew= {@nextnew,@ind+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!pref,!k)= @vec.
-end if.
!doend
release @v,@n,@next,@vec,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** СОЧЕТАНИЯ ПО K ЭЛЕМЕНТОВ ИЗ РАЗНЫХ МНОЖЕСТВ ********.
*/*!KO_dscombk(sets%last%k%name)*/*.
*Version 1.
*Образует сочетания по k элементов, принадлежащих разным множествам.
*Множества SETS это столбец или матрица: элементом множества выступает ряд.
*Множества должны быть собраны в SETS как стопка: сначала идут все элементы (ряды) одного множества,
*под ними все елементы второго множества, и так далее. Множеств может быть любое количество и число
*элементов в них может быть разное.
*LAST - вектор (ряд или столбец) длиной как число множеств в SETS и в соответствие им, содержащий
*индексы последних рядов каждого множества. Например, {3,6,11} означает: в SETS есть три множества,
*с элементами, занимающими ряды 1:3, 4:6, 7:11. 11 (последний элемент последнего множества) должно
*равняться числу рядов в SETS.
*K (целый скаляр от 2 до nc, где nc это число множеств, длина вектора LAST) - число элементов в сочетании.
*Результат NAME - матрица, каждый ряд которой есть сочетание из k элементов от разных множеств, не более
*одного элемента от множества. Если K = число множеств, то сочетание есть "по элементу от всех множеств".
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_dscombk(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @v= !2.
comp @last= !3.
comp @maxk= !4.
comp @n= nrow(@v).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp !5= @v.
loop @k= 2 to @maxk.
-comp @nextnew= 0.
-comp @vecnew= make(1,ncol(!5)+ncol(@v),0).
-loop @i= 1 to nrow(!5).
- do if @next(@i)<=@n.
-  comp @ind= @next(@i):@n.
-  comp @vecnew= {@vecnew;{make(ncol(@ind),1,1)*!5(@i,:),@v(@ind,:)}}.
-  comp @nextnew= {@nextnew,@explast(@ind)+1}.
- end if.
-end loop.
-comp @next= @nextnew(2:ncol(@nextnew)).
-comp !5= @vecnew(2:nrow(@vecnew),:).
end loop.
release @v,@last,@maxk,@n,@next,@k,@plast,@explast,@val,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** ВСЕ СОЧЕТАНИЯ ПО K=2,3... ЭЛЕМЕНТОВ ИЗ РАЗНЫХ МНОЖЕСТВ ********.
*/*!KO_dsallcomb(sets%last%maxk%prefix)*/*.
*Version 1.
*Образует, из нескольких множеств, сочетания по k=2, k=3, ..., k=MAXK элементов, принадлежащих
*разным множествам.
*Выдает все эти сочетания соответственно как матрицы PREFIX2, PREFIX3, и т.д.
*Т.е. эта функция подобна */*!KO_dscombk*/*, но выдает сразу все сочетания с разными k.
*Множества SETS это столбец или матрица: элементом множества выступает ряд.
*Множества должны быть собраны в SETS как стопка: сначала идут все элементы (ряды) одного множества,
*под ними все елементы второго множества, и так далее. Множеств может быть любое количество и число
*элементов в них может быть разное.
*LAST - вектор (ряд или столбец) длиной как число множеств в SETS и в соответствие им, содержащий
*индексы последних рядов каждого множества. Например, {3,6,11} означает: в SETS есть три множества,
*с элементами, занимающими ряды 1:3, 4:6, 7:11. 11 (последний элемент последнего множества) должно
*равняться числу рядов в SETS.
*MAXK - максимальное число элементов в сочетании. Это должно быть целое число (не имя и не выражение)
*от 2 до ns, где ns это число множеств, длина вектора LAST. Опционально число можно взять в кавычки
*или апострофы.
*PREFIX - приставка (опционально в кавычках или апострофах) в имена выдаваемых матриц - наборов сочетаний.
*В выдаваемых матрицах каждый ряд есть сочетание из k элементов от разных множеств, не более
*одного элемента от множества.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Совет: если вы хотите установить MAXK=ns, но число ns вам заранее не известно, задайте MAXK числом,
*заведомо превышающим возможное ns. Матрицы сверх PREFIXns не смогут быть созданы и ошибки не будет.

define !KO_dsallcomb(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!4) !let !pref= !unquote(!5)
comp @v= !2.
comp @last= !3.
comp @n= nrow(@v).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=ncol(@last).
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+ncol(@v),0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=@n.
-   comp @ind= @next(@i):@n.
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind,:)}}.
-   comp @nextnew= {@nextnew,@explast(@ind)+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!pref,!k)= @vec.
-end if.
!doend
release @v,@last,@n,@next,@vec,@plast,@explast,@val,@nextnew,@vecnew,@i,@ind.
!enddefine.


******** ГОРИЗОНТАЛЬНЫЕ АРИФМЕТИЧЕСКИЕ ОПЕРАЦИИ В СОЧЕТАНИЯХ ПЕРЕМЕННЫХ ********.
/*!KO_comboper(vars%last%maxk%return%oper%weights%alpha%prefix%prefix_res)*/*.
*Version 1.
*Столбцы матрицы VARS это числовые переменные (минимум 2 переменные).
*Функция рассматривает все сочетания переменных по K переменных в сочетании; K пробегает
*значения от 2 до MAXK. В каждом сочетании переменных вычисляется величина OPER горизонтально,
*т.е. внутри каждого наблюдения. Это и есть результат.
*Аргумент OPER - арифметическое действие, которое надо совершить в каждом сочетании переменных;
*заглавное кл. слово (опционально можно в кавычках/апострофах):
*"SUM" - сумма, операция сложения переменных: V(1)+V(2)+...+V(K)
*"DIF" - разница, операция вычитания переменных V(1)-V(2)-...-V(K)
*"PROD" - произведение, операция перемножения переменных V(1)*V(2)*...*V(K)
*"QUOT" - частное, операция деления переменных V(1)/V(2)/.../V(K) (нуждается в ненулевых данных)
*"WSUM" - взвешенная сумма: w(1)*V(1)+w(2)*V(2)+...+w(K)*V(K) (веса w - аргумент WEIGHTS)
*"WDIF" - взвешенная разница: w(1)*V(1)-w(2)*V(2)-...-w(K)*V(K) (веса w - аргумент WEIGHTS)
*"EWMA" - конечное (K-е) значение экспоненциально взвешенной скользящей средней: S(K);
* S(K)=(1-a)*S(K-1)+a*V(K), где S(K-1)=(1-a)*S(K-2)+a*V(K-1), где S(K-2)=...; и S(1)=V(1)
* (экспоненциальный вес a - аргумент ALPHA).
*Аргумент WEIGHTS - веса для операций WSUM и WDIF. Либо вектор-ряд длиной с число столбцов VARS,
*либо матрица размером как VARS; второй случай означает, что для разных наблюдений (рядов VARS)
*веса переменным заданы разные. Другие действия - не WSUM и WDIF - не пользуются аргументом:
*задайте любое значение, например 0.
*Аргумент ALPHA - экспоненциальный вес для операции EWMA, значение между 0 и 1. Это либо скаляр,
*либо вектор-столбец длиной с число рядов VARS; второй случай означает, что для разных наблюдений
*альфа-вес задан разный. Другие действия - не EWMA - не пользуются аргументом: задайте любое,
*значение, например 0.
*Аргументы LAST, MAXK, а также PREFIX - те же, что в функции /*!KO_dsallcomb*/, на которой построена
*данная функция.
*Аргумент LAST - параметр для разбивки переменных на множества. Только переменные, принадлежащие
*разным множествам, будут комбинироваться между собой. Это вектор (ряд или столбец) длиной как
*число множеств и в соответствие им, содержащий индексы (номера) замыкающих переменных каждого
*множества. Например, {3,6,11} означает: в VARS есть три множества, группы переменных со столбцами
*1:3, 4:6, 7:11. 11 (последнее значение в LAST) должно равняться числу столбцов в VARS.
*Если у вас нет групп переменных, т.е. каждая переменная есть свое собственное множество, задайте LAST
*как 1:ncol(VARS), т.е. вектор из чисел 1,2,...,число_переменных.
*Аргумент MAXK - максимальное нужное вам значение K, число переменных в сочетании. Это должно быть
*целое число (не имя и не выражение) от 2 до ns, где ns это число множеств, длина вектора LAST.
*Аргумент RETURN - для каких значений K вы хотите получить результаты. Если для всех K от 2 до
*min(MAXK,ns), т.е. все результаты, задайте RETURN как кл. слово ALL (заглавными).
*В противном случае закажите список из требуемых K, целые через пробел, например: 2 3 5.
*Тогда функция вычислит результаты только для этих K.
*Результаты: PREFIX, PREFIX_RES. Укажите на месте этих аргументов 2 приставки
*для имен выдаваемых матриц, которые будут состоять из этих приставок и окончания в виде числа K.
*Для каждого K будет выдана пара массивов:
*- Номера переменных, образующих сочетания - матрица с приставкой PREFIX.
*  Ряды в PREFIX это разные сочетания переменных VARS, элементы в ряду это номера переменных в сочетании.
*- Результаты операции OPER - матрица с приставкой PREFIX_RES. Ряды в ней отвечают наблюдениям данных
*  (рядам VARS), а столбцы отвечает сочетаниям переменных (рядам PREFIX).
*Задания аргументов MAXK, RETURN, PREFIX, PREFIX_RES вы вправе взять в кавычки/апострофы.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Совет: если вы хотите установить MAXK=ns, но число ns вам заранее не известно, задайте MAXK числом,
*заведомо превышающим возможное ns. Матрицы сверх PREFIXns не смогут быть созданы и ошибки не будет.

define !KO_comboper(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                    /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!4) !let !return= !unquote(!5) !let !oper= !unquote(!6) !let !comb= !unquote(!9) !let !a= !unquote(!10)
comp @vars= t(!2).
comp @last= !3.
comp @w= !7.
comp @ew= !8.
comp @v= t(1:nrow(@vars)).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=ncol(@last).
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+1,0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=nrow(@v).
-   comp @ind= @next(@i):nrow(@v).
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind)}}.
-   comp @nextnew= {@nextnew,@explast(@ind)+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!comb,!k)= @vec.
-end if.
!doend
release @v,@next,@vec,@plast,@explast,@val,@nextnew,@vecnew,@ind.
!if (!return='ALL') !then !let !ret='ALL' !else !let !ret= !conc(' ',!return,' ') !ifend
!if (!oper='SUM' !or !oper='WSUM') !then !let !op= '+'
!else !if (!oper='DIF' !or !oper='WDIF') !then !let !op= '-'
!else !if (!oper='PROD') !then !let !op= '&*'
!else !if (!oper='QUOT') !then !let !op= '/'
!ifend !ifend !ifend !ifend
!if (!oper='WSUM' !or !oper='WDIF') !then !let !wght= 1
-do if nrow(@w)=1.
- comp @w= t(@w)*make(1,ncol(@vars),1).
-else.
- comp @w= t(@w).
-end if.
!else !if (!oper='EWMA') !then !let !wght= 2
-do if nrow(@ew)*ncol(@ew)=1.
- comp @iscol= 0.
- comp @ewk= @ew.
-else.
- comp @iscol= 1.
-end if.
!ifend !ifend
!do !k= 2 !to !maxk !let !kk= !conc(!comb,!k) !let !intk= !conc(!a,!k)
-do if !k<=ncol(@last).
  !if (!ret='ALL' !or !index(!ret,!conc(' ',!k,' '))>0) !then
-  comp !intk= @vars(!kk(:,1),:).
   !if (!wght=1) !then
-   comp !intk= @w(!kk(:,1),:)&*!intk.
   !else !if (!wght=2) !then
-   do if @iscol.
-    comp @ewk= @ew(make(1,nrow(!kk),1),:).
-   end if.
   !ifend !ifend
-  loop @i= 2 to !k.
    !if (!wght=1) !then
-    comp !intk= !intk !op @w(!kk(:,@i),:)&*@vars(!kk(:,@i),:).
    !else !if (!wght=2) !then
-    comp !intk= (1-@ewk)&*!intk+@ewk&*@vars(!kk(:,@i),:).
-   !else
-    comp !intk= !intk !op @vars(!kk(:,@i),:).
    !ifend !ifend
-  end loop.
-  comp !intk= t(!intk).
  !else
-  release !kk.
  !ifend
-end if.
!doend
release @vars,@last,@i,@w,@ew !if (!wght=2) !then ,@ewk !ifend.
!enddefine.


******** КОЛИЧЕСТВА НА БАЗЕ ГОРИЗОНТАЛЬНЫХ СУММ ИЛИ СРЕДНИХ В СОЧЕТАНИЯХ ПЕРЕМЕННЫХ ********.
*/*!KO_turflike(vars%last%maxk%return%tosum%tocnt%cond%prefix%prefix_sum%prefix_cnt)*/*.
*Version 1.
*Столбцы матрицы VARS это числовые переменные (минимум 2 переменные).
*Функция рассматривает все сочетания переменных по K переменных в сочетании; K пробегает
*значения от 2 до MAXK. В каждом сочетании переменных вычисляются статистики hs (см. ниже)
*горизонтально, т.е. внутри каждого наблюдения. Затем функция (1) суммирует hs сквозь
*наблюдения, выдавая результат PREFIX_SUM; (2) подсчитывает наблюдения, где hs удовлетворяет
*условию COND, выдавая результат PREFIX_CNT. Таким образом, для каждого сочетания переменных,
*столбцов из VARS, вы получите значение "column_sum(hs)" и "column_count(hs COND = True)".
*Аргументы TOSUM и TOCNT. Здесь надо указать статистику hs в виде кл. слова заглавными буквами.
*Можно заказать одну или разные статистики в этих двух аргументах. Выбор из следующих
*статистик, вычислить внутри каждого наблюдения (ряда VARS):
*"SUM" - сумма значений данных
*"MEAN" - средняя значений данных
*"SUMVLD" - сумма валидных значений данных
*"MEANVLD" - средняя валидных значений данных
*"NVLD" - количество валидных значений данных
*В трех последних статистиках за невалидное значение принято значение: -999.
*Если валидных данных в наблюдении нет, MEANVLD наблюдения оценивается как 0.
*Аргумент COND - условие для подсчета числа наблюдений, в которых TOCNT-статистика удовлетворит ему.
*Состоит из одного оператора и одного операнда. Операторами могут быть: GT (или >), LT (или <),
*EQ (или =), GE, LE, NE. (>=, <=, <> использовать нельзя в данной функции.) Операндом может быть
*число или имя некоторой матрицы или выражение; перед операндом может стоять отрицательный знак.
*Примеры задания COND: GT 0, EQ -4.12, LE v3, LT x(:,1).
*Операнд с именем. Именем может быть скаляр (число), вектор-столбец длиной =nrow(VARS),
*вектор-ряд длиной >=min(MAXK,ns) [см. ниже об ns] либо матрица размером [nrow(VARS),>=min(MAXK,ns)].
*Вектор-столбец позволяет вам задать разные значения операнда для разных наблюдений данных (рядов VARS).
*Вектор-ряд позволяет вам задать разные значения операнда для разных K (числа переменных в сочетании);
*K-й элемент в векторе-ряде будет отвечать данному K. Операнд в виде матрицы указанного выше размера
*позволяет вам задать разные значения операнда и для разных наблюдений, и для разных K.
*Аргументы LAST, MAXK, а также PREFIX - те же, что в функции /*!KO_dsallcomb*/, на которой построена
*данная функция.
*Аргумент LAST - параметр для разбивки переменных на множества. Только переменные, принадлежащие
*разным множествам, будут комбинироваться между собой. Это вектор (ряд или столбец) длиной как
*число множеств и в соответствие им, содержащий индексы (номера) замыкающих переменных каждого
*множества. Например, {3,6,11} означает: в VARS есть три множества, группы переменных со столбцами
*1:3, 4:6, 7:11. 11 (последнее значение в LAST) должно равняться числу столбцов в VARS.
*Если у вас нет групп переменных, т.е. каждая переменная есть свое собственное множество, задайте LAST
*как 1:ncol(VARS), т.е. вектор из чисел 1,2,...,число_переменных.
*Аргумент MAXK - максимальное нужное вам значение K, число переменных в сочетании. Это должно быть
*целое число (не имя и не выражение) от 2 до ns, где ns это число множеств, длина вектора LAST.
*Аргумент RETURN - для каких значений K вы хотите получить результаты. Если для всех K от 2 до
*min(MAXK,ns), т.е. все результаты, задайте RETURN как кл. слово ALL (заглавными).
*В противном случае закажите список из требуемых K, целые через пробел, например: 2 3 5.
*Тогда функция вычислит результаты только для этих K.
*Результаты: PREFIX, PREFIX_SUM, PREFIX_CNT. Укажите на месте этих аргументов 3 приставки
*для имен выдаваемых матриц, которые будут состоять из этих приставок и окончания в виде числа K.
*Для каждого K будет выдана серия из 3 массивов:
*- Номера переменных, образующих сочетания - матрица с приставкой PREFIX.
*  Ряды в PREFIX это разные сочетания переменных VARS, элементы в ряду это номера переменных в сочетании.
*- Результаты суммирования TOSUM - вектор-столбец с приставкой PREFIX_SUM. Ряды в нем отвечают рядам
*  в PREFIX.
*- Результаты подсчета TOCNT - вектор-столбец с приставкой PREFIX_CNT. Ряды в нем отвечают рядам
*  в PREFIX.
*Если вы не желаете получать результат PREFIX_SUM или PREFIX_CNT, поствьте цифру 0 на место этого
*аргумента.
*Задания аргументов TOSUM, TOCNT, COND, MAXK, RETURN, PREFIX, PREFIX_SUM, PREFIX_CNT вы вправе
*взять в кавычки/апострофы.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Совет: если вы хотите установить MAXK=ns, но число ns вам заранее не известно, задайте MAXK числом,
*заведомо превышающим возможное ns. Матрицы сверх PREFIXns не смогут быть созданы и ошибки не будет.
*Частный случай: если VARS - двоичные переменные (0 vs 1), TOSUM=TOCNT="SUM" и условие COND="GT 0",
*то функция фактически выдает величины, вычисляемые в так. наз. TURF-анализе. PREFIX_SUM даст частоты
*"Объема" ("Frequency"), а PREFIX_CNT даст частоты "Охвата" ("Reach").

define !KO_turflike(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                    /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
!let !maxk= !unquote(!4) !let !return= !unquote(!5) !let !tosum= !unquote(!6) !let !tocnt= !unquote(!7) !let !cond= !unquote(!8)
!let !comb= !unquote(!9) !let !a= !unquote(!10) !let !b= !unquote(!11)
comp @vars= t(!2).
comp @last= !3.
comp @v= t(1:nrow(@vars)).
do if nrow(@last)>ncol(@last).
-comp @last= t(@last).
end if.
comp @plast= {0,@last}.
comp @explast= 0.
loop @i= 1 to ncol(@last).
-comp @val= @last(@i).
-comp @explast= {@explast,make(1,@val-@plast(@i),@val)}.
end loop.
comp @explast= @explast(2:ncol(@explast)).
comp @next= @explast+1.
comp @vec= @v.
!do !k= 2 !to !maxk
-do if !k<=ncol(@last).
- comp @nextnew= 0.
- comp @vecnew= make(1,ncol(@vec)+1,0).
- loop @i= 1 to nrow(@vec).
-  do if @next(@i)<=nrow(@v).
-   comp @ind= @next(@i):nrow(@v).
-   comp @vecnew= {@vecnew;{@vec(@i,:)(make(1,ncol(@ind),1),:),@v(@ind)}}.
-   comp @nextnew= {@nextnew,@explast(@ind)+1}.
-  end if.
- end loop.
- comp @next= @nextnew(2:ncol(@nextnew)).
- comp @vec= @vecnew(2:nrow(@vecnew),:).
- comp !conc(!comb,!k)= @vec.
-end if.
!doend
release @v,@next,@vec,@plast,@explast,@val,@nextnew,@vecnew,@ind.
!if (!return='ALL') !then !let !ret='ALL' !else !let !ret= !conc(' ',!return,' ') !ifend
!if (!b<>0) !then !let !op= !head(!cond)
-comp @x= t(!tail(!cond)).
!ifend
!let !sum= 0 !let !sumvld= 0 !let !nvld= 0
!if (!tosum='SUM' !or !tosum='MEAN' !or !tocnt='SUM' !or !tocnt='MEAN') !then !let !sum= 1 !ifend
!if (!tosum='SUMVLD' !or !tosum='MEANVLD' !or !tocnt='SUMVLD' !or !tocnt='MEANVLD') !then !let !sumvld= 1 !ifend
!if (!tosum='NVLD' !or !tosum='MEANVLD' !or !tocnt='NVLD' !or !tocnt='MEANVLD') !then !let !nvld= 1 !ifend
!if (!nvld !or !sumvld) !then
-comp @flagvld= @vars<>-999.
 !if (!sumvld) !then
- comp @varsvld= @vars&*@flagvld.
 !ifend
!ifend
!do !k= 2 !to !maxk !let !kk= !conc(!comb,!k)
-do if !k<=ncol(@last).
  !if (!ret='ALL' !or !index(!ret,!conc(' ',!k,' '))>0) !then
   !if (!sum) !then
-   comp @wsk= @vars(!kk(:,1),:).
   !ifend
   !if (!sumvld) !then
-   comp @wsvldk= @varsvld(!kk(:,1),:).
   !ifend
   !if (!nvld) !then
-   comp @wnvldk= @flagvld(!kk(:,1),:).
   !ifend
-  loop @i= 2 to !k.
    !if (!sum) !then
-    comp @wsk= @wsk+@vars(!kk(:,@i),:).
    !ifend
    !if (!sumvld) !then
-    comp @wsvldk= @wsvldk+@varsvld(!kk(:,@i),:).
    !ifend
    !if (!nvld) !then
-    comp @wnvldk= @wnvldk+@flagvld(!kk(:,@i),:).
    !ifend
-  end loop.
   !if (!b<>0) !then
-   do if nrow(@x)>1 and ncol(@x)>1.
-    comp @x_= @x(!k,:)(make(1,nrow(!kk),1),:).
-   else if ncol(@x)>1.
-    comp @x_= @x(make(1,nrow(!kk),1),:).
-   else if nrow(@x)>1.
-    comp @x_= @x(!k).
-   else.
-    comp @x_= @x.
-   end if.
   !ifend
   !if (!a<>0) !then
    !if (!tosum='SUM') !then
-    comp !conc(!a,!k)= rsum(@wsk).
    !else !if (!tosum='MEAN') !then
-    comp !conc(!a,!k)= rsum(@wsk/!k).
    !else !if (!tosum='SUMVLD') !then
-    comp !conc(!a,!k)= rsum(@wsvldk).
    !else !if (!tosum='MEANVLD') !then
-    comp !conc(!a,!k)= rsum(@wsvldk/(@wnvldk+(@wnvldk=0))).
    !else
-    comp !conc(!a,!k)= rsum(@wnvldk).
    !ifend !ifend !ifend !ifend
   !ifend
   !if (!b<>0) !then
    !if (!tosum='SUM') !then
-    comp !conc(!b,!k)= rsum(@wsk !op @x_).
    !else !if (!tosum='MEAN') !then
-    comp !conc(!b,!k)= rsum(@wsk/!k !op @x_).
    !else !if (!tosum='SUMVLD') !then
-    comp !conc(!b,!k)= rsum(@wsvldk !op @x_).
    !else !if (!tosum='MEANVLD') !then
-    comp !conc(!b,!k)= rsum(@wsvldk/(@wnvldk+(@wnvldk=0)) !op @x_).
    !else
-    comp !conc(!b,!k)= rsum(@wnvldk !op @x_).
    !ifend !ifend !ifend !ifend
   !ifend
  !else
-  release !kk.
  !ifend
-end if.
!doend
release @vars,@last,@i !if (!nvld !or !sumvld) !then ,@flagvld !ifend !if (!sumvld) !then ,@varsvld !ifend
 !if (!sum) !then ,@wsk !ifend !if (!sumvld) !then ,@wsvldk !ifend !if (!nvld) !then ,@wnvldk !ifend
 !if (!b<>0) !then ,@x,@x_ !ifend.
!enddefine.


******** СПАРИВАНИЕ ВЕНГЕРСКИМ АЛГОРИТМОМ ********.
*/*!KO_hungar(cost%name)*/*.
*Version 1.
*Венгерский или Kuhn-Munkres алгоритм спаривает взаимно-однозначно элементы между двумя множествами
*оптимально так, чтобы суммарная стоимость пар (или суммарное различие в парах) оказалась минимальна.
*В терминологии графов, он создает оптимальное паросочетание (optimal matching) из двудольного
*взвешенного или невзвешенного графа. (Отметим, что в случае невзвешенного графа вы можете использовать
*для решения этой задачи и алгоритм Хопкрофта-Карпа /*!KO_hopckarp*/.)
*Функция принимает в качестве входящей n x m матрицу стоимостей COST с неотрицательными элементами.
*Число рядов должно быть не меньше числа столбцов. Ряды - это одно множество элементов, столбцы - другое
*множество элементов. Значения матрицы понимаются как "стоимость" пары ряд-столбец или как "различие"
*между рядом и столбцом. Алгоритм находит каждому столбцу в пару один ряд.
*Результат NAME - трехстолбцовая матрица с числом рядов как число столбцов COST; ряды в ней это
*созданные пары. 2-й столбец ее это номер столбца входящей матрицы COST, 1-й столбец это номера ряда
*COST, а 3-й столбец это значение COST на их пересечении. Алгоритм минимизирует сумму в 3-м столбце.
*Если ваша входящая матрица являет собой не стоимости (различия), а выигрыши (сходства), переведите
*их в стоимости (различия), обратив знак и затем прибавив константу, чтобы сделать значения
*неотрицательными.
*Скорость алгоритма выше, когда число различных значений в COST невелико, поэтому имеет смысл подумать
*о предварительной дискретизации значений "стоимости".
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_hungar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @c= !2.
comp @c_= @c.
do if nrow(@c)>ncol(@c).
-comp @c= {@c,make(nrow(@c),nrow(@c)-ncol(@c),0)}.
end if.
comp @n= nrow(@c).
comp @rzero= make(1,@n,0).
comp @czero= make(@n,1,0).
comp @rnums= t(1:@n).
comp @cnums= 1:@n.
comp @c= @c-rmin(@c)*make(1,@n,1).
comp @m= (not @c)&*@cnums(make(1,@n,1),:).
loop @i= 1 to @n.
-comp @j= rmax(@m(@i,:)).
-do if @j.
- comp @m(@i,:)= @rzero.
- comp @m(:,@j)= @czero.
- comp @m(@i,@j)= 1.
-end if.
end loop.
loop.
-comp @ucr= make(@n,1,1).
-comp @ucc= not csum(@m).
-do if not any(@ucc).
- break.
-else.
- loop.
-  comp @i= 0.
-  loop @j= 1 to @n.
-   do if @ucc(@j).
-    comp @i= cmax((not @c(:,@j))&*@ucr&*@rnums).
-    do if @i.
-     break.
-    end if.
-   end if.
-  end loop.
-  do if @i.
-   comp @m(@i,@j)= -2.
-   do if any(@m(@i,:)=1).
-    comp @ucr(@i)= 0.
-    comp @ucc(rmax(@m(@i,:)&*@cnums))= 1.
-   else.
-    comp @m(@i,@j)= -3.
-    loop if any(@m(:,@j)=1).
-     comp @i2= cmax(@m(:,@j)&*@rnums).
-     comp @m(@i2,@j)= 0.
-     comp @j2= rmax((@m(@i2,:)=-2)&*@cnums).
-     comp @m(@i2,@j2)= -3.
-     comp @j= @j2.
-    end loop.
-    comp @m= @m&*(@m<>-3)+(@m=-3).
-    comp @m= @m&*(@m<>-2).
-    break.
-   end if.
-  else.
    !do !w !in ("'@ucr @rind' '@ucc @cind'")
     !let !w0= !unquote(!w) !let !w1= !head(!w0) !let !w2= !tail(!w0)
-    comp @nnz= msum(!w1).
-    do if @nnz=@n.
-     comp !w2= @cnums.
-    else if @nnz/@n>.8.
-     comp !w2= make(1,@nnz+1,0).
-     comp @cnt= 1.
-     loop @#= 1 to @n.
-      comp !w2(@cnt)= @#.
-      comp @cnt= @cnt+!w1(@#).
-     end loop.
-     comp !w2= !w2(1:@nnz).
-    else.
-     comp !w2= make(1,@nnz,0).
-     comp @cnt= 0.
-     loop @#= 1 to @n.
-      do if !w1(@#).
-       comp @cnt= @cnt+1.
-       comp !w2(@cnt)= @#.
-      end if.
-     end loop.
-    end if.
    !doend
-   comp @min= mmin(@c(@rind,@cind)).
-   comp @c= @c+(not @ucr)*make(1,@n,@min)-make(@n,1,@min)*@ucc.
-   release @nnz,@rind,@cind,@cnt,@#.
-  end if.
- end loop.
-end if.
end loop.
comp !3= make(ncol(@c_),3,0).
loop @j= 1 to ncol(@c_).
-comp @i= cmax(@m(:,@j)&*@rnums).
-comp !3(@j,:)= {@i,@j,@c_(@i,@j)}.
end loop.
release @c,@c_,@m,@n,@rzero,@czero,@rnums,@cnums,@ucr,@ucc,@i,@j.
!enddefine.


******** СПАРИВАНИЕ АЛГОРИТМОМ ХОПКРОФТА-КАРПА ********.
*/*!KO_hopckarp(adj%name)*/*.
*Version 1.
*Алгоритм Хопкрофта-Карпа (Hopcroft-Karp) так спаривает взаимно-однозначно элементы между двумя
*множествами, имеющими между собой связи, чтобы максимизировать число пар, т.е. число учтенных связей.
*В терминологии графов, он создает максимальное паросочетание (maximal matching) из двудольного
*невзвешенного графа. (Отметим, что вы можете использовать для решения этой задачи и венгерский
*алгоритм /*!KO_hungar*/.)
*Входящая матрица ADJ - это "список смежности" (adjacency list) размером n x m, где n и m - количества
*элементов в двух множествах. Каждый ряд ADJ это элемент из множества "n", и он содержит номера
*элементов множества "m", идущие слева направо утрамбованно. Например, пусть есть множества "n"
*(1,2,...,n) и "m" (1,2,...,m=7), где, в частности, 2-й элемент из "n" имеет связи с элементами 1,4 и 5
*множества "m". Тогда 2-й ряд ADJ должен выглядеть так: {1,4,5,0,0,0,0}. Порядок номеров не играет роли:
*{4,5,1,0,0,0,0}, например, тоже корректно. Число столбцов ADJ должно быть m, поэтому пустоты добиты
*нулями до 7. Список смежности соответствует двоичной матрице размером n x m, в которой единицы
*обозначают наличие связи, и является другим способом ее представить. Если ваши входящие - двоичная
*матрица, вы можете получить из нее список смежности функцией /*!KO_ram*/.
*Пустые (полностью из нулей) ряды в ADJ допускаются - они соответствуют изолированным, не имеющим связей,
*элементам из "n", - но вы можете захотеть сначала удалить их в интересах быстродействия.
*В ADJ должен быть хотя бы один ненулевой элемент.
*Если n и m сильно различаются, выгоднее с точки зрения быстродействия иметь ADJ широкую (n<m), нежели
*высокую (n>m).
*Результат NAME - двухстолбцовая матрица, являющая максимальное паросочетание; 1-й столбец это элементы
*множества "n", 2-й столбец - элементы множества "m"; каждый ряд это осуществленное спаривание.
*Число рядов в NAME есть, т.о., величина паросочетания.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_hopckarp(!pos= !token(1) /!pos= !charend('%') /!pos= !charend(')'))
comp @adj= !2.
comp @n= nrow(@adj).
comp @m= ncol(@adj).
comp @adj= @adj+@n*(@adj<>0).
comp @adj2= make(@m,@n,0).
comp @adj2= {make(@m,1,@n+@m+2),@adj2}.
do if ncol(@adj2)<@m.
-comp @adj= {@adj;{@adj2,make(@m,@m-ncol(@adj2),0)}}.
else if ncol(@adj2)>@m.
-comp @adj= {{@adj,make(@n,ncol(@adj2)-@m,0)};@adj2}.
-else.
-comp @adj= {@adj;@adj2}.
end if.
release @adj2.
comp @adj= {@adj;make(2,ncol(@adj),0)}.
comp @src= nrow(@adj)-1.
comp @snk= nrow(@adj).
comp @j= 0.
loop @i= 1 to @n.
-do if @adj(@i,1).
- comp @j= @j+1.
- comp @adj(@src,@j)= @i.
-end if.
end loop.
comp @nadj= rsum(@adj<>0).
comp @nums= 1:ncol(@adj).
comp !3= {-1,-1}.
comp @ginm= make(@n,1,0).
loop.
-comp @g= @adj.
-comp @ng= @nadj.
-do if nrow(!3)>1.
- comp @jj= {0;(@g(!3(2:nrow(!3),1),:)=(!3(2:nrow(!3),2)*make(1,ncol(@adj),1)))*t(@nums)}.
- comp @g(!3(2:nrow(!3),2),1)= !3(2:nrow(!3),1).
- loop @i= 2 to nrow(!3).
-  comp @grl= !3(@i,1).
-  comp @g(@grl,@jj(@i))= @g(@grl,@ng(@grl)).
-  comp @ng(@grl)= @ng(@grl)-1.
-  comp @g(@src,rmax((@g(@src,:)=@grl)&*@nums))= @g(@src,@ng(@src)).
-  comp @g(@src,@ng(@src))= 0.
-  comp @ng(@src)= @ng(@src)-1.
- end loop.
-end if.
-comp @b= make(1,@snk,1).
-comp @s= {@src,0}.
-comp @done= 1.
-loop if ncol(@s)>1.
- loop if @ng(@s(1)).
-  comp @top= @s(1).
-  comp @fst= @g(@top,@ng(@top)).
-  comp @ng(@top)= @ng(@top)-1.
-  do if @fst=@snk.
-   loop if ncol(@s)>2.
-    do if @ginm(@s(2)).
-     comp !3(@ginm(@s(2)),2)= @s(1).
-    else.
-     comp !3= {!3;{@s(2),@s(1)}}.
-     comp @ginm(@s(2))= nrow(!3).
-    end if.
-    comp @s= @s(3:ncol(@s)).
-   end loop.
-   comp @done= 0.
-  end if.
-  do if @b(@fst).
-   comp @s= {@fst,@s}.
-   comp @b(@fst)= 0.
-  end if.
- end loop.
- comp @s= @s(2:ncol(@s)).
-end loop.
end loop if @done.
comp !3= !3(2:nrow(!3),:).
comp !3(:,2)= !3(:,2)-@n.
release @adj,@n,@m,@src,@snk,@j,@nadj,@nums,@ginm,@done,@g,@ng,@jj,@i,@grl,@b,@s,@top,@fst.
!enddefine.


******** ПРОСТОЕ ЖАДНОЕ СПАРИВАНИЕ ********.
*/*!KO_greedm(mat%cg%name)*/*.
*Version 1.
*Простое жадное спаривание это пошаговый алгоритм, спаривающий элементы между двумя множествами так,
*чтобы минимизировать (или максимизировать) сумму значений в парах. Это та же задача, что выполняет
*венгерское спаривание (ф-ция /*!KO_hungar*/), но /*!KO_greedm*/ быстрее и не гарантирует глобального
*оптимума в смысле итоговой суммы. Алгоритм находит наименьший (или наибольший - как затребовано)
*элемент в матрице и записывает его ряд и столбец как пару (match), затем вырезает их из матрицы и
*в оставшейся матрице опять находит наименьший элемент, спаривает столбец и ряд, удаляет их,
*и так далее.
*Входящие:
*MAT - матрица размера не меньше чем 2x2 с любыми числами. Если число столбцов намного превышает число
*рядов, возможно имеет смысл транспонировать ее сначала, это может дать выигрыш в скорости.
*CG - цифра (не имя и не выражение) 0, 1 или 2 (цифру можно опционально взять в кавычки или апострофы).
*Если "0", то функция будет искать минимальные значения, т.е. относиться к MAT как к стоимости (cost),
*которую надо минимизировать в спаривании. Если "1", то функция будет искать максимальные значения,
*т.е. относиться к MAT как к выигрышу (gain), который надо максимизировать в спаривании.
*Если "2", то MAT это разреженная (sparse) неотрицательная gain-матрица: с положительными значениями
*в качестве выигрышей и нулевыми элементами, запрещающими спаривание, хотя бы одно значение должно быть
*больше 0.
*Результат NAME: трехстолбцовая матрица, где ряды - это пары ряд-столбец, выбранные алгоритмом:
*номер ряда входящей MAT записан в 1-м столбце NAME, номер спаренного с ним столбца входящей MAT записан
*во 2-м столбце NAME, а значение, лежащее на их пересечении в MAT - в 3-м столбце NAME. Сумму в этом 3-м
*столбце ф-ция и пытается минимизировать (максимизировать).
*В случае CG="2" некоторые последние ряды в NAME могут быть пусты, поскольку разреженная MAT может
*не дать спарить все ряды и столбцы.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_greedm(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
!let !cg= !unquote(!3)
comp !4= make(rmin({nrow(@mat),ncol(@mat)}),3,0).
comp @rns= 1:nrow(@mat).
comp @cns= 1:ncol(@mat).
loop @i= 1 to nrow(!4).
-comp @val= !if (!cg=0) !then mmin !else mmax !ifend (@mat).
 !if (!cg=2) !then
- do if not @val.
-  break.
- end if.
 !ifend
-comp @flag= @mat=@val.
-comp @ri= mmax(@flag&*(t((1:nrow(@mat)))*make(1,ncol(@mat),1))).
-comp @ci= mmax(@flag(@ri,:)&*(1:ncol(@mat))).
-comp !4(@i,:)= {@rns(@ri),@cns(@ci),@val}.
-do if @i<nrow(!4).
- do if @ri=1.
-  comp @rind= 2:nrow(@mat).
- else if @ri=nrow(@mat).
-  comp @rind= 1:(@ri-1).
- else.
-  comp @rind= {1:(@ri-1),(@ri+1):nrow(@mat)}.
- end if.
- do if @ci=1.
-  comp @cind= 2:ncol(@mat).
- else if @ci=ncol(@mat).
-  comp @cind= 1:(@ci-1).
- else.
-  comp @cind= {1:(@ci-1),(@ci+1):ncol(@mat)}.
- end if.
- comp @mat= @mat(@rind,@cind).
- comp @rns= @rns(@rind).
- comp @cns= @cns(@cind).
-end if.
end loop.
release @mat,@rns,@cns,@i,@val,@flag,@ri,@ci,@rind,@cind.
!enddefine.


******** ПРОСТОЕ ЖАДНОЕ СПАРИВАНИЕ (ДИСКРЕТНЫЕ ДАННЫЕ) ********.
*/*!KO_greedm2(mat%cg%name)*/*.
*Version 1.
*Эта функция дает тот же результат, что /*!KO_greedm*/, но она быстрее ее в ситуации, когда число
*различных значений в MAT невелико по сравнению с размером MAT. Например, когда MAT это целые
*числа из существенно ограниченного диапазона. И напротив, если различных значений в MAT много, как,
*например, при континуальных данных, то /*!KO_greedm2*/ будет медленнее, чем /*!KO_greedm*/.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_greedm2(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
!let !cg= !unquote(!3)
comp !4= make(rmin({nrow(@mat),ncol(@mat)}),3,0).
comp @rns= 1:nrow(@mat).
comp @cns= 1:ncol(@mat).
comp @cnt= 0.
comp @cind= 0.
loop.
-comp @val= !if (!cg=0) !then mmin !else mmax !ifend (@mat).
 !if (!cg=2) !then
- do if not @val.
-  break.
- end if.
 !ifend
-comp @rind= 0.
-comp @cnums= 1:ncol(@mat).
-loop @ri= nrow(@mat) to 1 by -1.
- comp @ci= rmax((@mat(@ri,:)=@val)&*@cnums).
- do if @ci.
-  comp @cnt= @cnt+1. 
-  comp !4(@cnt,:)= {@rns(@ri),@cns(@ci),@val}.
-  comp @cnums(@ci)= 0.
- else.
-  comp @rind= {@ri,@rind}.
- end if.
-end loop.
-do if @cnt=nrow(!4).
- break.
-end if.
-comp @rind= @rind(1:(ncol(@rind)-1)).
-comp @vec= @cnums<>0.
-comp @n= ncol(@vec).
-comp @nnz= msum(@vec).
-do if @nnz/@n>.8.
- comp @cind= make(1,@nnz+1,0).
- comp @c= 1.
- loop @i= 1 to @n.
-  comp @cind(@c)= @i.
-  comp @c= @c+@vec(@i).
- end loop.
- comp @cind= @cind(1:@nnz).
- release @c,@i.
-else.
- comp @cind= make(1,@nnz,0).
- comp @c= 0.
- loop @i= 1 to @n.
-  do if @vec(@i).
-   comp @c= @c+1.
-   comp @cind(@c)= @i.
-  end if.
- end loop.
- release @c,@i.
-end if.
-release @vec,@n,@nnz.
-comp @mat= @mat(@rind,@cind).
-comp @rns= @rns(@rind).
-comp @cns= @cns(@cind).
end loop.
release @mat,@rns,@cns,@cnt,@val,@cnums,@ri,@ci,@rind,@cind.
!enddefine.


******** АЛГОРИТМ ФЛОЙДА-УОРШАЛЛА: КРАТЧАЙШИЕ ПУТИ / ЛЕГЧАЙШИЕ ПРОХОДЫ ********.
*/*!KO_flowar(cost%task%path%name1%name2)*/.
*Version 1.
*Берет квадратную матрицу COST с неотрицательными элементами и нулевой диагональю;
*матрица может быть симметричной или асимметричной. Значения COST можно понимать как веса в графе
*или связи в сети, и ряды/столбцы матрицы - это вершины графа или узлы в сети.
*TASK (ключевое слово заглавными буквами, можно взять в кавычки или апострофы) - цель алгоритма:
*"SHWAY" - найти, из каждой точки в каждую точку, кратчайший путь (Shortest path aka Shortest
*way-модификация алгоритма). Значения COST принимаются за расстояния между точками:
*COST(i,j) это расстояние из i в j.
*Из каждой точки i в каждую точку j будет найден самый короткий путь. Результат NAME1 будет
*содержать суммарную длину этого кратчайшего пути из всякой точки-ряда i во всякую точку-столбец j.
*"EPASS" - найти, из каждой точки в каждую точку, путь из легчайших перегонов (Optimal routing
*aka Easiest passes-модификация алгоритма). Значения COST принимаются за трудность преодоления
*перегона из точки в точку: COST(i,j) это трудность из i в j.
*Из каждой точки i в каждую точку j будет найден самый легкий путь:
*все перегоны, из которых он будет состоять, минимально трудны. Результат NAME1 будет
*содержать трудность самого трудного перегона на этом легчайшем пути из всякой точки-ряда i
*во всякую точку-столбец j.
*Сами найденные (т.е. оптимальные) пути, траектории, будут записаны в "свернутом виде" в матрицу NAME2.
*Элемент (i,j) в NAME2 это номер точки, следующей за точкой i на всяком найденном пути, ведущем
*к точке j как к конечной точке пути. См. ПРИМЕР, показывающий, как извлечь всю траекторию пути
*из некоторой начальной точки U в некоторую конечную точку V=j.
*Записывать траекторию или нет - зависит от аргумента PATH. Аргумент PATH - это цифра 1 или 0
*(не имя и не выражение, цифру можно взять в кавычки или апострофы). Если 0, то траектории не будут
*записаны и матрица NAME2 выйдет как нулевой скаляр.
*Отсутствующие значения в COST и невозможный путь. Если во входящей матрице должно не быть связи
*между некоторыми точками - например, перегона из точки i в точку j не существует, то поставьте в
*ячейку (i,j) число X, не меньшее чем сумма всех существующих в матрице значений, т.е. X>=msum(COST).
*Тогда, если в некоторой ячейке (U,V) NAME1 вы увидите число X, то значит, путь из начальной точки U
*в конечную точку V - невозможен.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_flowar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                 /!pos= !charend('%') /!pos= !charend(')'))
comp !5= !2.
comp @n= ncol(!5).
!let !task= !unquote(!3)
!if (!task='EPASS') !then !let !upd= !conc('rmax({@dik,',!5,'(@k,@j)})')
!else !let !upd= !conc('@dik+',!5,'(@k,@j)')
!ifend
!let !path= !unquote(!4)
!if (!path) !then
-comp !6= (1:@n)(make(1,@n,1),:).
!else
-comp !6= 0.
!ifend
loop @k= 1 to @n.
-loop @i= 1 to @n.
- do if @i<>@k.
-  comp @dik= !5(@i,@k).
-  loop @j= 1 to @n.
-   do if !5(@i,@j)>!upd.
-    comp !5(@i,@j)= !upd.
     !if (!path) !then
-     comp !6(@i,@j)= !6(@i,@k).
     !ifend
-   end if.
-  end loop.
- end if.
-end loop.
end loop.
release @n,@k,@i,@j,@dik.
!enddefine.


******** АЛГОРИТМ ФЛОЙДА-УОРШАЛЛА (СИММЕТРИЧНАЯ МАТРИЦА) ********.
*/*!KO_sflowar(cost%task%name)*/.
*Version 1.
*Эта функция тождественна /*!KO_flowar*/, но рассчитана только на симметричную входящую
*матрицу COST. Эта функция несколько быстрее, чем /*!KO_flowar*/.
*Аргумент PATH и результат NAME2 - не предусмотрены.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.

define !KO_sflowar(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp !4= !2.
comp @n= ncol(!4).
!let !task= !unquote(!3)
!if (!task='EPASS') !then !let !upd= !conc('rmax({@dik,',!4,'(@k,@j)})')
!else !let !upd= !conc('@dik+',!4,'(@k,@j)')
!ifend
loop @k= 1 to @n.
-loop @i= 1 to @n.
- do if @i<>@k.
-  comp @dik= !4(@i,@k).
-  loop @j= @i+1 to @n.
-   do if !4(@i,@j)>!upd.
-    comp !4(@i,@j)= !upd.
-   end if.
-  end loop.
-  comp !4(:,@i)= t(!4(@i,:)).
- end if.
-end loop.
end loop.
release @n,@k,@i,@j,@dik.
!enddefine.


******** АЛГОРИТМ ДЕЙКСТРЫ: КРАТЧАЙШИЙ ПУТЬ / ЛЕГЧАЙШИЙ ПРОХОД ********.
*/*!KO_dijkstra(cost%start%stop%task%path%name1%name2)*/.
*Version 1.
*Берет квадратную матрицу COST с неотрицательными элементами; матрица может быть симметричной или
*асимметричной. Значения COST можно понимать как веса в графе или связи в сети, и ряды/столбцы
*матрицы - это вершины графа или узлы в сети. Диагональные элементы COST игнорируются функцией
*(они принимаются за 0).
*START - скаляр; номер точки (вершины), из которой находить пути в другие точки.
*TASK (ключевое слово заглавными буквами, можно взять в кавычки или апострофы) - цель алгоритма:
*"SHWAY" - найти из точки START в каждую точку кратчайший путь (Shortest path aka Shortest
*way-модификация алгоритма). Значения COST принимаются за расстояния между точками:
*COST(i,j) это расстояние из i в j.
*Из точки START в каждую точку j будет найден самый короткий путь. Результат NAME1 будет
*содержать суммарную длину этого кратчайшего пути из точки-ряда START во всякую точку-столбец j.
*"EPASS" - найти из точки START в каждую точку путь из легчайших перегонов (Optimal routing
*aka Easiest passes-модификация алгоритма). Значения COST принимаются за трудность преодоления
*перегона из точки в точку: COST(i,j) это трудность из i в j.
*Из точки START в каждую точку j будет найден самый легкий путь:
*все перегоны, из которых он будет состоять, минимально трудны. Результат NAME1 будет
*содержать трудность самого трудного перегона на этом легчайшем пути из точки-ряда START
*во всякую точку-столбец j.
*Сами найденные (т.е. оптимальные) пути, траектории, будут записаны в "свернутом виде" в вектор NAME2.
*Значение j-го элемента в NAME2 это номер точки, предшествующей точке j на всяком найденном пути,
*ведущем к точке j как к конечной точке пути. См. ПРИМЕР, показывающий, как извлечь всю траекторию пути
*из точки START в некоторую конечную точку V=j.
*Записывать траекторию или нет - зависит от аргумента PATH. Аргумент PATH - это цифра 1 или 0
*(не имя и не выражение, цифру можно взять в кавычки или апострофы). Если 0, то траектории не будут
*записаны и NAME2 выйдет как нулевой скаляр.
*Отсутствующие значения в COST и невозможный путь. Если во входящей матрице должно не быть связи
*между некоторыми точками - например, перегона из точки i в точку j не существует, то поставьте в
*ячейку (i,j) значение: -1. Тогда, если некоторый V-й элемент NAME1 оказался -1, то значит,
*путь из START в конечную точку V - невозможен.
*Аргумент STOP. Если вам нужны пути из START ко всем точкам, установите STOP на значение 0.
*Если же вас интересует путь из START только в одну определенную точку, то укажите ее
*номер тут. Тогда алгоритм прекратит работать, дойдя до этой точки; так можно сэкономить время.
*Результат NAME1 выйдет тогда не вектором, а скаляром. STOP должно отличаться от START.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если вам нужны пути из всякой точки во всякую точку, используйте функцию /*!KO_flowar*/.

define !KO_dijkstra(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%')
                   /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
comp @u= !3.
comp @stop= !4.
comp @n= ncol(@mat).
!let !task= !unquote(!5)
!let !path= !unquote(!6)
comp !7= make(1,@n,1E300).
!if (!path) !then
-comp !8= make(1,@n,0).
!else
-comp !8= 0.
!ifend
comp @q= 1:@n.
comp !7(@u)= 0.
loop if any(@q).
-comp @u=  mmax((!7=mmin(!7+(not @q)*1E300))&*@q). 
-do if @u=@stop.
- break.
-end if.
-comp @q(@u)= 0.
 !if (!task='EPASS') !then
- comp @alt= cmax({@mat(@u,:);make(1,ncol(@mat),!7(@u))}).
 !else
- comp @alt= @mat(@u,:)+!7(@u).
 !ifend
-comp @flag= @alt<!7 and @mat(@u,:)<>-1 and @q.
-comp !7= !7&*(not @flag)+@alt&*@flag.
 !if (!path) !then
- comp !8= !8&*(not @flag)+@u*@flag.
 !ifend
end loop.
do if @stop.
-comp !7= !7(@stop).
end if.
comp @q= !7=1E300.
comp !7= !7&*(not @q)-@q.
release @mat,@u,@stop,@n,@q,@alt,@flag.
!enddefine.


******** АЛГОРИТМ МИНИМАЛЬНОГО ОСТОВНОГО ДЕРЕВА ПРИМА ********.
*/*!KO_prim(mat%cg%start%name)*/*.
*Version 1.
*В однодольном взвешенном ненаправленном графе, заданным матрицей смежности MAT, находит
*минимальное (или максимальное) остовное дерево, начиная с вершины START (скаляр, номер вершины).
*MAT - квадратная n x n симметричная матрица. Ее диагональные элементы функция игнорирует.
*Аргумент CG: задает понимание MAT. CG - цифра (не имя и не выражение) 0, 1 или 2
*(цифру можно опционально взять в кавычки или апострофы).
*Если CG=0, то MAT понимается как матрица стоимостей (cost) или расстояний между вершинами
*полного взвешенного графа. Результат NAME это минимальное остовное дерево.
*Если CG=1, то MAT понимается как матрица выигрышей (gain) или сходств между вершинами
*полного взвешенного графа. Результат NAME это максимальное остовное дерево. В обоих этих случаях
*значения в MAT могут иметь любой знак.
*Если CG=2, то MAT понимается как разреженная (sparse) неотрицательная матрица выигрышей:
*с положительными значениями в качестве выигрышей и нулевыми элементами, соответствующими
*отсутствующим ребрам (неполный граф). Результат NAME тогда это максимальное остовное дерево из
*вершины start в условиях неполного графа. Если неполный граф не является связным графом
*(т.е. таким графом, где есть путь из всякой вершины во всякую вершину), то остовное дерево
*покроет только его часть - подграф, содержащий вершину start, - и в результате в NAME
*будут заполнены не все n-1 рядов (какие-то последние останутся пусты). Так алгоритм Прима
*можно использовать с CG=2 для диагностики связности неполного графа.
*Когда граф полный или неполный, но связный, остовное дерево покроет весь граф и величина
*(сумма весов) дерева не зависит от выбора вершины start.
*Результат NAME: имеет три столбца и n-1 рядов, ряды соответствуют ребрам, записанным в остовное
*дерево. Первые два элемента в ряду это вершины i и j, смежные ребру, а третье значение это
*вес ребра, т.е. значение (i,j) в MAT. Алгоритм Прима минимизирует (если MAT это cost-матрица) или
*максимизирует (если MAT это gain-матрица) сумму в 3-м столбце NAME.
*Для этой функции может требоваться предварительная установка предела числу циклов на достаточно большое
*число командой SET MXLOOPS.
*Если ваша MAT это cost-матрица и граф неполный (некоторые ребра отсутствуют), переведите ее в
*положительную gain-матрицу, обратив знак и прибавив положительную константу, а отсутствующие ребра
*пометьте нулями; далее используйте CG=2.

define !KO_prim(!pos= !token(1) /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend('%') /!pos= !charend(')'))
comp @mat= !2.
!let !cg= !unquote(!3)
comp @st= !4.
comp @n= ncol(@mat).
comp !5= make(@n-1,3,0).
comp @rns= @st.
do if @st=1.
-comp @cns= 2:@n.
else if @st=@n.
-comp @cns= 1:(@n-1).
else.
-comp @cns= {1:(@st-1),(@st+1):@n}.
end if.
comp @ri= 0.
comp @ci= 0.
loop @i= 1 to nrow(!5).
-comp @m= @mat(@rns,@cns).
-comp @val= !if (!cg=0) !then mmin !else mmax !ifend (@m).
 !if (!cg=2) !then
- do if not @val.
-  break.
- end if.
 !ifend
-comp @m= @m=@val.
-comp @ci= mmax(@m&*(1:ncol(@m))(make(1,nrow(@m),1),:)).
-comp @ri= mmax(@m(:,@ci)&*t(1:nrow(@m))).
-comp !5(@i,:)= {@rns(@ri),@cns(@ci),@val}.
-comp @rns= {@rns,@cns(@ci)}.
-do if @i<nrow(!5).
- do if @ci=1.
-  comp @cns= @cns(2:ncol(@cns)).
- else if @ci=ncol(@cns).
-  comp @cns= @cns(1:(ncol(@cns)-1)).
- else.
-  comp @cns= @cns({1:(@ci-1),(@ci+1):ncol(@cns)}).
- end if.
-end if.
end loop.
release @mat,@st,@n,@rns,@cns,@i,@m,@ci,@ri.
!enddefine.
